/*************************************************************************
 *
 *  Bench -- a benchmarking tool for Fortune's algorithm
 *  Copyright (C) 2012-2014 Jussi M. Hukkanen
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *************************************************************************
 *
 *  bench.c: main module
 *
 ************************************************************************/

#include <stdio.h>
#include <signal.h>
#include <sys/time.h>

//#define USE_RAND48

#include "fortune.h"
#include "opts.h"
#include "rnd.h"

#ifdef H_PLATFORM
#include H_PLATFORM
#endif


#define SITE_X_RANGE 1000
#define SITE_Y_RANGE 1000
#define SITE_X_MIN      0
#define SITE_Y_MIN      0


#ifndef VALGRIND

#ifndef TARGET_WINDOWS
volatile int timer_flag = 0;
#else
#ifdef SDL
#include <SDL.h>
Uint32 sdl_timer_target = 0;
#define timer_flag (SDL_GetTicks() > sdl_timer_target)
#endif
#endif

#else

#define DEFAULT_VALGRIND_RUNS 1000
#define MIN_VALGRIND_RUNS        1
#define MAX_VALGRIND_RUNS   100000
int valgrind_runs;
#define timer_flag ((valgrind_runs--) <= 0)

#endif

long long int benchmark_runs;

#define DEFAULT_BENCHMARK_SECONDS   10
#define MIN_BENCHMARK_SECONDS        1
#define MAX_BENCHMARK_SECONDS     3600

uint16_t benchmark_duration =
	#ifdef VALGRIND
	DEFAULT_VALGRIND_RUNS;
	#else
	DEFAULT_BENCHMARK_SECONDS;
	#endif


#define DEFAULT_SITES   100

int sites = DEFAULT_SITES;
long sites_total;

site_t *site; /* sites items */


#define DELAUNAY	(function & FORTUNE_DELAUNAY_MASK)
#define VORONOI		(function & FORTUNE_VORONOI_MASK)

uint8_t function;

#define DEFAULT_SEED 42

int32_t random_seed = DEFAULT_SEED;


#ifndef FORTUNE_NO_VORONOI
vertex_t *vlist;		/* array of Voronoi vertices */
boundary_t *blist;		/* array of Voronoi boundaries */
celldata_t *cell_list;	/* array of Voronoi cells */
int vnum, bnum;			/* number of vertices and boundaries */
long vnum_total, bnum_total;		/* accumulating counts */
#endif

#ifndef FORTUNE_NO_DELAUNAY
triangle_t *tlist;		/* array of Delaunay triangles */
connect_t *clist;		/* array of Delaunay connections */
int tnum, cnum;			/* number of triangles and connections */
long tnum_total, cnum_total;		/* accumulating counts */
#endif


#ifndef VALGRIND

#ifndef TARGET_WINDOWS

/* sighandler()

	sets the timer semaphore */

void sighandler(int sig) {

	timer_flag = 1;
}


/* init_timer()

	creates a handler for SIGALRM */

void init_timer(long int sec, long int usec) {

	struct sigaction sa;
	struct itimerval itv;

	timer_flag = 0;

	sigemptyset(&sa.sa_mask);
	sa.sa_flags = SA_RESTART;
	sa.sa_handler = sighandler;
	sigaction(SIGALRM, &sa, NULL);

	itv.it_interval.tv_sec = sec;
	itv.it_interval.tv_usec = usec;
	itv.it_value.tv_sec = itv.it_interval.tv_sec;
	itv.it_value.tv_usec = itv.it_interval.tv_usec;

	setitimer(ITIMER_REAL, &itv, NULL);
}

#else

#ifdef SDL	/* on Windows, we use SDL for timing */

/* init_timer()

	the Windows version sets a value
	  that SDL_GetTicks() must reach */

void init_timer(long int sec, long int usec) {

	sdl_timer_target = SDL_GetTicks() + (sec * 1000) + (usec / 1000);
}

#endif

#endif

#else

/* init_timer()

	dummy function to use with VALGRIND */

void init_timer(long int sec, long int usec) {

	valgrind_runs = sec;
}

#endif


/* create_random_sites()
*/

void create_random_sites(void) {

	uint16_t x, y;
	int i;

	#ifndef USE_RAND48
	#define RANDOM_COORD(R, M) (RND_FLOAT((float)R) + (float)M)
	#else
	#define RANDOM_COORD(R, M) (RND_FLOAT(R) + M)
	#endif

	for (i = 0; i < sites; i++) {
		site[i].x = RANDOM_COORD(SITE_X_RANGE, SITE_X_MIN);
		site[i].y = RANDOM_COORD(SITE_Y_RANGE, SITE_Y_MIN);
	}
}


/* parse_options()

	interprets the given list of command-line
	  options */

char parse_options(int argc, char *argv[]) {

	#define CMDLINE_OPTC_HELP		'h'
	#define CMDLINE_OPTS_HELP			"help"
	#define CMDLINE_OPTC_DELAUNAY	'd'
	#define CMDLINE_OPTS_DELAUNAY		"delaunay"
	#define CMDLINE_OPTC_VORONOI	'v'
	#define CMDLINE_OPTS_VORONOI		"voronoi"
	#define CMDLINE_OPTC_SITES		'n'
	#define CMDLINE_OPTS_SITES			"num"
	#define CMDLINE_OPTC_RNDSEED	's'
	#define CMDLINE_OPTS_RNDSEED		"seed"
	#define CMDLINE_OPTC_GRID		'g'
	#define CMDLINE_OPTS_GRID			"grid"
	#define CMDLINE_OPTC_SECONDS	't'
	#define CMDLINE_OPTS_SECONDS		"time"

	struct option long_opts[] = {
		{ CMDLINE_OPTS_HELP,		no_argument,		NULL,	CMDLINE_OPTC_HELP		},
		{ CMDLINE_OPTS_DELAUNAY,	no_argument,		NULL,	CMDLINE_OPTC_DELAUNAY	},
		{ CMDLINE_OPTS_VORONOI,		no_argument,		NULL,	CMDLINE_OPTC_VORONOI	},
		{ CMDLINE_OPTS_SITES,		required_argument,	NULL,	CMDLINE_OPTC_SITES		},
		{ CMDLINE_OPTS_RNDSEED,		required_argument,	NULL,	CMDLINE_OPTC_RNDSEED	},
		{ CMDLINE_OPTS_GRID,		required_argument,	NULL,	CMDLINE_OPTC_GRID		},
		{ CMDLINE_OPTS_SECONDS,		required_argument,	NULL,	CMDLINE_OPTC_SECONDS	},
		{ NULL, 0, NULL, 0 }
	};
	int c;
	int i = 0;
	int index = 0;
	char *optstring;

	optstring = make_opt_string(long_opts);
	if (optstring == NULL) {
		fprintf(stderr, "Out of memory reading command line options!\n");
		return RET_FORTUNE_NO_MEM;
	}

	while (1) {

		c = getopt_long(argc, argv, optstring, long_opts, &index);

		if (c == -1)
			break;

		if (optarg != NULL)
			i = atoi(optarg);

		switch (c) {

			case CMDLINE_OPTC_HELP:
				printf("bench - Copyright (C) 2012-2014 Jussi M. Hukkanen\n");
				printf("Run a benchmark test of Fortune's algorithm.\n");
				printf("This build uses the "
				       #ifdef USE_RAND48
				       "drand48()"
				       #else
				       "rand()"
				       #endif
				       " pseudo-random number\ngenerator.\n\n");
				printf("This program comes with ABSOLUTELY NO WARRANTY.\n"
				       "This is free software, and you are welcome to\n"
				       "redistribute it under the conditions of the\n"
				       "GNU General Public License, version 3 or later.\n"
				       "To view the License, see `gpl.txt'.\n\n");

				printf("Usage: %s [options]\n", argv[0]);
				#ifndef FORTUNE_NO_DELAUNAY
				#ifndef FORTUNE_NO_VORONOI
				printf("   -%c, --" CMDLINE_OPTS_DELAUNAY "\n", CMDLINE_OPTC_DELAUNAY);
				printf("   -%c, --" CMDLINE_OPTS_VORONOI "\n", CMDLINE_OPTC_VORONOI);
				printf("See `triangle --" CMDLINE_OPTS_HELP "' "
					   "for help on the above options.\n");
				#endif
				#endif
				printf("   -%c N, --" CMDLINE_OPTS_RNDSEED " N\n", CMDLINE_OPTC_RNDSEED);
				printf("        sets the seed value for the random number\n"
					   "        generator. N can be any non-negative integer.\n");
				printf("   -%c N, --" CMDLINE_OPTS_SITES " N\n", CMDLINE_OPTC_SITES);
				printf("        sets the number of sites created. N must be\n"
					   "        at least three. The default value is %u.\n",
								DEFAULT_SITES);
				printf("   -%c N, --" CMDLINE_OPTS_SECONDS " N\n", CMDLINE_OPTC_SECONDS);
				#ifdef VALGRIND
				printf("        sets the number of test runs.\n"
					   "        N must be between %u and %u; default is %u.\n",
								MIN_VALGRIND_RUNS, MAX_VALGRIND_RUNS,
								DEFAULT_VALGRIND_RUNS);
				#else
				printf("        sets the duration of the test in seconds.\n"
					   "        N must be between %u and %u; default is %u.\n",
								MIN_BENCHMARK_SECONDS, MAX_BENCHMARK_SECONDS,
								DEFAULT_BENCHMARK_SECONDS);
				#endif
				return RET_FORTUNE_NO_TRIANGLES;

			case CMDLINE_OPTC_DELAUNAY:
				#ifdef FORTUNE_NO_DELAUNAY
				printf("This build does not create Delaunay triangles.\n");
				printf("Re-compile without the FORTUNE_NO_DELAUNAY flag.\n");
				return RET_FORTUNE_PARAM_ERROR;
				#else
				function |= FORTUNE_DELAUNAY_MASK;
				break;
				#endif

			case CMDLINE_OPTC_VORONOI:
				#ifdef FORTUNE_NO_VORONOI
				printf("This build does not create Voronoi diagrams.\n");
				printf("Re-compile without the FORTUNE_NO_VORONOI flag.\n");
				return RET_FORTUNE_PARAM_ERROR;
				#else
				function |= FORTUNE_VORONOI_MASK;
				break;
				#endif

			case CMDLINE_OPTC_SITES:
				if (i >= 3)
					sites = i;
				break;

			case CMDLINE_OPTC_RNDSEED:
				if (i > 0)
					random_seed = i;
				break;

			case CMDLINE_OPTC_SECONDS:
				if (
					#ifdef VALGRIND
					i >= MIN_VALGRIND_RUNS && i <= MAX_VALGRIND_RUNS
					#else
					i >= MIN_BENCHMARK_SECONDS && i <= MAX_BENCHMARK_SECONDS
					#endif
					)
					benchmark_duration = i;
				break;
		}
	}

	free(optstring);

	return 0;
}


/* main()
*/

int main(int argc, char *argv[]) {

	int i;

	function = 0;

	i = parse_options(argc, argv);
	if (i != 0)
		return i;

	#ifdef TARGET_WINDOWS

	/* MinGW doesn't recognise the 'll'
	     (long long) modifier in format strings
	     sent to printf(), so we have to use
	     this alternative */

	#define ITERATION_FORMAT "%11I64d"

	#ifndef SDL
    fputs("The Windows version of this program requires SDL.\n"
          "Please rebuild.\n", stderr);
    return 0;
    #else
	if (SDL_Init(SDL_INIT_TIMER) == -1) {
		fprintf(stderr, "ERROR: Could not initialize SDL: %s\n", SDL_GetError());
		return -1;
	}
	atexit(SDL_Quit);
    #endif

    #else

	#define ITERATION_FORMAT "%11lld"

	#endif

	if (!function) {
		#ifndef FORTUNE_NO_DELAUNAY
		function = FORTUNE_DELAUNAY_MASK;
		#else
		function = FORTUNE_VORONOI_MASK;
		#endif
	}

	site = (site_t *)malloc(sites * sizeof(site_t));

	if (site == NULL) {
		fprintf(stderr,
			"Not enough memory for %u sites! (" SIZEOF_FORMAT " bytes needed)\n",
			sites, sites * sizeof(site_t));
		return 1;
	}

	RND_SEED(random_seed);

	#ifdef VALGRIND
	#define PRINT_TEST_DURATION {						\
		printf("    Testing %u run%c...",				\
			benchmark_duration,							\
			(benchmark_duration == 1) ? '.' : 's');		\
	}
	#define PRINT_TEST_RESULTS {						\
		printf("\n");									\
	}
	#else
	#define PRINT_TEST_DURATION {						\
		printf("    Test duration %u second%c...",		\
			benchmark_duration,							\
			(benchmark_duration == 1) ? '.' : 's');		\
	}
	#define PRINT_TEST_RESULTS {						\
		printf(ITERATION_FORMAT " runs;",				\
			benchmark_runs);							\
		printf(ITERATION_FORMAT " runs/sec.\n",			\
			(benchmark_runs +							\
			(benchmark_duration >> 1)) /				\
			benchmark_duration);						\
	}
	#endif

	#if 1
	#define PRINT_STATISTICS(WHAT, STR, PADDING) {		\
		printf("%15ld " STR ";" PADDING, WHAT);			\
		printf("%8.3f " STR "/site\n",					\
			((double)WHAT) / sites_total);				\
	}
	#else
	#define PRINT_STATISTICS(WHAT, STR, PADDING) { }
	#endif

	#define DO_BENCHMARK(FUNC, STR, SETUP, CLEANUP) {	\
														\
		printf(STR " benchmark [%u sites]\n", sites);	\
		PRINT_TEST_DURATION;							\
		fflush(stdout);									\
														\
		benchmark_runs = 1;								\
		sites_total = 0;								\
														\
		init_timer(benchmark_duration, 0);				\
		i = RET_FORTUNE_OK;								\
														\
		while (!timer_flag) {							\
														\
			create_random_sites();						\
														\
			SETUP;										\
														\
			i = FUNC;									\
			if (i != RET_FORTUNE_OK)					\
				break;									\
														\
			CLEANUP;									\
														\
			benchmark_runs++;							\
														\
			if (benchmark_runs < 0) {					\
				printf("\nCPU too fast!");				\
				break;									\
			}											\
		}												\
														\
		switch (i) {									\
			case RET_FORTUNE_OK:						\
				break;									\
			case RET_FORTUNE_NO_MEM:					\
				printf("\nOut of memory!\n");			\
				break;									\
			case RET_FORTUNE_NO_TRIANGLES:				\
				printf("\nNot enough sites!\n");		\
				break;									\
			case RET_FORTUNE_PARAM_ERROR:				\
				printf("\nInvalid parameter!\n");		\
				break;									\
			default:									\
				printf("\nError %d\n", i);				\
		}												\
														\
		benchmark_runs--;								\
														\
		PRINT_TEST_RESULTS;								\
	}

	#ifndef FORTUNE_NO_DELAUNAY

	if (DELAUNAY) {

		#define DELAUNAY_TOTALS {	\
			sites_total += sites;	\
			tnum_total += tnum;		\
			cnum_total += cnum;		\
		}

		tnum_total = cnum_total = 0;

		DO_BENCHMARK(Delaunay(site, sites, &clist, &cnum, &tlist, &tnum),
			"Delaunay()",
			{ clist = NULL; tlist = NULL; },
			#ifdef FORTUNE_MASTER_ARRAYS
			DELAUNAY_TOTALS
			#else
			{ DELAUNAY_TOTALS; Delaunay_cleanup(clist, tlist) }
			#endif
			);

		PRINT_STATISTICS(tnum_total, "triangles", "  ");
		PRINT_STATISTICS(cnum_total, "connections", "");

		#ifdef FORTUNE_MASTER_ARRAYS
		Delaunay_cleanup(clist, tlist);
		Voronoi_cleanup(NULL, NULL, cell_list);
		#endif
	}

	#endif

	#ifndef FORTUNE_NO_VORONOI

	if (VORONOI) {

		#define VORONOI_TOTALS {	\
			sites_total += sites;	\
			vnum_total += vnum;		\
			bnum_total += bnum;		\
		}

		vnum_total = bnum_total = 0;

		DO_BENCHMARK(Voronoi(site, sites, &vlist, &vnum, &blist, &bnum, &cell_list, NULL),
			"Voronoi()",
			{ vlist = NULL; blist = NULL; cell_list = NULL; },
			#ifdef FORTUNE_MASTER_ARRAYS
			VORONOI_TOTALS
			#else
			{ VORONOI_TOTALS; Voronoi_cleanup(vlist, blist, cell_list) }
			#endif
			);

		PRINT_STATISTICS(vnum_total, "vertices", "   ");
		PRINT_STATISTICS(bnum_total, "boundaries", " ");

		#ifdef FORTUNE_MASTER_ARRAYS
		Voronoi_cleanup(vlist, blist, cell_list);
		#endif
	}

	#endif

	i = RET_FORTUNE_OK;

	/* free memory and exit */

	free(site);

	return i;
}

