/*************************************************************************
 *
 *  Bitmap -- a utility for creating arrays of sites for `triangle'
 *  Copyright (C) 2012-2014 Jussi M. Hukkanen
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *************************************************************************
 *
 *  bitmap.c: entire program
 *
 ************************************************************************/

#include <stdio.h>
#include <string.h>
#include <float.h>
#include "opts.h"
#include "rnd.h"

#ifndef NO_SDL_IMAGE
#include <math.h>		/* declares pow() */
#include <SDL.h>
#include <SDL_image.h>
#endif

#ifdef H_PLATFORM
#include H_PLATFORM
#endif


#define RET_OK        0
#define RET_NO_FILE  -1
#define RET_FILE_ERR -2
#define RET_NO_MEM   -3
#ifndef NO_SDL_IMAGE
#define RET_SDL_ERR  -9
#endif

char *input_file = NULL;

int width, height;

int32_t random_seed = 42;
float range = 0.5;

#define FRAME_OFF	0x00
#define FRAME_ON	0x01
#define FRAME_OLD	0x02

char frame = FRAME_ON;

#define DEFAULT_FRAME_SCALE	 1.1
#define DEFAULT_FRAME_DIV	18.0

float frame_scale = DEFAULT_FRAME_SCALE;
float frame_div = DEFAULT_FRAME_DIV;

#ifndef NO_SDL_IMAGE

short levels[256];				/* adjustment table for 256 shades of grey */

#define DIFFUSE_BITS	6		/* added precision */
#define LEVELS_SCALE	(1 << (8 + DIFFUSE_BITS))
#define LEVELS_WHITE	(LEVELS_SCALE - 1)
#define LEVELS_HALF		(1 << (7 + DIFFUSE_BITS))

#define DEFAULT_OUTPUT_BLACK	170
#define MIN_INPUT_GAMMA			0.05
#define MAX_INPUT_GAMMA			20.0		/* == 1.0 / MIN_INPUT_GAMMA */

Uint8 input_black = 0;
Uint8 input_white = 255;
float input_gamma = 1.0;
Uint8 output_black = DEFAULT_OUTPUT_BLACK;
Uint8 output_white = 255;

#define DEFAULT_WEIGHT_RED		 9797		/* 0.299 */
#define DEFAULT_WEIGHT_GREEN	19235		/* 0.587 */
#define DEFAULT_WEIGHT_BLUE		 3736		/* 0.114 */

Uint16 Rweight, Gweight, Bweight;

int custom_weight_sum = 0;	/* nonzero = user-defined weights */

#define DITHER_ORDERED			0
#define DITHER_FLOYD_STEINBERG	1
#define DITHER_SIERRA_2			2
#define DITHER_SIERRA_3			3
#define DITHER_MAX				3

#define DEFAULT_DITHER	DITHER_SIERRA_3

char dither = DEFAULT_DITHER;

#if 0
#define BAYER_SIZE	4
Uint8 dither_matrix[BAYER_SIZE * BAYER_SIZE] = {
	 15, 136,  45, 166,
	196,  75, 226, 105,
	 60, 181,  30, 151,
	241, 120, 211,  90
};
#else
#define BAYER_SIZE	8
Uint8 dither_matrix[BAYER_SIZE * BAYER_SIZE] = {
	  4, 130,  36, 161,  12, 138,  44, 169,
	193,  67, 224,  99, 200,  75, 232, 106,
	 51, 177,  20, 146,  59, 185,  28, 153,
	240, 114, 208,  83, 248, 122, 216,  91,
	 16, 142,  48, 173,   8, 134,  40, 165,
	204,  79, 236, 110, 197,  71, 228, 102,
	 63, 189,  32, 157,  55, 181,  24, 150,
	252, 126, 220,  95, 244, 118, 212,  87
};

#endif

#define DEFAULT_ORIENTATION		0
#define ORIENTATION_270			1
#define ORIENTATION_180			2
#define ORIENTATION_90			3
#define ORIENTATION_FLIP		4

Uint8 transform = DEFAULT_ORIENTATION;

#define PREVIEW_OFF				0
#define PREVIEW_ON				1
#define PREVIEW_HISTOGRAM		2

char preview_window = PREVIEW_OFF;

#define DEFAULT_HISTOGRAM_W		79
#define DEFAULT_HISTOGRAM_H		49
#define MIN_HISTOGRAM_SIZE		 5

Uint8 histogram_w = DEFAULT_HISTOGRAM_W - 1;
Uint8 histogram_h = DEFAULT_HISTOGRAM_H - 1;

#endif

int x, y;

float fwidth = 2.0;
float fheight = 2.0;
float xf, yf;
float xfstart;
float xfstep, yfstep;
float xsmin, ysmin;
float xsrange, ysrange;
float site_x, site_y;

#define SITE_LOOP_INIT {				\
	if (width >= height)				\
		fwidth = (fheight * width)		\
			/ height;					\
	else								\
		fheight = (fwidth * height)		\
			/ width;					\
	xfstart = -fwidth / 2.0;			\
	xfstep =  fwidth  / width;			\
	yfstep = -fheight / height;			\
	xsrange = xfstep * range;			\
	ysrange = yfstep * range;			\
	xsmin = (xfstep - xsrange) / 2.0;	\
	ysmin = (yfstep - ysrange) / 2.0;	\
	xfstart += xsmin;					\
	xf = xfstart;						\
	yf = ysmin + (fheight / 2.0);		\
}

#define SITE_LOOP_ADVANCE {				\
	xf += xfstep;						\
}

#define SITE_LOOP_NEWLINE {				\
	xf = xfstart;						\
	yf += yfstep;						\
}

#define OUTPUT_COORD(X, Y) {			\
	printf("%10.6f,%10.6f\n", X, Y);	\
}

#define OUTPUT_SITE {					\
	site_x = xf + RND_FLOAT(xsrange);	\
	site_y = yf + RND_FLOAT(ysrange);	\
	OUTPUT_COORD(site_x, site_y);		\
}


#ifndef NO_SDL_IMAGE

/* wait_for_key()
*/

void wait_for_key(void) {

	SDL_Event event;

	while (SDL_WaitEvent(&event) > 0)
		if (event.type == SDL_KEYDOWN || event.type == SDL_QUIT)
			break;

	return;
}


/* create_adjustment_table()

	fills the levels lookup with values
	  based on the input_* and output_*
	  variables */

void create_adjustment_table() {

	float k, kstep, range;
	float factor;
	Uint16 v, w;
	short d;

	if (input_white < input_black)
		input_white = input_black;

	factor = LEVELS_WHITE / 255.0;

	for (d = (output_black * factor) + 0.5, v = 0; v < input_black; v++)
		levels[v] = d;

	if (input_white > input_black) {

		range = (output_white - output_black) * factor;

		for (k = 0.0, kstep = 1.0 / (input_white - input_black),
			w = output_black * factor;
			v < input_white; v++, k += kstep)

			levels[v] = w + (range * pow(k, input_gamma)) + 0.5;
	}

	for (d = (output_white * factor) + 0.5; v < 256; v++)
		levels[v] = d;

	if (preview_window & PREVIEW_HISTOGRAM) {
		#if 0
		for (v = 0; v < 256; v++) printf("%u ", levels[v]); puts("");
		#else
		for (d = LEVELS_WHITE; d >= 0; d -= LEVELS_WHITE / histogram_h) {
			for (v = w = 0; v <= histogram_w; v++, w += 255)
				putchar((levels[w / histogram_w] >= d) ? '|' : ' ');
			puts("");
		}
		#endif
	}
}


/* convert_image()

	reads and converts any image file
	  supported by SDL_image */

char convert_image() {

	SDL_Surface *image;
	SDL_Surface *preview = NULL;
	SDL_Surface *converted_image;
	Uint8 *greyscale;
	short *diffused = NULL;	/* buffer used in error diffusion */
	SDL_PixelFormat *format;
	//#define PREVIEW_COLOURS 256
	#define PREVIEW_COLOURS 2
	SDL_Color preview_palette[PREVIEW_COLOURS];
	Uint16 pitch;
	Uint8 bytes;
	Uint8 *p;				/* pointer to image data */
	Uint8 *q;				/* another one */
	Uint8 *c;				/* pointer to greyscale */
	Uint8 r, g, b;			/* red, green and blue components */
	Uint32 u, v;			/* offsets to diffused */
	int e, es;				/* quantization error, times 16 or 32 */
	int step_minor, step_major;
	int dim_minor, dim_major;
	char ret = RET_OK;

	//#define TGA_TO_STDOUT 1

	image = IMG_Load(input_file);
	if (image == NULL) {
		fprintf(stderr, "Unable to read `%s': %s",
			input_file, SDL_GetError());
		return RET_FILE_ERR;
	}

	width = image->w;
	height = image->h;
	pitch = image->pitch;
	format = image->format;
	bytes = format->BytesPerPixel;

	#if 0
	#ifndef TGA_TO_STDOUT
	printf("%s - %u x %u pixels, %u bytes per pixel (pitch = %u)",
		input_file, width, height, bytes, image->pitch);
	if (bytes == 1) {
		if (format->palette == NULL)
			printf(", no palette");
		else
			printf(", %u colours in palette", format->palette->ncolors);
	}
	printf("\n");
	printf("R / G / B = %u / %u / %u (sum = %u)\n",
		Rweight, Gweight, Bweight, custom_weight_sum);
	#endif
	#endif

	/* create preview image buffer and window */

	greyscale = (Uint8 *)malloc(width * height * sizeof(Uint8));
	if (greyscale == NULL) {
		ret = RET_NO_MEM;
		goto done_with_input_image;
	}

	/* create error diffusion buffer */

	if (dither != DITHER_ORDERED) {

		diffused = (short *)malloc(width * height * sizeof(short));
		if (diffused == NULL) {
			ret = RET_NO_MEM;
			goto done_with_greyscale_image;
		}
	}

	#ifndef TGA_TO_STDOUT
	if (preview_window) {

		preview = SDL_SetVideoMode(width, height, 8, SDL_HWSURFACE);

		if (preview == NULL) {
			fprintf(stderr, "ERROR: Could not set %u" "x%u" "x%u video mode: %s\n",
				width, height, 8, SDL_GetError());
			ret = RET_SDL_ERR;
			goto done_with_greyscale_image;
		}

		#if PREVIEW_COLOURS > 2
		for (e = 0; e < PREVIEW_COLOURS; e++) {
			preview_palette[e].r =
			preview_palette[e].g =
			preview_palette[e].b = e;
			preview_palette[e].unused = 0;
		}
		#else
		preview_palette[0].r = preview_palette[0].g = preview_palette[0].b = 0;
		preview_palette[1].r = preview_palette[1].g = preview_palette[1].b = 0xFF;
		#endif
		SDL_SetColors(preview, preview_palette, 0, PREVIEW_COLOURS);

		SDL_WM_SetCaption(input_file, NULL);
	}
	#endif

	#define OUTPUT_PIXEL { *(c++) = v; }

	/* read and handle one scanline at a time */

	#define DEFAULT_GREY(R, G, B) {							\
		v = (R * Rweight) + (G * Gweight) + (B * Bweight);	\
		v >>= 15;											\
	}

	#define CUSTOM_GREY(R, G, B) {							\
		v = (R * Rweight) + (G * Gweight) + (B * Bweight);	\
		v /= custom_weight_sum;								\
	}

	SDL_LockSurface(image);

	#define READ_8BPP {						\
		SDL_GetRGB(*q, format, &r, &g, &b);	\
	}

	#define READ_OTHER {					\
		SDL_GetRGB(*(Uint32 *)q,			\
			format, &r, &g, &b);			\
	}

	#define READ_IMAGE_LOOP(READER, HANDLER) {				\
		p = (Uint8 *)image->pixels;							\
		c = greyscale;										\
		for (y = height; y; y--, p += image->pitch) {		\
			for (x = width, q = p; x; x--, q += bytes) {	\
				READER;										\
				HANDLER(r, g, b);							\
				OUTPUT_PIXEL;								\
			}												\
		}													\
	}

	if (custom_weight_sum) {

		if (bytes == 1) {
			READ_IMAGE_LOOP(READ_8BPP, CUSTOM_GREY);
		}
		else {
			READ_IMAGE_LOOP(READ_OTHER, CUSTOM_GREY);
		}
	}
	else {

		if (bytes == 1) {
			READ_IMAGE_LOOP(READ_8BPP, DEFAULT_GREY);
		}
		else {
			READ_IMAGE_LOOP(READ_OTHER, DEFAULT_GREY);
		}
	}

	SDL_UnlockSurface(image);

	/* output greyscale data */

	if (dither == DITHER_ORDERED) {

		for (y = 0, c = greyscale; y < height; y++) {

			if (!(y & (BAYER_SIZE - 1)))
				p = dither_matrix;
			else
				p += BAYER_SIZE;

			for (x = 0; x < width; x++, c++) {
				e = levels[*c] >> DIFFUSE_BITS;
				*c = (e >= p[x & (BAYER_SIZE - 1)]) ?
					(PREVIEW_COLOURS - 1) : 0;
			}
		}
	}
	else {

		/* copy greyscale to diffused */

		for (v = width * height, u = 0; u < v; u++)
			diffused[u] = levels[greyscale[u]];

		/* setup error diffusion algorithm */

		switch (transform) {

			case ORIENTATION_90:
				step_minor = width;		step_major = -1;
				dim_minor = height;		dim_major = width;
				u = width - 1;
				break;

			case ORIENTATION_90 | ORIENTATION_FLIP:
				step_minor = -width;	step_major = -1;
				dim_minor = height;		dim_major = width;
				u = (width * height) - 1;
				break;

			case ORIENTATION_180:
				step_minor = -1;		step_major = -width;
				dim_minor = width;		dim_major = height;
				u = (width * height) - 1;
				break;

			case ORIENTATION_180 | ORIENTATION_FLIP:
				step_minor = 1;			step_major = -width;
				dim_minor = width;		dim_major = height;
				u = width * (height - 1);
				break;

			case ORIENTATION_270:
				step_minor = -width;	step_major = 1;
				dim_minor = height;		dim_major = width;
				u = width * (height - 1);
				break;

			case ORIENTATION_270 | ORIENTATION_FLIP:
				step_minor = width;		step_major = 1;
				dim_minor = height;		dim_major = width;
				u = 0;
				break;

			case ORIENTATION_FLIP:
				step_minor = -1;		step_major = width;
				dim_minor = width;		dim_major = height;
				u = width - 1;
				break;

			default:
				step_minor = 1;			step_major = width;
				dim_minor = width;		dim_major = height;
				u = 0;
		}

		#define HANDLE_PIXEL(INDEX, BITS) {	\
			v = INDEX;						\
			e = diffused[v];				\
			if (e < LEVELS_HALF) {			\
				diffused[v] = 0;			\
			}								\
			else {							\
				e -= LEVELS_WHITE;			\
				diffused[v] =				\
					PREVIEW_COLOURS - 1;	\
			}								\
			es = e << BITS;					\
		}

		#define FAKE_SPREAD(WHAT) {			\
			es -= WHAT;						\
		}

		#if 0	/* clamp values inside 0..255 */
		#define SPREAD(WHAT, BITS, WHERE) {	\
			f = diffused[v + (WHERE)] +		\
				((WHAT) >> BITS);			\
			diffused[v + (WHERE)] =			\
				(f < 0) ? 0 :				\
				((f > 0xFF) ? 0xFF : f);	\
			FAKE_SPREAD(WHAT);				\
		}
		#else
		#define SPREAD(WHAT, BITS, WHERE) {	\
			diffused[v + (WHERE)] +=		\
				(WHAT) >> BITS;				\
			FAKE_SPREAD(WHAT);				\
		}
		#endif

		switch (dither) {

			case DITHER_FLOYD_STEINBERG:

				/*       x-1   x     x+1
				   y                7/16
				   y+1  3/16  5/16  1/16
				*/

				#define FS_HANDLE_PIXEL(INDEX)	HANDLE_PIXEL(INDEX, 4)

				#define FS_DO_RIGHT			SPREAD((e << 3) - e, 4, step_minor)
				#define FS_DO_DOWN_LEFT		SPREAD((e << 1) + e, 4, step_major - step_minor)
				#define FS_DO_DOWN			SPREAD((e << 2) + e, 4, step_major)
				#define FS_DO_DOWN_RIGHT	SPREAD( es         , 4, step_major + step_minor)
				#define FS____RIGHT			FAKE_SPREAD((e << 3) - e)
				#define FS____DOWN_LEFT		FAKE_SPREAD((e << 1) + e)
				#define FS____DOWN			FAKE_SPREAD((e << 2) + e)
				#define FS____DOWN_RIGHT	{ }

				/* handle all rows except the last one */

				for (y = dim_major - 1; y > 0; y--, u += step_major) {

					FS_HANDLE_PIXEL(u);
					FS_DO_RIGHT;
					FS____DOWN_LEFT;
					FS_DO_DOWN;
					FS_DO_DOWN_RIGHT;

					for (x = dim_minor - 2; x > 0; x--) {

						FS_HANDLE_PIXEL(v + step_minor);
						FS_DO_RIGHT;
						FS_DO_DOWN_LEFT;
						FS_DO_DOWN;
						FS_DO_DOWN_RIGHT;
					}

					FS_HANDLE_PIXEL(v + step_minor);
					FS____RIGHT;
					FS_DO_DOWN_LEFT;
					FS_DO_DOWN;
					FS____DOWN_RIGHT;

				}

				/* handle last row */

				for (x = dim_minor - 1; x > 0; x--, u += step_minor) {

					FS_HANDLE_PIXEL(u);
					FS_DO_RIGHT;
				}

				FS_HANDLE_PIXEL(u);

				break;

			case DITHER_SIERRA_2:

				/*       x-2   x-1   x     x+1   x+2
				   y                      4/16  3/16
				   y+1  1/16  2/16  3/16  2/16  1/16
				*/

				#define S2_HANDLE_PIXEL(INDEX)	HANDLE_PIXEL(INDEX, 4)

				#define S2_DO_RIGHT_1		SPREAD( e << 2     , 4, step_minor)
				#define S2_DO_RIGHT_2		SPREAD((e << 1) + e, 4, step_minor + step_minor)
				#define S2_DO_DOWN_LEFT_2	SPREAD(           e, 4, step_major - step_minor - step_minor)
				#define S2_DO_DOWN_LEFT_1	SPREAD( e << 1     , 4, step_major - step_minor)
				#define S2_DO_DOWN			SPREAD((e << 1) + e, 4, step_major)
				#define S2_DO_DOWN_RIGHT_1	SPREAD( e << 1     , 4, step_major + step_minor)
				#define S2_DO_DOWN_RIGHT_2	SPREAD( es         , 4, step_major + step_minor + step_minor)
				#define S2____RIGHT_1		FAKE_SPREAD( e << 2     )
				#define S2____RIGHT_2		FAKE_SPREAD((e << 1) + e)
				#define S2____DOWN_LEFT_2	FAKE_SPREAD(           e)
				#define S2____DOWN_LEFT_1	FAKE_SPREAD( e << 1     )
				#define S2____DOWN			FAKE_SPREAD((e << 1) + e)
				#define S2____DOWN_RIGHT_1	FAKE_SPREAD( e << 1     )
				#define S2____DOWN_RIGHT_2	{ }

				/* handle all rows except the last one */

				for (y = dim_major - 1; y > 0; y--, u += step_major) {

					S2_HANDLE_PIXEL(u);
					S2_DO_RIGHT_1;		S2_DO_RIGHT_2;
					S2____DOWN_LEFT_2;	S2____DOWN_LEFT_1;
					S2_DO_DOWN;
					S2_DO_DOWN_RIGHT_1;	S2_DO_DOWN_RIGHT_2;

					S2_HANDLE_PIXEL(v + step_minor);
					S2_DO_RIGHT_1;		S2_DO_RIGHT_2;
					S2____DOWN_LEFT_2;	S2_DO_DOWN_LEFT_1;
					S2_DO_DOWN;
					S2_DO_DOWN_RIGHT_1;	S2_DO_DOWN_RIGHT_2;

					for (x = dim_minor - 4; x > 0; x--) {

						S2_HANDLE_PIXEL(v + step_minor);
						S2_DO_RIGHT_1;		S2_DO_RIGHT_2;
						S2_DO_DOWN_LEFT_2;	S2_DO_DOWN_LEFT_1;
						S2_DO_DOWN;
						S2_DO_DOWN_RIGHT_1;	S2_DO_DOWN_RIGHT_2;
					}

					S2_HANDLE_PIXEL(v + step_minor);
					S2_DO_RIGHT_1;		S2____RIGHT_2;
					S2_DO_DOWN_LEFT_2;	S2_DO_DOWN_LEFT_1;
					S2_DO_DOWN;
					S2_DO_DOWN_RIGHT_1;	S2____DOWN_RIGHT_2;

					S2_HANDLE_PIXEL(v + step_minor);
					S2____RIGHT_1;		S2____RIGHT_2;
					S2_DO_DOWN_LEFT_2;	S2_DO_DOWN_LEFT_1;
					S2_DO_DOWN;
					S2____DOWN_RIGHT_1;	S2____DOWN_RIGHT_2;
				}

				/* handle last row */

				for (x = dim_minor - 2, v = u; x > 0; x--, u += step_minor) {

					S2_HANDLE_PIXEL(u);
					S2_DO_RIGHT_1;		S2_DO_RIGHT_2;
				}

				S2_HANDLE_PIXEL(v + step_minor);
				S2_DO_RIGHT_1;

				S2_HANDLE_PIXEL(v + step_minor);

				break;

			case DITHER_SIERRA_3:

				/*       x-2   x-1   x     x+1   x+2
				   y                      5/32  3/32
				   y+1  2/32  4/32  5/32  4/32  2/32
				   y+2        2/32  3/32  2/32
				*/

				#define S3_HANDLE_PIXEL(INDEX)	HANDLE_PIXEL(INDEX, 5)

				#define S3_DO_RIGHT_1			SPREAD((e << 2) + e, 5, step_minor)
				#define S3_DO_RIGHT_2			SPREAD((e << 1) + e, 5, step_minor + step_minor)
				#define S3_DO_DOWN_1_LEFT_2		SPREAD( e << 1     , 5, step_major - step_minor - step_minor)
				#define S3_DO_DOWN_1_LEFT_1		SPREAD( e << 2     , 5, step_major - step_minor)
				#define S3_DO_DOWN_1			SPREAD((e << 2) + e, 5, step_major)
				#define S3_DO_DOWN_1_RIGHT_1	SPREAD( e << 2     , 5, step_major + step_minor)
				#define S3_DO_DOWN_1_RIGHT_2	SPREAD( e << 1     , 5, step_major + step_minor + step_minor)
				#define S3_DO_DOWN_2_LEFT_1		SPREAD( e << 1     , 5, step_major + step_major - step_minor)
				#define S3_DO_DOWN_2			SPREAD((e << 1) + e, 5, step_major + step_major)
				#define S3_DO_DOWN_2_RIGHT_1	SPREAD( es         , 5, step_major + step_major + step_minor)

				#define S3____RIGHT_1			FAKE_SPREAD((e << 2) + e)
				#define S3____RIGHT_2			FAKE_SPREAD((e << 1) + e)
				#define S3____DOWN_1_LEFT_2		FAKE_SPREAD( e << 1     )
				#define S3____DOWN_1_LEFT_1		FAKE_SPREAD( e << 2     )
				#define S3____DOWN_1			FAKE_SPREAD((e << 2) + e)
				#define S3____DOWN_1_RIGHT_1	FAKE_SPREAD( e << 2     )
				#define S3____DOWN_1_RIGHT_2	FAKE_SPREAD( e << 1     )
				#define S3____DOWN_2_LEFT_1		FAKE_SPREAD( e << 1     )
				#define S3____DOWN_2			FAKE_SPREAD((e << 1) + e)
				#define S3____DOWN_2_RIGHT_1	{ }

				/* handle all rows except the last two */

				for (y = dim_major - 2; y > 0; y--, u += step_major) {

					S3_HANDLE_PIXEL(u);
					S3_DO_RIGHT_1;			S3_DO_RIGHT_2;
					S3____DOWN_1_LEFT_2;	S3____DOWN_1_LEFT_1;
					S3_DO_DOWN_1;
					S3_DO_DOWN_1_RIGHT_1;	S3_DO_DOWN_1_RIGHT_2;
					S3____DOWN_2_LEFT_1;
					S3_DO_DOWN_2;
					S3_DO_DOWN_2_RIGHT_1;

					S3_HANDLE_PIXEL(v + step_minor);
					S3_DO_RIGHT_1;			S3_DO_RIGHT_2;
					S3____DOWN_1_LEFT_2;	S3_DO_DOWN_1_LEFT_1;
					S3_DO_DOWN_1;
					S3_DO_DOWN_1_RIGHT_1;	S3_DO_DOWN_1_RIGHT_2;
					S3_DO_DOWN_2_LEFT_1;
					S3_DO_DOWN_2;
					S3_DO_DOWN_2_RIGHT_1;

					for (x = dim_minor - 4; x > 0; x--) {

						S3_HANDLE_PIXEL(v + step_minor);
						S3_DO_RIGHT_1;			S3_DO_RIGHT_2;
						S3_DO_DOWN_1_LEFT_2;	S3_DO_DOWN_1_LEFT_1;
						S3_DO_DOWN_1;
						S3_DO_DOWN_1_RIGHT_1;	S3_DO_DOWN_1_RIGHT_2;
						S3_DO_DOWN_2_LEFT_1;
						S3_DO_DOWN_2;
						S3_DO_DOWN_2_RIGHT_1;
					}

					S3_HANDLE_PIXEL(v + step_minor);
					S3_DO_RIGHT_1;			S3____RIGHT_2;
					S3_DO_DOWN_1_LEFT_2;	S3_DO_DOWN_1_LEFT_1;
					S3_DO_DOWN_1;
					S3_DO_DOWN_1_RIGHT_1;	S3____DOWN_1_RIGHT_2;
					S3_DO_DOWN_2_LEFT_1;
					S3_DO_DOWN_2;
					S3_DO_DOWN_2_RIGHT_1;

					S3_HANDLE_PIXEL(v + step_minor);
					S3____RIGHT_1;			S3____RIGHT_2;
					S3_DO_DOWN_1_LEFT_2;	S3_DO_DOWN_1_LEFT_1;
					S3_DO_DOWN_1;
					S3____DOWN_1_RIGHT_1;	S3____DOWN_1_RIGHT_2;
					S3_DO_DOWN_2_LEFT_1;
					S3_DO_DOWN_2;
					S3____DOWN_2_RIGHT_1;
				}

				/* handle second-to-last row */

				S3_HANDLE_PIXEL(u);
				S3_DO_RIGHT_1;			S3_DO_RIGHT_2;
				S3____DOWN_1_LEFT_2;	S3____DOWN_1_LEFT_1;
				S3_DO_DOWN_1;
				S3_DO_DOWN_1_RIGHT_1;	S3_DO_DOWN_1_RIGHT_2;

				S3_HANDLE_PIXEL(v + step_minor);
				S3_DO_RIGHT_1;		S3_DO_RIGHT_2;
				S3____DOWN_1_LEFT_2;	S3_DO_DOWN_1_LEFT_1;
				S3_DO_DOWN_1;
				S3_DO_DOWN_1_RIGHT_1;	S3_DO_DOWN_1_RIGHT_2;

				for (x = dim_minor - 4; x > 0; x--) {

					S3_HANDLE_PIXEL(v + step_minor);
					S3_DO_RIGHT_1;			S3_DO_RIGHT_2;
					S3_DO_DOWN_1_LEFT_2;	S3_DO_DOWN_1_LEFT_1;
					S3_DO_DOWN_1;
					S3_DO_DOWN_1_RIGHT_1;	S3_DO_DOWN_1_RIGHT_2;
				}

				S3_HANDLE_PIXEL(v + step_minor);
				S3_DO_RIGHT_1;			S3____RIGHT_2;
				S3_DO_DOWN_1_LEFT_2;	S3_DO_DOWN_1_LEFT_1;
				S3_DO_DOWN_1;
				S3_DO_DOWN_1_RIGHT_1;

				S3_HANDLE_PIXEL(v + step_minor);
				S3____RIGHT_1;			S3____RIGHT_2;
				S3_DO_DOWN_1_LEFT_2;	S3_DO_DOWN_1_LEFT_1;
				S3_DO_DOWN_1;

				/* handle last row */

				u += step_major;

				for (x = dim_minor - 2, v = u; x > 0; x--, u += step_minor) {

					S3_HANDLE_PIXEL(u);
					S3_DO_RIGHT_1;		S3_DO_RIGHT_2;
				}

				S3_HANDLE_PIXEL(v + step_minor);
				S3_DO_RIGHT_1;

				S3_HANDLE_PIXEL(v + step_minor);

				break;
		}

		/* copy diffused back to greyscale */

		for (v = width * height, u = 0; u < v; u++)
			greyscale[u] = diffused[u];
	}

	#ifdef TGA_TO_STDOUT

	#define fputw(X, F) {				\
	   fputc( (X)        & 0xFF, F);	\
	   fputc( (X) >>  8        , F);	\
	}

	fputc(0, stdout);			/*  0: id length */
	fputc(0, stdout);			/*  1: palette type */
	fputc(3, stdout);			/*  2: image type */
	fputw(0, stdout);			/*  3: offset into palette */
	fputw(0, stdout);			/*  5: colours in palette */
	fputc(0, stdout);			/*  7: palette entry size */
	fputw(0, stdout);			/*  8: left */
	fputw(0, stdout);			/* 10: top */
	fputw(width, stdout);		/* 12: width */
	fputw(height, stdout);		/* 14: height */
	fputc(8, stdout);			/* 16: bpp */
	fputc(0x20, stdout);		/* 17: descriptor bits */

	#if PREVIEW_COLOURS > 2
	fwrite(greyscale, width * height, sizeof(Uint8), stdout);
	#else
	for (v = width * height, c = greyscale; v; v--, c++)
		fputc((*c << 8) - *c, stdout);
	#endif

	#else

	if (preview_window) {

		/* show results in a window */

		converted_image = SDL_CreateRGBSurfaceFrom(greyscale,
			width, height, 8, width, 0, 0, 0, 0);

		if (converted_image == NULL) {
			fprintf(stderr, "ERROR: Could not create %u" "x%u" " surface: %s\n",
				width, height, SDL_GetError());
			ret = RET_SDL_ERR;
			goto done_with_greyscale_image;
		}

		SDL_SetColors(converted_image, preview_palette, 0, PREVIEW_COLOURS);

		SDL_BlitSurface(converted_image, NULL, preview, NULL);
		SDL_UpdateRect(preview, 0, 0, 0, 0);

		if (preview_window & PREVIEW_HISTOGRAM) {
			for (v = width * height, u = 0, c = greyscale; v; v--, c++)
				if (!(*c))
					u++;
			printf("%u sites", u);
			fflush(stdout);
		}

		wait_for_key();

		if (preview_window & PREVIEW_HISTOGRAM)
			puts("");

		SDL_FreeSurface(converted_image);
	}
	else {

		/* print out a list of sites */

		SITE_LOOP_INIT;

		for (y = height, c = greyscale; y; y--) {

			for (x = width; x; x--, c++) {
				if (!(*c))
					OUTPUT_SITE;
				SITE_LOOP_ADVANCE;
			}

			SITE_LOOP_NEWLINE;
		}
	}

	#endif

	done_with_greyscale_image:

	if (diffused != NULL)
		free(diffused);

	free(greyscale);

	done_with_input_image:

	SDL_FreeSurface(image);

	return ret;
}

#endif


/* convert_pbm()

	reads and converts a black-and-white pbm image */

char convert_pbm(void) {

	FILE *f;
	char *line = NULL;
	size_t len = 0;
	ssize_t n;
	int i;
	char ret = RET_OK;

	f = fopen(input_file, "rb");
	if (f == NULL) {
		fprintf(stderr, "File `%s' could not be read.\n", input_file);
		return RET_FILE_ERR;
	}

	i = 0;
	if (fgetc(f) == 'P')
		if (fgetc(f) == '1')
			if (fgetc(f) == 0x0a)
				i = 1;

	if (!i) {
		fprintf(stderr, "PBM header string `P1' not found.\n");
		ret = RET_FILE_ERR;
		goto done_with_pbm_file;
	}

	/* read image dimensions */

	width = 0;
	height = 0;

	while (1) {

		n = getline(&line, &len, f);
		if (n < 0)
			goto eof_found;

		if (line[0] == '#')
			continue;

		if (sscanf(line, "%u %u", &width, &height) == 2)
			break;
	}

	if (width < 1 || height < 1) {
		fprintf(stderr, "No valid image dimensions found in `%s'.\n", input_file);
		ret = RET_FILE_ERR;
		goto done_with_pbm_file;
	}

	/* read and convert pixel data */

	SITE_LOOP_INIT;

	x = y = 0;

	while (y < height) {

		n = getline(&line, &len, f);
		if (n < 0)
			goto eof_found;

		if (line[0] == '#')
			continue;

		for (i = 0; i < n; i++) {

			if (line[i] < ' ')
				break;

			if (line[i] != '0')
				OUTPUT_SITE;

			x++;
			if (x >= width) {
				x = 0;
				y++;
				if (y >= height)
					i = n;
				else
					SITE_LOOP_NEWLINE;
			}
			else
				SITE_LOOP_ADVANCE;
		}
	}

	goto done_with_pbm_file;

	/* eof reached */

	eof_found:

	fprintf(stderr, "Unexpected end-of-file encountered.\n");

	ret = RET_FILE_ERR;

	done_with_pbm_file:

	if (line != NULL)
		free(line);

	fclose(f);

	return ret;
}


/* parse_options()
*/

char parse_options(int argc, char *argv[]) {

	#define CMDLINE_OPTC_HELP		'h'
	#define CMDLINE_OPTS_HELP			"help"
	#define CMDLINE_OPTC_RNDSEED	's'
	#define CMDLINE_OPTS_RNDSEED		"seed"
	#define CMDLINE_OPTC_RANGE		'r'
	#define CMDLINE_OPTS_RANGE			"range"
	#define CMDLINE_OPTC_NOFRAME	'f'
	#define CMDLINE_OPTS_NOFRAME		"no-frame"
	#define CMDLINE_OPTC_OLDMODE	'F'
	#define CMDLINE_OPTS_OLDMODE		"old-frame"
	#define CMDLINE_OPTC_SCALE		'S'
	#define CMDLINE_OPTS_SCALE			"frame-scale"
	#define CMDLINE_OPTC_DIV		'D'
	#define CMDLINE_OPTS_DIV			"frame-steps"
	#ifndef NO_SDL_IMAGE
	#define CMDLINE_OPTC_PREVIEW	'p'
	#define CMDLINE_OPTS_PREVIEW		"preview"
	#define CMDLINE_OPTC_HIST		'H'
	#define CMDLINE_OPTS_HIST			"histogram"
	#define CMDLINE_OPTC_DITHER		'd'
	#define CMDLINE_OPTS_DITHER			"dither"
	#define CMDLINE_OPTC_ROTATE		'o'
	#define CMDLINE_OPTS_ROTATE			"orientation"
	#define CMDLINE_OPTC_MIRROR		'm'
	#define CMDLINE_OPTS_MIRROR			"mirror"
	#define CMDLINE_OPTC_RWEIGHT	'R'
	#define CMDLINE_OPTS_RWEIGHT		"weight-red"
	#define CMDLINE_OPTC_GWEIGHT	'G'
	#define CMDLINE_OPTS_GWEIGHT		"weight-green"
	#define CMDLINE_OPTC_BWEIGHT	'B'
	#define CMDLINE_OPTS_BWEIGHT		"weight-blue"
	#define CMDLINE_OPTC_INPUT		'I'
	#define CMDLINE_OPTS_INPUT			"input"
	#define CMDLINE_OPTC_OUTPUT		'O'
	#define CMDLINE_OPTS_OUTPUT			"output"
	#define CMDLINE_OPTC_GAMMA		'g'
	#define CMDLINE_OPTS_GAMMA			"gamma"
	#endif

	struct option long_opts[] = {
		{ CMDLINE_OPTS_HELP,    no_argument,       NULL, CMDLINE_OPTC_HELP     },
		{ CMDLINE_OPTS_RNDSEED, required_argument, NULL, CMDLINE_OPTC_RNDSEED  },
		{ CMDLINE_OPTS_RANGE,   required_argument, NULL, CMDLINE_OPTC_RANGE    },
		{ CMDLINE_OPTS_NOFRAME, no_argument,       NULL, CMDLINE_OPTC_NOFRAME  },
		{ CMDLINE_OPTS_OLDMODE, no_argument,       NULL, CMDLINE_OPTC_OLDMODE  },
		{ CMDLINE_OPTS_SCALE,   required_argument, NULL, CMDLINE_OPTC_SCALE    },
		{ CMDLINE_OPTS_DIV,     required_argument, NULL, CMDLINE_OPTC_DIV      },
		#ifndef NO_SDL_IMAGE
		{ CMDLINE_OPTS_PREVIEW, no_argument,       NULL, CMDLINE_OPTC_PREVIEW  },
		{ CMDLINE_OPTS_HIST,    optional_argument, NULL, CMDLINE_OPTC_HIST     },
		{ CMDLINE_OPTS_DITHER,  required_argument, NULL, CMDLINE_OPTC_DITHER   },
		{ CMDLINE_OPTS_ROTATE,  required_argument, NULL, CMDLINE_OPTC_ROTATE   },
		{ CMDLINE_OPTS_MIRROR,  no_argument,       NULL, CMDLINE_OPTC_MIRROR   },
		{ CMDLINE_OPTS_RWEIGHT, required_argument, NULL, CMDLINE_OPTC_RWEIGHT  },
		{ CMDLINE_OPTS_GWEIGHT, required_argument, NULL, CMDLINE_OPTC_GWEIGHT  },
		{ CMDLINE_OPTS_BWEIGHT, required_argument, NULL, CMDLINE_OPTC_BWEIGHT  },
		{ CMDLINE_OPTS_INPUT,   required_argument, NULL, CMDLINE_OPTC_INPUT    },
		{ CMDLINE_OPTS_OUTPUT,  required_argument, NULL, CMDLINE_OPTC_OUTPUT   },
		{ CMDLINE_OPTS_GAMMA,   required_argument, NULL, CMDLINE_OPTC_GAMMA    },
		#endif
		{ NULL, 0, NULL, 0 }
	};
	int c;
	int i = 0;
	int j = 0;
	float f = 0.0;
	int index = 0;
	char ret = RET_OK;
	char *optstring;

	optstring = make_opt_string(long_opts);
	if (optstring == NULL) {
		fprintf(stderr, "Out of memory reading command line options!\n");
		return RET_NO_MEM;
	}

	input_file = get_input_file_name(argc, argv);

	while (1) {

		c = getopt_long(argc, argv, optstring, long_opts, &index);

		if (c == -1)
			break;

		if (optarg != NULL)
			i = atoi(optarg);

		switch (c) {

			case CMDLINE_OPTC_HELP:

				printf("bitmap - Copyright (C) 2012-2014 Jussi M. Hukkanen\n");
				printf("Fill a rectangular area on a plane with sites (points),\n"
					   "based on the pixel data of a source image. The output\n"
					   "of this program can be used as input to `triangle'.\n");
				#ifdef NO_SDL_IMAGE
				printf("The image must be in black-and-white ASCII NetPBM (.pbm)\n"
					   "format, where each pixel is represented by a '0' or a '1'.\n"
					   "For each '1' pixel, a site is generated.\n");
				#endif
				printf("\nThis program comes with ABSOLUTELY NO WARRANTY.\n"
				       "This is free software, and you are welcome to\n"
				       "redistribute it under the conditions of the\n"
				       "GNU General Public License, version 3 or later.\n"
				       "To view the License, see `gpl.txt'.\n\n");

				printf("Usage: %s FILE [options]\n", argv[0]);
				printf("   -%c N, --" CMDLINE_OPTS_RNDSEED " N\n", CMDLINE_OPTC_RNDSEED);
				printf("        sets the seed value for the random number\n"
					   "        generator. N can be any non-negative integer.\n");
				printf("   -%c N, --" CMDLINE_OPTS_RANGE " N\n", CMDLINE_OPTC_RANGE);
				printf("        sets the scattering range of the sites.\n"
					   "        N == 0.0: each site is placed in the exact center\n"
					   "        of the corresponding pixel in the source image.\n"
					   "        0.0 < N <= 1.0: each site is placed in a random\n"
					   "        location inside the pixel.\n"
					   "        N > 1.0: the sites are scattered more aggressively\n"
					   "        and the result will appear fuzzier.\n");
				printf("   -%c, --" CMDLINE_OPTS_NOFRAME "\n", CMDLINE_OPTC_NOFRAME);
				printf("        disables framing. Without this option, a number of\n"
					   "        extra sites is added as a frame around the image.\n");
				printf("   -%c N, --" CMDLINE_OPTS_SCALE " N\n", CMDLINE_OPTC_SCALE);
				printf("        sets the scale factor of the frame to N.\n"
					   "        The default is %3.1f.\n", DEFAULT_FRAME_SCALE);
				printf("   -%c N, --" CMDLINE_OPTS_DIV " N\n", CMDLINE_OPTC_DIV);
				printf("        sets the number of segments in the shorter edge\n"
					   "        of the frame to N. The longer edge is divided into\n"
					   "        a number of similarly sized segments. N must be\n"
					   "        a positive integer and defaults to %u.\n", (int)DEFAULT_FRAME_DIV);
				printf("   -%c, --" CMDLINE_OPTS_OLDMODE "\n", CMDLINE_OPTC_OLDMODE);
				printf("        forces the default frame to be drawn using the\n"
					   "        old, buggy method. This option should not be used\n"
					   "        unless backwards compatibility is required. This\n"
					   "        option ignores --" CMDLINE_OPTS_SCALE " and --" CMDLINE_OPTS_DIV ".\n");
				#ifndef NO_SDL_IMAGE
				printf("   -%c, --" CMDLINE_OPTS_PREVIEW "\n", CMDLINE_OPTC_PREVIEW);
				printf("        only displays the converted image; no sites are\n"
					   "        printed to standard output.\n");
				printf("   -%c[=WIDTHxHEIGHT], --" CMDLINE_OPTS_HIST "[=WIDTHxHEIGHT]\n", CMDLINE_OPTC_HIST);
				printf("        (only with --" CMDLINE_OPTS_PREVIEW ") display a histogram in\n"
					   "        standard output. If WIDTH and HEIGHT are omitted,\n"
					   "        the default size is %u" "x%u.\n",
					   DEFAULT_HISTOGRAM_W, DEFAULT_HISTOGRAM_H);
				printf("   -%c BLACK,WHITE, --" CMDLINE_OPTS_OUTPUT " BLACK,WHITE\n", CMDLINE_OPTC_OUTPUT);
				printf("        specify the black and white points of the output,\n"
					   "        in the range from 0 to 255. The intensity scale of\n"
					   "        the output image is scaled between these values.\n"
					   "        For best results, BLACK should be at least 128;\n"
					   "        values below this may result in a very long list\n"
					   "        of sites, especially with a large input image.\n"
					   "        The defaults are %u for BLACK and 255 for WHITE.\n",
					   DEFAULT_OUTPUT_BLACK);
				printf("   -%c BLACK,WHITE, --" CMDLINE_OPTS_INPUT " BLACK,WHITE\n", CMDLINE_OPTC_INPUT);
				printf("        specify the black and white points of the input,\n"
					   "        in the range from 0 to 255. Every pixel with an\n"
					   "        intensity outside this range will be clipped,\n"
					   "        resulting in flat single-colour areas. Note that\n"
					   "        WHITE must be greater than or equal to BLACK.\n"
					   "        The defaults are 0 for BLACK and 255 for WHITE.\n");
				printf("   -%c N, --" CMDLINE_OPTS_GAMMA " N\n", CMDLINE_OPTC_GAMMA);
				printf("        sets input gamma, which makes the converted image\n"
					   "        lighter (if N > 1.0) or darker (if N < 1.0).\n"
					   "        The default is 1.0 and the allowed range is\n"
					   "        %4.2f <= N <= %4.2f.\n",
					   MIN_INPUT_GAMMA, MAX_INPUT_GAMMA);
				printf("   -%c N, --" CMDLINE_OPTS_RWEIGHT " N\n", CMDLINE_OPTC_RWEIGHT);
				printf("   -%c N, --" CMDLINE_OPTS_GWEIGHT " N\n", CMDLINE_OPTC_GWEIGHT);
				printf("   -%c N, --" CMDLINE_OPTS_BWEIGHT " N\n", CMDLINE_OPTC_BWEIGHT);
				printf("        specify the weighting coefficients for the red,\n"
					   "        green and blue channels when converting RGB colour\n"
					   "        images to greyscale. N must be between 0 and 1.\n"
					   "        The default values produce well balanced output\n"
					   "        suitable for most cases. If only one or two of\n"
					   "        these options are given, the omitted values are\n"
					   "        set to zero, so that e.g. using `--" CMDLINE_OPTS_BWEIGHT " 1'\n"
					   "        causes the output to be based only on the blue\n"
					   "        channel of the input.\n");
				#define DITHER_STR_DEFAULT " (default)"
				printf("   -%c N, --" CMDLINE_OPTS_DITHER " N\n", CMDLINE_OPTC_DITHER);
				printf("        selects a dithering algorithm. N must be one of:\n"
					   "        0 - ordered, using a %u" "x%u Bayer matrix"
					   #if DEFAULT_DITHER == DITHER_ORDERED
					   DITHER_STR_DEFAULT
					   #endif
					   ".\n        1 - Floyd-Steinberg error diffusion"
					   #if DEFAULT_DITHER == DITHER_FLOYD_STEINBERG
					   DITHER_STR_DEFAULT
					   #endif
					   ".\n        2 - Two-line Sierra error diffusion"
					   #if DEFAULT_DITHER == DITHER_SIERRA_2
					   DITHER_STR_DEFAULT
					   #endif
					   ".\n        3 - Three-line Sierra error diffusion"
					   #if DEFAULT_DITHER == DITHER_SIERRA_3
					   DITHER_STR_DEFAULT
					   #endif
					   ".\n", BAYER_SIZE, BAYER_SIZE);
				printf("   -%c N, --" CMDLINE_OPTS_ROTATE " N\n", CMDLINE_OPTC_ROTATE);
				printf("   -%c, --" CMDLINE_OPTS_MIRROR "\n", CMDLINE_OPTC_MIRROR);
				printf("        change the orientation of the error diffusion\n"
					   "        algorithms. The input image is rotated N degrees\n"
					   "        clockwise before dithering; N must be 90, 180 or\n"
					   "        270. The --" CMDLINE_OPTS_MIRROR " option reverses the direction\n"
					   "        in which the dithering is performed, i.e. right-\n"
					   "        to-left, bottom-to-top etc. These two options have\n"
					   "        no effect on ordered dithering.\n");
				#endif

				ret = RET_NO_FILE;
				break;

			case CMDLINE_OPTC_RNDSEED:

				if (optarg != NULL)
					if (i >= 0)
						random_seed = i;
				break;

			case CMDLINE_OPTC_RANGE:

				if (optarg != NULL) {

					f = strtod(optarg, NULL);
					if (f > -FLT_EPSILON)
						range = f;
				}
				break;

			case CMDLINE_OPTC_NOFRAME:

				frame = FRAME_OFF;
				break;

			case CMDLINE_OPTC_OLDMODE:

				frame = FRAME_ON | FRAME_OLD;
				break;

			case CMDLINE_OPTC_SCALE:

				if (optarg != NULL) {

					f = strtod(optarg, NULL);
					if (f > FLT_EPSILON)
						frame_scale = f;
				}
				break;

			case CMDLINE_OPTC_DIV:

				if (i > 0)
					frame_div = i;
				break;

			#ifndef NO_SDL_IMAGE

			case CMDLINE_OPTC_DITHER:

				if (optarg != NULL) {
					if (i < 0 || i > DITHER_MAX)
						i = DEFAULT_DITHER;
					dither = i;
				}
				break;

			case CMDLINE_OPTC_ROTATE:

				if (optarg != NULL) {
					if (i >= 0 && (i % 90) == 0) {
						transform &= ORIENTATION_FLIP;
						transform |= i / 90;
					}
				}
				break;

			case CMDLINE_OPTC_PREVIEW:

				preview_window ^= PREVIEW_ON;
				break;

			case CMDLINE_OPTC_HIST:

				if (optarg != NULL) {

					if (sscanf((*optarg == '=') ? (optarg + 1) : optarg,
						"%dx%d", &i, &j) == 2) {

						histogram_w = (i < MIN_HISTOGRAM_SIZE) ? (MIN_HISTOGRAM_SIZE - 1) :
							((i > 256) ? 255 : i - 1);
						histogram_h = (j < MIN_HISTOGRAM_SIZE) ? (MIN_HISTOGRAM_SIZE - 1) :
							((j > 256) ? 255 : j - 1);
					}
				}

				preview_window ^= PREVIEW_HISTOGRAM;
				break;

			case CMDLINE_OPTC_MIRROR:

				transform ^= ORIENTATION_FLIP;
				break;

			case CMDLINE_OPTC_RWEIGHT:
			case CMDLINE_OPTC_GWEIGHT:
			case CMDLINE_OPTC_BWEIGHT:

				if (optarg != NULL) {

					f = strtod(optarg, NULL);
					i = 32768.0 * ((f < 0.0) ? 0.0 : ((f > 1.0) ? 1.0 : f));
					custom_weight_sum = 1;

					switch (c) {
						case CMDLINE_OPTC_RWEIGHT: Rweight = i; break;
						case CMDLINE_OPTC_GWEIGHT: Gweight = i; break;
						default: Bweight = i;
					}
				}
				break;

			case CMDLINE_OPTC_OUTPUT:
			case CMDLINE_OPTC_INPUT:

				if (optarg != NULL) {

					if (sscanf(optarg, "%d,%d", &i, &j) == 2) {

						i = (i < 0) ? 0 : ((i > 255) ? 255 : i);
						j = (j < 0) ? 0 : ((j > 255) ? 255 : j);

						if (c == CMDLINE_OPTC_OUTPUT) {
							output_black = i;
							output_white = j;
						}
						else {
							input_black = i;
							input_white = (j < i) ? i : j;
						}
					}
				}
				break;

			case CMDLINE_OPTC_GAMMA:

				if (optarg != NULL) {

					f = strtod(optarg, NULL);
					if (f > MIN_INPUT_GAMMA - FLT_EPSILON &&
						f < MAX_INPUT_GAMMA + FLT_EPSILON)
						input_gamma = 1.0 / f;
				}
				break;

			#endif
		}
	}

	free(optstring);

	return ret;
}


/* main()
*/

int main(int argc, char *argv[]) {

	int i;
	int ret = RET_OK;

	/* get command line options */

	i = parse_options(argc, argv);
	if (i != 0)
		return i;

	/* open input file for reading */

	if (input_file == NULL) {
		fprintf(stderr, "No valid input file specified; see `%s -h'.\n", argv[0]);
		return RET_NO_FILE;
	}

	RND_SEED(random_seed);

	#ifndef NO_SDL_IMAGE

	/* initialize SDL */

	if (SDL_Init(SDL_INIT_VIDEO) == -1) {
		fprintf(stderr, "ERROR: Could not initialize SDL: %s\n", SDL_GetError());
		return -1;
	}

	atexit(SDL_Quit);

	/* do not display histogram if preview is disabled */

	if (!(preview_window & PREVIEW_ON))
		preview_window = PREVIEW_OFF;

	/* set up RGB component weights and adjustment lookup;
	     if the user has specified no weight coefficients
	     or only zero values, use defaults */

	if (custom_weight_sum)
		custom_weight_sum = Rweight + Gweight + Bweight;

	if (!custom_weight_sum) {
		Rweight = DEFAULT_WEIGHT_RED;
		Gweight = DEFAULT_WEIGHT_GREEN;
		Bweight = DEFAULT_WEIGHT_BLUE;
	}

	create_adjustment_table();

	#endif

	/* read and convert input image */

	ret =
		#ifdef NO_SDL_IMAGE
		convert_pbm();
		#else
		convert_image();
		#endif

	if (ret != RET_OK)
		goto all_done;

	/* add frame if needed */

	#ifndef NO_SDL_IMAGE
	if (preview_window)
		goto all_done;
	#endif

	switch (frame) {

		case FRAME_ON | FRAME_OLD:

			/* old method for framing */

			i = ((fwidth / fheight) * DEFAULT_FRAME_DIV) + 0.5;
			xfstart = (-fwidth * DEFAULT_FRAME_SCALE) / 2.0;
			xfstep = (fwidth * DEFAULT_FRAME_SCALE) / i;
			yf = (fheight * DEFAULT_FRAME_SCALE) / 2.0;

			for (xf = xfstart; i; i--, xf += xfstep) {
				OUTPUT_COORD(xf,  yf);
				OUTPUT_COORD(xf, -yf);
			}

			yfstep = (-fheight * DEFAULT_FRAME_SCALE) / DEFAULT_FRAME_DIV;
			xf = (fwidth * DEFAULT_FRAME_SCALE) / 2.0;
			yf += yfstep;

			for (i = DEFAULT_FRAME_DIV - 2.0; i; i--, yf += yfstep) {
				OUTPUT_COORD( xf, yf);
				OUTPUT_COORD(-xf, yf);
			}

			break;

		case FRAME_ON:

			/* new, more flexible method for framing */

			if (width >= height) {

				i = ((fwidth / fheight) * frame_div) + 0.5;
				xfstart = (-fwidth * frame_scale) / 2.0;
				xfstep = (fwidth * frame_scale) / i;
				yf = (fheight * frame_scale) / 2.0;

				for (xf = xfstart, i++; i; i--, xf += xfstep) {
					OUTPUT_COORD(xf,  yf);
					OUTPUT_COORD(xf, -yf);
				}

				yfstep = (-fheight * frame_scale) / frame_div;
				xf = (fwidth * frame_scale) / 2.0;
				yf += yfstep;

				for (i = frame_div - 1.0; i; i--, yf += yfstep) {
					OUTPUT_COORD( xf, yf);
					OUTPUT_COORD(-xf, yf);
				}
			}
			else {

				xfstart = (-fwidth * frame_scale) / 2.0;
				xfstep = (fwidth * frame_scale) / frame_div;
				yf = (fheight * frame_scale) / 2.0;

				for (xf = xfstart, i = frame_div, i++; i; i--, xf += xfstep) {
					OUTPUT_COORD(xf,  yf);
					OUTPUT_COORD(xf, -yf);
				}

				i = ((fheight / fwidth) * frame_div) + 0.5;
				yfstep = (-fheight * frame_scale) / i;
				xf = (fwidth * frame_scale) / 2.0;
				yf += yfstep;

				for (i--; i; i--, yf += yfstep) {
					OUTPUT_COORD( xf, yf);
					OUTPUT_COORD(-xf, yf);
				}
			}

			break;
	}

	all_done:

	return ret;
}
