/*************************************************************************
 *
 *  Dot -- an implementation of a rubberband algorithm
 *  Copyright (C) 2012-2014 Jussi M. Hukkanen
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *************************************************************************
 *
 *  dot.c: main interface
 *
 ************************************************************************/

#include <stdio.h>
#include "opts.h"

#include "rubber.h"
#ifndef NO_VGA
#include H_GFX
#endif

#ifdef H_PLATFORM
#include H_PLATFORM
#endif

char *window_title = "dot";


#define DEFAULT_LIMIT_MARGIN  (1.0 / 3.0)
#define LIMIT_LEFT    (WINDOW_LEFT   + limit_margin)
#define LIMIT_RIGHT   (WINDOW_RIGHT  - limit_margin)
#define LIMIT_TOP     (WINDOW_TOP    - limit_margin)
#define LIMIT_BOTTOM  (WINDOW_BOTTOM + limit_margin)

site_coord_t limit_margin = DEFAULT_LIMIT_MARGIN;

#define MIN_DISTANCE         0.05
#define DEFAULT_DISTANCE     0.25
#define DEFAULT_DISTANCE_DEV 0.1

site_coord_t distance_param = DEFAULT_DISTANCE;
site_coord_t distance_dev_param = DEFAULT_DISTANCE_DEV;

#ifdef SVGALIB
#define DIRECT_TO_PHYSICAL 1
#endif


//#define IMMEDIATE 1


#define DEFAULT_MAX_SITES   1000

int max_sites = DEFAULT_MAX_SITES;

#ifdef IMMEDIATE
site_t *site;
int sites;
#endif

int32_t random_seed = 0;
char show_vga_caps = 0;
char *binary_name;


#ifndef NO_VGA

#define SITE_COLOUR                  7
#define RUBBERBAND_COLOUR_OUTSIDE   10
#define RUBBERBAND_COLOUR_INSIDE     2

char quiet_mode = 1;

#else

#define WINDOW_RIGHT  (4.0 / 3.0)
#define WINDOW_LEFT   (-WINDOW_RIGHT)
#define WINDOW_TOP    ( 1.0)
#define WINDOW_BOTTOM (-1.0)
#define WINDOW_WIDTH  (WINDOW_RIGHT  - WINDOW_LEFT)
#define WINDOW_HEIGHT (WINDOW_TOP    - WINDOW_BOTTOM)

#define quiet_mode 1

#endif


#ifndef NO_VGA

/* draw_limit_area()

    draws a gray frame around the limit area */

char draw_limit_area(void) {

    #ifndef VALGRIND

    int lx, rx, ty, by;

    ty = PIXEL_Y(LIMIT_TOP);
    by = PIXEL_Y(LIMIT_BOTTOM);
    lx = PIXEL_X(LIMIT_LEFT);
    rx = PIXEL_X(LIMIT_RIGHT);

    if (lx >=  WIDTH) return 0;
    if (rx <       0) return 0;
    if (ty >= HEIGHT) return 0;
    if (by <       0) return 0;

    #define CLAMP_MIN(V, MIN) ((V) >= MIN ? (V) : MIN)
    #define CLAMP_MAX(V, MAX) ((V) <= MAX ? (V) : MAX)

    #ifdef SDL

    rectangleColor(scr_virtual,
        CLAMP_MIN(lx, 0), CLAMP_MIN(ty, 0),
        CLAMP_MAX(rx, WIDTH - 1), CLAMP_MAX(by, HEIGHT - 1),
        sdl_gfx_colour(8));

    #elif defined(SVGALIB)

	gl_setcontext(scr_virtual);

    if (ty >= 0)
      gl_hline(CLAMP_MIN(lx, 0), ty, CLAMP_MAX(rx, WIDTH - 1), 8);
    if (by < HEIGHT)
      gl_hline(CLAMP_MIN(lx, 0), by, CLAMP_MAX(rx, WIDTH - 1), 8);
    if (lx >= 0)
      gl_line(lx, CLAMP_MIN(ty, 0), lx, CLAMP_MAX(by, HEIGHT - 1), 8);
    if (rx < WIDTH)
      gl_line(rx, CLAMP_MIN(ty, 0), rx, CLAMP_MAX(by, HEIGHT - 1), 8);

    #endif

    #endif

    return 0;
}


#define plot_site(S, C) \
    plot_dot_ex(PIXEL_X(S->x), PIXEL_Y(S->y), \
                PLOT_DOT_HOLLOW, C, BRIGHT_COLOUR(C))

#define plot_site_physical(S, C) \
    plot_dot_ex(PIXEL_X(S->x), PIXEL_Y(S->y), \
                PLOT_DOT_HOLLOW | PLOT_DOT_PHYSICAL, \
                C, BRIGHT_COLOUR(C))


#if 0

/* draw_rubberband_segment()

    draws a line between rubberband sites r1 and r2
      into scr_virtual */

static void draw_rubberband_segment(rubberband_t *r1, rubberband_t *r2,
                                    uint8_t colour) {

    if (r1 != NULL && r2 != NULL)
      draw_boundary(r1->site->x, r1->site->y,
                    r2->site->x, r2->site->y,
                    0, colour);
}

#endif


/* draw_rubberband()

    draws the entire rubberband;
    if virtual is nonzero, draw_rubberband() draws
      into scr_virtual;
    limit is the maximum number of rubberband
      segments to draw, in case something goes wrong
      and the functions ends up in an infinite loop */

static void draw_rubberband(uint8_t virtual, int limit) {

    rubberband_t *r;
    rubberband_t *rn;
    uint8_t flags = (virtual != 0) ? 0 : DRAW_BOUNDARY_PHYSICAL;

    for (r = rubberband_start; limit; limit--) {

      rn = r->right;
      if (rn == NULL)
        break;

      draw_boundary(r->site->x, r->site->y, rn->site->x, rn->site->y,
        flags,
        (r->outside >= RUBBERBAND_OUTSIDE) ?
          RUBBERBAND_COLOUR_OUTSIDE : RUBBERBAND_COLOUR_INSIDE
        );

      if (rn == rubberband_start)
        break;
      r = rn;
    }
}

#endif


/* parse_options()

    interprets the given list of command-line
      options */

char parse_options(int argc, char *argv[]) {

    #define CMDLINE_OPTC_HELP       'h'
    #define CMDLINE_OPTS_HELP           "help"
    #define CMDLINE_OPTC_SITES      'n'
    #define CMDLINE_OPTS_SITES          "num"
    #define CMDLINE_OPTC_RNDSEED    's'
    #define CMDLINE_OPTS_RNDSEED        "seed"
    #define CMDLINE_OPTC_DISTANCE   'r'
    #define CMDLINE_OPTS_DISTANCE       "range"
    #define CMDLINE_OPTC_DEVIATE    'd'
    #define CMDLINE_OPTS_DEVIATE        "deviate"
    #define CMDLINE_OPTC_AREA       'a'
    #define CMDLINE_OPTS_AREA           "area"
    #define CMDLINE_OPTC_VGAMODE    'm'
    #define CMDLINE_OPTS_VGAMODE        "mode"
    #define CMDLINE_OPTC_SDLMODE    'w'
    #define CMDLINE_OPTS_SDLMODE        "window"
	#define CMDLINE_OPTC_VGAINFO    'I'
	#define CMDLINE_OPTS_VGAINFO        "info"
    #define CMDLINE_OPTC_PREVIEW    'p'
    #define CMDLINE_OPTS_PREVIEW        "preview"

    struct option long_opts[] = {
      { CMDLINE_OPTS_HELP,     no_argument,       NULL, CMDLINE_OPTC_HELP     },
      { CMDLINE_OPTS_SITES,    required_argument, NULL, CMDLINE_OPTC_SITES    },
      { CMDLINE_OPTS_RNDSEED,  required_argument, NULL, CMDLINE_OPTC_RNDSEED  },
      { CMDLINE_OPTS_DISTANCE, required_argument, NULL, CMDLINE_OPTC_DISTANCE },
      { CMDLINE_OPTS_DEVIATE,  required_argument, NULL, CMDLINE_OPTC_DEVIATE  },
      { CMDLINE_OPTS_AREA,     required_argument, NULL, CMDLINE_OPTC_AREA     },
      #ifndef NO_VGA
      #ifndef VALGRIND
      #ifdef SVGALIB
      { CMDLINE_OPTS_VGAMODE,  required_argument, NULL, CMDLINE_OPTC_VGAMODE  },
	  #endif
	  #ifdef SDL
      { CMDLINE_OPTS_SDLMODE,  required_argument, NULL, CMDLINE_OPTC_SDLMODE  },
	  #endif
      { CMDLINE_OPTS_VGAINFO,  no_argument,       NULL, CMDLINE_OPTC_VGAINFO  },
	  #endif
      { CMDLINE_OPTS_PREVIEW,  no_argument,       NULL, CMDLINE_OPTC_PREVIEW  },
      #endif
      { NULL, 0, NULL, 0 }
    };
    int c;
    int i = 0;
    float f = 0.0;
    char *modestr;
    int index = 0;
    char *optstring;

    optstring = make_opt_string(long_opts);
    if (optstring == NULL) {
      fprintf(stderr, "Out of memory reading command line options!\n");
      return RET_RUBBERBAND_NO_MEM;
    }

    while (1) {

      c = getopt_long(argc, argv, optstring, long_opts, &index);

      if (c == -1)
        break;

      if (optarg != NULL) {
        i = strtol(optarg, NULL, 10);
        f = strtod(optarg, NULL);
      }

      switch (c) {

        case CMDLINE_OPTC_HELP:
          printf("dot - Copyright (C) 2012-2014 Jussi M. Hukkanen\n");
          printf("Fill a rectangular area on a plane with sites\n"
                 "(points) placed at random, with a more or less\n"
                 "even distribution. The output of this program\n"
                 "can be used as input to `triangle'.\n\n");
          printf("This program comes with ABSOLUTELY NO WARRANTY.\n"
                 "This is free software, and you are welcome to\n"
                 "redistribute it under the conditions of the\n"
                 "GNU General Public License, version 3 or later.\n"
                 "To view the License, see `gpl.txt'.\n\n");

          printf("Usage: %s [options]\n", binary_name);
          printf("   -%c N, --" CMDLINE_OPTS_RNDSEED " N\n", CMDLINE_OPTC_RNDSEED);
          printf("        sets the seed value for the random number\n"
                 "        generator. N can be any non-negative integer.\n");
          printf("   -%c N, --" CMDLINE_OPTS_SITES " N\n", CMDLINE_OPTC_SITES);
          printf("        stops after N sites have been created.\n"
                 "        N must be at least 3; the default is %d.\n",
                          DEFAULT_MAX_SITES);
          printf("   -%c N, --" CMDLINE_OPTS_DISTANCE " N\n", CMDLINE_OPTC_DISTANCE);
          printf("        specifies the average distance between sites.\n"
                 "        N must be at least %.3f; the default is %.3f.\n",
                          MIN_DISTANCE, DEFAULT_DISTANCE);
          printf("   -%c N, --" CMDLINE_OPTS_DEVIATE " N\n", CMDLINE_OPTC_DEVIATE);
          printf("        specifies the maximum deviation in distance.\n"
                 "        N must be at least zero; the default is %.3f.\n",
                          DEFAULT_DISTANCE_DEV);
          printf("   -%c N, --" CMDLINE_OPTS_AREA " N\n", CMDLINE_OPTC_AREA);
          printf("        specifies the size of the area to be filled.\n"
                 "        N must be above zero; 1.0 fills the entire\n"
                 "        screen. The default is %.3f.\n",
                          1.0 - DEFAULT_LIMIT_MARGIN);

          #ifndef NO_VGA
          printf("   -%c, --" CMDLINE_OPTS_PREVIEW "\n", CMDLINE_OPTC_PREVIEW);
          printf("        displays the generated sites on the screen\n"
                 "        instead of printing them in standard output.\n");
          #ifndef VALGRIND
          #ifdef SVGALIB
          printf("   -%c MODE, --" CMDLINE_OPTS_VGAMODE " MODE\n", CMDLINE_OPTC_VGAMODE);
          printf("        (only with --" CMDLINE_OPTS_PREVIEW ") selects a screen mode\n"
                 "        from the following:");
          modestr = list_vga_modes();
          if (modestr != NULL)
            printf("        The default is %s.\n", modestr);
          #endif
          #ifdef SDL
          printf("   -%c WIDTHxHEIGHT, --" CMDLINE_OPTS_SDLMODE " WIDTHxHEIGHT\n",
            CMDLINE_OPTC_SDLMODE);
          printf("        (only with --" CMDLINE_OPTS_PREVIEW ") selects a screen size.\n"
                 "        The default is %u" "x%u.\n",
            SCREEN_DEFAULT_W, SCREEN_DEFAULT_H);
          #endif
          printf("   -%c, --" CMDLINE_OPTS_VGAINFO "\n",
            CMDLINE_OPTC_VGAINFO);
          printf("        displays information about the video driver\n");
          printf("        and exits.\n");
          #endif
          #endif

          return RET_RUBBERBAND_FINISHED;

        case CMDLINE_OPTC_SITES:
          if (optarg != NULL)
            if (i >= 3)
              max_sites = i;
          break;

        case CMDLINE_OPTC_RNDSEED:
          if (optarg != NULL)
            if (i >= 0)
              random_seed = i;
          break;

        case CMDLINE_OPTC_DISTANCE:
          if (optarg != NULL)
            if (f > MIN_DISTANCE - FLT_EPSILON)
              distance_param = f;
          break;

        case CMDLINE_OPTC_DEVIATE:
          if (optarg != NULL)
            if (f > -FLT_EPSILON)
              distance_dev_param = f;
          break;

        case CMDLINE_OPTC_AREA:
          if (optarg != NULL)
            if (f > FLT_EPSILON)
              limit_margin = 1.0 - f;
          break;

        #ifndef NO_VGA
        #ifdef SVGALIB
        case CMDLINE_OPTC_VGAMODE:
            if (find_vga_mode(optarg) != 0)
                return RET_RUBBERBAND_PARAM_ERROR;
            break;
        #endif
        #ifdef SDL
        case CMDLINE_OPTC_SDLMODE:
            if (parse_sdl_video_mode(optarg) != 0)
                return RET_RUBBERBAND_PARAM_ERROR;
            break;
        #endif

        case CMDLINE_OPTC_VGAINFO:
            quiet_mode = 0;
            show_vga_caps = 1;
            break;

        case CMDLINE_OPTC_PREVIEW:
            quiet_mode = 0;
            break;
        #endif
      }
    }

    free(optstring);

    return 0;
}


/* main()
*/

int main(int argc, char *argv[]) {

    uint16_t x;
    int i;
    int n;
    char ret;
    site_list_t *sl;
    #ifdef IMMEDIATE
    site_t *sp;
    #endif
    uint8_t b1;

    binary_name = argv[0];

    #ifndef NO_VGA
    #ifdef SVGALIB
	vga_disabledriverreport();
    #endif
    #endif

    i = parse_options(argc, argv);
    if (i != 0)
      return i;

    #ifndef NO_VGA

    /* enter graphics mode */

    if (!quiet_mode) {

        i = init_vga(INIT_VGA_DRAW_FRAME
            #ifndef DIRECT_TO_PHYSICAL
            | INIT_VGA_DOUBLE_BUFFER
            #endif
            );
        if (i != 0)
            return i;

        get_vga_accel_caps();

        if (show_vga_caps)
        	goto return_to_text_mode;

        #if 0
        if (scr_physical->format->palette == NULL) {
            printf("No palette data found\n");
        }
        else {
            for (i = 0; i < scr_physical->format->palette->ncolors; i++)
                printf("#%02x%02x%02x ",
                    scr_physical->format->palette->colors[i].r,
                    scr_physical->format->palette->colors[i].g,
                    scr_physical->format->palette->colors[i].b);
            printf("\n");
        }
        #endif

        draw_limit_area();
    }
    #ifdef SDL
    else {

        /* since we are not calling init_vga(), we must
             call sdl_get_window_size() separately
             in order to properly initialize the
             WINDOW_* values */

        sdl_get_window_size();
    }

    #endif

    #endif

    if (distance_dev_param > distance_param)
      distance_dev_param = distance_param;

    #ifdef IMMEDIATE

    /* in immediate mode, call rubberband()
         and display the results;
       the shape of the rubberband can not
         be drawn, because rubberband()
         destroys it before returning */

    i = rubberband(&site, &sites,
      distance_param, distance_dev_param,
      LIMIT_LEFT, LIMIT_TOP, LIMIT_RIGHT, LIMIT_BOTTOM,
      random_seed);
    if (i != RET_RUBBERBAND_OK)
      goto algorithm_done;

    #ifndef NO_VGA
    if (!quiet_mode) {
      for (i = 0; i < sites; i++) {
          sp = &site[i];
          plot_site(sp, SITE_COLOUR);
      }
      blit_screen(scr_virtual);
      dot_getch();
    }
    #endif

    #else

    /* in non-immediate mode, show how the
         algorithm progresses */

    #if 0
    printf("x1 =%10.6f  y1 =%10.6f\n", LIMIT_LEFT, LIMIT_TOP);
    printf("x2 =%10.6f  y2 =%10.6f\n", LIMIT_RIGHT, LIMIT_BOTTOM);
    #endif

    i = rubberband_init(distance_param, distance_dev_param,
      LIMIT_LEFT, LIMIT_TOP, LIMIT_RIGHT, LIMIT_BOTTOM,
      random_seed);
    if (i != RET_RUBBERBAND_OK) {
      fprintf(stderr, "rubberband_init() returned %d\n", i);
      goto algorithm_done;
    }

    #ifndef NO_VGA
    if (!quiet_mode) {
      for (sl = site_list; sl; sl = sl->next)
         plot_site(sl->site, SITE_COLOUR);
    }
    #endif

    for (i = 2, b1 = quiet_mode ^ 1; i < max_sites; i++) {

        ret = rubberband_add(&n);
        if (ret != RET_RUBBERBAND_OK) {
            if (ret != RET_RUBBERBAND_FINISHED)
                fprintf(stderr, "rubberband_add() returned %d\n", ret);
            break;
        }

        #ifndef NO_VGA

        if (b1) {

            for (sl = site_list; n > 0 && sl != NULL; n--, sl = sl->next)
                plot_site(sl->site, SITE_COLOUR);

            #ifndef DIRECT_TO_PHYSICAL
            copy_screen(scr_virtual, scr_frame);
            #else
            blit_screen(scr_virtual);
            #endif

            draw_rubberband(0, max_sites);

            #ifndef DIRECT_TO_PHYSICAL
            blit_screen(scr_frame);
            #endif

            switch (dot_getch()) {
                #ifdef SDL
                case -1:
                #endif
                case 27:
                    i = max_sites;
                    break;
                case 10: /* Enter */
                    b1 = 0;
                    break;
            }
        }

        #endif
    }

    #ifndef NO_VGA
    if (b1 == 0 && quiet_mode == 0) {
      for (sl = site_list; sl; sl = sl->next)
          plot_site(sl->site, SITE_COLOUR);
      draw_rubberband(1, max_sites);
      blit_screen(scr_virtual);
      dot_getch();
    }
    #endif

    #endif  /* IMMEDIATE */

    #define SITE_TO_STDOUT(S) printf("%10.6f,%10.6f\n", S->x, S->y)

    if (quiet_mode) {
      #ifdef IMMEDIATE
      for (i = 0; i < sites; i++) {
          sp = &site[i];
          SITE_TO_STDOUT(sp);
      }
      #else
      for (sl = site_list; sl; sl = sl->next)
          SITE_TO_STDOUT(sl->site);
      #endif
    }

    algorithm_done:

    /* free memory, return to text mode and exit */

    #ifdef IMMEDIATE
    free(site);
    i = RET_RUBBERBAND_OK;
    #else
    i = rubberband_exit();
    #endif

    return_to_text_mode:

    #ifndef NO_VGA
    if (!quiet_mode)
      exit_vga();

    if (show_vga_caps) {
    	report_vga_accel_caps();
		i = RET_RUBBERBAND_OK;
	}
    #endif

    return i;
}

