/*************************************************************************
 *
 *  An implementation of Fortune's algorithm
 *  Copyright (C) 2012-2014 Jussi M. Hukkanen
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *************************************************************************
 *
 *  dotcell.c: polygon renderer module
 *
 ************************************************************************/

#include "dotcell.h"


#if 0
#define DEBUG_POLYGON 1
//#define DEBUG_POLYGON_VERBOSE 1
#include <stdio.h>
#endif

#if 0
#define REPORT_EXIT_STATUS 1
#define TAG ">>> "
#ifndef DEBUG_POLYGON
#include <stdio.h>
#endif
#endif


#define ALLOCATE_ONCE 1
static int *x1_master = NULL;
static int *x2_master = NULL;
static int *x_tmp_master = NULL;

#define SHRINK_MAX 10

static short last_shrink = -1;  /* set by first call to paint_cell() */

static vertex_t v_inf_1 = { 0.0, 0.0 };
static vertex_t v_inf_2 = { 0.0, 0.0 };

static vertex_t v_top_left     = { 0.0, 0.0 };
static vertex_t v_top_right    = { 0.0, 0.0 };
static vertex_t v_bottom_left  = { 0.0, 0.0 };
static vertex_t v_bottom_right = { 0.0, 0.0 };

static boundary_t boundary_edge[4] = {
    { &v_top_left,     &v_top_right,    NULL },
    { &v_top_right,    &v_bottom_right, NULL },
    { &v_bottom_right, &v_bottom_left,  NULL },
    { &v_bottom_left,  &v_top_left,     NULL }
};

static cellboundary_t b_edge[4] = {
    { &boundary_edge[0], NULL, NULL, NULL },
    { &boundary_edge[1], NULL, NULL, NULL },
    { &boundary_edge[2], NULL, NULL, NULL },
    { &boundary_edge[3], NULL, NULL, NULL }
};


/* if COPY_INFINITE_CELL is #defined, then whenever
     paint_cell() draws an infinite cell, it
     operates on a temporary copy of the cell's
     boundary list;
   without COPY_INFINITE_CELL, paint_cell() turns
     the actual cell temporarily finite and restores
     its shape when done; currently this has some
     serious issues and should be avoided */

#define COPY_INFINITE_CELL 1
static boundary_t boundary_inf[2] = {
    { NULL, NULL, NULL },
    { NULL, NULL, NULL }
};


#if 0
#define REPORT_VISIBLE_SCANLINES 1
int cell_scanlines; /**** for debugging ****/
#endif


static site_t triangle_site = {
    0.0, 0.0
};

static vertex_t triangle_vertex[3] = {
    { 0.0, 0.0 },
    { 0.0, 0.0 },
    { 0.0, 0.0 }
};

static boundary_t triangle_boundary[3] = {
    { &triangle_vertex[0], &triangle_vertex[1], NULL },
    { &triangle_vertex[1], &triangle_vertex[2], NULL },
    { &triangle_vertex[2], &triangle_vertex[0], NULL }
};

static cellboundary_t triangle_cellboundary[3] = {
    #ifndef FORTUNE_MASTER_ARRAYS
    /* boundary,             neighbour, left, right */
    { &triangle_boundary[0], NULL,      NULL, NULL },
    { &triangle_boundary[1], NULL,      NULL, NULL },
    { &triangle_boundary[2], NULL,      NULL, NULL }
    #else
    /* boundary,             neighbour, left, right, i */
    { &triangle_boundary[0], NULL,      NULL, NULL,  0 },
    { &triangle_boundary[1], NULL,      NULL, NULL,  0 },
    { &triangle_boundary[2], NULL,      NULL, NULL,  0 }
    #endif
};

static celldata_t triangle_celldata = {
    &triangle_site,                 /* s */
    &triangle_cellboundary[0],      /* blist */
    3,                              /* sides */
    0,                              /* infinity */
    0,                              /* error */
    NULL,                           /* bloop */
    NULL                            /* extra */
};


#define LESS_THAN(A, B)     ((A) < (B))
#define GREATER_THAN(A, B)  ((A) > (B))


#define VECTOR_ZONE_UP        0x01
#define VECTOR_ZONE_DOWN      0x02
#define VECTOR_ZONE_LEFT      0x04
#define VECTOR_ZONE_RIGHT     0x08

#define VECTOR_OUTSIDE_TOP    0x10
#define VECTOR_OUTSIDE_BOTTOM 0x20
#define VECTOR_OUTSIDE_LEFT   0x40
#define VECTOR_OUTSIDE_RIGHT  0x80

#define VECTOR_OUTSIDE  (VECTOR_OUTSIDE_TOP  | VECTOR_OUTSIDE_BOTTOM | \
                         VECTOR_OUTSIDE_LEFT | VECTOR_OUTSIDE_RIGHT)


/* vector_zone()

    returns one of the VECTOR_ZONE_*
      values depending on the direction
      of the given vector */

static uint8_t vector_zone(vertex_t *v) {

    vertex_coord_t x = v->x;
    vertex_coord_t y = v->y;
    char flags = 0;

    #define VEC_X_NEG 0x01
    #define VEC_Y_NEG 0x02

    if (x < 0.0) {
        x = -x;
        flags |= VEC_X_NEG;
    }

    if (y < 0.0) {
        y = -y;
        flags |= VEC_Y_NEG;
    }

    if (y > x) {
        return (flags & VEC_Y_NEG) ?
            #ifdef FORTUNE_POSITIVE_Y_DOWN
            VECTOR_ZONE_UP : VECTOR_ZONE_DOWN;
            #else
            VECTOR_ZONE_DOWN : VECTOR_ZONE_UP;
            #endif
    }

    return (flags & VEC_X_NEG) ?
        VECTOR_ZONE_LEFT : VECTOR_ZONE_RIGHT;
}


/* make_inf_vertex()

    finds the point where an infinite line
      drawn from vertex v to the direction
      of vector vec hits the edge of the
      visible screen;
    the coordinates are stored in *t;
    *zt receives one of the VECTOR_ZONE_*
      values, depending on which of the
      four edges the line hits first;
    the value returned in *zt has one or two
      of the VECTOR_OUTSIDE_* bits set if the
      line is entirely outside the visible
      screen area */

static void make_inf_vertex(vertex_t *t, vertex_t *v,
                            vertex_t *vec, uint8_t *zt) {

    char zone = vector_zone(vec);
    char out_flag = 0;

    #define OUT_F 0.01

    #define CHECK_IF_OUTSIDE(CONDITION, OUT) {  \
        if (CONDITION)                          \
            out_flag |= OUT;                    \
    }

    #define SKIP_IF_OUTSIDE(OUT) {              \
        if (out_flag & OUT) {                   \
            t->x = v->x + (vec->x * OUT_F);     \
            t->y = v->y + (vec->y * OUT_F);     \
            return;                             \
        }                                       \
    }

    #if 0
    #define CHECK_EDGE(CONDITION, ZONE) {       \
        if (CONDITION) {                        \
            zone = ZONE;                        \
            break;                              \
        }                                       \
    }
    #define SET_ZONE { *zt = zone; }
    #else
    #define CHECK_EDGE(CONDITION, ZONE) { }
    #define SET_ZONE { }
    #endif

    #define DO_COMPONENT(PRIM, SEC, MARGIN) {               \
        t->PRIM = MARGIN;                                   \
        t->SEC = v->SEC;                                    \
        if (vec->PRIM <= -VERTEX_COORD_EPSILON ||           \
            vec->PRIM >=  VERTEX_COORD_EPSILON)             \
            t->SEC +=                                       \
            ((MARGIN - v->PRIM) * (vec->SEC / vec->PRIM));  \
    }

    #ifdef DEBUG_POLYGON
    fprintf(stderr, "zone = ");
    #endif

    CHECK_IF_OUTSIDE(Y_ABOVE(v->y, v_top_left.y),    VECTOR_OUTSIDE_TOP);
    CHECK_IF_OUTSIDE(Y_BELOW(v->y, v_bottom_left.y), VECTOR_OUTSIDE_BOTTOM);
    CHECK_IF_OUTSIDE((v->x < v_top_left.x),          VECTOR_OUTSIDE_LEFT);
    CHECK_IF_OUTSIDE((v->x > v_top_right.x),         VECTOR_OUTSIDE_RIGHT);

    *zt = zone | out_flag;

    switch (zone) {

        case VECTOR_ZONE_UP:
            #ifdef DEBUG_POLYGON
            fprintf(stderr, "up, ");
            #endif
            SKIP_IF_OUTSIDE(VECTOR_OUTSIDE_TOP);
            DO_COMPONENT(y, x, v_top_left.y);
            CHECK_EDGE((t->x < v_top_left.x),  VECTOR_ZONE_LEFT);
            CHECK_EDGE((t->x > v_top_right.x), VECTOR_ZONE_RIGHT);
            SET_ZONE;
            return;

        case VECTOR_ZONE_DOWN:
            #ifdef DEBUG_POLYGON
            fprintf(stderr, "down, ");
            #endif
            SKIP_IF_OUTSIDE(VECTOR_OUTSIDE_BOTTOM);
            DO_COMPONENT(y, x, v_bottom_left.y);
            CHECK_EDGE((t->x < v_top_left.x),  VECTOR_ZONE_LEFT);
            CHECK_EDGE((t->x > v_top_right.x), VECTOR_ZONE_RIGHT);
            SET_ZONE;
            return;

        case VECTOR_ZONE_LEFT:
            #ifdef DEBUG_POLYGON
            fprintf(stderr, "left, ");
            #endif
            SKIP_IF_OUTSIDE(VECTOR_OUTSIDE_LEFT);
            DO_COMPONENT(x, y, v_top_left.x);
            CHECK_EDGE(Y_ABOVE(t->y, v_top_left.y),    VECTOR_ZONE_UP);
            CHECK_EDGE(Y_BELOW(t->y, v_bottom_left.y), VECTOR_ZONE_DOWN);
            SET_ZONE;
            return;

        case VECTOR_ZONE_RIGHT:
            #ifdef DEBUG_POLYGON
            fprintf(stderr, "right, ");
            #endif
            SKIP_IF_OUTSIDE(VECTOR_OUTSIDE_RIGHT);
            DO_COMPONENT(x, y, v_top_right.x);
            CHECK_EDGE(Y_ABOVE(t->y, v_top_left.y),    VECTOR_ZONE_UP);
            CHECK_EDGE(Y_BELOW(t->y, v_bottom_left.y), VECTOR_ZONE_DOWN);
            SET_ZONE;
            return;
    }
}


#ifdef ALLOCATE_ONCE

/* destroy_scanline_arrays()

    releases the memory allocated by
      allocate_scanline_arrays() */

void destroy_scanline_arrays(void) {

    if (x1_master != NULL)
        free(x1_master);
    if (x2_master != NULL)
        free(x2_master);
    if (x_tmp_master != NULL)
        free(x_tmp_master);
}


/* allocate_scanline_arrays()

    reserves memory for the x1_master,
      x2_master and x_tmp_master arrays
      and returns zero on success */

static char allocate_scanline_arrays(void) {

    int i;

    #define SCANLINES (SCREEN_HEIGHT + (SHRINK_MAX << 1))

    atexit(destroy_scanline_arrays);

    x1_master = (int *)malloc(SCANLINES * sizeof(int));
    if (x1_master == NULL)
        return 1;

    x2_master = (int *)malloc(SCANLINES * sizeof(int));
    if (x2_master == NULL)
        return 1;

    x_tmp_master = (int *)malloc(SCANLINES * sizeof(int));
    if (x_tmp_master == NULL)
        return 1;

    for (i = SCANLINES; i; ) {
        i--;
        x1_master[i] = -1;
        x2_master[i] = -2;
        x_tmp_master[i] = -1;
    }

    return 0;
}

#endif


/* paint_triangle()

    draws the Delaunay triangle ta using
      paint_cell();
    the three corner sites of the triangle
      are assumed to be in clockwise order,
      however the first site doesn't have
      to be on top if flags contains
      DRAW_CELL_ROTATED */

char paint_triangle(triangle_t *ta, uint8_t flags, uint8_t shrink,
                    short fill, short stroke, uint32_t *count, uint8_t *clip) {

    /* if paint_triangle() is being called
         for the first time, set up the
         left and right pointers in
         triangle_cellboundary */

    if (triangle_cellboundary[0].left == NULL) {
        triangle_cellboundary[0].left  = &triangle_cellboundary[2];
        triangle_cellboundary[0].right = &triangle_cellboundary[1];
        triangle_cellboundary[1].left  = &triangle_cellboundary[0];
        triangle_cellboundary[1].right = &triangle_cellboundary[2];
        triangle_cellboundary[2].left  = &triangle_cellboundary[1];
        triangle_cellboundary[2].right = &triangle_cellboundary[0];
    }

    /* copy site coordinates to vertex coordinates */

    triangle_vertex[0].x = ta->site[0]->x;
    triangle_vertex[0].y = ta->site[0]->y;
    triangle_vertex[1].x = ta->site[1]->x;
    triangle_vertex[1].y = ta->site[1]->y;
    triangle_vertex[2].x = ta->site[2]->x;
    triangle_vertex[2].y = ta->site[2]->y;

    /* draw it and return */

    return paint_cell(&triangle_celldata,
        flags, shrink, fill, stroke, count, clip);
}


/* paint_cell()

    fills the area of cell ca with the given
      colour and draws a stroke around the
      cell with another colour;
    if fill is negative, only the stroke is
      drawn; if stroke is negative, the fill
      colour is used for the stroke;
    if shrink is greater than zero, the cell
      area is shrunk by that many pixels;
    paint_cell() uses hardware acceleration
      if possible;
    the first boundary in the list of the
      cell is assumed to be the top boundary
      unless the DRAW_CELL_ROTATED bit is
      given in flags;
    if flags does not contain the
      DRAW_CELL_PHYSICAL bit, the cell is
      drawn into scr_virtual;
    if count is not NULL, the number of pixels
      in the visible area of the cell is
      stored in *count; however, if only the
      stroke is drawn, *count is unchanged;
    if clip is not NULL, a combination of
      zero or more CLIP_CELL_* bits is stored
      in *clip;
    the return value is zero if the operation
      was successful, positive if there is not
      enough memory to hold an x coordinate
      pair (start and end) for each scanline,
      or negative if the cell is entirely
      outside the screen or invisible for some
      other reason */

char paint_cell(celldata_t *ca, uint8_t flags, uint8_t shrink,
                short fill, short stroke, uint32_t *count, uint8_t *clip) {

    char ret = 0;
    int scan, scanmax;
    int x, y;
    int xu, xd, yd;
    int ytop, ybottom;
    int x1, y1;
    int x2, y2;
    int *x1_list = NULL;
    int *x2_list = NULL;
    int *x_tmp = NULL;
    int *x1_p;
    int *x2_p;
    uint32_t total_pixels;
    cellboundary_t *blist = NULL;
    cellboundary_t *b;
    cellboundary_t *b_inf_1 = NULL;
    cellboundary_t *b_inf_2 = NULL;
    cellboundary_t *b_start_left = NULL;
    cellboundary_t *b_start_right = NULL;
    uint8_t vz1, vz2;
    uint8_t vz1o, vz2o;
    vertex_t *v1;
    vertex_t *v2;
    vertex_t *vtop;
    vertex_t *vbottom;
    vertex_coord_t v_tmpx, v_tmpy;
    float r;
    uint8_t partial = 0;

    #ifdef SDL

    /**** TODO: USE_SDL_POLYGON is currently disabled, because
            SDL polygons sometimes fill the entire screen ****/

    //#define USE_SDL_POLYGON     /* call polygonColor() when applicable */

    SDL_Surface *where =
        (flags & DRAW_CELL_PHYSICAL) ? scr_physical : scr_virtual;
    #ifndef VALGRIND
    Sint16 *sdl_polygon_x;
    Sint16 *sdl_polygon_y;
    Uint32 fill_rgba = 0;
    Uint32 stroke_rgba = 0;
    #define draw_scanline(X1, Y, X2, C) hlineColor(where, X1, X2, Y, C##_rgba)
    #else
    #define draw_scanline(X1, Y, X2, C) { }
    #endif

    #elif defined(SVGALIB)

    #define draw_scanline(X1, Y, X2, C) gl_hline(X1, Y, X2, C)

    #endif

    #define MASTER_TOP      (-SHRINK_MAX)
    #define MASTER_BOTTOM   (SCREEN_HEIGHT + SHRINK_MAX)

    #if 0
    #define RET_CELL_ERROR      (-10)
    #else
    #define RET_CELL_ERROR      RET_FORTUNE_OK
    #endif

    #ifdef DEBUG_POLYGON
    #if 0
    #define FORMAT_VERTEX(V) \
        fprintf(stderr, "%8.5f, %8.5f", V->x, V->y)
    #else
    #define FORMAT_VERTEX(V) {                  \
        int vertex_pixel_x = PIXEL_X(V->x);     \
        int vertex_pixel_y = PIXEL_Y(V->y);     \
        fprintf(stderr, "%d,%d",                \
            vertex_pixel_x, vertex_pixel_y);    \
    }
    #endif
    fprintf(stderr, "<site %4ld,%4ld>  %d sides",
        PIXEL_X(ca->s->x), PIXEL_Y(ca->s->y), ca->sides);
    if (ca->infinity)
        fprintf(stderr, ", inf");
    fprintf(stderr, "\n");
    #endif

    #define EXIT_WITH_CODE(CODE) {          \
        ret = CODE; goto paint_cell_done;   \
    }

    #ifdef REPORT_VISIBLE_SCANLINES
    cell_scanlines = 0;
    #endif

    if (fill < 0)
        if (stroke < 0) {
            #ifdef REPORT_EXIT_STATUS
            fprintf(stderr, TAG "no fill or stroke colour specified\n");
            #endif
            return 0;
        }

    if (ca->error != 0) {
        #ifdef REPORT_EXIT_STATUS
        fprintf(stderr, TAG);
        #endif
        #if defined(REPORT_EXIT_STATUS) || defined(DEBUG_POLYGON)
        fprintf(stderr, "cell structure marked as broken\n");
        #endif
        return RET_CELL_ERROR;
    }

    if (ca->sides < 2) {
        #ifdef REPORT_EXIT_STATUS
        fprintf(stderr, TAG "only %d sides in cell\n", ca->sides);
        #endif
        return -1;
    }

    if (stroke == fill)
        stroke = -1;

    #ifdef SDL
    #ifndef VALGRIND
    if (fill >= 0)
        fill_rgba = sdl_gfx_colour(fill);
    if (stroke >= 0)
        stroke_rgba = sdl_gfx_colour(stroke);
    #endif
    #endif

    #ifdef ALLOCATE_ONCE
    if (x1_master == NULL) {
        if (allocate_scanline_arrays() != 0) {
            #ifdef REPORT_EXIT_STATUS
            fprintf(stderr, TAG "allocate_scanline_arrays() returned nonzero\n");
            #endif
            return 1;
        }
    }

    if (shrink > SHRINK_MAX)
        shrink = SHRINK_MAX;
    #endif

    /* update vertices at corners of screen if necessary */

    if (shrink != last_shrink) {

        #ifdef DEBUG_POLYGON_VERBOSE
        fprintf(stderr, "update vertices at screen corners");
        #endif

        r = (WINDOW_WIDTH * (shrink + 2.0)) / SCREEN_WIDTH;
        v_top_left.x    = v_bottom_left.x  = WINDOW_LEFT   - r;
        v_top_right.x   = v_bottom_right.x = WINDOW_RIGHT  + r;

        r = (WINDOW_HEIGHT * (shrink + 2.0)) / SCREEN_HEIGHT;
        v_top_left.y    = v_top_right.y    = WINDOW_TOP    + r;
        v_bottom_left.y = v_bottom_right.y = WINDOW_BOTTOM - r;

        #ifdef DEBUG_POLYGON_VERBOSE
        fprintf(stderr, "...done\n");
        #endif

        last_shrink = shrink;
    }

    /* find top and bottom y coordinates */

    if (ca->infinity) {

        /* cell is infinite */

        x1 = ca->sides;

        #ifdef COPY_INFINITE_CELL

        /* make a temporary copy of the boundary list */

        #ifdef DEBUG_POLYGON_VERBOSE
        fprintf(stderr, "create temporary blist array\n");
        #endif

        blist = (cellboundary_t *)malloc(x1 * sizeof(cellboundary_t));
        if (blist == NULL) {
            #ifdef REPORT_EXIT_STATUS
            fprintf(stderr, TAG "insufficient memory for temporary blist array\n");
            #endif
            return 1;
        }

        for (x = 1; x < x1 - 1; x++) {
            blist[x].left  = &blist[x - 1];
            blist[x].right = &blist[x + 1];
        }
        blist[0].left  = &blist[x];
        blist[0].right = &blist[1];
        blist[x].left  = &blist[x - 1];
        blist[x].right = &blist[0];

        for (b = ca->blist, x = y = 0; x < x1; b = b->right, x++) {

            if (b->boundary->v2 == NULL && y < 2) {
                blist[x].boundary = &boundary_inf[y];
                boundary_inf[y].v1 = b->boundary->v1;
                boundary_inf[y].v2 = NULL;
                boundary_inf[y].vector = b->boundary->vector;
                y++;
            }
            else
                blist[x].boundary = b->boundary;

            blist[x].neighbour = b->neighbour;
        }

        #else

        blist = ca->blist;

        #endif

        #ifdef DEBUG_POLYGON_VERBOSE
        for (b = blist, x = ca->sides; x; x--, b = b->left) {
            fprintf(stderr, "%8.5f, %8.5f -> ",
                b->boundary->v1->x, b->boundary->v1->y);
            if (b->boundary->v2 != NULL)
                fprintf(stderr, "%8.5f, %8.5f",
                    b->boundary->v2->x, b->boundary->v2->y);
            else {
                if (b->boundary->vector->x > -VERTEX_COORD_EPSILON &&
                    b->boundary->vector->x <  VERTEX_COORD_EPSILON)
                    fprintf(stderr, " 0.0");
                else
                    fprintf(stderr, "%cInf",
                        (b->boundary->vector->x < 0.0) ? '-' : '+');
                fprintf(stderr, ", ");
                if (b->boundary->vector->y > -VERTEX_COORD_EPSILON &&
                    b->boundary->vector->y <  VERTEX_COORD_EPSILON)
                    fprintf(stderr, " 0.0");
                else
                    fprintf(stderr, "%cInf",
                        (b->boundary->vector->y < 0.0) ? '-' : '+');
            }
            fprintf(stderr, "\n");
        }
        #endif

        /* point b_inf_1 and b_inf_2 to the 'left' and 'right'
             infinite boundary, respectively */

        if (x1 > 2) {

            for (b = blist, x = 0; b->boundary->v2 != NULL; x++) {
                if (x >= ca->sides) { /**** eeek! ****/
                    if (count != NULL)
                        *count = 0;
                    #ifdef REPORT_EXIT_STATUS
                    fprintf(stderr, TAG "no infinite boundaries found in infinite cell\n");
                    #endif
                    EXIT_WITH_CODE(-1);
                }
                b = b->left;
            }

            if (b->left->boundary->v2 == NULL) {
                b_inf_1 = b->left;
                b_inf_2 = b;
            }
            else {
                b_inf_1 = b;
                b_inf_2 = b->right;
            }
        }
        else {

            /* special case where a cell consists of
                 only two infinite boundaries */

            for (b_inf_1 = blist; b_inf_1->boundary->v2 != NULL; )
                b_inf_1 = b_inf_1->left;

            for (b_inf_2 = b_inf_1->left; b_inf_2->boundary->v2 != NULL; )
                b_inf_2 = b_inf_2->left;

            v_tmpx = b_inf_1->boundary->v1->x;
            v_tmpy = b_inf_1->boundary->v1->y;
            v_inf_1.x = v_tmpx + b_inf_1->boundary->vector->x;
            v_inf_1.y = v_tmpy + b_inf_1->boundary->vector->y;
            v_inf_2.x = v_tmpx + b_inf_2->boundary->vector->x;
            v_inf_2.y = v_tmpy + b_inf_2->boundary->vector->y;

            if (LEFT_SIDE(v_tmpx, v_tmpy, v_inf_1.x, v_inf_1.y,
                v_inf_2.x, v_inf_2.y)) {
                b = b_inf_1;
                b_inf_1 = b_inf_2;
                b_inf_2 = b;
            }
        }

        /* place v_inf_1 and v_inf_2 on the screen edges */

        if (b_inf_1->boundary->vector == NULL ||
            b_inf_2->boundary->vector == NULL) {

            #if 0
            for (b = blist, x = ca->sides; x; x--, b = b->left) {
                fprintf(stderr, "%8.5f, %8.5f -> ",
                    b->boundary->v1->x, b->boundary->v1->y);
                if (b->boundary->v2 != NULL)
                    fprintf(stderr, "%8.5f, %8.5f",
                        b->boundary->v2->x, b->boundary->v2->y);
                else {
                    if (b->boundary->vector->x > -VERTEX_COORD_EPSILON &&
                        b->boundary->vector->x <  VERTEX_COORD_EPSILON)
                        fprintf(stderr, " 0.0");
                    else
                        fprintf(stderr, "%cInf",
                            (b->boundary->vector->x < 0.0) ? '-' : '+');
                    fprintf(stderr, ", ");
                    if (b->boundary->vector->y > -VERTEX_COORD_EPSILON &&
                        b->boundary->vector->y <  VERTEX_COORD_EPSILON)
                        fprintf(stderr, " 0.0");
                    else
                        fprintf(stderr, "%cInf",
                            (b->boundary->vector->y < 0.0) ? '-' : '+');
                }
                fprintf(stderr, "\n");
            }
            #endif

            #ifdef REPORT_EXIT_STATUS
            if (b_inf_1->boundary->vector == NULL)
                fprintf(stderr, TAG "error creating b_inf_1\n");
            if (b_inf_2->boundary->vector == NULL)
                fprintf(stderr, TAG "error creating b_inf_2\n");
            #endif

            EXIT_WITH_CODE(9);
        }

        make_inf_vertex(&v_inf_1,
            b_inf_1->boundary->v1,
            b_inf_1->boundary->vector, &vz1);

        #ifdef DEBUG_POLYGON
        #ifdef DEBUG_POLYGON_VERBOSE
        FORMAT_VERTEX(b_inf_1->boundary->v1);
        fprintf(stderr, " + %8.5f, %8.5f",
            b_inf_1->boundary->vector->x, b_inf_1->boundary->vector->y);
        fprintf(stderr, " [b_inf_1]\n");
        #endif
        v1 = &v_inf_1;
        FORMAT_VERTEX(v1);
        fprintf(stderr, " [v_inf_1]  vz1 = 0x%02X\n", vz1);
        #endif

        make_inf_vertex(&v_inf_2,
            b_inf_2->boundary->v1,
            b_inf_2->boundary->vector, &vz2);

        #ifdef DEBUG_POLYGON
        #ifdef DEBUG_POLYGON_VERBOSE
        FORMAT_VERTEX(b_inf_2->boundary->v1);
        fprintf(stderr, " + %8.5f, %8.5f",
            b_inf_2->boundary->vector->x, b_inf_2->boundary->vector->y);
        fprintf(stderr, " [b_inf_2]\n");
        #endif
        v2 = &v_inf_2;
        FORMAT_VERTEX(v2);
        fprintf(stderr, " [v_inf_2]  vz2 = 0x%02X\n", vz2);
        #endif

        b_inf_1->boundary->v2 = &v_inf_1;
        b_inf_2->boundary->v2 = &v_inf_2;
        b_inf_1->boundary->vector = NULL;
        b_inf_2->boundary->vector = NULL;

        /* insert extra segments into boundary list;
             this is not necessary if the line from
             v_inf_1 to v_inf_2 is entirely outside
             the screen */

        #define CONNECT_FIRST_TO_EDGE {             \
            b_inf_1->right = &b_edge[0];            \
            boundary_edge[0].v1 = &v_inf_1;         \
            b_edge[0].left = b_inf_1;               \
            v1 = &v_inf_1;                          \
            x = 0;                                  \
        }
        #define CONNECT_EDGES(V) {                  \
            if (LEFT_SIDE(v1->x, v1->y,             \
                v_inf_2.x, v_inf_2.y, V.x, V.y)) {  \
                b_edge[x].right = &b_edge[x + 1];   \
                boundary_edge[x].v2 = &V;           \
                x++;                                \
                boundary_edge[x].v1 = &V;           \
                v1 = &V;                            \
                b_edge[x].left = &b_edge[x - 1];    \
            }                                       \
        }
        #define CONNECT_EDGE_TO_LAST {              \
            b_edge[x].right = b_inf_2;              \
            boundary_edge[x].v2 = &v_inf_2;         \
            b_inf_2->left = &b_edge[x];             \
        }

        CONNECT_FIRST_TO_EDGE;

        vz1o = vz1 & VECTOR_OUTSIDE;
        vz2o = vz2 & VECTOR_OUTSIDE;

        if (vz1o > 0 && vz2o > 0) {
            if (v_inf_1.x < v_top_left.x)
                if (v_inf_2.x < v_top_left.x)
                    goto boundaries_connected;
            if (v_inf_1.x > v_top_right.x)
                if (v_inf_2.x > v_top_right.x)
                    goto boundaries_connected;
            if (Y_ABOVE(v_inf_1.y, v_top_left.y))
                if (Y_ABOVE(v_inf_2.y, v_top_left.y))
                    goto boundaries_connected;
            if (Y_BELOW(v_inf_1.y, v_bottom_left.y))
                if (Y_BELOW(v_inf_2.y, v_bottom_left.y))
                    goto boundaries_connected;
        }

        vz1 &= ~VECTOR_OUTSIDE;
        vz2 &= ~VECTOR_OUTSIDE;

        switch (vz1) {

            case VECTOR_ZONE_UP:
                if (vz2 == VECTOR_ZONE_UP)    break;
                CONNECT_EDGES(v_top_right);
                if (vz2 == VECTOR_ZONE_RIGHT) break;
                CONNECT_EDGES(v_bottom_right);
                if (vz2 == VECTOR_ZONE_DOWN)  break;
                CONNECT_EDGES(v_bottom_left);
                break;

            case VECTOR_ZONE_RIGHT:
                if (vz2 == VECTOR_ZONE_RIGHT) break;
                CONNECT_EDGES(v_bottom_right);
                if (vz2 == VECTOR_ZONE_DOWN)  break;
                CONNECT_EDGES(v_bottom_left);
                if (vz2 == VECTOR_ZONE_LEFT)  break;
                CONNECT_EDGES(v_top_left);
                break;

            case VECTOR_ZONE_DOWN:
                if (vz2 == VECTOR_ZONE_DOWN)  break;
                CONNECT_EDGES(v_bottom_left);
                if (vz2 == VECTOR_ZONE_LEFT)  break;
                CONNECT_EDGES(v_top_left);
                if (vz2 == VECTOR_ZONE_UP)    break;
                CONNECT_EDGES(v_top_right);
                break;

            case VECTOR_ZONE_LEFT:
                if (vz2 == VECTOR_ZONE_LEFT)  break;
                CONNECT_EDGES(v_top_left);
                if (vz2 == VECTOR_ZONE_UP)    break;
                CONNECT_EDGES(v_top_right);
                if (vz2 == VECTOR_ZONE_RIGHT) break;
                CONNECT_EDGES(v_bottom_right);
                break;
        }

        boundaries_connected:

        CONNECT_EDGE_TO_LAST;

        #define SETUP_VERTEX_LOOP {             \
            b = blist;                          \
            v1 = b->boundary->v1;               \
            v2 = b->boundary->v2;               \
            vtop = v1;                          \
            vbottom = vtop;                     \
            CHECK_BOTH(v2);                     \
        }
        #define CHECK_Y_TOP(V) {                \
            if (Y_ABOVE(V->y, vtop->y))         \
                vtop = V;                       \
        }
        #define CHECK_Y_BOTTOM(V) {             \
            if (Y_BELOW(V->y, vbottom->y))      \
                vbottom = V;                    \
        }
        #define CHECK_BOTH(V) {                 \
            CHECK_Y_TOP(V);                     \
            CHECK_Y_BOTTOM(V);                  \
        }

        /* there is no need to look for NULL boundaries
             as long as the error flag of the cell is
             checked at the start of paint_cell() */

        #ifdef DEBUG_POLYGON
        #define CHECK_NULL_BOUNDARY {           \
            if (b == NULL) {                    \
                fprintf(stderr,                 \
                    "*** WARNING: 'NULL' "      \
                    "boundary encountered; "    \
                    "aborting\n");              \
                EXIT_WITH_CODE(RET_CELL_ERROR); \
            }                                   \
        }
        #else
        #define CHECK_NULL_BOUNDARY { }
        #endif

        #define CHECK_VERTEX_Y(V, METHOD) {     \
            if (b->boundary->V != v1 &&         \
                b->boundary->V != v2) {         \
                METHOD(b->boundary->V);         \
            }                                   \
        }
        #define CHECK_VERTICES(METHOD) {        \
            CHECK_NULL_BOUNDARY;                \
            CHECK_VERTEX_Y(v1, METHOD);         \
            CHECK_VERTEX_Y(v2, METHOD);         \
            v1 = b->boundary->v1;               \
            v2 = b->boundary->v2;               \
        }

        SETUP_VERTEX_LOOP;

        for (b = b->left; b != blist; b = b->left)
            CHECK_VERTICES(CHECK_BOTH);

        /* determine which boundaries to start
             drawing left and right side from;
           with infinite cells, this always has
             to be done the hard way, i.e. it's
             not safe to assume that the first
             boundary is the 'top' one, because
             the working copy of the list is not
             the same as the source */

        #ifdef REPORT_EXIT_STATUS
        #define GET_START_BOUNDARY_ERR(SIDE) {  \
            fprintf(stderr, TAG                 \
                "top vertex not found in ");    \
            if (ca->infinity)                   \
                fprintf(stderr, "in");          \
            fprintf(stderr, "finite cell when " \
                "trying to determine " #SIDE    \
                "\n");                          \
        }
        #else
        #define GET_START_BOUNDARY_ERR(SIDE) { }
        #endif

        #define GET_START_BOUNDARY(WHICH, D) {  \
            for (b = blist, WHICH = NULL,       \
                x = 0; /*x <= ca->sides*/; x++) {   \
                if (b->boundary->v1 == vtop ||  \
                    b->boundary->v2 == vtop)    \
                    WHICH = b;                  \
                else if (WHICH != NULL)         \
                    break;                      \
                b = b->D;                       \
                if (b == blist)                 \
                    break;                      \
            }                                   \
            /*if (x > ca->sides)*/              \
            if (WHICH == NULL) /**** eeek! ****/ \
                GET_START_BOUNDARY_ERR(WHICH);  \
        }

        GET_START_BOUNDARY(b_start_left,  left);
        GET_START_BOUNDARY(b_start_right, right);
    }
    else {

        /* cell is finite */

        blist = ca->blist;

        SETUP_VERTEX_LOOP;

        if (flags & DRAW_CELL_ROTATED) {

            /* the first boundary can be anywhere */

            for (b = b->left, x = ca->sides - 1; x; x--, b = b->left)
                CHECK_VERTICES(CHECK_BOTH);

            GET_START_BOUNDARY(b_start_left,  left);
            GET_START_BOUNDARY(b_start_right, right);
        }
        else {

            /* the first boundary must be on 'top' */

            for (b = b->left, x = ca->sides - 1; x; x--, b = b->left)
                CHECK_VERTICES(CHECK_Y_BOTTOM);

            b_start_right = blist;
            b_start_left = blist->left;
        }
    }

    #if 1   /**** debug start ****/
    #ifdef DEBUG_POLYGON

    FORMAT_VERTEX(vtop);
    fprintf(stderr, " [vtop]\n");
    FORMAT_VERTEX(vbottom);
    fprintf(stderr, " [vbottom]\n");

    #define PRINT_VERTEX(P, S) {                        \
        FORMAT_VERTEX(b->boundary->v1);                 \
        fprintf(stderr, "..");                          \
        FORMAT_VERTEX(b->boundary->v2);                 \
    }
    #define PRINT_VERTEX_SEPARATOR fprintf(stderr, "  ")
    #define FINISH_DEBUG_VERTEX_LOOP fprintf(stderr, ".\n")

    #define PRINT_VERTEX_LOOP(START, DIR, PRIM, SEC) {  \
        for (b = START, v2 = NULL; ; ) {                \
            PRINT_VERTEX(PRIM, SEC);                    \
            b = b->DIR;                                 \
            if (b == START) break;                      \
            PRINT_VERTEX_SEPARATOR;                     \
        }                                               \
        FINISH_DEBUG_VERTEX_LOOP;                       \
    }

    fprintf(stderr, "cw:  "); PRINT_VERTEX_LOOP(blist, right, v2, v1);

    if (b_start_left != NULL) {
        FORMAT_VERTEX(b_start_left->boundary->v1);
        fprintf(stderr, " -> ");
        FORMAT_VERTEX(b_start_left->boundary->v2);
        fprintf(stderr, " [b_start_left]\n");
    }
    if (b_start_right != NULL) {
        FORMAT_VERTEX(b_start_right->boundary->v1);
        fprintf(stderr, " -> ");
        FORMAT_VERTEX(b_start_right->boundary->v2);
        fprintf(stderr, " [b_start_right]\n");
    }

    #endif
    #endif  /**** debug end ****/

    if (b_start_left == NULL || b_start_right == NULL) {
        if (count != NULL)
            *count = 0;
        EXIT_WITH_CODE(-1);
    }

    ytop = PIXEL_Y(vtop->y);
    ybottom = PIXEL_Y(vbottom->y);

    if (ytop >= SCREEN_HEIGHT - shrink || ybottom < shrink ||
        ybottom < ytop) {
        #ifdef DEBUG_POLYGON_VERBOSE
        fprintf(stderr, "entire polygon is outside screen\n");
        #endif
        #ifdef REPORT_EXIT_STATUS
        fprintf(stderr, TAG "entire polygon is outside screen\n");
        #endif
        if (count != NULL)
            *count = 0;
        EXIT_WITH_CODE(-1);
    }

    #ifdef SDL
    #ifdef USE_SDL_POLYGON
    #ifndef VALGRIND

    /* if the calling application doesn't need the pixel count
         and clipping information and no shrink width has been
         requested, we can draw the polygon using SDL routines */

    if (count == NULL && clip == NULL && shrink == 0) {

        #ifdef DEBUG_POLYGON
        fprintf(stderr, "draw polygon using SDL\n");
        #endif

        for (b = blist->left, xu = 1; b != blist; b = b->left)
            xu++;

        #ifdef DEBUG_POLYGON_VERBOSE
        fprintf(stderr, "%d vertices\n", xu);
        fprintf(stderr, "allocate temporary sdl_polygon_x array\n");
        #endif

        sdl_polygon_x = alloca(xu * sizeof(Sint16));

        if (sdl_polygon_x == NULL) {

            #ifdef DEBUG_POLYGON_VERBOSE
            fprintf(stderr, "...failed!\n");
            #endif
        }
        else {

            #ifdef DEBUG_POLYGON_VERBOSE
            fprintf(stderr, "allocate temporary sdl_polygon_y array\n");
            #endif

            sdl_polygon_y = alloca(xu * sizeof(Sint16));

            if (sdl_polygon_y == NULL) {

                #ifdef DEBUG_POLYGON_VERBOSE
                fprintf(stderr, "...failed!\n");
                #endif
            }
            else {

                b = blist;

                #ifdef DEBUG_POLYGON_VERBOSE
                fprintf(stderr, "fill sdl_polygon_x and sdl_polygon_y\n");
                #endif

                v_tmpx = v_tmpy = 0.0;

                for (x = 0; x < xu; x++, b = b->right) {

                    v1 = b->right->boundary->v1;
                    v2 = b->right->boundary->v2;

                    if (b->boundary->v1 == v2 || b->boundary->v2 == v2)
                        v1 = v2;

                    sdl_polygon_x[x] = PIXEL_X(v1->x);
                    sdl_polygon_y[x] = PIXEL_Y(v1->y);

                    v_tmpx += v1->x;
                    v_tmpy += v1->y;
                }

                /* draw filled polygon */

                if (filledPolygonColor(where,
                    sdl_polygon_x, sdl_polygon_y,
                    xu, fill_rgba) != 0) {
                    #ifdef REPORT_EXIT_STATUS
                    fprintf(stderr, TAG "filledPolygonColor() returned nonzero\n");
                    #endif
                    EXIT_WITH_CODE(1);
                }

                if (stroke >= 0) {

                    /* draw stroke */

                    if (
                        #ifndef DOTSDL_BOUNDARY_ANTIALIAS
                        polygonColor
                        #else
                        aapolygonColor
                        #endif
                        (where,
                        sdl_polygon_x, sdl_polygon_y,
                        xu, stroke_rgba) != 0) {
                        #ifdef REPORT_EXIT_STATUS
                        fprintf(stderr, TAG "filledPolygonColor() returned nonzero\n");
                        #endif
                        EXIT_WITH_CODE(1);
                    }
                }

                EXIT_WITH_CODE(0);
            }
        }
    }

    #endif
    #endif
    #endif

    /* allocate memory for scanline array */

    scanmax = ybottom - ytop + 1;

    if (shrink > 0)
        if (scanmax <= (shrink << 1)) {
            if (count != NULL)
                *count = 0;
            #ifdef REPORT_EXIT_STATUS
            fprintf(stderr, TAG "shrink value makes polygon invisible\n");
            #endif
            EXIT_WITH_CODE(-1);
        }

    partial = 0;
    if (ytop < -SHRINK_MAX)
        partial |= CLIP_CELL_TOP;
    if (ybottom >= SCREEN_HEIGHT + SHRINK_MAX)
        partial |= CLIP_CELL_BOTTOM;

    #ifdef ALLOCATE_ONCE

    if (partial & CLIP_CELL_TOP) {

        x1_list = x1_master;
        x2_list = x2_master;

        if (shrink > 0)
            x_tmp = x_tmp_master;
    }
    else {

        x1_list = &x1_master[ytop + SHRINK_MAX];
        x2_list = &x2_master[ytop + SHRINK_MAX];

        if (shrink > 0)
            x_tmp = &x_tmp_master[ytop + SHRINK_MAX];
    }

    #endif

    #ifndef ALLOCATE_ONCE

    x1_list = (int *)malloc(scanmax * sizeof(int));
    if (x1_list == NULL) {
        #ifdef REPORT_EXIT_STATUS
        fprintf(stderr, TAG "insufficient memory for x1_list array\n");
        #endif
        EXIT_WITH_CODE(1);
    }

    x2_list = (int *)malloc(scanmax * sizeof(int));
    if (x2_list == NULL) {
        free(x1_list);
        #ifdef REPORT_EXIT_STATUS
        fprintf(stderr, TAG "insufficient memory for x2_list array\n");
        #endif
        EXIT_WITH_CODE(1);
    }

    if (shrink > 0) {
        x_tmp = (int *)malloc(scanmax * sizeof(int));
        if (x_tmp == NULL) {
            free(x1_list);
            free(x2_list);
            #ifdef REPORT_EXIT_STATUS
            fprintf(stderr, TAG "insufficient memory for x_tmp array\n");
            #endif
            EXIT_WITH_CODE(1);
        }
    }

    for (scan = 0; scan < scanmax; scan++)
        x1_list[scan] = x2_list[scan] = -1;

    #ifdef DEBUG_POLYGON
    fprintf(stderr, "%d scanlines allocated\n", scanmax);
    #endif

    #endif

    #ifdef DEBUG_POLYGON
    fprintf(stderr, "partial = 0x%02X\n", partial);
    #endif

    /* set up left and right endpoints of
         each scanline; the method used here is
         based on Bresenham's line algorithm */

    #define GET_NEXT_V2 {                   \
        v1 = b->boundary->v1;               \
        v2 = b->boundary->v2;               \
        if (Y_ABOVE(v2->y, v1->y)) {        \
            v1 = v2;                        \
            v2 = b->boundary->v1;           \
        }                                   \
    }

    #define START_SIDE(START) {             \
        b = START;                          \
        GET_NEXT_V2;                        \
        x1 = PIXEL_X(v1->x);                \
        y1 = PIXEL_Y(v1->y);                \
        x2 = PIXEL_X(v2->x);                \
        y2 = PIXEL_Y(v2->y);                \
        scan = y1 - ytop;                   \
    }

    #define GET_NEXT_X1_AND_X2(DIR) {       \
        b = b->DIR;                         \
        GET_NEXT_V2;                        \
        x1 = x2;                            \
        y1 = y2;                            \
        x2 = PIXEL_X(v2->x);                \
        y2 = PIXEL_Y(v2->y);                \
    }

    #define NEXT_BOUNDARY(DIR, DONE) {      \
        /*if (v2 == vbottom) break;*/       \
        if (y2 >= ybottom)                  \
            goto DONE;                      \
        GET_NEXT_X1_AND_X2(DIR);            \
        /*if (y2 < y1) break;*/             \
    }

    #define CHECK_TOP(DIR, COMP, DONE) {    \
        while (y1 == y2) {                  \
            if (COMP(x1, x2))               \
                x2 = x1;                    \
            if (COMP(b->boundary->v1->x,    \
                b->boundary->v2->x))        \
                v2 = b->boundary->v1;       \
            else                            \
                v2 = b->boundary->v2;       \
            NEXT_BOUNDARY(DIR, DONE);       \
        }                                   \
    }

    #ifdef DEBUG_POLYGON
    #if 0
    #define PRINT_BOUNDARY_INFO {           \
        FORMAT_VERTEX(b->boundary->v1);     \
        fprintf(stderr, " -> ");            \
        FORMAT_VERTEX(b->boundary->v2);     \
        fprintf(stderr, " [b]   ");         \
        FORMAT_VERTEX(v2);                  \
        fprintf(stderr, " [v2]\n");         \
    }
    #else
    #define PRINT_BOUNDARY_INFO {           \
        fprintf(stderr, "%d,%d -> %d,%d\n", \
            x1, y1, x2, y2);                \
    }
    #endif
    #else
    #define PRINT_BOUNDARY_INFO { }
    #endif

    #define DO_SCANLINES(WHERE, XLOWER, PREFIX, SUFFIX) {       \
        PRINT_BOUNDARY_INFO;                                    \
        yd = y2 - y1 + 1;                                       \
        if (yd <= 0) { ret = 112; goto polygon_done; }          \
        if (x1 < x2) {                                          \
            xd = x2 - x1 + 1;                                   \
            xu = 1;                                             \
        }                                                       \
        else {                                                  \
            xd = x1 - x2 + 1;                                   \
            xu = -1;                                            \
        }                                                       \
        if (xd > yd) {                                          \
            for (y = 0, x = x1; x1 != x2 + xu; x1 += xu) {      \
                y += yd;                                        \
                if (y >= xd) {                                  \
                    if (xu == XLOWER) {                         \
                        PREFIX                                  \
                        WHERE[scan++] = x + (XLOWER);           \
                        SUFFIX                                  \
                        x = x1;                                 \
                    }                                           \
                    else {                                      \
                        PREFIX                                  \
                        WHERE[scan++] = x1;                     \
                        SUFFIX                                  \
                    }                                           \
                    y -= xd;                                    \
                    y1++;                                       \
                }                                               \
            }                                                   \
        }                                                       \
        else {                                                  \
            for (x = 0; y1 <= y2; y1++) {                       \
                PREFIX                                          \
                WHERE[scan++] = x1;                             \
                SUFFIX                                          \
                x += xd;                                        \
                if (x >= yd) {                                  \
                    x -= yd;                                    \
                    x1 += xu;                                   \
                }                                               \
            }                                                   \
        }                                                       \
        scan--;                                                 \
    }

    #define PREFIX_NOP
    #define SUFFIX_NOP

    #ifdef ALLOCATE_ONCE

    #define PREFIX_TOP      if (y1 >= MASTER_TOP) {
    #define PREFIX_BOTTOM   if (y1 <  MASTER_BOTTOM) {
    #define PREFIX_BOTH     PREFIX_TOP PREFIX_BOTTOM
    #define SUFFIX_TOP      }
    #define SUFFIX_BOTTOM   } else break;
    #define SUFFIX_BOTH     SUFFIX_BOTTOM SUFFIX_TOP

    #define DO_SCANLINES_IN_PARTS(DIR, WHERE, XLOWER, SKIP)     \
    if (partial & CLIP_CELL_TOP) {                              \
        while (y2 < MASTER_TOP)                                 \
            NEXT_BOUNDARY(DIR, SKIP);                           \
        if (Y_BELOW(v2->y, v1->y)) {                            \
            r = (v_top_left.y - v1->y) / (v2->y - v1->y);       \
            x1 *= 1.0 - r;                                      \
            x1 += x2 * r;                                       \
        }                                                       \
        /*if (y2 > y1)                                            \
            x1 += ((x2 - x1) * (MASTER_TOP - y1)) / (y2 - y1);*/  \
        scan = 0;                                               \
        y1 = MASTER_TOP;                                        \
        if (y2 >= MASTER_BOTTOM) {                              \
            DO_SCANLINES(WHERE, XLOWER,                         \
                PREFIX_BOTH, SUFFIX_BOTH);                      \
            goto SKIP;                                          \
        }                                                       \
        else {                                                  \
            DO_SCANLINES(WHERE, XLOWER,                         \
                PREFIX_TOP, SUFFIX_TOP);                        \
            NEXT_BOUNDARY(DIR, SKIP);                           \
        }                                                       \
    }                                                           \
    if (partial & CLIP_CELL_BOTTOM) {                           \
        while (y2 < MASTER_BOTTOM) {                            \
            DO_SCANLINES(WHERE, XLOWER,                         \
                PREFIX_NOP, SUFFIX_NOP);                        \
            NEXT_BOUNDARY(DIR, SKIP);                           \
        }                                                       \
        DO_SCANLINES(WHERE, XLOWER,                             \
            PREFIX_BOTTOM, SUFFIX_BOTTOM);                      \
    }                                                           \
    else {                                                      \
        while (1) {                                             \
            DO_SCANLINES(WHERE, XLOWER,                         \
                PREFIX_NOP, SUFFIX_NOP);                        \
            NEXT_BOUNDARY(DIR, SKIP);                           \
        }                                                       \
    }

    #endif

    #ifdef DEBUG_POLYGON_VERBOSE
    fprintf(stderr, "fill x1_list\n");
    #endif

    START_SIDE(b_start_left);
    CHECK_TOP(left, LESS_THAN, left_top_checked);
    left_top_checked:

    #ifdef ALLOCATE_ONCE
    DO_SCANLINES_IN_PARTS(left, x1_list, 1, left_side_done);
    #else
    while (1) {
        DO_SCANLINES(x1_list, 1, PREFIX_NOP, SUFFIX_NOP);
        NEXT_BOUNDARY(left, left_side_done);
    }
    #endif
    left_side_done:

    #ifdef DEBUG_POLYGON_VERBOSE
    fprintf(stderr, "fill x2_list\n");
    #endif

    START_SIDE(b_start_right);
    CHECK_TOP(right, GREATER_THAN, right_top_checked);
    right_top_checked:

    #ifdef ALLOCATE_ONCE
    DO_SCANLINES_IN_PARTS(right, x2_list, -1, right_side_done);
    #else
    while (1) {
        DO_SCANLINES(x2_list, -1, PREFIX_NOP, SUFFIX_NOP);
        NEXT_BOUNDARY(right, right_side_done);
    }
    #endif
    right_side_done:

    #ifdef ALLOCATE_ONCE
    if (partial & CLIP_CELL_TOP)
        ytop = MASTER_TOP;
    if (partial & CLIP_CELL_BOTTOM)
        ybottom = MASTER_BOTTOM - 1;
    scanmax = ybottom - ytop + 1;
    #endif

    /* shrink fill area by one or more pixels if needed */

    #define SHRINK_SIDE(WHERE, COMP, I, REMOVED) { \
        for (y1 = 0, y2 = shrink << 1,             \
            y = shrink; y < scanmax - shrink;      \
            y1++, y2++, y++) {                     \
            x1 = WHERE[y1];                        \
            x2 = WHERE[y2];                        \
            x = WHERE[y] + (I);                    \
            if (COMP(x1, x))                       \
                x = x1;                            \
            if (COMP(x2, x))                       \
                x = x2;                            \
            x_tmp[y] = x;                          \
        }                                          \
        for (scan = 0; scan < shrink; scan++)      \
            WHERE[scan] = REMOVED;                 \
        for ( ; scan < scanmax - shrink; scan++)   \
            WHERE[scan] = x_tmp[scan];             \
        for ( ; scan < scanmax; scan++)            \
            WHERE[scan] = REMOVED;                 \
    }

    if (shrink > 0) {

        #ifdef DEBUG_POLYGON_VERBOSE
        fprintf(stderr, "shrink by %u pixel", shrink);
        if (shrink != 1) fprintf(stderr, "s");
        fprintf(stderr, "\n");
        #endif

        SHRINK_SIDE(x1_list, GREATER_THAN, shrink, -1);
        SHRINK_SIDE(x2_list, LESS_THAN, -shrink, -2);

        for (y = 0, scan = 0; scan < scanmax;
            y++, ytop++, scan++) {
            if (x1_list[scan] <= x2_list[scan])
                break;
            x1_list[scan] = x2_list[scan] = -1;
        }

        for (scan = scanmax - 1; scan >= 0; ybottom--, scan--) {
            if (x1_list[scan] <= x2_list[scan])
                break;
            x1_list[scan] = x2_list[scan] = -1;
        }

        if (ybottom < ytop) {
            if (count != NULL)
                *count = 0;
            #ifdef REPORT_EXIT_STATUS
            fprintf(stderr, TAG "shrink value moved ybottom above ytop\n");
            #endif
            goto polygon_done;
        }

        x1_p = &x1_list[y];
        x2_p = &x2_list[y];
        scanmax = ybottom - ytop + 1;
    }
    else {

        x1_p = x1_list;
        x2_p = x2_list;
    }

    /**** debug ****/
    #if 0
    #ifdef DEBUG_POLYGON
    if (ca->infinity) {
        for (y = ytop, scan = 0; y <= ybottom;
            y++, scan++) {
            #if 1
            if (y < 0) continue;
            if (y >= SCREEN_HEIGHT) continue;
            #endif
            fprintf(stderr, "%6d ..%6d\n",
                x1_list[scan], x2_list[scan]);
        }
        fprintf(stderr, "(%d scanlines)\n", scanmax);
    }
    #endif
    #endif

    /* select only those scanlines with at least
         one visible pixel */

    y1 = SCREEN_HEIGHT;
    y2 = -1;

    for (y = ytop, scan = 0; y <= ybottom; y++, scan++) {

        if (x1_p[scan] < SCREEN_WIDTH && x2_p[scan] >= 0) {
            y1 = y;
            break;
        }
    }

    #ifdef DEBUG_POLYGON_VERBOSE
    fprintf(stderr, "y1 = %d\n", y1);
    #endif

    if (y1 >= SCREEN_HEIGHT) {
        #ifdef REPORT_EXIT_STATUS
        fprintf(stderr, TAG "entire polygon is below screen area\n");
        #endif
        goto polygon_done;
    }

    for (y = ybottom, scan = scanmax - 1; y >= ytop; y--, scan--) {

        if (x1_p[scan] < SCREEN_WIDTH && x2_p[scan] >= 0) {
            y2 = y;
            break;
        }
    }

    #ifdef DEBUG_POLYGON_VERBOSE
    fprintf(stderr, "y2 = %d\n", y2);
    #endif

    if (y2 < 0) {
        #ifdef REPORT_EXIT_STATUS
        fprintf(stderr, TAG "entire polygon is above screen area\n");
        #endif
        goto polygon_done;
    }

    yd = y1 - ytop;

    if (y1 < 0) {
        yd -= y1;
        y1 = 0;
    }
    if (y2 >= SCREEN_HEIGHT)
        y2 = SCREEN_HEIGHT - 1;

    #ifdef REPORT_VISIBLE_SCANLINES
    cell_scanlines = y2 - y1 + 1;
    #endif

    /* fill polygon area;
       use svgalib hardware acceleration only
         if no stroke is to be drawn */

    #ifdef SVGALIB
    gl_setcontext(
        (flags & DRAW_CELL_PHYSICAL) == 0 ? scr_virtual :
        #ifdef DIRECT_TO_PHYSICAL
        scr_physical
        #else
        scr_frame
        #endif
        );
    #endif

    if (fill < 0)
        goto draw_stroke;

    #ifdef SVGALIB

    if (!get_vga_accel_caps_called)
        get_vga_accel_caps();

    if (stroke < 0) {

        if (vga_accel_caps &
            (ACCELFLAG_DRAWHLINELIST | ACCELFLAG_SETFGCOLOR)) {

            for (y = y1, scan = yd, total_pixels = 0;
                y <= y2; y++, scan++) {

                x1 = x1_p[scan];
                x2 = x2_p[scan];

                if (x1 >= SCREEN_WIDTH) {
                    x1_p[scan] =  0;
                    x2_p[scan] = -1;
                    partial |= CLIP_CELL_RIGHT;
                    continue;
                }
                if (x2 < 0) {
                    x1_p[scan] =  0;
                    x2_p[scan] = -1;
                    partial |= CLIP_CELL_LEFT;
                    continue;
                }
                if (x2 < x1)
                    x2_p[scan] = x1;
                if (x1 < 0) {
                    x1_p[scan] = 0;
                    partial |= CLIP_CELL_LEFT;
                }
                if (x2 >= SCREEN_WIDTH) {
                    x2_p[scan] = SCREEN_WIDTH - 1;
                    partial |= CLIP_CELL_RIGHT;
                }

                total_pixels += x2_p[scan] - x1_p[scan] + 1;
            }

            #ifdef DEBUG_POLYGON_VERBOSE
            fprintf(stderr, "fill scanlines (accelerated)\n");
            #endif

            if (!vga_accel(ACCELFLAG_SETFGCOLOR, fill)) {
                if (!vga_accel(ACCELFLAG_DRAWHLINELIST,
                    y1, y2 - y1 + 1, &x1_p[yd], &x2_p[yd])) {
                    if (count != NULL)
                        *count = total_pixels;
                    goto polygon_done;
                }
            }

            #ifdef DEBUG_POLYGON_VERBOSE
            fprintf(stderr, "...failed!\n");
            #endif
        }
    }

    #endif

    #ifdef DEBUG_POLYGON_VERBOSE
    fprintf(stderr, "fill scanlines (non-accelerated)\n");
    #endif

    for (y = y1, scan = yd, total_pixels = 0;
        y <= y2; y++, scan++) {

        x1 = x1_p[scan];
        x2 = x2_p[scan];

        if (x1 >= SCREEN_WIDTH) {
            partial |= CLIP_CELL_RIGHT;
            continue;
        }
        if (x2 < 0) {
            partial |= CLIP_CELL_LEFT;
            continue;
        }
        if (x2 < x1)
            continue;
        if (x1 < 0) {
            x1 = 0;
            partial |= CLIP_CELL_LEFT;
        }
        if (x2 >= SCREEN_WIDTH) {
            x2 = SCREEN_WIDTH - 1;
            partial |= CLIP_CELL_RIGHT;
        }

        draw_scanline(x1, y, x2, fill);

        total_pixels += x2 - x1 + 1;
    }

    #ifdef DEBUG_POLYGON_VERBOSE
    fprintf(stderr, "total_pixels = %u\n", total_pixels);
    #endif

    if (count != NULL)
        *count = total_pixels;

    if (stroke < 0)
        goto polygon_done;

    /* draw stroke without acceleration */

    draw_stroke:

    y = y1;
    scan = yd;

    #ifdef DEBUG_POLYGON_VERBOSE
    fprintf(stderr, "draw stroke");
    #endif

    /* top scanline */

    if (y1 == ytop) {

        #ifdef DEBUG_POLYGON_VERBOSE
        fprintf(stderr, " top");
        #endif

        x1 = x1_p[0];
        x2 = x2_p[0];

        if (x1 < SCREEN_WIDTH && x2 >= 0) {
            if (x1 < 0)
                x1 = 0;
            if (x2 >= SCREEN_WIDTH)
                x2 = SCREEN_WIDTH - 1;
            draw_scanline(x1, y1, x2, stroke);
        }

        y1++;
        yd++;
    }

    /* bottom scanline */

    if (y2 == ybottom) {

        #ifdef DEBUG_POLYGON_VERBOSE
        fprintf(stderr, ", bottom");
        #endif

        x1 = x1_p[y2 - ytop];
        x2 = x2_p[y2 - ytop];

        if (x1 < SCREEN_WIDTH && x2 >= 0) {
            if (x1 < 0)
                x1 = 0;
            if (x2 >= SCREEN_WIDTH)
                x2 = SCREEN_WIDTH - 1;
            draw_scanline(x1, y2, x2, stroke);
        }

        y2--;
    }

    /* left side */

    y = y1;
    scan = yd;
    xu = x1_p[scan - 1];
    x1 = x1_p[scan];

    if (x1 > xu)
        goto left_lower_part;

    #ifdef DEBUG_POLYGON_VERBOSE
    fprintf(stderr, ", left upper");
    #endif

    for ( ; y <= y2; y++, xu = x1, x1 = xd) {

        scan++;
        xd = x1_p[scan];

        if (xd > x1) {

            x = (xu > xd) ? xu : xd;
            if (x > 0)
                draw_scanline((x1 < 0) ? 0 : x1, y,
                    (x >= SCREEN_WIDTH) ? SCREEN_WIDTH - 1 : x - 1,
                    stroke);

            scan--;
            break;
        }

        if (x1 < 0) {

            if (xu > 0)
                draw_scanline(0, y, xu - 1, stroke);
        }
        else {

            x = (x1 == xu) ? x1 : xu - 1;
            if (x >= SCREEN_WIDTH)
                x = SCREEN_WIDTH - 1;

            draw_scanline(x1, y, x, stroke);
        }
    }

    left_lower_part:

    #ifdef DEBUG_POLYGON_VERBOSE
    fprintf(stderr, ", left lower");
    #endif

    for ( ; y <= y2; y++, x1 = xd) {

        scan++;
        xd = x1_p[scan];

        if (x1 < 0) {

            if (xd > 0)
                draw_scanline(0, y,
                    (xd >= SCREEN_WIDTH) ? SCREEN_WIDTH - 1 : xd - 1,
                    stroke);
        }
        else {

            x = (x1 == xd) ? x1 : xd - 1;
            if (x >= SCREEN_WIDTH)
                x = SCREEN_WIDTH - 1;

            draw_scanline(x1, y, x, stroke);
        }
    }

    /* right side */

    y = y1;
    scan = yd;
    xu = x2_p[scan - 1];
    x2 = x2_p[scan];

    if (x2 < xu)
        goto right_lower_part;

    #ifdef DEBUG_POLYGON_VERBOSE
    fprintf(stderr, ", right upper");
    #endif

    for ( ; y <= y2; y++, xu = x2, x2 = xd) {

        scan++;
        xd = x2_p[scan];

        if (xd < x2) {

            x = (xu < xd) ? xu : xd;
            if (x < SCREEN_WIDTH - 1)
                draw_scanline((x < 0) ? 0 : x + 1, y,
                    (x2 >= SCREEN_WIDTH) ? SCREEN_WIDTH - 1 : x2,
                    stroke);

            scan--;
            break;
        }

        if (x2 >= SCREEN_WIDTH) {

            if (xu < SCREEN_WIDTH - 1)
                draw_scanline(xu + 1, y, SCREEN_WIDTH - 1, stroke);
        }
        else {

            x = (x2 == xu) ? x2 : xu + 1;
            if (x < 0)
                x = 0;

            draw_scanline(x, y, x2, stroke);
        }
    }

    right_lower_part:

    #ifdef DEBUG_POLYGON_VERBOSE
    fprintf(stderr, ", right lower\n");
    #endif

    for ( ; y <= y2; y++, x2 = xd) {

        scan++;
        xd = x2_p[scan];

        if (x2 >= SCREEN_WIDTH) {

            if (xd < SCREEN_WIDTH - 1)
                draw_scanline((xd < 0) ? 0 : xd + 1, y,
                    SCREEN_WIDTH - 1, stroke);
        }
        else {

            x = (x2 == xd) ? x2 : xd + 1;
            if (x < 0)
                x = 0;

            draw_scanline(x, y, x2, stroke);
        }
    }

    /* done */

    polygon_done:

    /**** debug ****/
    #if 0
    #define PLOT_VERTEX(V) {                    \
        if (b->boundary->V != v1 &&             \
            b->boundary->V != v2) {             \
            x1 = PIXEL_X(b->boundary->V->x);    \
            if (x1 < 0) x1 = 0;                 \
            if (x1 >= SCREEN_WIDTH)             \
                x1 = SCREEN_WIDTH - 1;          \
            y1 = PIXEL_Y(b->boundary->V->y);    \
            if (y1 < 0) y1 = 0;                 \
            if (y1 >= SCREEN_HEIGHT)            \
                y1 = SCREEN_HEIGHT - 1;         \
            plot_dot_ex(x1, y1,                 \
                PLOT_DOT_HOLLOW | flags,        \
                x, (x == 8) ? 7 : x + 8);       \
        }                                       \
    }
    if (ca->infinity)
        for (b = b_start_left, v1 = v2 = NULL, x = 0; ; x &= 7) {
            x++;
            PLOT_VERTEX(v1);
            PLOT_VERTEX(v2);
            v1 = b->boundary->v1;
            v2 = b->boundary->v2;
            b = b->left;
            if (b == b_start_left) break;
        }
    #endif

    #ifndef ALLOCATE_ONCE
    #ifdef DEBUG_POLYGON_VERBOSE
    fprintf(stderr, "release memory\n");
    #endif
    free(x1_list);
    free(x2_list);
    if (shrink > 0) if (x_tmp != NULL)
        free(x_tmp);
    #endif

    if (clip != NULL)
        *clip = partial;

    paint_cell_done:

    if (ca->infinity) {
        #ifdef COPY_INFINITE_CELL
        if (blist != NULL)
            free(blist);
        #else
        b_inf_1->boundary->v2 = NULL;
        b_inf_1->right = b_inf_2;
        b_inf_2->boundary->v2 = NULL;
        b_inf_2->left = b_inf_1;
        #endif
    }

    #ifdef DEBUG_POLYGON_VERBOSE
    fprintf(stderr, "ret = %d\n", ret);
    #endif

    return ret;
}
