/*************************************************************************
 *
 *  An implementation of Fortune's algorithm
 *  Copyright (C) 2012-2014 Jussi M. Hukkanen
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *************************************************************************
 *
 *  dotsdl.c: graphics output using SDL
 *
 ************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <float.h>   /* needed for FLT_EPSILON */
#include <string.h>  /* declares strcmp() */

#include "dotsdl.h"


int WIDTH = SCREEN_DEFAULT_W;
int HEIGHT = SCREEN_DEFAULT_H;
int BPP = SCREEN_DEFAULT_BPP;


static char sdl_get_window_size_called = 0;


extern char *window_title;  /* each program must define this (can be NULL) */


/* sdl_get_window_size()

    initializes the WINDOW_* variables */

void sdl_get_window_size() {

    #ifdef WINDOW_COORDINATES_FLOAT
    if (WIDTH > HEIGHT) {
        WINDOW_TOP = 1.0;
        WINDOW_RIGHT = (float)WIDTH / HEIGHT;
    }
    else {
        WINDOW_RIGHT = 1.0;
        WINDOW_TOP = (float)HEIGHT / WIDTH;
    }
    WINDOW_LEFT = -WINDOW_RIGHT;
    #ifdef FORTUNE_POSITIVE_Y_DOWN
    WINDOW_TOP = -WINDOW_TOP;
    #endif
    WINDOW_BOTTOM = -WINDOW_TOP;
    #else
    WINDOW_LEFT = 0;
    WINDOW_RIGHT = WIDTH;
    #ifdef FORTUNE_POSITIVE_Y_DOWN
    WINDOW_TOP = 0;
    WINDOW_BOTTOM = HEIGHT;
    #else
    WINDOW_TOP = HEIGHT;
    WINDOW_BOTTOM = 0;
    #endif
    #endif

    sdl_get_window_size_called = 1;
}


#ifndef VALGRIND

#if SCREEN_DEFAULT_BPP == 8
SDL_Color vga_palette[] = {
    #include "vga_standard_pal.txt"
};
#else
Uint32 vga_palette[] = {
    #include "vga_standard_pal_hex.txt"
};
#endif


static const SDL_VideoInfo *videoinfo;

static Uint32 video_mode_flags = SDL_HWSURFACE;


/* sdl_gfx_colour()

    helper function to retrieve a 32-bit
      (0xRRGGBBAA) colour value from the 8-bit
      palette */

Uint32 sdl_gfx_colour(uint8_t colour) {

    #if SCREEN_DEFAULT_BPP == 8
    SDL_Color *pal_colour = &(scr_physical->format->palette->colors[colour]);
    return
        (pal_colour->r << 24) |
        (pal_colour->g << 16) |
        (pal_colour->b <<  8) |
        0xFF; /* fully opaque */
    #else
    return vga_palette[colour];
    #endif
}


/* sdl_key_to_ascii()

    in SDL, the Enter key is assigned the ASCII
      code of 13 (`\r'); this helper function
      translates it to code 10 (`\n') for
      svgalib compatibility */

char sdl_key_to_ascii(SDLKey key) {

    return (key == SDLK_RETURN) ? 10 : (char)key;
}


/* dot_getch()
   dot_getkey()

    replacements for svgalib functions
      vga_getch() and vga_getkey();
    note that the type of their return values
      should in fact be SDLKey, but we can use
      char for svgalib compatibility, because
      SDLKey values up to SDLK_DELETE map
      neatly to ASCII;
    these functions return -1 if the user quits
      the program */

char dot_getch(void) {

    SDL_Event event;

	while (SDL_WaitEvent(&event) > 0) {
        if (event.type == SDL_KEYDOWN)
            return sdl_key_to_ascii(event.key.keysym.sym);
        if (event.type == SDL_QUIT)
            return -1;
    }

    return 0;
}

char dot_getkey(void) {

    SDL_Event event;

    if (SDL_PollEvent(&event)) {
		if (event.type == SDL_KEYDOWN)
			return sdl_key_to_ascii(event.key.keysym.sym);
    }

    if (SDL_QuitRequested())
        return -1;

    return 0;
}


/* get_vga_accel_caps()

    retrieves info about the selected video mode;
    returns nonzero if no info is found */

char get_vga_accel_caps(void) {

    videoinfo = SDL_GetVideoInfo();
    if (videoinfo == NULL)
        return -1;

    return 0;
}


/* report_vga_accel_caps()

    displays the findings of get_vga_accel_caps() */

void report_vga_accel_caps(void) {

    #define CHECK_FLAG(WHICH) {         \
        if (videoinfo->WHICH)           \
            printf("YES  ");            \
        else                            \
            printf(" no  ");            \
        printf(#WHICH "\n");            \
    }

    CHECK_FLAG(hw_available);
    CHECK_FLAG(wm_available);
    CHECK_FLAG(blit_hw);
    CHECK_FLAG(blit_hw_CC);
    CHECK_FLAG(blit_hw_A);
    CHECK_FLAG(blit_sw);
    CHECK_FLAG(blit_sw_CC);
    CHECK_FLAG(blit_sw_A);
    CHECK_FLAG(blit_fill);

    printf("%u KB video memory\n", videoinfo->video_mem);
    printf("%u bits per pixel\n", videoinfo->vfmt->BitsPerPixel);
}


/* parse_sdl_video_mode()

    reads screen mode info from the given
      string, which should be in the format
      `WIDTHxHEIGHT' */

char parse_sdl_video_mode(char *str) {

    int n;

    n = sscanf(str, "%dx%d", &WIDTH, &HEIGHT);

    if (n != 2) {
        fprintf(stderr,
            "Invalid screen mode identifier `%s'.\n", str);
        return -1;
	}

    /* SDL_VideoModeOK() throws strange segfaults, so let's just
         hope the requested mode is supported */

    #if 0
    n = SDL_VideoModeOK(WIDTH, HEIGHT, SCREEN_DEFAULT_BPP, video_mode_flags);

    if (n == 0) {
        fprintf(stderr,
            "Screen mode %s not supported.\n", str);
        return 1;
    }
    #endif

    return 0;
}


/* init_vga()

    sets up graphics mode */

int init_vga(uint8_t flags) {

    int i, n;
    Uint32 sdl_video_init_flags = video_mode_flags;
    SDL_Palette *pal_physical;
    SDL_Palette *pal_virtual;

    #if 1
    if (flags & INIT_VGA_DOUBLE_BUFFER)
        sdl_video_init_flags |= SDL_DOUBLEBUF;
    #endif

    if (!sdl_get_window_size_called)
        sdl_get_window_size();

	if (SDL_Init(SDL_INIT_VIDEO | SDL_INIT_TIMER) == -1) {
		fprintf(stderr, "ERROR: Could not initialize SDL: %s\n", SDL_GetError());
		return -1;
	}

	atexit(SDL_Quit);

    /* create scr_physical */

	scr_physical = SDL_SetVideoMode(WIDTH, HEIGHT, BPP, sdl_video_init_flags);

	if (scr_physical == NULL) {
		fprintf(stderr, "ERROR: Could not set %u" "x%u" "x%u video mode: %s\n",
            WIDTH, HEIGHT, BPP, SDL_GetError());
		return -1;
	}

    #if SCREEN_DEFAULT_BPP == 8
    SDL_SetColors(scr_physical, vga_palette, 0, 256);
    #endif

    /* create scr_virtual */

	scr_virtual = SDL_CreateRGBSurface(SDL_HWSURFACE, WIDTH, HEIGHT, BPP,
        0, 0, 0, 0);    /* assuming BPP == 8 */

	if (scr_virtual == NULL) {
		fprintf(stderr, "ERROR: Could not create scr_virtual: %s\n",
            SDL_GetError());
		return -1;
	}

    /* copy palette from scr_physical to scr_virtual */

    #if SCREEN_DEFAULT_BPP == 8
    #if 1
    SDL_SetColors(scr_virtual, vga_palette, 0, 256);
    #else
    pal_physical = scr_physical->format->palette;
    pal_virtual = scr_virtual->format->palette;

    n = pal_virtual->ncolors;
    if (pal_physical->ncolors < n);
        n = pal_physical->ncolors;

    for (i = 0; i < n; i++) {
        pal_virtual->colors[i].r = pal_physical->colors[i].r;
        pal_virtual->colors[i].g = pal_physical->colors[i].g;
        pal_virtual->colors[i].b = pal_physical->colors[i].b;
    }
    #endif
    #endif

    /* set window title */

    if (window_title != NULL)
        SDL_WM_SetCaption(window_title, NULL);

    /* draw blue frame around screen */

    if (flags & INIT_VGA_DRAW_FRAME) {
        rectangleColor(scr_virtual, 0, 0, WIDTH - 1, HEIGHT - 1,
            sdl_gfx_colour(1));
    }

    return 0;
}


/* exit_vga()

    does nothing when using SDL */

char exit_vga(void) {

    return 0;
}


/* copy_screen()

    copies virtual screen src to dst;
    when using SDL, scr_frame is NULL,
      so everything copied to scr_frame
      is directed to scr_physical */

char copy_screen(SDL_Surface *src, SDL_Surface *dst) {

    if (dst == NULL)
        dst = scr_physical;
    if (src != NULL)
        SDL_BlitSurface(src, NULL, dst, NULL);
    if (dst == scr_physical)
        SDL_Flip(dst);

    return 0;
}


/* fill_screen()

    fills virtual screen dst with the given colour */

char fill_screen(SDL_Surface *dst, uint8_t colour) {

    if (dst == NULL)
        dst = scr_physical;

    SDL_FillRect(dst, NULL, colour);

    if (dst == scr_physical)
        SDL_Flip(dst);

    return 0;
}


/* plot_dot_ex()

    draws a dot at the given coordinates */

char plot_dot_ex(int x, int y, uint8_t flags, uint8_t c1, uint8_t c2) {

    char left, center, right;
    SDL_Surface *where;
    Uint8 *p;
    Uint32 rgba1, rgba2;

    if (x < -1) return 0;
    if (y < -1) return 0;
    if (x > SCREEN_WIDTH ) return 0;
    if (y > SCREEN_HEIGHT) return 0;

    where = (flags & PLOT_DOT_PHYSICAL) ? scr_physical : scr_virtual;

    #if 0   /* write pixels directly; only in 8-bpp mode */

    left   = (x > 0) ? 1 : 0;
    center = (x >= 0) ? ((x <= SCREEN_WIDTH - 1) ? 1 : 0) : 0;
    right  = (x < SCREEN_WIDTH - 1) ? 1 : 0;

	if (SDL_MUSTLOCK(where))
		SDL_LockSurface(where);

    p = (Uint8 *)where->pixels + (y * where->pitch);

    if (y > 0) {
        p -= where->pitch;
    	if (flags & PLOT_DOT_HOLLOW) {
		    if (left)                       p[x - 1] = c1;
            if (center)                     p[x    ] = c2;
		    if (right)                      p[x + 1] = c1;
        }
        else
            if (center)                     p[x    ] = c1;
        p += where->pitch;
    }
    if (y < SCREEN_HEIGHT - 1) {
        p += where->pitch;
    	if (flags & PLOT_DOT_HOLLOW) {
    		if (left)                       p[x - 1] = c1;
            if (center)                     p[x    ] = c2;
		    if (right)                      p[x + 1] = c1;
        }
        else
            if (center)                     p[x    ] = c1;
        p -= where->pitch;
    }
	if (flags & PLOT_DOT_HOLLOW)
      c1 = c2;

    if (y >= 0 && y < SCREEN_HEIGHT) {
        if (left)                           p[x - 1] = c1;
	    if (!(flags & PLOT_DOT_HOLLOW)) {
            if (center)                     p[x    ] = c2;
	    }
        if (right)                          p[x + 1] = c1;
    }

	if (SDL_MUSTLOCK(where))
		SDL_UnlockSurface(where);

    if (flags & PLOT_DOT_PHYSICAL)
    	SDL_UpdateRect(where, x - 1, y - 1, 3, 3);

    #else

    rgba1 = sdl_gfx_colour(c1);
    rgba2 = sdl_gfx_colour(c2);

	if (flags & PLOT_DOT_HOLLOW) {
	    pixelColor(where, x - 1, y - 1, rgba1);
	    pixelColor(where, x + 1, y - 1, rgba1);
	    pixelColor(where, x - 1, y + 1, rgba1);
	    pixelColor(where, x + 1, y + 1, rgba1);
        rgba1 = rgba2;
	}
    else
        pixelColor(where, x    , y    , rgba2);

    pixelColor(where, x    , y - 1, rgba1);
    pixelColor(where, x - 1, y    , rgba1);

    pixelColor(where, x + 1, y    , rgba1);
    pixelColor(where, x    , y + 1, rgba1);

    #endif

    return 0;
}


/* draw_boundary_ex()

    draws a line from x1,y1 to x2,y2;
    flags is a combination of the DRAW_BOUNDARY_* bits;
    part is a factor (0..1) specifying how much of the
      line to draw; 1 means full length */

char draw_boundary_ex(float x1, float y1, float x2, float y2,
                      uint8_t flags, uint8_t colour, float part) {

    float x;
    int xs1, ys1, xs2, ys2;
    SDL_Surface *where;

    if (flags & DRAW_BOUNDARY_RELATIVE) {

      if (x2 > -FLT_EPSILON && x2 < FLT_EPSILON)
        if (y2 > -FLT_EPSILON && y2 < FLT_EPSILON)
          return 0; /* should never happen */

      if (x2 < 0.0) {       /* left */

        if (x1 < WINDOW_LEFT + FLT_EPSILON)
          return 0;

        if (y2 < 0.0) {     /* down */

          if (y1 < WINDOW_BOTTOM + FLT_EPSILON)
            return 0;

          if (-x2 < -y2) {
            x2 = x1 + (x2 * (WINDOW_BOTTOM - y1) / y2);
            y2 = WINDOW_BOTTOM;
          }
          else {
            y2 = y1 + (y2 * (WINDOW_LEFT - x1) / x2);
            x2 = WINDOW_LEFT;
          }
        }
        else {              /* up */

          if (y1 > WINDOW_TOP - FLT_EPSILON)
            return 0;

          if (-x2 < y2) {
            x2 = x1 + (x2 * (WINDOW_TOP - y1) / y2);
            y2 = WINDOW_TOP;
          }
          else {
            y2 = y1 + (y2 * (WINDOW_LEFT - x1) / x2);
            x2 = WINDOW_LEFT;
          }
        }
      }
      else {                /* right */

        if (x1 > WINDOW_RIGHT - FLT_EPSILON)
          return 0;

        if (y2 < 0.0) {     /* down */

          if (y1 < WINDOW_BOTTOM + FLT_EPSILON)
            return 0;

          if (x2 < -y2) {
            x2 = x1 + (x2 * (WINDOW_BOTTOM - y1) / y2);
            y2 = WINDOW_BOTTOM;
          }
          else {
            y2 = y1 + (y2 * (WINDOW_RIGHT - x1) / x2);
            x2 = WINDOW_RIGHT;
          }
        }
        else {              /* up */

          if (y1 > WINDOW_TOP - FLT_EPSILON)
            return 0;

          if (x2 < y2) {
            x2 = x1 + (x2 * (WINDOW_TOP - y1) / y2);
            y2 = WINDOW_TOP;
          }
          else {
            y2 = y1 + (y2 * (WINDOW_RIGHT - x1) / x2);
            x2 = WINDOW_RIGHT;
          }
        }
      }
    }

    if (part < 1.0) {
      x = 1.0 - part;
      x2 = (x * x1) + (part * x2);
      y2 = (x * y1) + (part * y2);
    }

    xs1 = PIXEL_X(x1);
    xs2 = PIXEL_X(x2);
    ys1 = PIXEL_Y(y1);
    ys2 = PIXEL_Y(y2);

    #define CLIP_X(LIMIT, XPART, XTARGET, YTARGET) { \
        if (x2 > x1)                                 \
          x = (XPART) / (x2 - x1);                   \
        else                                         \
          x = (XPART) / (x1 - x2);                   \
        x = (x * y2) + ((1.0 - x) * y1);             \
        XTARGET = LIMIT;                             \
        YTARGET = PIXEL_Y(x);                        \
    }

    #define CLIP_Y(LIMIT, YPART, XTARGET, YTARGET) { \
        if (y2 > y1)                                 \
          x = (YPART) / (y2 - y1);                   \
        else                                         \
          x = (YPART) / (y1 - y2);                   \
        x = (x * x2) + ((1.0 - x) * x1);             \
        XTARGET = PIXEL_X(x);                        \
        YTARGET = LIMIT;                             \
    }

    if (xs1 < 0) {
      if (xs2 < 0) return 0;
      CLIP_X(0, WINDOW_LEFT - x1, xs1, ys1);
    }

    if (xs1 >= SCREEN_WIDTH) {
      if (xs2 >= SCREEN_WIDTH) return 0;
      CLIP_X(SCREEN_WIDTH - 1, x1 - WINDOW_RIGHT, xs1, ys1);
    }

    if (ys1 < 0) {
      if (ys2 < 0) return 0;
      CLIP_Y(0, y1 - WINDOW_TOP, xs1, ys1);
    }

    if (ys1 >= SCREEN_HEIGHT) {
      if (ys2 >= SCREEN_HEIGHT) return 0;
      CLIP_Y(SCREEN_HEIGHT - 1, WINDOW_BOTTOM - y1, xs1, ys1);
    }

    if (xs2 < 0)
      CLIP_X(0, x1 - WINDOW_LEFT, xs2, ys2);

    if (xs2 >= SCREEN_WIDTH)
      CLIP_X(SCREEN_WIDTH - 1, WINDOW_RIGHT - x1, xs2, ys2);

    if (ys2 < 0)
      CLIP_Y(0, WINDOW_TOP - y1, xs2, ys2);

    if (ys2 >= SCREEN_HEIGHT)
      CLIP_Y(SCREEN_HEIGHT - 1, y1 - WINDOW_BOTTOM, xs2, ys2);

    where = (flags & DRAW_BOUNDARY_PHYSICAL) ? scr_physical : scr_virtual;

    #ifndef DOTSDL_BOUNDARY_ANTIALIAS
    lineColor
    #else
    aalineColor
    #endif
        (where, xs1, ys1, xs2, ys2, sdl_gfx_colour(colour));

    return 0;
}


#else  /* VALGRIND */

/* dummy replacements for svgalib functions */

char dot_getch(void) { return 0; }
char dot_getkey(void) { return 0; }

/* dummy replacements for functions in dotsdl.c */

char get_vga_accel_caps(void) { return 0; }

void report_vga_accel_caps(void) { }

char parse_sdl_video_mode(char *str) {
    #if 1
    if (sscanf(str, "%d" "x%d", &WIDTH, &HEIGHT) != 2)
        return -1;
    #else
    if (sscanf(str, "%d" "x%d" "x%d", &WIDTH, &HEIGHT, &BPP) != 3)
        return -1;
    if (BPP != 8 && BPP != 15 && BPP != 16 && BPP != 24 && BPP != 32)
        return 1;
    #endif
    if (WIDTH < 128 || HEIGHT < 128)
        return 1;
    return 0;
}

int init_vga(uint8_t flags) {
    SCREEN_WIDTH = WIDTH;
    SCREEN_HEIGHT = HEIGHT;
    return 0;
}

char exit_vga(void) { return 0; }

char copy_screen(SDL_Surface *src, SDL_Surface *dst) { return 0; }

char fill_screen(SDL_Surface *dst, uint8_t colour) { return 0; }

char plot_dot_ex(int x, int y, uint8_t flags, uint8_t c1, uint8_t c2) { return 0; }

char draw_boundary_ex(float x1, float y1, float x2, float y2,
                      uint8_t flags, uint8_t colour, float part) { return 0; }

#endif
