/*************************************************************************
 *
 *  An implementation of Fortune's algorithm
 *  Copyright (C) 2012-2014 Jussi M. Hukkanen
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *************************************************************************
 *
 *  dotsdl.h: header file for graphics output using SDL
 *
 ************************************************************************/

#ifndef _DOTSDL_H_INCLUDED
#define _DOTSDL_H_INCLUDED

#ifndef VALGRIND
#include <SDL.h>
#include <SDL_gfxPrimitives.h>
#endif

#include <stdint.h>
#include <math.h>	/* declares lrintf() */


#define WINDOW_WIDTH  (WINDOW_RIGHT  - WINDOW_LEFT)
#define WINDOW_HEIGHT (WINDOW_TOP    - WINDOW_BOTTOM)

#define PIXEL_X_(x) (((x) - WINDOW_LEFT) * SCREEN_WIDTH ) / WINDOW_WIDTH
#define PIXEL_Y_(y) ((WINDOW_TOP - (y))  * SCREEN_HEIGHT) / WINDOW_HEIGHT
#define WORLD_X_(x) WINDOW_LEFT  + (((x) * WINDOW_WIDTH ) / SCREEN_WIDTH )
#define WORLD_Y_(y) WINDOW_TOP   - (((y) * WINDOW_HEIGHT) / SCREEN_HEIGHT)

/* when using SDL, the screen (i.e. window) can be
    any crazy size, so WINDOW_LEFT, WINDOW_RIGHT,
    WINDOW_TOP and WINDOW_BOTTOM cannot be macros
    as in dotvga.h; instead, they are variables
    which are initialized in init_vga() */

#if 1 /* floating-point coordinates */

#define WINDOW_COORDINATES_FLOAT 1

float WINDOW_LEFT, WINDOW_RIGHT;
float WINDOW_TOP, WINDOW_BOTTOM;

#define PIXEL_X(x) lrintf(PIXEL_X_(x))
#define PIXEL_Y(y) lrintf(PIXEL_Y_(y))
#define WORLD_X(x) ((x < 0) ? WORLD_X_(x - 0.5) : WORLD_X_(x + 0.5))
#define WORLD_Y(y) ((y < 0) ? WORLD_Y_(y - 0.5) : WORLD_Y_(y + 0.5))

#else /* integer coordinates */

#define WINDOW_COORDINATES_INT 1

int WINDOW_LEFT, WINDOW_RIGHT;
int WINDOW_TOP, WINDOW_BOTTOM;

#define PIXEL_X(x) PIXEL_X_(x)
#define PIXEL_Y(y) PIXEL_Y_(y)
#define WORLD_X(x) WORLD_X_(x)
#define WORLD_Y(y) WORLD_Y_(y)

#endif


/* comment the following line to replace calls to aalineColor()
     with lineColor() in draw_boundary_ex() and calls to
     aapolygonColor() with polygonColor() in paint_cell() */

#define DOTSDL_BOUNDARY_ANTIALIAS


#ifdef VALGRIND

typedef char SDL_Surface;

#define hlineColor(surface, x1, x2, y, colour) { }
#define pixelColor(surface, x, y, colour) { }
#define rectangleColor(surface, x1, y1, x2, y2, colour) { }
#define gfxPrimitivesSetFont(font, w, h) { }
#define stringColor(surface, x, y, string, colour) { }

#endif


#define SCREEN_DEFAULT_W    640
#define SCREEN_DEFAULT_H    480
#define SCREEN_DEFAULT_BPP  32//8

int WIDTH, HEIGHT, BPP;

#define SCREEN_WIDTH   WIDTH
#define SCREEN_HEIGHT  HEIGHT

SDL_Surface *scr_physical;
SDL_Surface *scr_virtual;
//SDL_Surface *scr_frame;
#define scr_frame   NULL


#define BRIGHT_COLOUR(X) ((X) | 8)


void sdl_get_window_size();
#ifndef VALGRIND
Uint32 sdl_gfx_colour(uint8_t);
char sdl_key_to_ascii(SDLKey);
#else
#define sdl_gfx_colour(C) (0)
#endif

int init_vga(uint8_t);
#define INIT_VGA_DOUBLE_BUFFER 0x01	 /* use hardware flipping if possible */
#define INIT_VGA_DRAW_FRAME	   0x02  /* draw blue frame around screen */

char exit_vga(void);
char get_vga_accel_caps(void);
void report_vga_accel_caps(void);
char parse_sdl_video_mode(char *);

char dot_getch(void);
char dot_getkey(void);

char copy_screen(SDL_Surface *, SDL_Surface *);
#define blit_screen(S) copy_screen(S, scr_physical)

char fill_screen(SDL_Surface *, uint8_t);

char plot_dot_ex(int, int, uint8_t, uint8_t, uint8_t);
#define PLOT_DOT_HOLLOW        0x01  /* circle instead of dot */
#define PLOT_DOT_PHYSICAL      0x02  /* use scr_physical, not scr_virtual */

char draw_boundary_ex(float, float, float, float, uint8_t, uint8_t, float);
#define DRAW_BOUNDARY_RELATIVE 0x01  /* x2,y2 is relative to x1,y1 */
#define DRAW_BOUNDARY_PHYSICAL 0x02  /* use scr_physical, not scr_virtual */
#define draw_boundary(X1, Y1, X2, Y2, FLAGS, COLOUR) \
     draw_boundary_ex(X1, Y1, X2, Y2, FLAGS, COLOUR, 1.0)

#endif /* _DOTSDL_H_INCLUDED */
