/*************************************************************************
 *
 *  An implementation of Fortune's algorithm
 *  Copyright (C) 2012-2014 Jussi M. Hukkanen
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *************************************************************************
 *
 *  dotvga.c: graphics output using svgalib
 *
 ************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <float.h>   /* needed for FLT_EPSILON */
#include <string.h>  /* declares strcmp() */

#include "dotvga.h"


#ifndef VALGRIND

static GraphicsContext *scr_screen; /* == scr_physical or scr_frame */


#define SCREEN_MODE_DEF(MODE) { MODE, #MODE }

static struct screen_mode_s {
    int mode;
    char *id;
} screen_mode[] = {
    /*
    SCREEN_MODE_DEF(G320x200x16),
    SCREEN_MODE_DEF(G640x200x16),
    SCREEN_MODE_DEF(G640x350x16),
    SCREEN_MODE_DEF(G640x480x16),
    */
    SCREEN_MODE_DEF(G320x200x256),
    SCREEN_MODE_DEF(G320x240x256),
    SCREEN_MODE_DEF(G320x400x256),
    SCREEN_MODE_DEF(G320x480x256),
    SCREEN_MODE_DEF(G400x300x256),
    SCREEN_MODE_DEF(G512x384x256),
    SCREEN_MODE_DEF(G640x480x256),
    SCREEN_MODE_DEF(G800x600x256),
    SCREEN_MODE_DEF(G1024x768x256),
    /*
    SCREEN_MODE_DEF(G1152x864x256),
    SCREEN_MODE_DEF(G1280x1024x256),
    SCREEN_MODE_DEF(G1600x1200x256),
    SCREEN_MODE_DEF(G1920x1440x256),
    */
    { 0, NULL }
};

int SCREEN_MODE = SCREEN_MODE_DEFAULT;


int vga_accel_caps = 0;
char get_vga_accel_caps_called = 0;
static int vga_chipset_num = 0;


/* get_vga_accel_caps()

    determines the availability of hardware accelerated
      features in the selected video mode;
    returns nonzero if no capability flags are found */

char get_vga_accel_caps(void) {

    get_vga_accel_caps_called = 1;

    vga_modeinfo *info = vga_getmodeinfo(SCREEN_MODE);

    vga_chipset_num = vga_getcurrentchipset();

    if (!(info->flags & HAVE_EXT_SET)) {
        return -1;
    }

    vga_accel_caps = vga_ext_set(VGA_EXT_AVAILABLE, VGA_AVAIL_ACCEL);

    return 0;
}


/* report_vga_accel_caps()

    displays the findings of get_vga_accel_caps() */

void report_vga_accel_caps(void) {

    printf("vga_getcurrentchipset() = %d\n", vga_chipset_num);

    #define CHECK_FLAG(WHICH) {         \
        if (vga_accel_caps & WHICH)     \
            printf("YES  ");            \
        else                            \
            printf(" no  ");            \
        printf(#WHICH "\n");            \
    }

    CHECK_FLAG(ACCELFLAG_FILLBOX);
    CHECK_FLAG(ACCELFLAG_SCREENCOPY);
    CHECK_FLAG(ACCELFLAG_PUTIMAGE);
    CHECK_FLAG(ACCELFLAG_DRAWLINE);
    CHECK_FLAG(ACCELFLAG_SETFGCOLOR);
    CHECK_FLAG(ACCELFLAG_SETBGCOLOR);
    CHECK_FLAG(ACCELFLAG_SETTRANSPARENCY);
    CHECK_FLAG(ACCELFLAG_SETRASTEROP);
    CHECK_FLAG(ACCELFLAG_PUTBITMAP);
    CHECK_FLAG(ACCELFLAG_SCREENCOPYBITMAP);
    CHECK_FLAG(ACCELFLAG_DRAWHLINELIST);
    CHECK_FLAG(ACCELFLAG_SETMODE);
    CHECK_FLAG(ACCELFLAG_SYNC);
    CHECK_FLAG(ACCELFLAG_SETOFFSET);
    CHECK_FLAG(ACCELFLAG_SCREENCOPYMONO);
    CHECK_FLAG(ACCELFLAG_POLYLINE);
    CHECK_FLAG(ACCELFLAG_POLYHLINE);
    CHECK_FLAG(ACCELFLAG_POLYFILLMODE);
}


/* list_vga_modes()

	prints the contents of the screen_mode
	  table into stdout;
	returns a pointer to a string containing
	  the name of the default mode or NULL if
	  the default is not found in the table */

char *list_vga_modes(void) {

	int mode, default_mode;

    for (default_mode = -1, mode = 0; ; mode++) {

       if (screen_mode[mode].id == NULL)
         break;

       if (!(mode & 1))
         printf("\n        ");

       printf("%16s", screen_mode[mode].id);

       if (screen_mode[mode].mode == SCREEN_MODE_DEFAULT)
         default_mode = mode;
    }

    printf("\n");

    if (default_mode == -1)
      return NULL;

    return screen_mode[default_mode].id;
}


/* find_vga_mode()

	looks for the given screen mode identifier
	  string in the screen_mode table and sets
	  SCREEN_MODE accordingly;
	returns -1 if the string is not found or
	  1 if the specified screen mode is not
	  supported; otherwise, returns zero */

char find_vga_mode(char *str) {

	int mode;

    for (mode = 0; ; mode++) {

       if (screen_mode[mode].id == NULL) {
         fprintf(stderr,
           "Screen mode identifier `%s' not recognised.\n", str);
         return -1;
	   }

       if (strcmp(str, screen_mode[mode].id) == 0) {

         if (vga_hasmode(screen_mode[mode].mode) == 0) {
           fprintf(stderr,
             "Screen mode %s not supported.\n", str);
           return 1;
         }

         SCREEN_MODE = screen_mode[mode].mode;
         break;
       }
    }

    return 0;
}


/* init_vga()

    sets up graphics mode */

int init_vga(uint8_t flags) {

    int ret;

	ret = vga_init();
	if (ret != 0) {
		fprintf(stderr, "ERROR: vga_init() returned %d\n", ret);
		return ret;
	}

	ret = vga_setmode(SCREEN_MODE);
	if (ret != 0) {
		fprintf(stderr, "ERROR: vga_setmode() returned %d\n", ret);
		return ret;
	}

	ret = gl_setcontextvga(SCREEN_MODE);
	if (ret != 0) {
		vga_setmode(TEXT);
		fprintf(stderr, "ERROR: gl_setcontextvga() returned %d\n", ret);
		return ret;
	}

	scr_physical = gl_allocatecontext();
	gl_getcontext(scr_physical);

    SCREEN_WIDTH = WIDTH;
    SCREEN_HEIGHT = HEIGHT;

    ret = gl_setcontextvgavirtual(SCREEN_MODE);
	if (ret != 0) {
		vga_setmode(TEXT);
		fprintf(stderr,
            "ERROR: gl_setcontextvgavirtual() returned %d\n"
            "when trying to create scr_virtual",
            ret);
		return ret;
	}

	scr_virtual = gl_allocatecontext();
	gl_getcontext(scr_virtual);

    if (flags & INIT_VGA_DOUBLE_BUFFER) {

        ret = gl_setcontextvgavirtual(SCREEN_MODE);
		if (ret != 0) {
			vga_setmode(TEXT);
			fprintf(stderr,
                "ERROR: gl_setcontextvgavirtual() returned %d\n"
                "when trying to create scr_frame",
                ret);
			return ret;
		}

	    scr_frame = gl_allocatecontext();
	    gl_getcontext(scr_frame);

        scr_screen = scr_frame;
    }
    else
        scr_screen = scr_physical;

    /* draw blue frame around screen */

    if (flags & INIT_VGA_DRAW_FRAME) {

   	    gl_setcontext(scr_virtual);

        gl_hline(0,          0, WIDTH - 1, 1);
        gl_hline(0, HEIGHT - 1, WIDTH - 1, 1);
        gl_line(        0, 1,         0, HEIGHT - 2, 1);
        gl_line(WIDTH - 1, 1, WIDTH - 1, HEIGHT - 2, 1);
    }

    return 0;
}


/* exit_vga()

    returns to text mode */

char exit_vga(void) {

    vga_setmode(TEXT);

    return 0;
}


/* copy_screen()

    copies virtual screen src to dst */

char copy_screen(GraphicsContext *src, GraphicsContext *dst) {

    gl_setcontext(src);
    gl_copyscreen(dst);

    return 0;
}


/* fill_screen()

    fills virtual screen dst with the given colour */

char fill_screen(GraphicsContext *dst, uint8_t colour) {

    gl_setcontext(dst);
    gl_fillbox(0, 0, WIDTH, HEIGHT, colour);

    return 0;
}


/* plot_dot_ex()

    draws a dot at the given coordinates */

char plot_dot_ex(int x, int y, uint8_t flags, uint8_t c1, uint8_t c2) {

    char left, center, right;

    if (x < -1) return 0;
    if (y < -1) return 0;
    if (x > SCREEN_WIDTH ) return 0;
    if (y > SCREEN_HEIGHT) return 0;

	if (flags & PLOT_DOT_PHYSICAL) {
        gl_setcontext(scr_screen);
    }
    else {
        gl_setcontext(scr_virtual);
	}

    left   = (x > 0) ? 1 : 0;
    center = (x >= 0) ? ((x <= SCREEN_WIDTH - 1) ? 1 : 0) : 0;
    right  = (x < SCREEN_WIDTH - 1) ? 1 : 0;

	if (flags & PLOT_DOT_HOLLOW) {
      if (y > 0) {
		if (left)    gl_setpixel(x - 1, y - 1, c1);
		if (right)   gl_setpixel(x + 1, y - 1, c1);
  	  }
      if (y < SCREEN_HEIGHT - 1) {
		if (left)    gl_setpixel(x - 1, y + 1, c1);
		if (right)   gl_setpixel(x + 1, y + 1, c1);
	  }
      c1 = c2;
	}

    if (y > 0) {
      if (center) 	 gl_setpixel(x    , y - 1, c1);
    }
    if (y >= 0 && y < SCREEN_HEIGHT) {
      if (left)      gl_setpixel(x - 1, y    , c1);
	  if (!(flags & PLOT_DOT_HOLLOW)) {
        if (center)  gl_setpixel(x    , y    , c2);
	  }
      if (right)     gl_setpixel(x + 1, y    , c1);
    }
    if (y < SCREEN_HEIGHT - 1) {
      if (center)    gl_setpixel(x    , y + 1, c1);
    }

    return 0;
}


/* draw_boundary_ex()

    draws a line from x1,y1 to x2,y2;
    flags is a combination of the DRAW_BOUNDARY_* bits;
    part is a factor (0..1) specifying how much of the
      line to draw; 1 means full length */

char draw_boundary_ex(float x1, float y1, float x2, float y2,
                      uint8_t flags, uint8_t colour, float part) {

    float x;
    int xs1, ys1, xs2, ys2;

    if (flags & DRAW_BOUNDARY_PHYSICAL) {
        gl_setcontext(scr_screen);
    }
    else {
        gl_setcontext(scr_virtual);
    }

    if (flags & DRAW_BOUNDARY_RELATIVE) {

      if (x2 > -FLT_EPSILON && x2 < FLT_EPSILON)
        if (y2 > -FLT_EPSILON && y2 < FLT_EPSILON)
          return 0; /* should never happen */

      if (x2 < 0.0) {       /* left */

        if (x1 < WINDOW_LEFT + FLT_EPSILON)
          return 0;

        if (y2 < 0.0) {     /* down */

          if (y1 < WINDOW_BOTTOM + FLT_EPSILON)
            return 0;

          if (-x2 < -y2) {
            x2 = x1 + (x2 * (WINDOW_BOTTOM - y1) / y2);
            y2 = WINDOW_BOTTOM;
          }
          else {
            y2 = y1 + (y2 * (WINDOW_LEFT - x1) / x2);
            x2 = WINDOW_LEFT;
          }
        }
        else {              /* up */

          if (y1 > WINDOW_TOP - FLT_EPSILON)
            return 0;

          if (-x2 < y2) {
            x2 = x1 + (x2 * (WINDOW_TOP - y1) / y2);
            y2 = WINDOW_TOP;
          }
          else {
            y2 = y1 + (y2 * (WINDOW_LEFT - x1) / x2);
            x2 = WINDOW_LEFT;
          }
        }
      }
      else {                /* right */

        if (x1 > WINDOW_RIGHT - FLT_EPSILON)
          return 0;

        if (y2 < 0.0) {     /* down */

          if (y1 < WINDOW_BOTTOM + FLT_EPSILON)
            return 0;

          if (x2 < -y2) {
            x2 = x1 + (x2 * (WINDOW_BOTTOM - y1) / y2);
            y2 = WINDOW_BOTTOM;
          }
          else {
            y2 = y1 + (y2 * (WINDOW_RIGHT - x1) / x2);
            x2 = WINDOW_RIGHT;
          }
        }
        else {              /* up */

          if (y1 > WINDOW_TOP - FLT_EPSILON)
            return 0;

          if (x2 < y2) {
            x2 = x1 + (x2 * (WINDOW_TOP - y1) / y2);
            y2 = WINDOW_TOP;
          }
          else {
            y2 = y1 + (y2 * (WINDOW_RIGHT - x1) / x2);
            x2 = WINDOW_RIGHT;
          }
        }
      }
    }

    if (part < 1.0) {
      x = 1.0 - part;
      x2 = (x * x1) + (part * x2);
      y2 = (x * y1) + (part * y2);
    }

    xs1 = PIXEL_X(x1);
    xs2 = PIXEL_X(x2);
    ys1 = PIXEL_Y(y1);
    ys2 = PIXEL_Y(y2);

    #define CLIP_X(LIMIT, XPART, XTARGET, YTARGET) { \
        if (x2 > x1)                                 \
          x = (XPART) / (x2 - x1);                   \
        else                                         \
          x = (XPART) / (x1 - x2);                   \
        x = (x * y2) + ((1.0 - x) * y1);             \
        XTARGET = LIMIT;                             \
        YTARGET = PIXEL_Y(x);                        \
    }

    #define CLIP_Y(LIMIT, YPART, XTARGET, YTARGET) { \
        if (y2 > y1)                                 \
          x = (YPART) / (y2 - y1);                   \
        else                                         \
          x = (YPART) / (y1 - y2);                   \
        x = (x * x2) + ((1.0 - x) * x1);             \
        XTARGET = PIXEL_X(x);                        \
        YTARGET = LIMIT;                             \
    }

    if (xs1 < 0) {
      if (xs2 < 0) return 0;
      CLIP_X(0, WINDOW_LEFT - x1, xs1, ys1);
    }

    if (xs1 >= SCREEN_WIDTH) {
      if (xs2 >= SCREEN_WIDTH) return 0;
      CLIP_X(SCREEN_WIDTH - 1, x1 - WINDOW_RIGHT, xs1, ys1);
    }

    if (ys1 < 0) {
      if (ys2 < 0) return 0;
      CLIP_Y(0, y1 - WINDOW_TOP, xs1, ys1);
    }

    if (ys1 >= SCREEN_HEIGHT) {
      if (ys2 >= SCREEN_HEIGHT) return 0;
      CLIP_Y(SCREEN_HEIGHT - 1, WINDOW_BOTTOM - y1, xs1, ys1);
    }

    if (xs2 < 0)
      CLIP_X(0, x1 - WINDOW_LEFT, xs2, ys2);

    if (xs2 >= SCREEN_WIDTH)
      CLIP_X(SCREEN_WIDTH - 1, WINDOW_RIGHT - x1, xs2, ys2);

    if (ys2 < 0)
      CLIP_Y(0, WINDOW_TOP - y1, xs2, ys2);

    if (ys2 >= SCREEN_HEIGHT)
      CLIP_Y(SCREEN_HEIGHT - 1, y1 - WINDOW_BOTTOM, xs2, ys2);

    gl_line(xs1, ys1, xs2, ys2, colour);

    return 0;
}


/* save_pbm_screenshot()

    writes the contents of the graphics context src
      into a one-bit-per-pixel PBM image;
    save_pbm_screenshot() only saves pixels with
      the colour value specified in mask; these
      pixels are `on' in the resulting image file,
      while everything else is `off';
    the return value is zero on success, negative if
      there is not enough memory and positive if an
      error occurs */

char save_pbm_screenshot(char *name, GraphicsContext *src, uint8_t mask) {

    FILE *f;
    char *buffer;
    int x, y;
    int len;
    char *p;

    #define PBM_LINE_LEN 70

    /* create buffer */

    gl_setcontext(src);

    if (BYTESPERPIXEL != 1)
        return 127;

    buffer = (char *)malloc(WIDTH * HEIGHT);
    if (buffer == NULL)
        return -1;

    /* retrieve contents of src */

    gl_getbox(0, 0, WIDTH, HEIGHT, buffer);

    /* open image file for writing */

    f = fopen(name, "wb");
    if (f == NULL) {
        free(buffer);
        return 1;
    }

    /* output header */

    fprintf(f, "P1\n%u %u\n", WIDTH, HEIGHT);

    /* translate image data */

    #define DO_PIXEL {                      \
        fputc((*p == mask) ? '1' : '0', f); \
        p++;                                \
    }

    p = buffer;
    x = PBM_LINE_LEN;
    y = HEIGHT;

    while (1) {

        len = WIDTH;

        while (len >= PBM_LINE_LEN) {
            len -= x;
            for ( ; x; x--)
                DO_PIXEL;
            fputc('\n', f);
            x = PBM_LINE_LEN;
        }

        for (x = len; x; x--)
            DO_PIXEL;

        x = PBM_LINE_LEN - len;
        y--;
        if (!y)
            break;
    }

    /* return */

    free(buffer);
    fclose(f);

    return 0;
}


#else  /* VALGRIND */

static int WIDTH  = VALGRIND_WIDTH_DEFAULT;
static int HEIGHT = VALGRIND_HEIGHT_DEFAULT;

int vga_accel_caps = 0;
char get_vga_accel_caps_called = 1;

/* dummy replacements for svgalib functions */

char vga_getch(void) { return 0; }
char vga_getkey(void) { return 0; }
void gl_setcontext(GraphicsContext *gc) { }
int vga_accel(unsigned operation, ...) { return 0; }

/* dummy replacements for functions in dotvga.c */

char get_vga_accel_caps(void) { return 0; }

void report_vga_accel_caps(void) { }

char *list_vga_modes(void) {
    printf("\n");
    return NULL;
}

char find_vga_mode(char *str) {
    int w, h, c;
    if (sscanf(str, "G%d" "x%d" "x%d", &w, &h, &c) != 3)
        return -1;
    if (w < 128 || h < 128)
        return 1;
    WIDTH = w;
    HEIGHT = h;
    return 0;
}

int init_vga(uint8_t flags) {
    SCREEN_WIDTH = WIDTH;
    SCREEN_HEIGHT = HEIGHT;
    return 0;
}

char exit_vga(void) { return 0; }

char copy_screen(GraphicsContext *src, GraphicsContext *dst) { return 0; }

char fill_screen(GraphicsContext *dst, uint8_t colour) { return 0; }

char plot_dot_ex(int x, int y, uint8_t flags, uint8_t c1, uint8_t c2) { return 0; }

char draw_boundary_ex(float x1, float y1, float x2, float y2,
                      uint8_t flags, uint8_t colour, float part) { return 0; }

char save_pbm_screenshot(char *name, GraphicsContext *src, uint8_t mask) { return 0; }

#endif
