/*************************************************************************
 *
 *  An implementation of Lloyd's algorithm
 *  Copyright (C) 2013-2014 Jussi M. Hukkanen
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *************************************************************************
 *
 *  even.c: main interface
 *
 ************************************************************************/

#include <stdio.h>
#include <string.h>
#include "lloyd.h"
#include "opts.h"

#ifdef H_PLATFORM
#include H_PLATFORM
#endif


int sites;

site_t *site; /* sites items */

typedef struct site_list_s {
	site_coord_t x;
	site_coord_t y;
	struct site_list_s *next;
} site_list_t;

site_list_t *site_list;

char *site_list_file = NULL;


/* read_sites_from_file()

	reads site_list_file and fills site_list
	  with coordinates found */

char read_sites_from_file(void) {

	FILE *fh;
	char *line = NULL;
	size_t size;
	ssize_t len;
	int n;
	site_list_t *sl;
	site_coord_t x, y;

	sites = 0;
	site_list = NULL;

	if (site_list_file == NULL) {
		fh = stdin;
	}
	else {
		fh = fopen(site_list_file, "rb");
		if (fh == NULL) {
			fprintf(stderr, "File `%s' could not be read.\n", site_list_file);
			return 1;
		}
	}

	while (1) {

		len = getline(&line, &size, fh);
		if (len <= 0)
			break;

		if (line[0] == '\n')
			break;

		n = sscanf(line, "%f,%f", &x, &y);
		if (n < 2) {
			if (n < 0)
				break;
			continue;
		}

		sl = malloc(sizeof(site_list_t));
		if (sl == NULL) {
			fprintf(stderr, "Out of memory allocating another site!\n");
			break;
		}
		sl->x = x;
		sl->y = y;
		sl->next = site_list;
		site_list = sl;

		sites++;
	}

	if (fh != stdin)
		fclose(fh);

	if (line != NULL)
		free(line);

	return 0;
}


/* interpret_site_list()

	fills the site array with the coordinate
	  pairs in site_list and frees site_list */

void interpret_site_list(void) {

	site_list_t *sn;
	int s;

	for (s = 0; s < sites; s++) {

		sn = site_list->next;
		site[s].x = site_list->x;
		site[s].y = site_list->y;

		free(site_list);

		if (sn == NULL)
			break;

		site_list = sn;
	}
}


/* parse_options()

	interprets the given list of command-line
	  options */

char parse_options(int argc, char *argv[]) {

	#define CMDLINE_OPTC_HELP       'h'
	#define CMDLINE_OPTS_HELP           "help"

	struct option long_opts[] = {
		{ CMDLINE_OPTS_HELP,  no_argument,       NULL, CMDLINE_OPTC_HELP  },
		{ NULL, 0, NULL, 0 }
	};
	int c;
	float f = 0.0;
	int index = 0;
	char ret = RET_FORTUNE_OK;
	char *optstring;

	optstring = make_opt_string(long_opts);
	if (optstring == NULL) {
		fprintf(stderr, "Out of memory reading command line options!\n");
		return RET_FORTUNE_NO_MEM;
	}

	site_list_file = get_input_file_name(argc, argv);

	while (1) {

		c = getopt_long(argc, argv, optstring, long_opts, &index);

		if (c == -1)
			break;

		switch (c) {

			case CMDLINE_OPTC_HELP:
				printf("even - Copyright (C) 2013-2014 Jussi M. Hukkanen\n");
				printf("Draw a Voronoi diagram based on a given list\n"
				       "of points on a plane and run Lloyd's relaxation\n"
				       "algorithm to even out the resulting cells.\n\n");
				printf("This program comes with ABSOLUTELY NO WARRANTY.\n"
				       "This is free software, and you are welcome to\n"
				       "redistribute it under the conditions of the\n"
				       "GNU General Public License, version 3 or later.\n"
				       "To view the License, see `gpl.txt'.\n\n");

				printf("Usage: %s [FILE]\n", argv[0]);
				printf("FILE should contain a list of sites as one pair\n"
				       "of x,y coordinates per line. If no FILE is given,\n"
				       "the list is read from standard input. In other\n"
				       "words, it makes sense to pipe the data from\n"
				       "another program such as `bitmap' or `sites'.\n");

				ret = RET_FORTUNE_NO_TRIANGLES;
				break;
		}
	}

	free(optstring);

	return ret;
}


/* main()
*/

int main(int argc, char *argv[]) {

	int i, j;
	int s;
	site_coord_t xd, yd;

	i = parse_options(argc, argv);
	if (i != 0)
		return i;

	i = read_sites_from_file();
	if (i != 0)
		return i;

	site = (site_t *)malloc(sites * sizeof(site_t));

	if (site == NULL) {
		fprintf(stderr,
			"Not enough memory for %u sites! (" SIZEOF_FORMAT " bytes needed)\n",
			sites, sites * sizeof(site_t));
		return RET_FORTUNE_NO_MEM;
	}

	interpret_site_list();
	/*
	if (site_list_file != NULL)
		free(site_list_file);
	*/

	for (i = 0, s = sites; i < s; i++) {

		xd = site[i].x;
		yd = site[i].y;

		for (j = i + 1; j < s; j++) {
			if (site[j].x > xd - SITE_COORD_EPSILON)
				if (site[j].x < xd + SITE_COORD_EPSILON)
					if (site[j].y > yd - SITE_COORD_EPSILON)
						if (site[j].y < yd + SITE_COORD_EPSILON) {
							s--;
							site[j].x = site[s].x;
							site[j].y = site[s].y;
						}
		}
	}

	if (s < 3) {
		fprintf(stderr, "Less than three valid coordinate pairs entered!\n");
		free(site);
		return RET_FORTUNE_NO_TRIANGLES;
	}

	sites = s;

	/* run Lloyd's algorithm */

	i = Lloyd(site, sites, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
	if (i != RET_FORTUNE_OK)
		goto clean_up_and_exit;

	/* print results in stdout;
	     the site array is printed backwards
	     because the order of the sites has
	     been reversed when reading the
	     input */

	for (i = sites; i; ) {
		i--;
		printf("%10.6f,%10.6f\n", site[i].x, site[i].y);
	}

	i = RET_FORTUNE_OK;

	clean_up_and_exit:

	free(site);

	return i;
}

