/*************************************************************************
 *
 *  An implementation of Fortune's algorithm
 *  Copyright (C) 2012-2014 Jussi M. Hukkanen
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *************************************************************************
 *
 *  fortune.c: main module
 *
 ************************************************************************/

#define FORTUNE_PUBLIC
#include "fortune.h"

#if 1
#define DEBUG_CELLS 1
#include <stdio.h>
#endif


//#define FIND_ZERO_LENGTH_BOUNDARIES

#if 1
#define VERTEX_DISTANCE_MIN  VERTEX_COORD_EPSILON
#elif 1
#define VERTEX_DISTANCE_MIN (VERTEX_COORD_EPSILON * 5.0)
#else
#define VERTEX_DISTANCE_MIN (VERTEX_COORD_EPSILON * 10.0)
#endif


#ifdef FORTUNE_DEBUG_MODE
static uint16_t next_debug_id;
static uint16_t next_triangle_debug_id;
#endif

static char delete_circle(event_t *, beach_arc_t *);

static uint32_t triangles;
static uint32_t conns;

#ifndef FORTUNE_NO_VORONOI
static char create_vertices;
static uint32_t vertices;
static int cells;
static celldata_t *cell_master = NULL;
#endif

#ifdef FORTUNE_MASTER_ARRAYS

#ifndef FORTUNE_NO_VORONOI

static vertex_t *vertex_master = NULL;
static boundary_t *boundary_master = NULL;
static cellboundary_t *cellboundary_master = NULL;
static int vertex_master_size = 0;
static int boundary_master_size = 0;
static int cellboundary_master_size = 0;
static int cellboundary_master_ptr = 0;
static int cell_master_size = 0;

/* destroy_Voronoi_masters()

    releases the memory allocated to the
      vertex, boundary and cell master
      arrays;
    this function is called automagically
      at program termination */

void destroy_Voronoi_masters(void) {

    if (vertex_master != NULL)
        free(vertex_master);
    if (boundary_master != NULL)
        free(boundary_master);
    if (cell_master != NULL)
        free(cell_master);
    if (cellboundary_master != NULL)
        free(cellboundary_master);

    vertex_master = NULL;
    boundary_master = NULL;
    cellboundary_master = NULL;
    cell_master = NULL;
    vertex_master_size = 0;
    boundary_master_size = 0;
    cellboundary_master_size = 0;
    cellboundary_master_ptr = 0;
    cell_master_size = 0;
}

#endif

#ifndef FORTUNE_NO_DELAUNAY

static triangle_t *triangle_master = NULL;
static connect_t *connect_master = NULL;
static int triangle_master_size = 0;
static int connect_master_size = 0;

/* destroy_Delaunay_masters()

    releases the memory allocated to the
      triangle and connection master
      arrays;
    this function is called automagically
      at program termination */

void destroy_Delaunay_masters(void) {

    if (connect_master != NULL)
        free(connect_master);
    if (triangle_master != NULL)
        free(triangle_master);

    triangle_master = NULL;
    connect_master = NULL;
    triangle_master_size = 0;
    connect_master_size = 0;
}

#endif

#endif  /* FORTUNE_MASTER_ARRAYS */

/**** TODO: it might be a good idea to have
        all internal data of the algorithm
        (the event queue and beach line as
        well as vertex, triangle, circle and
        connection lists) in master arrays ****/


/* add_event()

    adds a new site event or circle event
      in the event queue;
    returns a pointer to the event or NULL
      if there is not enough memory */

static event_t *add_event(vertex_coord_t x, vertex_coord_t y, char type) {

    event_t *e;
    event_t *q;
    event_t *p;
    char found;

    e = (event_t *)malloc(sizeof(event_t));

    if (e == NULL)
      return NULL;

    e->sdata = NULL;
    e->cdata = NULL;

    switch (type) {

      case FORTUNE_SITE_EVENT:
          e->sdata = (site_event_t *)malloc(sizeof(site_event_t));
          if (e->sdata == NULL) {
            free(e);
            return NULL;
          }
          break;

      default:
          e->cdata = (circle_event_t *)malloc(sizeof(circle_event_t));
          if (e->cdata == NULL) {
            free(e);
            return NULL;
          }
          break;
    }


    e->type = type;
    e->x = x;
    e->y = y;

    #define COMPARE_WITH_EVENT(E, ACTION) {            \
        if (E == NULL) {                               \
          ACTION;                                      \
        }                                              \
        else                                           \
          if (E->y > y) {                              \
            E->prev = e;                               \
            ACTION;                                    \
          }                                            \
          else                                         \
            if (y > E->y - VERTEX_COORD_EPSILON)       \
              if (y < E->y + VERTEX_COORD_EPSILON)     \
                /* y coordinate equal */               \
                if (x < E->x + VERTEX_COORD_EPSILON) { \
                  E->prev = e;                         \
                  ACTION;                              \
                }                                      \
    }

    Fortune_events_left++;

    found = 0;

    COMPARE_WITH_EVENT(event_queue, { found = 1; } );

    if (found) {
      e->prev = NULL;
      e->next = event_queue;
      event_queue = e;
      return e;
    }

    p = event_queue;

    while (1) {
      q = p->next;
      COMPARE_WITH_EVENT(q, break);
      p = q;
    }

    p->next = e;
    e->prev = p;
    e->next = q;

    return e;
}


/* delete_event()

    deletes the given event e from the queue */

static char delete_event(event_t *e) {

    if (e->type == FORTUNE_CIRCLE_EVENT) {
      delete_circle(e, e->cdata->left);
      delete_circle(e, e->cdata->middle);
      delete_circle(e, e->cdata->right);
    }

    if (e == event_queue) {

      if (e->next != NULL) {
        event_queue = e->next;
        event_queue->prev = NULL;
      }
      else
        event_queue = NULL;
    }
    else {

      e->prev->next = e->next;

      if (e->next != NULL)
        e->next->prev = e->prev;
    }

    if (e->sdata != NULL)
      free(e->sdata);
    if (e->cdata != NULL)
      free(e->cdata);

    free(e);

    Fortune_events_left--;

    return RET_FORTUNE_OK;
}


/* add_beach_arc()

    adds an arc segment to the end of the beach line;
      cell is a pointer to the Voronoi cell around
      site s;
    returns a pointer to the beach_arc_t structure
      or NULL if there is not enough memory */

static beach_arc_t *add_beach_arc(site_t *s
                                  #ifndef FORTUNE_NO_VORONOI
                                  , celldata_t *cell
                                  #endif
                                  ) {

    beach_arc_t *b;
    beach_arc_t *p;
    beach_arc_t *q;

    b = (beach_arc_t *)malloc(sizeof(beach_arc_t));

    if (b == NULL)
      return NULL;

    b->s = s;
    b->split_left  = NULL;
    b->split_right = NULL;
    b->circles = NULL;

    #ifndef FORTUNE_NO_VORONOI
    b->cell = cell;
    #endif

    #ifdef FORTUNE_DEBUG_MODE
    b->debug_id = next_debug_id++;
    #endif

    if (beach_line == NULL) {
      b->left = NULL;
      beach_line = b;
    }
    else {
      for (p = beach_line; p; ) {
         q = p->right;
         if (q == NULL)
           break;
         p = q;
      }
      p->right = b;
      b->left = p;
    }

    b->right = NULL;

    return b;
}


/* delete_beach_arc()

    deletes the given arc segment from the beach line */

static char delete_beach_arc(beach_arc_t *b) {

    circle_list_t *c;
    circle_list_t *cn;

    if (b == beach_line) {

      beach_line = b->right;

      if (beach_line != NULL)
        beach_line->left = NULL;
    }
    else {

      b->left->right = b->right;

      if (b->right != NULL)
        b->right->left = b->left;
    }

    for (c = b->circles; c != NULL; ) {
       cn = c->next;
       free(c);
       c = cn;
    }

    free(b);

    return RET_FORTUNE_OK;
}


/* split_beach_arc()

    splits the beach arc segment p in two and
      creates a new segment between the two parts;
      cell is a pointer to the Voronoi cell around
      site s;
    returns a pointer to the beach_arc_t structure
      or NULL if there is not enough memory */

static beach_arc_t *split_beach_arc(site_t *s
                                    #ifndef FORTUNE_NO_VORONOI
                                    , celldata_t *cell
                                    #endif
                                    , beach_arc_t *p) {

    beach_arc_t *b;
    beach_arc_t *c;
    beach_arc_t *q;

    b = (beach_arc_t *)malloc(sizeof(beach_arc_t));

    if (b == NULL)
        return NULL;

    b->s = s;

    q = p->right;
    b->left  = p;
    p->right = b;
    b->right = q; /* just in case we run out of memory */
    b->circles = NULL;

    c = (beach_arc_t *)malloc(sizeof(beach_arc_t));

    if (c == NULL)
        return NULL;

    c->s = p->s;

    #ifndef FORTUNE_NO_VORONOI
    b->cell = cell;
    c->cell = p->cell;
    #endif

    b->right = c;
    c->left  = b;
    c->right = q;

    b->split_right = NULL;
    b->split_left  = NULL;

    if (q != NULL) {
      q->left = c;
      q->split_left = s;
    }
    if (p->left != NULL) {
      p->left->split_right = s;
    }

    c->split_right = p->split_right;
    p->split_right = s;
    c->split_left  = s;

    c->circles = NULL;

    #ifdef FORTUNE_DEBUG_MODE
    b->debug_id = next_debug_id++;
    c->debug_id = next_debug_id++;
    #endif

    return b;
}


/* get_beach_arc()

    finds the beach arc segment in the beach line
      at the given x coordinate, when the sweep
      line is at the given y coordinate */

static beach_arc_t *get_beach_arc(site_coord_t x, site_coord_t y) {

    beach_arc_t *b;
    beach_arc_t *c;
    site_coord_t xd, yd;
    double yarc, ymin;

    b = beach_line;

    if (b == NULL)
        return NULL;

    ymin = -1.0;
    c = NULL;

    for ( ; b; b = b->right) {

        if (b->split_left != NULL)
            if (x < b->split_left->x)
                continue;

        if (b->split_right != NULL)
            if (x >= b->split_right->x)
                continue;

        xd = x - b->s->x;
        yd = y - b->s->y;
        if (yd > -SITE_COORD_EPSILON && yd < SITE_COORD_EPSILON)
            continue;

        yarc = ((xd * xd) / (2.0 * yd)) + (yd / 2.0);

        if (ymin < 0 || yarc < ymin) {
            ymin = yarc;
            c = b;
        }
    }

    return c;
}


/* positive_side_of_line()

    returns nonzero if site sc is on the
      'positive' side of a line drawn from
      site sa to site sb;
    the 'positive' side of a line is the
      right side when FORTUNE_POSITIVE_Y_DOWN
      is defined and the left side when
      FORTUNE_POSITIVE_Y_UP is defined */

static char positive_side_of_line(site_t *sa, site_t *sb, site_t *sc) {

    site_coord_t x1, y1;
    site_coord_t x2, y2;

    x1 = sa->x; y1 = sa->y;
    x2 = sb->x; y2 = sb->y;

    return (LINE_TEST(x1, y1, x2, y2, sc->x, sc->y) > 0) ? 1 : 0;
}


#if 0

/* find_connection()

    returns a pointer to the connection
      between sites sa and sb if such a
      connection exists and NULL otherwise */

static connect_list_t *find_connection(site_t *sa, site_t *sb) {

    connect_list_t *cl;

    for (cl = connect_list; cl != NULL; cl = cl->next) {
        if (cl->s1 == sa)
            if (cl->s2 == sb)
                return cl;
        if (cl->s1 == sb)
            if (cl->s2 == sa)
                return cl;
    }

    return NULL;
}

#endif


/* add_triangle()

    creates a new triangle list item and
      returns a pointer to it;
    sa, sb and sc are beach arc segments whose
      corresponding sites form the corners
      of the triangle;
    add_triangle() orders the sites by
      y coordinate so that site[0] is the
      topmost and site[1] and site[2] are
      in clockwise order */

static triangle_list_t *add_triangle(beach_arc_t *sa, beach_arc_t *sb, beach_arc_t *sc) {

    site_t *sitea = sa->s;
    site_t *siteb = sb->s;
    site_t *sitec = sc->s;
    triangle_list_t *t;
    triangle_list_t *n;
    site_coord_t x1, y1;
    site_coord_t x2, y2;
    connect_list_t *c;
    uint8_t cn;

    t = (triangle_list_t *)malloc(sizeof(triangle_list_t));

    if (t == NULL)
        return NULL;

    /* put the sites in clockwise order */

    #define ASSIGN_TRIANGLE_SITES_ONLY(S0, S1, S2) { \
        t->site[0] = S0->s;                          \
        t->site[1] = S1->s;                          \
        t->site[2] = S2->s;                          \
        goto triangle_sites_ordered;                 \
    }

    #ifdef FORTUNE_NO_VORONOI
    #define ASSIGN_TRIANGLE_SITES(S0, S1, S2)   \
        ASSIGN_TRIANGLE_SITES_ONLY(S0, S1, S2)
    #else
    #define ASSIGN_TRIANGLE_SITES(S0, S1, S2) { \
        t->cell[0] = S0->cell;                  \
        t->cell[1] = S1->cell;                  \
        t->cell[2] = S2->cell;                  \
        ASSIGN_TRIANGLE_SITES_ONLY(S0, S1, S2); \
    }
    #endif

    #define SITE_DIFF(S1, S2, S3) {                         \
        x1 = S1->s->x; y1 = S1->s->y;                       \
        x2 = S2->s->x; y2 = S2->s->y;                       \
        if (RIGHT_SIDE(x1, y1, x2, y2, S3->s->x, S3->s->y)) \
          ASSIGN_TRIANGLE_SITES(S1, S2, S3);                \
        ASSIGN_TRIANGLE_SITES(S1, S3, S2);                  \
    }

    if (ABOVE_OR_EQUAL(sitea, siteb)) {
      if (ABOVE_OR_EQUAL(sitea, sitec))
        SITE_DIFF(sa, sb, sc);  /* sa on top */
      SITE_DIFF(sc, sa, sb);    /* sc on top */
    }

    if (sitea->y > siteb->y - SITE_COORD_EPSILON &&
        sitea->y < siteb->y + SITE_COORD_EPSILON) {
      if (sitea->x <= siteb->x - SITE_COORD_EPSILON) {

        /* sa (left) and sb (right) on same line */

        if (ABOVE_OR_EQUAL(sitec, sitea))
          ASSIGN_TRIANGLE_SITES(sc, sb, sa);
        ASSIGN_TRIANGLE_SITES(sa, sb, sc);
      }

      /* sb (left) and sa (right) on same line */

      if (ABOVE_OR_EQUAL(sitec, sitea))
        ASSIGN_TRIANGLE_SITES(sc, sa, sb);
      ASSIGN_TRIANGLE_SITES(sb, sa, sc);
    }

    if (ABOVE_OR_EQUAL(siteb, sitec)) {
      if (ABOVE_OR_EQUAL(siteb, sitea))
        SITE_DIFF(sb, sa, sc);  /* sb on top */
      SITE_DIFF(sa, sb, sc);    /* sa on top */
    }

    if (siteb->y > sitec->y - SITE_COORD_EPSILON &&
        siteb->y < sitec->y + SITE_COORD_EPSILON) {
      if (siteb->x <= sitec->x - SITE_COORD_EPSILON) {

        /* sb (left) and sc (right) on same line */

        if (ABOVE_OR_EQUAL(sitea, siteb))
          ASSIGN_TRIANGLE_SITES(sa, sc, sb);
        ASSIGN_TRIANGLE_SITES(sb, sc, sa);
      }

      /* sc (left) and sb (right) on same line */

      if (ABOVE_OR_EQUAL(sitea, siteb))
        ASSIGN_TRIANGLE_SITES(sa, sb, sc);
      ASSIGN_TRIANGLE_SITES(sc, sb, sa);
    }

    if (ABOVE_OR_EQUAL(sitec, sitea)) {
      if (ABOVE_OR_EQUAL(sitec, siteb))
        SITE_DIFF(sc, sa, sb);  /* sc on top */
      SITE_DIFF(sb, sa, sc);    /* sb on top */
    }

    if (sitec->y > sitea->y - SITE_COORD_EPSILON &&
        sitec->y < sitea->y + SITE_COORD_EPSILON) {
      if (sitec->x <= sitea->x - SITE_COORD_EPSILON) {

        /* sc (left) and sa (right) on same line */

        if (ABOVE_OR_EQUAL(siteb, sitec))
          ASSIGN_TRIANGLE_SITES(sb, sa, sc);
        ASSIGN_TRIANGLE_SITES(sc, sa, sb);
      }

      /* sa (left) and sc (right) on same line */

      if (ABOVE_OR_EQUAL(siteb, sitec))
        ASSIGN_TRIANGLE_SITES(sb, sc, sa);
      ASSIGN_TRIANGLE_SITES(sa, sc, sb);
    }

    triangle_sites_ordered:

    /* assign connections and links to neighbour triangles;
       stop looking for neighbours as soon as three have been found */

    #define ASSIGN_NEIGHBOUR_ONLY(THIS, THAT) { \
        t->neighbour[THIS] = n;                 \
        n->neighbour[THAT] = t;                 \
        t->neighbour_side[THIS] = THAT;         \
        n->neighbour_side[THAT] = THIS;         \
        cn++;                                   \
    }

    /*
    #ifdef FORTUNE_NO_DELAUNAY
    #define ASSIGN_NEIGHBOUR(THIS, THAT)    \
        ASSIGN_NEIGHBOUR_ONLY(THIS, THAT)
    #else
    #define ASSIGN_NEIGHBOUR(THIS, THAT) {  \
        ASSIGN_NEIGHBOUR_ONLY(THIS, THAT);  \
        t->conn[THIS] = n->conn[THAT];      \
    }
    #endif
    */
    #define ASSIGN_NEIGHBOUR(THIS, THAT) {  \
        ASSIGN_NEIGHBOUR_ONLY(THIS, THAT);  \
        t->conn[THIS] = n->conn[THAT];      \
    }

    t->neighbour[0] = NULL;
    t->neighbour[1] = NULL;
    t->neighbour[2] = NULL;

    t->conn[0] = NULL;
    t->conn[1] = NULL;
    t->conn[2] = NULL;

    for (n = triangle_list, cn = 0; n != NULL && cn < 3; n = n->next) {
        if (t->site[0] == n->site[0])
            if (t->site[1] == n->site[2])
                ASSIGN_NEIGHBOUR(0, 2);
        if (t->site[0] == n->site[2])
            if (t->site[1] == n->site[1])
                ASSIGN_NEIGHBOUR(0, 1);
        if (t->site[1] == n->site[0])
            if (t->site[2] == n->site[2])
                ASSIGN_NEIGHBOUR(1, 2);
        if (t->site[1] == n->site[2])
            if (t->site[2] == n->site[1])
                ASSIGN_NEIGHBOUR(1, 1);
        if (t->site[1] == n->site[1])
            if (t->site[2] == n->site[0])
                ASSIGN_NEIGHBOUR(1, 0);
        if (t->site[2] == n->site[2])
            if (t->site[0] == n->site[1])
                ASSIGN_NEIGHBOUR(2, 1);
        if (t->site[2] == n->site[1])
            if (t->site[0] == n->site[0])
                ASSIGN_NEIGHBOUR(2, 0);
    }

    /* create new connections as needed */

    for (cn = 0; cn < 3; cn++) {

        if (t->conn[cn] != NULL)
            continue;

        c = (connect_list_t *)malloc(sizeof(connect_list_t));

        if (c == NULL) {
            free(t);
            return NULL;
        }

        t->conn[cn] = c;

        c->s1 = t->site[cn];
        c->s2 = t->site[(cn + 1) % 3];

        #ifndef FORTUNE_NO_VORONOI
        c->v1 = NULL;   /* set by update_boundaries() */
        c->v2 = NULL;
        c->index = conns;
        c->side = cn;
        #endif

        c->next = connect_list;
        connect_list = c;
        conns++;
    }

    /* update triangle_list */

    t->next = triangle_list;
    triangle_list = t;

    #ifdef FORTUNE_DEBUG_MODE
    t->debug_id = next_triangle_debug_id++;
    #endif

    triangles++;

    return t;
}


/* destroy_triangle_list()

    releases the memory occupied by
      triangle_list and connect_list */

static char destroy_triangle_list(void) {

    triangle_list_t *t;
    connect_list_t *c;

    while (triangle_list != NULL) {
      t = triangle_list->next;
      free(triangle_list);
      triangle_list = t;
    }

    while (connect_list != NULL) {
      c = connect_list->next;
      free(connect_list);
      connect_list = c;
    }

    triangles = 0;
    conns = 0;

    return RET_FORTUNE_OK;
}


#ifndef FORTUNE_NO_VORONOI

/* vertices_equal_or_too_close()

    returns nonzero if the vertex coordinate
      pairs (x1,y1) and (x2,y2) are within
      the VERTEX_DISTANCE_MIN limit from
      each other */

static char vertices_equal_or_too_close(vertex_coord_t x1, vertex_coord_t y1,
                                        vertex_coord_t x2, vertex_coord_t y2) {

    if (x1 > x2 - VERTEX_DISTANCE_MIN)
        if (x1 < x2 + VERTEX_DISTANCE_MIN)
            if (y1 > y2 - VERTEX_DISTANCE_MIN)
                if (y1 < y2 + VERTEX_DISTANCE_MIN)
                    return 1;

    return 0;
}


/* add_vertex()

    adds a new vertex in the vertex list;
    returns a pointer to the vertex or NULL
      if there is not enough memory;
    if the list already contains a vertex in
      the exact same coordinates, add_vertex()
      returns that one instead of creating a
      new one */

static vertex_list_t *add_vertex(vertex_coord_t x, vertex_coord_t y) {

    vertex_list_t *v;

    #ifndef FIND_ZERO_LENGTH_BOUNDARIES

    for (v = vertex_list; v != NULL; v = v->next) {
        if (vertices_equal_or_too_close(v->x, v->y, x, y)) {
            #if DEBUG_CELLS > 1
            fprintf(stderr, "vertex already exists at %10.6f,%10.6f\n", v->x, v->y);
            #endif
            return v;
        }
    }

    #endif

    v = (vertex_list_t *)malloc(sizeof(vertex_list_t));

    if (v == NULL)
      return NULL;

    v->x = x;
    v->y = y;

    v->index = vertices;
    vertices++;

    v->next = vertex_list;
    vertex_list = v;

    return v;
}


/* add_triangle_to_loop()

    inserts the triangle t at the correct
      position in the boundary loop of cell;
    cl and cr are the left and right
      connections in t, looking from the site
      in cell;
    as a special case, t can be NULL when
      a placeholder for an infinite boundary
      (but only one) needs to be inserted
      into the loop of an infinite cell */

static boundary_loop_t *add_triangle_to_loop(triangle_list_t *t,
                                             celldata_t *cell,
                                             connect_list_t *cl,
                                             connect_list_t *cr) {

    boundary_loop_t *b;
    boundary_loop_t *p;
    boundary_loop_t *n;

    b = (boundary_loop_t *)malloc(sizeof(boundary_loop_t));

    if (b == NULL)
        return NULL;

    b->tri = t;
    b->conn_l = cl;
    b->conn_r = cr;
    b->left = NULL;
    b->right = NULL;
    b->first = 0;

    if (cell->bloop == NULL) {      /* first one in the loop */
        cell->sides = 0;
        goto boundary_loop_updated;
    }

    /* see if neighbours of t already exist in the loop */

    for (p = cell->bloop; p; p = p->next) {

        if (p->conn_r == cl) {
            b->left = p;
            p->right = b;
        }

        if (p->conn_l == cr) {
            b->right = p;
            p->left = b;
        }
    }

    boundary_loop_updated:

    b->next = cell->bloop;
    cell->bloop = b;
    cell->sides++;

    return b;
}


/* update_boundaries()

    adds the new triangle t and the corresponding
      connections to the boundary loops of the
      three sites that form t */

static char update_boundaries(triangle_list_t *t) {

    uint8_t cl; /* left connection, looking from site */

    /* update vertex information for the
         Voronoi boundaries that correspond
         to the three connections */

    for (cl = 0; cl < 3; cl++) {

        if (t->neighbour[cl] == NULL) { /* no neighbour -> infinite boundary */

            t->conn[cl]->v1 = t->v;
            t->conn[cl]->v2 = NULL;
        }
        else {                         /* neighbour found -> finite boundary */

            if (t->conn[cl]->v2 == NULL)
                t->conn[cl]->v2 = t->v;
            else
                t->conn[cl]->v1 = t->v;
        }
    }

    /* update boundary loops of the three sites */

    for (cl = 0; cl < 3; cl++) {

        if (add_triangle_to_loop(t, t->cell[cl], t->conn[cl], t->conn[(cl + 2) % 3]) == NULL)
            return RET_FORTUNE_NO_MEM;
    }

    return RET_FORTUNE_OK;
}


/* count_boundaries()

    when all events have been handled, this function
      must be called to ensure the number of cell
      boundaries in each cell is correct;
    add_triangle_to_loop() keeps track of boundaries
      in the loop of a cell and updates the sides
      field in celldata_t, but this value must be
      incremented by one if the cell is infinite;
    count_boundaries() reads the boundary loop of
      cell and updates the sides and infinity fields
      as needed;
    count_boundaries() also sets the starting point
      of the boundary loop; with finite cells, this
      is the topmost vertex in the loop, whereas
      the starting point of an infinite cell is
      either this vertex or, if one or two infinite
      boundaries point 'up', the rightmost such
      boundary */

static char count_boundaries(celldata_t *cell) {

    boundary_loop_t *b;
    boundary_loop_t *bn;
    boundary_loop_t *i_start;   /* first infinite boundary in loop */
    boundary_loop_t *i_end;     /* last infinite boundary in loop */
    vertex_coord_t ytop;        /* y coordinate of topmost vertex */
    int i, j;
    uint8_t side;
    char ret = RET_FORTUNE_OK;
    char k;

    if (cell->bloop == NULL) {
        #ifdef DEBUG_CELLS
        fprintf(stderr, "!!! no boundaries\n");
        #endif
        return RET_FORTUNE_PARAM_ERROR;
    }

    /* i = number of NULL left links;
       j = number of NULL right links;
       if both are zero, the cell is finite;
         otherwise both must be one */

    for (i = j = 0, i_start = i_end = NULL, b = cell->bloop;
        b != NULL; b = b->next) {

        if (b->left == NULL) {
            i++;
            i_start = b;
        }

        if (b->right == NULL) {
            j++;
            i_end = b;
        }
    }

    if (i == 0 && j == 0) {

        /* finite */

        cell->infinity = 0;
    }
    else {

        if (i != 1 || j != 1 || i_start == NULL || i_end == NULL) {
            #ifdef DEBUG_CELLS
            fprintf(stderr, "!!! boundaries incorrectly linked\n");
            #endif
            return RET_FORTUNE_PARAM_ERROR;
        }

        /* infinite */

        cell->infinity = 1;

        /* add an empty placeholder for the rightmost infinite boundary;
           this way the cell->sides gets incremented and there will
             be exactly cell->sides items in the loop;
           the conn_r member in the placeholder is the conn_l of
             the neighbour triangle */

        b = add_triangle_to_loop(NULL, cell, NULL, i_start->conn_l);

        if (b == NULL) {
            #ifdef DEBUG_CELLS
            fprintf(stderr, "!!! out of memory when inserting placeholder into boundary loop\n");
            #endif
            return RET_FORTUNE_NO_MEM;
        }

        i_start = b;

        i_start->left = i_end;
        i_end->right = i_start;

        j = 0;

        /* checking the direction of an infinite boundary
             based on the side field of connect_list_t
             is acceptable, because the connection does
             not link two triangles together */

        #define ACCEPT_UPWARD_INF(BOUNDARY) {           \
            BOUNDARY->first = 1;                        \
            j = 1;                                      \
        }

        #define CHECK_UPWARD_INF(BOUNDARY, TRI, CONN) { \
            side = CONN->side;                          \
            if (TRI->site[side]->x <                    \
                TRI->site[(side + 1) % 3]->x)           \
                ACCEPT_UPWARD_INF(BOUNDARY);            \
        }

        CHECK_UPWARD_INF(i_start, i_start->right->tri, i_start->conn_r);

        if (!j)
            CHECK_UPWARD_INF(i_end, i_end->tri, i_end->conn_r);
    }

    /* look for zero-length boundaries and remove them */

    for (b = cell->bloop, i = cell->sides; i > 0; i--) {

        if (b->conn_r->v2 == NULL) {
            b = b->right;
            continue;
        }

        k = 0;
        #ifdef FIND_ZERO_LENGTH_BOUNDARIES
        if (vertices_equal_or_too_close(b->conn_r->v1->x, b->conn_r->v1->y,
            b->conn_r->v2->x, b->conn_r->v2->y))
            k = 1;
        #else
        if (b->conn_r->v1 == b->conn_r->v2)
            k = 1;
        #endif

        if (k) {

            cell->sides--;

            b->left->right = b->right;
            b->right->left = b->left;

            bn = b->right;

            if (cell->bloop == b) {
                free(cell->bloop);
                cell->bloop = bn;
            }

            #if DEBUG_CELLS > 1
            fprintf(stderr, "!!! connection %d creates a zero-length boundary\n",
                b->conn_r->index);
            #endif

            b = bn;

            continue;
        }

        b = b->right;
    }

    if (!j) {

        /* set topmost vertex as the starting point;
           the top edge of the cell may consist of one
             or more perfectly horizontal boundaries;
             in such a case, we need to find the leftmost
             vertex with that y coordinate */

        b = cell->bloop;
        if (b->tri == NULL)
            b = b->right;

        ytop = b->tri->v->y;

        for (i = cell->sides - 1; i > 0; i--) {

            bn = b->right;
            if (bn->tri == NULL)
                bn = bn->right;

            if (!Y_ABOVE(bn->tri->v->y, ytop))
                break;

            b = bn;
            ytop = b->tri->v->y;
        }

        for (i = cell->sides - 1; i > 0; i--) {

            bn = b->left;
            if (bn->tri == NULL) {
                if (bn == i_start)
                    break;
                bn = bn->left;
            }

            if (Y_ABOVE(ytop, bn->tri->v->y))
                break;

            b = bn;
            ytop = b->tri->v->y;
        }

        b->first = 1;
    }

    return ret;
}


/* convert_boundaries()

    creates a list of cellboundary_t objects
      based on the boundary loop of the given
      cell and attaches it to the cell;
    ba is the list of Voronoi boundaries that
      the cell boundaries should be linked to */

static char convert_boundaries(celldata_t *cell, boundary_t *ba) {

    boundary_loop_t *b;
    connect_list_t *c;
    cellboundary_t *n;
    cellboundary_t *prev;
    int i, j;
    uint8_t side;
    char ret = RET_FORTUNE_OK;

    if (cell->error) {
        #ifdef DEBUG_CELLS
        fprintf(stderr, "!!! broken cell structure, skipping\n");
        #endif
        return RET_FORTUNE_PARAM_ERROR;
    }

    #if 0
    #if DEBUG_CELLS > 1
    fprintf(stderr, "!!! bloop:");
    for (b = cell->bloop, i = cell->sides; i > 0; i--, b = b->right) {
        fprintf(stderr, " c%d", b->conn_r->index);
        if (b->first)
            fprintf(stderr, "[1]");
    }
    fprintf(stderr, "\n");
    ret = RET_FORTUNE_PARAM_ERROR;
    #endif
    #endif

    #ifdef FORTUNE_MASTER_ARRAYS
    #define CREATE_CELLBOUNDARY(TARGET) {                       \
        TARGET = &cellboundary_master[cellboundary_master_ptr]; \
        TARGET->i = cellboundary_master_ptr;                    \
        cellboundary_master_ptr++;                              \
    }
    #else
    #define CREATE_CELLBOUNDARY(TARGET) {                       \
        TARGET = (cellboundary_t *)                             \
            malloc(sizeof(cellboundary_t));                     \
        if (TARGET == NULL)                                     \
            return RET_FORTUNE_NO_MEM;                          \
    }
    #endif

    CREATE_CELLBOUNDARY(cell->blist);

    for (prev = cell->blist, i = cell->sides - 1; i > 0; i--) {
        CREATE_CELLBOUNDARY(n);
        n->left = prev;
        prev->right = n;
        prev = n;
    }

    prev->right = cell->blist;
    cell->blist->left = prev;

    for (b = cell->bloop, i = cell->sides; i > 0; i--, b = b->right)
        if (b->first) break;

    #ifdef DEBUG_CELLS
    if (!i) {
        fprintf(stderr, "!!! no boundary marked first in the loop\n");
        return RET_FORTUNE_PARAM_ERROR;
    }
    #endif

    i = cell->sides;
    n = cell->blist;

    for ( ; i > 0; i--, b = b->right, n = n->right) {

        c = b->conn_r;
        n->boundary = &ba[c->index];
        n->neighbour = NULL;

        if (b->tri == NULL) {

            for (side = 0; side < 3; side++) {
                if (cell->s == b->right->tri->site[side]) {
                    n->neighbour = b->right->tri->cell[(side + 1) % 3];
                    break;
                }
            }
        }
        else {

            for (side = 0; side < 3; side++) {
                if (cell->s == b->tri->site[side]) {
                    n->neighbour = b->tri->cell[(side + 2) % 3];
                    break;
                }
            }
        }
    }

    return ret;
}


/* destroy_vertex_list()

    releases the memory occupied by the
      vertex list and boundary list, as
      well as each site's cellboundary loop */

static char destroy_vertex_list(void) {

    vertex_list_t *n;
    boundary_loop_t *c;
    int i;

    while (vertex_list != NULL) {
        n = vertex_list->next;
        free(vertex_list);
        vertex_list = n;
    }

    vertices = 0;

    for (i = 0; i < cells; i++) {
        while (cell_master[i].bloop != NULL) {
            c = cell_master[i].bloop;
            c = c->next;
            free(cell_master[i].bloop);
            cell_master[i].bloop = c;
        }
    }

    return RET_FORTUNE_OK;
}


#ifndef FORTUNE_MASTER_ARRAYS

/* destroy_cellboundary_list()

    releases the memory occupied by the
      cell boundary list of cell c */

char destroy_cellboundary_list(celldata_t *c) {

    cellboundary_t *start = c->blist;
    cellboundary_t *b;
    cellboundary_t *n;
    int i;

    if (start == NULL)
      return RET_FORTUNE_OK;

    for (b = start, i = c->sides; i; i--) {
      n = b->right;
      free(b);
      if (n == NULL)
        break;
      if (n == start)
        break;
      b = n;
    }

    c->blist = NULL;
    c->sides = 0;

    return RET_FORTUNE_OK;
}

#endif      /* !FORTUNE_MASTER_ARRAYS */

#endif      /* !FORTUNE_NO_VORONOI */


/* add_circle()

    adds a link to the given circle event e
      to the circle event list of beach arc b */

static circle_list_t *add_circle(event_t *e, beach_arc_t *b) {

    circle_list_t *c;

    c = (circle_list_t *)malloc(sizeof(circle_list_t));

    if (c == NULL)
      return NULL;

    c->next = b->circles;
    c->e = e;
    b->circles = c;

    return c;
}


/* delete_circle()

    deletes the link to the given circle event e
      from the list of beach arc b */

static char delete_circle(event_t *e, beach_arc_t *b) {

    circle_list_t *c;
    circle_list_t *n;
    circle_list_t *p;

    if (b == NULL)
      return RET_FORTUNE_OK;

    for (c = b->circles, p = NULL; ; c = n) {

       if (c == NULL)
         break;

       n = c->next;

       if (c->e != e) {
         p = c;
         continue;
       }

       if (p == NULL)
         b->circles = n;
       else
         p->next = n;

       free(c);

       /* there are not supposed to be more than one
            link to a certain circle event in a beach
            arc's list, which means that we should be
            able to stop when the first such link is
            found; however, let's play it safe */
    }

    return RET_FORTUNE_OK;
}


/* make_line()

    fills the line_t structure pointed to by line
      with the equation of a line equidistant
      from sites sa and sb */

static char make_line(line_t *line, site_t *sa, site_t *sb) {

    double dx = sb->x - sa->x;
    double dy = sb->y - sa->y;

    line->A = dx;
    line->B = dy;
    line->C = (-dx * (sa->x + (dx / 2.0))) + (-dy * (sa->y + (dy / 2.0)));

    return RET_FORTUNE_OK;
}


/* get_intersection()

    finds the intersection of two lines: the one
      between the beach arcs middle->left and
      middle and the one between middle and
      middle->right;
    xt, yt and yct should point to vertex_coord_t
      variables in which information about the
      resulting circle event is stored;
    xt and yt receive the coordinates of the
      event and yct receives the y coordinate
      of the circle's center;
    if the lines do not intersect,
      RET_FORTUNE_PARALLEL is returned */

static char get_intersection(beach_arc_t *middle,
                             vertex_coord_t *xt, vertex_coord_t *yt,
                             vertex_coord_t *yct) {

    vertex_coord_t x, y, yc;
    line_t *line1;
    line_t *line2;
    double k1, c1, k2, c2;

    line1 = &(middle->left->line);
    line2 = &(middle->line);

    #define DBL_ZERO(X) ((X) > -DBL_EPSILON && (X) < DBL_EPSILON)

    if (DBL_ZERO(line1->B)) {        /* line1 vertical */

      if (DBL_ZERO(line2->B))
        return RET_FORTUNE_PARALLEL;

      if (DBL_ZERO(line1->A)) /* impossible! */
        return RET_FORTUNE_PARALLEL;

      x = -line1->C / line1->A;
      y = (line2->C + (line2->A * x)) / (-line2->B);
    }
    else {

      if (DBL_ZERO(line2->B)) {      /* line2 vertical */

        if (DBL_ZERO(line2->A)) /* impossible! */
          return RET_FORTUNE_PARALLEL;

        x = -line2->C / line2->A;
        y = (line1->C + (line1->A * x)) / (-line1->B);
      }
      else {

        if (DBL_ZERO(line1->A)) {    /* line1 horizontal */

          if (DBL_ZERO(line2->A))
            return RET_FORTUNE_PARALLEL;

          y = -line1->C / line1->B;
          x = (line2->C + (line2->B * y)) / (-line2->A);
        }
        else {

          if (DBL_ZERO(line2->A)) {  /* line2 horizontal */

            y = -line2->C / line2->B;
            x = (line1->C + (line1->B * y)) / (-line1->A);
          }
          else {

            /* the vast majority of cases:
                 line1->A, line1->B, line2->A and line2->B
                 are all nonzero */

            k1 = -line1->A / line1->B;
            k2 = -line2->A / line2->B;
            k1 -= k2;

            if (DBL_ZERO(k1))       /* parallel */
              return RET_FORTUNE_PARALLEL;

            c1 = line1->C / line1->B;
            c2 = line2->C / line2->B;

            x = (c1 - c2) / k1;
            y = k2 * x - c2;
          }
        }
      }
    }

    *xt = x;
    *yct = yc = y;

    x -= middle->s->x;
    y -= middle->s->y;

    *yt = yc + sqrt((x * x) + (y * y));

    return RET_FORTUNE_OK;
}


/* new_circle_event()

    finds out if the three beach arc segments
      mid->left, mid and mid->right form a
      reasonable triple for a new circle event
      and, if they do, creates the event;
    the creation of the new event is aborted
      if any of the three arc segments correspond
      to the site pointed to by ignore */

static char new_circle_event(beach_arc_t *mid, site_t *ignore) {

    char ret;
    vertex_coord_t ix, iy, icy;
    event_t *e;
    circle_list_t *c;
    circle_event_t *d;
    site_t *sleft;
    site_t *smid;
    site_t *sright;
    site_t *cleft;
    site_t *cmid;
    site_t *cright;

    /* exit if any of the three segments is NULL */

    if (mid == NULL)
      return RET_FORTUNE_PARALLEL;
    if (mid->left == NULL)
      return RET_FORTUNE_PARALLEL;
    if (mid->right == NULL)
      return RET_FORTUNE_PARALLEL;

    sleft = mid->left->s;
    smid = mid->s;
    sright = mid->right->s;

    if (ignore != NULL) {
      if (sleft == ignore)
        return RET_FORTUNE_EVENT_EXISTS;
      if (smid == ignore)
        return RET_FORTUNE_EVENT_EXISTS;
      if (sright == ignore)
        return RET_FORTUNE_EVENT_EXISTS;
    }

    /* exit if sright is not on the positive side of line sleft->smid */

    if (!positive_side_of_line(sleft, smid, sright))
      return RET_FORTUNE_EVENT_EXISTS;

    /* get vertex coordinates */

    ret = get_intersection(mid, &ix, &iy, &icy);

    if (ret != RET_FORTUNE_OK)
      return ret;

    /* create event */

    e = add_event(ix, iy, FORTUNE_CIRCLE_EVENT);

    if (e == NULL)
      return RET_FORTUNE_NO_MEM;

    d = e->cdata;
    d->left   = mid->left;
    d->middle = mid;
    d->right  = mid->right;
    d->center_y = icy;

    if (add_circle(e, mid->left) == NULL)
      ret = RET_FORTUNE_NO_MEM;
    else
      if (add_circle(e, mid) == NULL)
        ret = RET_FORTUNE_NO_MEM;
      else
        if (add_circle(e, mid->right) == NULL)
          ret = RET_FORTUNE_NO_MEM;

    if (ret != RET_FORTUNE_OK)
      delete_event(e);

    return ret;
}


/* Fortune_handle_event()

    handles the next event in the event queue
      and deletes it;
    when the last event in the queue is handled,
      Fortune_handle_event() returns
      RET_FORTUNE_QUEUE_EMPTY */

char Fortune_handle_event(void) {

    event_t *e;
    event_t *en;
    beach_arc_t *b;
    beach_arc_t *bl;
    beach_arc_t *br;
    site_t *s;
    vertex_coord_t sweep;
    circle_event_t *d = NULL;
    circle_list_t *c;
    circle_list_t *cn;
    char ret;
    #ifndef FORTUNE_NO_DELAUNAY
    uint8_t tc, mask;
    site_coord_t sx1, sy1, sx2, sy2;
    vertex_coord_t vx, vy;
    #endif
    #ifndef FORTUNE_NO_VORONOI
    vertex_list_t *vl;
    #endif

    if (event_queue == NULL)
        return RET_FORTUNE_QUEUE_EMPTY; /* should never happen! */

    sweep = event_queue->y;

    if (event_queue->type == FORTUNE_CIRCLE_EVENT) {

        /* handle circle event */

        d = event_queue->cdata;

        b  = d->middle;
        bl = d->left;
        br = d->right;

        /* create a new triangle list item */

        if (add_triangle(bl, b, br) == NULL)
            return RET_FORTUNE_NO_MEM;

        #ifndef FORTUNE_NO_VORONOI
        if (create_vertices) {

            vl = add_vertex(event_queue->x, d->center_y);

            if (vl == NULL)
                return RET_FORTUNE_NO_MEM;

            triangle_list->v = vl;

            ret = update_boundaries(triangle_list);
            if (ret != RET_FORTUNE_OK)
                return ret;
        }
        #endif

        triangle_list->tags = 0;

        /* store the shape of the triangle
             in the tags bits */

        #ifndef FORTUNE_NO_DELAUNAY
        sx1 = triangle_list->site[0]->x;
        sy1 = triangle_list->site[0]->y;
        vx = event_queue->x;
        vy = d->center_y;

        for (tc = 1, mask = 0x01; ; mask <<= 1) {

            sx2 = triangle_list->site[tc]->x;
            sy2 = triangle_list->site[tc]->y;

            if (LEFT_SIDE(sx1, sy1, sx2, sy2, vx, vy)) {
                triangle_list->tags = mask;
                break;    /* there can be only one! */
            }

            if (!tc)
                break;

            tc = (tc == 1) ? 2 : 0;
            sx1 = sx2;
            sy1 = sy2;
        }
        #endif

        /* remove all circle events
             involving beach arc segment b */

        while (b->circles != NULL)
            delete_event(b->circles->e);

        make_line(&(bl->line), bl->s, br->s);

        s = b->s;
        delete_beach_arc(b);

        ret = new_circle_event(bl, s);

        if (ret == RET_FORTUNE_NO_MEM)
            return ret;

        ret = new_circle_event(br, s);

        if (ret == RET_FORTUNE_NO_MEM)
            return ret;

        return (event_queue == NULL) ?
            RET_FORTUNE_QUEUE_EMPTY : RET_FORTUNE_OK;
    }

    /* handle site event */

    s = event_queue->sdata->s;

    b = get_beach_arc(s->x, sweep);

    if (b == NULL) {

        b = add_beach_arc(s
            #ifndef FORTUNE_NO_VORONOI
            , event_queue->sdata->c
            #endif
            );

        if (b == NULL)
            return RET_FORTUNE_NO_MEM;

        if (b->left != NULL)
            make_line(&(b->left->line), s, b->left->s);
    }
    else {

        bl = b;
        b = split_beach_arc(s
            #ifndef FORTUNE_NO_VORONOI
            , event_queue->sdata->c
            #endif
            , bl);

        if (b == NULL)
            return RET_FORTUNE_NO_MEM;

        /* remove all circle events in which bl
             is the middle arc;
           update each circle event in which bl
             is the left arc */

        for (c = bl->circles, cn = NULL; c; ) {

            e = c->e;

            ret = 1;

            if (e->cdata->middle == bl) {

                delete_event(e);
                ret = 0;
            }
            else {

                if (e->cdata->left == bl) {

                    if (add_circle(e, b->right) == NULL)
                        return RET_FORTUNE_NO_MEM;

                    e->cdata->left = b->right;

                    if (cn == NULL)
                        bl->circles = c->next;
                    else
                        cn->next = c->next;

                    free(c);

                    ret = 0;
                }
            }

            if (ret) {

                cn = c;
                c = c->next;
            }
            else {

                if (cn == NULL)
                    c = bl->circles;
                else
                    c = cn->next;
            }
        }

        b->right->line.A = b->left->line.A;
        b->right->line.B = b->left->line.B;
        b->right->line.C = b->left->line.C;

        make_line(&(b->left->line), s, b->left->s);
        make_line(&(b->line), s, b->right->s);
    }

    delete_event(event_queue);

    /* create circle events */

    ret = new_circle_event(b->left, NULL);

    if (ret == RET_FORTUNE_NO_MEM)
        return ret;

    ret = new_circle_event(b->right, NULL);

    if (ret == RET_FORTUNE_NO_MEM)
        return ret;

    return (event_queue == NULL) ?
        RET_FORTUNE_QUEUE_EMPTY : RET_FORTUNE_OK;
}


/* Fortune_init()

    initializes the event queue;
    returns RET_FORTUNE_NO_MEM if the queue
      cannot fit in memory;
    function_mask is a combination of
      FORTUNE_DELAUNAY_MASK and FORTUNE_VORONOI_MASK;
      at least one of the two must be given */

char Fortune_init(site_t *s, int total_sites, uint8_t function_mask) {

    int i;
    event_t *e;
    event_t *n;
    site_coord_t ex, ey, nx, ny;
    beach_arc_t *b;

    if (!function_mask)
        return RET_FORTUNE_PARAM_ERROR;

    Fortune_events_left = 0;

    event_queue = NULL;
    beach_line = NULL;
    triangle_list = NULL;
    #ifndef FORTUNE_NO_VORONOI
    vertex_list = NULL;
    create_vertices = function_mask & FORTUNE_VORONOI_MASK;
    vertices = 0;
    #endif

    #ifdef FORTUNE_DEBUG_MODE
    next_debug_id = 1;
    next_triangle_debug_id = 1;
    #endif

    triangles = 0;
    conns = 0;

    /* create cell_master array */

    #ifndef FORTUNE_NO_VORONOI

    if (function_mask & FORTUNE_VORONOI_MASK) {

        cells = total_sites;

        #ifdef FORTUNE_MASTER_ARRAYS

        if (cell_master == NULL) {

            cell_master_size = cells;
            cell_master =
                (celldata_t *)malloc(cell_master_size * sizeof(celldata_t));

            atexit(destroy_Voronoi_masters);
        }
        else {

            if (cell_master_size < cells) {

                cell_master_size = cells;
                cell_master =
                    realloc(cell_master, cell_master_size * sizeof(celldata_t));
            }
        }

        #else

        cell_master = (celldata_t *)malloc(cells * sizeof(celldata_t));

        #endif

        if (cell_master == NULL)
            goto init_out_of_memory;

        for (i = 0; i < cells; i++) {
            cell_master[i].s = &s[i];
            cell_master[i].blist = NULL;
            cell_master[i].sides = 0;
            cell_master[i].infinity = 0;
            cell_master[i].error = 0;
            cell_master[i].bloop = NULL;
            cell_master[i].extra = NULL;
        }
    }

    #endif

    #ifndef FORTUNE_NO_DELAUNAY
    #ifdef FORTUNE_MASTER_ARRAYS

    if (function_mask & FORTUNE_DELAUNAY_MASK)
        atexit(destroy_Delaunay_masters);

    #endif
    #endif

    /* create a site event for each site */

    for (i = 0; i < total_sites; i++) {

        e = add_event(s[i].x, s[i].y, FORTUNE_SITE_EVENT);

        if (e == NULL)
            goto init_out_of_memory;

        e->sdata->s = &s[i];
        #ifndef FORTUNE_NO_VORONOI
        if (function_mask & FORTUNE_VORONOI_MASK)
            e->sdata->c = &cell_master[i];
        #endif
    }

    /* remove duplicate events */

    e = event_queue;

    if (e == NULL)
        return RET_FORTUNE_NO_TRIANGLES;

    ex = e->x;
    ey = e->y;

    while (1) {

        n = e->next;

        if (n == NULL)
            break;

        nx = n->x;
        ny = n->y;

        if (ex > nx - SITE_COORD_EPSILON)
            if (ex < nx + SITE_COORD_EPSILON)
                if (ey > ny - SITE_COORD_EPSILON)
                    if (ey < ny + SITE_COORD_EPSILON) {
                        delete_event(e);
                        total_sites--;
                    }

        e = n;
        ex = nx;
        ey = ny;
    }

    if (total_sites < 3) {
        Fortune_exit();
        return RET_FORTUNE_NO_TRIANGLES;
    }

    /* initialization done */

    return RET_FORTUNE_OK;

    /* error */

    init_out_of_memory:

    Fortune_exit();

    return RET_FORTUNE_NO_MEM;
}


/* Fortune_exit()

    releases all memory reserved by the algorithm */

char Fortune_exit(void) {

    while (event_queue != NULL)
        delete_event(event_queue);

    while (beach_line != NULL)
        delete_beach_arc(beach_line);

    destroy_triangle_list();

    #ifndef FORTUNE_NO_VORONOI
    destroy_vertex_list();
    #endif

    return RET_FORTUNE_OK;
}


#ifndef FORTUNE_NO_VORONOI

/* Voronoi_finish()

    finalizes the Voronoi diagram by converting
      existing Delaunay connections to boundaries
      and storing the results as arrays of
      vertex_t and boundary_t structures;
    the shapes of Voronoi cells are stored in a
      celldata_t array; cdout may be NULL if the
      cell data is not needed;
    the number of non-infinite vertices (which
      equals the number of Delaunay triangles if
      no two triangles form a rectangle; such a
      case would produce two vertices at the same
      location and one would be discarded) is
      stored in the variable pointed to by
      actual_vertices (which may be NULL if this
      information is not needed);
    the total vertex and boundary counts are stored
      in *total_vertices and *total_boundaries */

char Voronoi_finish(vertex_t **vout, int *total_vertices,
                    boundary_t **bout, int *total_boundaries,
                    celldata_t **cdout,
                    int *actual_vertices) {

    triangle_list_t *t;
    triangle_list_t *n;
    vertex_list_t *vl;
    uint8_t side, mask;
    int boundaries = 0;
    int vi, bi;
    int ci;
    line_t line;
    vertex_t *v1;
    vertex_t *v2;
    vertex_coord_t vx, vy;
    vertex_t *va = NULL;
    boundary_t *ba = NULL;
    celldata_t *ca = NULL;
    connect_list_t *c;
    char ret;

    if (vout == NULL)
      return RET_FORTUNE_PARAM_ERROR;
    if (bout == NULL)
      return RET_FORTUNE_PARAM_ERROR;
    if (total_vertices == NULL)
      return RET_FORTUNE_PARAM_ERROR;
    if (total_boundaries == NULL)
      return RET_FORTUNE_PARAM_ERROR;

    #ifndef FORTUNE_MASTER_ARRAYS
    if (*vout != NULL)
      return RET_FORTUNE_PARAM_ERROR;
    if (*bout != NULL)
      return RET_FORTUNE_PARAM_ERROR;
    #endif

    /* store the number of non-infinite vertices;
         infinite vertices are added in the array
         after the non-infinite ones */

    if (actual_vertices != NULL)
      *actual_vertices = vertices;

    /* calculate number of Voronoi boundaries */

    for (t = triangle_list; t; t = t->next)
       for (side = 0, mask = 0x10; side < 3; side++, mask <<= 1) {

          if (t->tags & mask)
            continue;

          n = t->neighbour[side];

          if (n == NULL) {
            vertices++; /* vertex needed for infinite boundary */
            boundaries++;
            continue;
          }

          n->tags |= 0x10 << t->neighbour_side[side];
          boundaries++;
       }

    /* create arrays for vertices and boundaries */

    #ifdef FORTUNE_MASTER_ARRAYS

    if (vertex_master == NULL) {
        vertex_master_size = vertices;
        vertex_master =
            (vertex_t *)malloc(vertices * sizeof(vertex_t));
    }
    else {
        if (vertex_master_size < vertices) {
            vertex_master_size = vertices;
            vertex_master =
                realloc(vertex_master, vertices * sizeof(vertex_t));
        }
    }

    if (vertex_master == NULL)
        return RET_FORTUNE_NO_MEM;

    va = vertex_master;

    if (boundary_master == NULL) {
        boundary_master_size = boundaries;
        boundary_master =
            (boundary_t *)malloc(boundaries * sizeof(boundary_t));
    }
    else {
        if (boundary_master_size < boundaries) {
            boundary_master_size = boundaries;
            boundary_master =
                realloc(boundary_master, boundaries * sizeof(boundary_t));
        }
    }

    if (boundary_master == NULL) {
        free(vertex_master);
        return RET_FORTUNE_NO_MEM;
    }

    ba = boundary_master;

    #define V_CLEANUP_ON_ERROR { /*Voronoi_cleanup(va, ba, cell_master);*/ }

    #else

    va = (vertex_t *)malloc(vertices * sizeof(vertex_t));
    if (va == NULL)
        return RET_FORTUNE_NO_MEM;

    ba = (boundary_t *)malloc(boundaries * sizeof(boundary_t));
    if (ba == NULL) {
        free(va);
        return RET_FORTUNE_NO_MEM;
    }

    #define V_CLEANUP_ON_ERROR {                            \
        for (ci = 0; ci < cells; ci++)                      \
            destroy_cellboundary_list(&cell_master[ci]);    \
        free(cell_master);                                  \
        free(va);                                           \
        free(ba);                                           \
    }

    #endif

    for (vi = 0; vi < vertices; vi++) {
        va[vi].x = 0.0;
        va[vi].y = 0.0;
        va[vi].extra = NULL;
    }

    for (bi = 0; bi < boundaries; bi++) {
        ba[bi].v1 = &va[vertices - 1];
        ba[bi].v2 = &va[vertices - 1];
        ba[bi].vector = NULL;
        ba[bi].extra = NULL;
    }

    /* store vertices */

    for (vl = vertex_list; vl; vl = vl->next) {
        va[vl->index].x = vl->x;
        va[vl->index].y = vl->y;
    }

    /* one Voronoi boundary per Delaunay connection */

    for (c = connect_list, vi = vertices - 1; c; c = c->next) {

        if (c->v2 != NULL) {

            ba[c->index].v1 = &va[c->v1->index];
            ba[c->index].v2 = &va[c->v2->index];

            continue;
        }

        /* infinite boundary */

        #if 0
        #if DEBUG_CELLS > 1
        if (c->s1->x > c->s2->x - SITE_COORD_EPSILON &&
            c->s1->x < c->s2->x + SITE_COORD_EPSILON) {
            fprintf(stderr, "%4d.\t", c->index);
            if (ABOVE(c->s1, c->s2)) {
                fprintf(stderr, "s1 *ABOVE* s2\n");
            }
            else {
                if (ABOVE(c->s2, c->s1))
                    fprintf(stderr, "s1 *BELOW* s2\n");
            }
        }
        if (c->s1->y > c->s2->y - SITE_COORD_EPSILON &&
            c->s1->y < c->s2->y + SITE_COORD_EPSILON) {
            fprintf(stderr, "%4d.\t", c->index);
            if (c->s1->x < c->s2->x) {
                fprintf(stderr, "s1 on the *LEFT*  side of s2\n");
            }
            else {
                if (c->s1->x > c->s2->x)
                    fprintf(stderr, "s1 on the *RIGHT* side of s2\n");
                else
                    fprintf(stderr, "!!! s1 == s2\n");
            }
        }
        #endif
        #endif

        #define MAKE_INF_BOUNDARY(START, END) {     \
            make_line(&line, c->START, c->END);     \
            vx = line.B;                            \
            vy = -line.A;                           \
        }

        #define INVERT_VECTOR_IF(CONDITION) {       \
            if (CONDITION) {                        \
                vx = -vx;                           \
                vy = -vy;                           \
            }                                       \
        }

        switch (c->side) {

            case 0:
                if (ABOVE(c->s2, c->s1)) {
                    MAKE_INF_BOUNDARY(s2, s1);
                }
                else {
                    MAKE_INF_BOUNDARY(s1, s2);
                }
                INVERT_VECTOR_IF(vx < VERTEX_COORD_EPSILON);
                break;

            case 2:
                if (ABOVE(c->s2, c->s1)) {
                    MAKE_INF_BOUNDARY(s1, s2);
                }
                else {
                    MAKE_INF_BOUNDARY(s2, s1);
                }
                INVERT_VECTOR_IF(vx > -VERTEX_COORD_EPSILON);
                break;

            default:
                if (ABOVE(c->s2, c->s1)) {
                    MAKE_INF_BOUNDARY(s2, s1);
                    INVERT_VECTOR_IF(vx > -VERTEX_COORD_EPSILON);
                }
                else {
                    MAKE_INF_BOUNDARY(s1, s2);
                    INVERT_VECTOR_IF(vx < VERTEX_COORD_EPSILON);
                }
        }

        va[vi].x = vx;
        va[vi].y = vy;

        ba[c->index].v1 = &va[c->v1->index];
        ba[c->index].v2 = NULL;
        ba[c->index].vector = &va[vi];

        vi--;
    }

    #ifdef FIND_ZERO_LENGTH_BOUNDARIES

    /* look for zero-length boundaries */

    for (bi = 0; bi < boundaries; bi++) {

        v1 = ba[bi].v1;
        v2 = ba[bi].v2;

        if (v2 == NULL) /* skip infinite boundary */
            continue;
        if (v1 == v2)   /* already handled */
            continue;
        if (!vertices_equal_or_too_close(v1->x, v1->y, v2->x, v2->y))
            continue;

        /* the vertices ba[bi].v1 and ba[bi].v2 are
             in the exact same location, so search
             through the boundary list and replace
             every occurrence of ba[bi].v2 with
             ba[bi].v1;
           the duplicate vertex and the zero-length
             boundary remain in the lists, but they
             are not used anywhere */

        #define REPLACE_V2_WITH_V1 {    \
            if (ba[ci].v1 == v2)        \
                ba[ci].v1 = v1;         \
            if (ba[ci].v2 == v2)        \
                ba[ci].v2 = v1;         \
        }

        for (ci = 0; ci < bi; ci++)
            REPLACE_V2_WITH_V1;

        ba[bi].v2 = v1;

        for (ci = bi + 1; ci < boundaries; ci++)
            REPLACE_V2_WITH_V1;
    }

    #endif

    if (cdout == NULL)
        goto return_with_arrays;

    /* convert the boundary loop of each cell
         to a list of cellboundary_t objects */

    #ifdef FORTUNE_MASTER_ARRAYS
    cellboundary_master_ptr = 0;
    #endif

    for (ci = 0; ci < cells; ci++) {

        ret = count_boundaries(&cell_master[ci]);

        if (ret != RET_FORTUNE_OK) {

            if (ret != RET_FORTUNE_PARAM_ERROR)
                return ret;

            #ifdef DEBUG_CELLS
            fprintf(stderr, "!!!   in cell %d <%f,%f>\n",
                ci, cell_master[ci].s->x, cell_master[ci].s->y);
            #endif

            cell_master[ci].error = 1;

            continue;
        }

        #ifdef FORTUNE_MASTER_ARRAYS
        cellboundary_master_ptr += cell_master[ci].sides;
        #endif
    }

    #ifdef FORTUNE_MASTER_ARRAYS

    if (cellboundary_master == NULL) {
        cellboundary_master_size = cellboundary_master_ptr;
        cellboundary_master =
            (cellboundary_t *)malloc(cellboundary_master_size *
                sizeof(cellboundary_t));
    }
    else {
        if (cellboundary_master_size < cellboundary_master_ptr) {
            cellboundary_master_size = cellboundary_master_ptr;
            cellboundary_master =
                realloc(cellboundary_master, cellboundary_master_size *
                    sizeof(cellboundary_t));
        }
    }

    if (cellboundary_master == NULL) {
        V_CLEANUP_ON_ERROR;
        return RET_FORTUNE_NO_MEM;
    }

    cellboundary_master_ptr = 0;

    #endif

    for (ci = 0; ci < cells; ci++) {

        if (cell_master[ci].error)
            continue;

        ret = convert_boundaries(&cell_master[ci], ba);

        if (ret != RET_FORTUNE_OK) {

            if (ret != RET_FORTUNE_PARAM_ERROR)
                return ret;

            #ifdef DEBUG_CELLS
            fprintf(stderr, "!!!   in cell %d <%f,%f>\n",
                ci, cell_master[ci].s->x, cell_master[ci].s->y);
            #endif

            cell_master[ci].error = 1;
        }
    }

    /* done */

    *cdout = cell_master;

    /* return */

    return_with_arrays:

    *vout = va;
    *bout = ba;

    *total_vertices = vertices;
    *total_boundaries = boundaries;

    return RET_FORTUNE_OK;
}


/* Voronoi()

    creates the Voronoi diagram based on the given
      set of sites s;
    total_sites is the number of sites in s;
    Voronoi() stores an array of vertex_t
      structures in *vout and another array of
      boundary_t structures in *bout;
    if cdout is not NULL, *cdout is set to point
      to the cell_master array, which has been
      created by Fortune_init() and filled by
      Voronoi_finish();
    if FORTUNE_MASTER_ARRAYS is not #defined,
      the vertex, boundary and cell arrays must be
      destroyed later using Voronoi_cleanup(),
      although free() could be used to destroy the
      vertex and boundary arrays;
    the numbers of items in the vertex and
      boundary arrays are stored in
      *total_vertices and *total_boundaries;
    the number of items in the cdout array always
      equals total_sites;
    the number of non-infinite vertices is stored
      in *actual_vertices; infinite vertices are
      placed in the vertex array after the non-
      infinite ones */

char Voronoi(site_t *s, int total_sites,
             vertex_t **vout, int *total_vertices,
             boundary_t **bout, int *total_boundaries,
             celldata_t **cdout,
             int *actual_vertices) {

    char ret;

    ret = Fortune_init(s, total_sites, FORTUNE_VORONOI_MASK);

    if (ret != RET_FORTUNE_OK)
      return ret;

    while (ret == RET_FORTUNE_OK)
      ret = Fortune_handle_event();

    if (ret != RET_FORTUNE_QUEUE_EMPTY)
      return ret;

    ret = Voronoi_finish(vout, total_vertices,
                         bout, total_boundaries,
                         cdout, actual_vertices);

    Fortune_exit();

    return ret;
}


#ifndef FORTUNE_MASTER_ARRAYS

/* Voronoi_cleanup()

    releases the memory allocated to the
      vertex, boundary and cell arrays */

char Voronoi_cleanup(vertex_t *va, boundary_t *ba, celldata_t *ca) {

    int c;

    if (va != NULL)
      free(va);

    if (ba != NULL)
      free(ba);

    if (ca != NULL) {
      for (c = 0; c < cells; c++) {
         destroy_cellboundary_list(&ca[c]);
      }
      free(ca);
    }

    return RET_FORTUNE_OK;
}

#endif

#endif


#ifndef FORTUNE_NO_DELAUNAY

/* Delaunay_finish()

    finalizes the Delaunay triangulation by
      storing the contents of triangle_list
      in an array of triangle_t structures
      and the connections that make up the
      triangles in an array of connect_t
      structures */

char Delaunay_finish(connect_t **cout, int *total_conns,
                     triangle_t **tout, int *total_triangles) {

    triangle_list_t *t;
    triangle_list_t *m;
    triangle_list_t *n;
    uint8_t side;
    int ti, ni, ci;
    connect_t *ca;
    connect_t *cn;
    triangle_t *ta;

    if (cout == NULL)
      return RET_FORTUNE_PARAM_ERROR;
    if (*cout != NULL)
      return RET_FORTUNE_PARAM_ERROR;
    if (total_conns == NULL)
      return RET_FORTUNE_PARAM_ERROR;
    if (tout == NULL)
      return RET_FORTUNE_PARAM_ERROR;
    if (*tout != NULL)
      return RET_FORTUNE_PARAM_ERROR;
    if (total_triangles == NULL)
      return RET_FORTUNE_PARAM_ERROR;

    /* create arrays for connections and triangles */

    #ifdef FORTUNE_MASTER_ARRAYS

    if (connect_master == NULL) {
        connect_master_size = conns;
        connect_master =
            (connect_t *)malloc(conns * sizeof(connect_t));
    }
    else {
        if (connect_master_size < conns) {
            connect_master_size = conns;
            connect_master =
                realloc(connect_master, conns * sizeof(connect_t));
        }
    }

    if (connect_master == NULL)
        return RET_FORTUNE_NO_MEM;

    ca = connect_master;

    if (triangle_master == NULL) {
        triangle_master_size = triangles;
        triangle_master =
            (triangle_t *)malloc(triangles * sizeof(triangle_t));
    }
    else {
        if (triangle_master_size < triangles) {
            triangle_master_size = triangles;
            triangle_master =
                realloc(triangle_master, triangles * sizeof(triangle_t));
        }
    }

    if (triangle_master == NULL) {
        free(connect_master);
        return RET_FORTUNE_NO_MEM;
    }

    ta = triangle_master;

    #else

    ca = (connect_t *)malloc(conns * sizeof(connect_t));

    if (ca == NULL)
        return RET_FORTUNE_NO_MEM;

    ta = (triangle_t *)malloc(triangles * sizeof(triangle_t));

    if (ta == NULL) {
       free(ca);
       return RET_FORTUNE_NO_MEM;
    }

    #endif

    /* store triangles and connections */

    for (t = triangle_list, ti = ci = 0; t; t = t->next, ti++) {

        ta[ti].shape = t->tags & 0x07;
        ta[ti].extra = NULL;

        for (side = 0; side < 3; side++) {

            ta[ti].site[side] = t->site[side];
            cn = NULL;

            if (t->conn[side] != NULL) {
                if (ci < conns) { /* unnecessary, but just to be safe */
                    cn = &ca[ci];
                    ci++;
                    cn->s1 = t->conn[side]->s1;
                    cn->s2 = t->conn[side]->s2;
                    cn->extra = NULL;
                    ta[ti].conn[side] = cn;
                    t->conn[side] = NULL;
                }
            }

            m = t->neighbour[side];

            if (m == NULL) {
                ta[ti].neighbour[side] = NULL;
                ta[ti].neighbour_side[side] = 0;
                continue;
            }

            m->conn[t->neighbour_side[side]] = NULL;

            for (n = triangle_list, ni = 0; n; n = n->next, ni++)
                if (m == n) {
                    ta[ti].neighbour[side] = &ta[ni];
                    ta[ti].neighbour_side[side] = t->neighbour_side[side];
                    if (cn != NULL)
                        ta[ni].conn[t->neighbour_side[side]] = cn;
                    break;
                }
        }
    }

    *cout = ca;
    *total_conns = conns;
    *tout = ta;
    *total_triangles = triangles;

    return RET_FORTUNE_OK;
}


/* Delaunay()

    performs the Delaunay triangulation of
      the given set of sites s;
    total_sites is the number of sites in s;
    Delaunay() creates an array of triangle_t
      structures and another array of
      connect_t structures and stores these
      in *tout and *cout;
    if FORTUNE_MASTER_ARRAYS is not #defined,
      the arrays must be later destroyed using
      either free() or Delaunay_cleanup();
    the number of triangles and connections
      in the resulting arrays are stored in
      thes variables pointed to by
      total_triangles and total_conns */

char Delaunay(site_t *s, int total_sites,
              connect_t **cout, int *total_conns,
              triangle_t **tout, int *total_triangles) {

    char ret;

    ret = Fortune_init(s, total_sites, FORTUNE_DELAUNAY_MASK);

    if (ret != RET_FORTUNE_OK)
      return ret;

    while (ret == RET_FORTUNE_OK)
      ret = Fortune_handle_event();

    if (ret != RET_FORTUNE_QUEUE_EMPTY)
      return ret;

    ret = Delaunay_finish(cout, total_conns, tout, total_triangles);

    Fortune_exit();

    return ret;
}


#ifndef FORTUNE_MASTER_ARRAYS

/* Delaunay_cleanup()

    releases the memory allocated to the
      triangle and connection arrays */

char Delaunay_cleanup(connect_t *ca, triangle_t *ta) {

    if (ca != NULL)
      free(ca);

    if (ta != NULL)
      free(ta);

    return RET_FORTUNE_OK;
}

#endif

#endif
