/*************************************************************************
 *
 *  An implementation of Fortune's algorithm
 *  Copyright (C) 2012-2014 Jussi M. Hukkanen
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *************************************************************************
 *
 *  fortune.h: main header file
 *
 ************************************************************************/

#ifndef _FORTUNE_H_INCLUDED
#define _FORTUNE_H_INCLUDED

#include <stdlib.h>
#include <stdint.h>
#include <math.h>   /* needed for sqrt() */
#include <float.h>  /* needed for FLT_EPSILON */


#define RET_FORTUNE_OK              0
#define RET_FORTUNE_NO_MEM         -2
#define RET_FORTUNE_NO_TRIANGLES   -3 /* less than three sites given */
#define RET_FORTUNE_PARALLEL       -4
#define RET_FORTUNE_QUEUE_EMPTY    -5
#define RET_FORTUNE_EVENT_EXISTS   -6 /* duplicate circle event */
#define RET_FORTUNE_PARAM_ERROR    -7

#define FORTUNE_DELAUNAY_MASK    0x01
#define FORTUNE_VORONOI_MASK     0x02


//#define FORTUNE_DEBUG_MODE  1

/* fortune.c can be compiled with FORTUNE_NO_VORONOI
     or FORTUNE_NO_DELAUNAY defined if an application
     only needs to create Voronoi diagrams or Delaunay
     triangles and not the other */

#ifdef FORTUNE_NO_VORONOI
#ifdef FORTUNE_NO_DELAUNAY
#undef FORTUNE_NO_DELAUNAY
#endif
#endif


/* an application may define either the macro
     FORTUNE_POSITIVE_Y_DOWN or FORTUNE_POSITIVE_Y_UP
     to specify the orientation of its coordinate
     system;
   by default, the positive y axis is assumed to point
     'up', which is the opposite of the screen
     coordinate system;
   this only affects the concepts of 'left', 'right'
     and 'clockwise', which are needed when
     add_triangle() orders the corner sites of a
     new Delaunay triangle in clockwise order;
   if an application uses top-down coordinates, it
     must define FORTUNE_POSITIVE_Y_DOWN even if it
     also defines FORTUNE_NO_DELAUNAY, because
     Voronoi_finish() assumes the corners of triangles
     to be in clockwise order;
   the sweep line always moves towards positive y */

#ifdef FORTUNE_POSITIVE_Y_DOWN
#ifdef FORTUNE_POSITIVE_Y_UP
#undef FORTUNE_POSITIVE_Y_UP
#endif
#else
#ifndef FORTUNE_POSITIVE_Y_UP
#define FORTUNE_POSITIVE_Y_UP
#endif
#endif

#define LINE_TEST(X1, Y1, X2, Y2, X3, Y3) \
    ((Y1) - (Y2)) * (X3) +                \
    ((X2) - (X1)) * (Y3) +                \
    ((Y2) * (X1)) - ((X2) * (Y1))

#ifdef FORTUNE_POSITIVE_Y_DOWN

#define  LEFT_SIDE(X1, Y1, X2, Y2, X3, Y3)    \
        (LINE_TEST(X1, Y1, X2, Y2, X3, Y3) < 0)
#define RIGHT_SIDE(X1, Y1, X2, Y2, X3, Y3)    \
        (LINE_TEST(X1, Y1, X2, Y2, X3, Y3) > 0)

#define Y_ABOVE(Y1, Y2) (Y1 < Y2)
#define Y_BELOW(Y1, Y2) (Y1 > Y2)
#define ABOVE(S1, S2) Y_ABOVE(S1->y, S2->y)
#define ABOVE_OR_EQUAL(S1, S2) (S1->y < S2->y + SITE_COORD_EPSILON)

#else

#define  LEFT_SIDE(X1, Y1, X2, Y2, X3, Y3)    \
        (LINE_TEST(X1, Y1, X2, Y2, X3, Y3) > 0)
#define RIGHT_SIDE(X1, Y1, X2, Y2, X3, Y3)    \
        (LINE_TEST(X1, Y1, X2, Y2, X3, Y3) < 0)

#define Y_ABOVE(Y1, Y2) (Y1 > Y2)
#define Y_BELOW(Y1, Y2) (Y1 < Y2)
#define ABOVE(S1, S2) Y_ABOVE(S1->y, S2->y)
#define ABOVE_OR_EQUAL(S1, S2) (S1->y > S2->y - SITE_COORD_EPSILON)

#endif


/* the application may define FORTUNE_DEF_HEADER as
     a file containing the type definitions site_t,
     site_coord_t and vertex_coord_t;
   vertex_coord_t must be float or double; in the latter
     case, VERTEX_COORD_EPSILON should be defined as
     DBL_EPSILON in the file */

#ifdef FORTUNE_DEF_HEADER

#include FORTUNE_DEF_HEADER

#else

typedef float site_coord_t;

typedef float vertex_coord_t;

typedef struct site_s {
    site_coord_t x;
    site_coord_t y;
} site_t;

#endif

#define SITE_COORD_EPSILON   FLT_EPSILON
#define SITE_COORD_ZERO 	 0.0

#ifndef VERTEX_COORD_EPSILON
#define VERTEX_COORD_EPSILON FLT_EPSILON
#endif


/* the application may attach its own extra data
     structures to each Delaunay triangle or
     Voronoi cell; this data should be pointed to
     by the extra member in triangle_t and
     celldata_t below;
   the application is responsible of destroying the
     data when it is not needed anymore; fortune.c
     will not take care of it! */

/**** TODO: the structures below should have links
        both ways; i.e. connect_t should point to
        two triangle_t's, vertex_t to three (or more!)
        celldata_t's and/or boundary_t's and boundary_t
        to two celldata_t's and/or cellboundary_t's ****/

#ifndef FORTUNE_NO_DELAUNAY

typedef struct connect_s {          /* Delaunay connection (boundary) */
    site_t *s1;
    site_t *s2;
    void *extra;            /* pointer to extra data */
} connect_t;

typedef struct triangle_s {         /* Delaunay triangle */
    site_t *site[3];        /* pointers to sites, clockwise from top */
    connect_t *conn[3];     /* pointers to connections, clockwise from top */
    struct triangle_s *neighbour[3]; /* neighbour[x] points to triangle
                                          beyond the edge that connects
                                          site[x] and site[(x + 1) % 3] */
    uint8_t neighbour_side[3];  /* the side of the neighbour triangle */
    uint8_t shape;              /* one bit for each connection:
                                     1 = the opposite corner is obtuse;
                                   i.e. if shape == 0, the triangle is either
                                     acute or right-angled */
    void *extra;                /* pointer to extra data */
} triangle_t;

#endif

#ifndef FORTUNE_NO_VORONOI

typedef struct vertex_s {           /* Voronoi vertex */
    vertex_coord_t x;
    vertex_coord_t y;
    void *extra;            /* pointer to data defined by the application */
} vertex_t;

typedef struct boundary_s {         /* Voronoi boundary */
    vertex_t *v1;
    vertex_t *v2;      /* if NULL, the boundary is infinite; in that
                            case, vector must be non-NULL */
    vertex_t *vector;  /* contains a vector that points away from v1 */
    void *extra;            /* pointer to extra data */
} boundary_t;

struct celldata_s;

typedef struct cellboundary_s {     /* Voronoi boundary of a specific cell */
    boundary_t *boundary;
    struct celldata_s *neighbour;
    struct cellboundary_s *left;
    struct cellboundary_s *right;
    #ifdef FORTUNE_MASTER_ARRAYS
    int i;  /* index to cellboundary_master */
    #endif
} cellboundary_t;

typedef struct celldata_s {         /* Voronoi cell */
    site_t *s;              /* pointer to site */
    cellboundary_t *blist;  /* list of boundaries, topmost first */
    int sides;              /* number of items in blist */
    char infinity;          /* if nonzero, this cell stretches to infinity */
    char error;             /* if nonzero, the cell structure is broken */
    void *bloop;            /* list of boundary_loop_t objects used when
                                 creating cells; in an infinite cell,
                                 a placeholder item is inserted into bloop
                                 so that the number of items equals sides */
    void *extra;            /* pointer to extra data */
} celldata_t;

#endif


/* an application may define FORTUNE_PUBLIC before
     including fortune.h if it needs to track the
     progress of the algorithm; if FORTUNE_PUBLIC
     is not defined, all the types and variables
     below are only available to fortune.c */

#ifdef FORTUNE_PUBLIC

#define FORTUNE_SITE_EVENT    0
#define FORTUNE_CIRCLE_EVENT  1

typedef struct site_event_s {
    site_t *s;
    #ifndef FORTUNE_NO_VORONOI
    celldata_t *c;
    #endif
} site_event_t;

struct beach_arc_s;

typedef struct circle_event_s {
    struct beach_arc_s *left;
    struct beach_arc_s *middle;
    struct beach_arc_s *right;
    vertex_coord_t center_y;
} circle_event_t;

typedef struct event_s {
    char type;
    vertex_coord_t x, y;
    site_event_t *sdata;
    circle_event_t *cdata;
    struct event_s *prev;
    struct event_s *next;
} event_t;

event_t *event_queue;

uint32_t Fortune_events_left; /* number of unhandled events in the queue;
                                   not actually needed anywhere */


#ifndef FORTUNE_NO_VORONOI

typedef struct vertex_list_s {
    vertex_coord_t x;
    vertex_coord_t y;
    int index;
    struct vertex_list_s *next;
} vertex_list_t;

vertex_list_t *vertex_list;

struct connect_list_s;
struct triangle_list_s;

typedef struct boundary_loop_s {
    struct triangle_list_s *tri;    /* corresponding Delaunay triangle */
    struct connect_list_s *conn_l;  /* left connection, looking from site */
    struct connect_list_s *conn_r;  /* right connection, looking from site */
    struct boundary_loop_s *left;   /* next in loop, counter-clockwise */
    struct boundary_loop_s *right;  /* next in loop, clockwise */
    struct boundary_loop_s *next;   /* next in linked list */
    char first;                     /* if nonzero, the loop starts here */
} boundary_loop_t;

#endif

typedef struct connect_list_s {
    site_t *s1;
    site_t *s2;
    #ifndef FORTUNE_NO_VORONOI
    vertex_list_t *v1;  /* start vertex of corresponding Voronoi boundary */
    vertex_list_t *v2;  /* end vertex of boundary or NULL if infinite */
    int index;
    char side;  /* if v2 == NULL, side of triangle, otherwise negative */
    #endif
    struct connect_list_s *next;
} connect_list_t;

connect_list_t *connect_list;

typedef struct triangle_list_s {
    site_t *site[3];
    connect_list_t *conn[3];
    struct triangle_list_s *neighbour[3];
    uint8_t neighbour_side[3];
    uint8_t tags;           /* x111x111
                                | | +-+-- shape bits as in triangle_t
                                +-+------ used when counting boundaries */
    #ifndef FORTUNE_NO_VORONOI
    vertex_list_t *v;       /* vertex at the center of the circumcircle */
    celldata_t *cell[3];    /* cells corresponding to each site */
    #endif
    #ifdef FORTUNE_DEBUG_MODE
    uint16_t debug_id;
    #endif
    struct triangle_list_s *next;
} triangle_list_t;

triangle_list_t *triangle_list;


typedef struct line_s {
    double A, B, C;  /* Ax + Bx + C = 0 */
} line_t;

typedef struct circle_list_s {
    event_t *e;
    struct circle_list_s *next;
} circle_list_t;

typedef struct beach_arc_s {
    site_t *s;
    site_t *split_left;
    site_t *split_right;
    line_t line; /* line between this site and the one to the right */
    #ifndef FORTUNE_NO_VORONOI
    celldata_t *cell; /* cell corresponding to site s */
    #endif
    #ifdef FORTUNE_DEBUG_MODE
    uint16_t debug_id;
    #endif
    circle_list_t *circles; /* list of circle events involving this arc */
    struct beach_arc_s *left;
    struct beach_arc_s *right;
} beach_arc_t;

beach_arc_t *beach_line;

#endif /* FORTUNE_PUBLIC */


/* function declarations */

char Fortune_init(site_t *, int, uint8_t); /* set, number of sites, function */
char Fortune_exit(void);

char Fortune_handle_event(void);

#ifndef FORTUNE_NO_VORONOI

/* either call
     Voronoi()
   or
     Fortune_init() with FORTUNE_VORONOI_MASK as the function,
     Fortune_handle_event() repeatedly,
     Voronoi_finish() and
     Fortune_exit();
   if FORTUNE_MASTER_ARRAYS is not #defined,
     Voronoi_cleanup() must be called when done with the data;
   if application-specific data is attached to the extra fields
     of the cells, it is *not* released by Voronoi_cleanup() */

char Voronoi(site_t *, int, vertex_t **, int *, boundary_t **, int *,
             celldata_t **, int *);
char Voronoi_finish(vertex_t **, int *, boundary_t **, int *,
                    celldata_t **, int *);
#ifndef FORTUNE_MASTER_ARRAYS
char Voronoi_cleanup(vertex_t *, boundary_t *, celldata_t *);
#else
#define Voronoi_cleanup(V, B, C) (RET_FORTUNE_OK)
#endif

#endif

#ifndef FORTUNE_NO_DELAUNAY

/* either call
     Delaunay()
   or
     Fortune_init() with FORTUNE_DELAUNAY_MASK as the function,
     Fortune_handle_event() repeatedly,
     Delaunay_finish() and
     Fortune_exit();
   if FORTUNE_MASTER_ARRAYS is not #defined,
     Delaunay_cleanup() must be called when done with the data;
   if application-specific data is attached to the extra fields
     of the triangles, it is *not* released by Delaunay_cleanup() */

char Delaunay(site_t *, int, connect_t **, int *, triangle_t **, int *);
char Delaunay_finish(connect_t **, int *, triangle_t **, int *);
#ifndef FORTUNE_MASTER_ARRAYS
char Delaunay_cleanup(connect_t *, triangle_t *);
#else
#define Delaunay_cleanup(C, T) (RET_FORTUNE_OK)
#endif

#endif


#endif /* _FORTUNE_H_INCLUDED */

