/*************************************************************************
 *
 *  An implementation of Lloyd's algorithm
 *  Copyright (C) 2013-2014 Jussi M. Hukkanen
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *************************************************************************
 *
 *  lloyd.c: main module
 *
 ************************************************************************/

#include "lloyd.h"


/* Lloyd()

	runs one iteration of Lloyd's algorithm;
	site_in must be a list of sites and size
	  their number;
	if site_out is NULL, the site data in
	  site_in is overwritten; otherwise,
	  site_out must have room for size sites;
	the rest of the arguments are the same
	  as those in Voronoi_finish(), i.e. if
	  v_out, vnum_out, b_out, bnum_out and
	  c_out are all non-NULL, they receive
	  pointers to the new vertex, boundary
	  and cell lists and their respective
	  counts; the number of cells will be
	  equal to size;
	if actual_vnum_out is non-NULL, it
	  receives the number of non-infinite
	  vertices */

char Lloyd(site_t *site_in, int size, site_t *site_out,
		   vertex_t **v_out, int *vnum_out,
		   boundary_t **b_out, int *bnum_out,
		   celldata_t **c_out, int *actual_vnum_out) {

	vertex_t *vdata;
	boundary_t *bdata;
	celldata_t *cdata;
	cellboundary_t *cb;
	int vnum, bnum;
	int actual_vnum;
	int i, j;
	int s;
	char ret;
	vertex_t *v;
	vertex_coord_t cx, cy;

	if (site_in == NULL)
		return RET_FORTUNE_PARAM_ERROR;

	if (site_out == NULL)
		site_out = site_in;

	/* create Voronoi diagram */

	ret = Voronoi(site_in, size, &vdata, &vnum, &bdata, &bnum, &cdata, &actual_vnum);
	if (ret != RET_FORTUNE_OK)
		return ret;

	/* move each site to the center
	     of its corresponding cell */

	for (i = 0; i < size; i++) {

		v = NULL;
		cx = cy = 0.0;
		s = 0;
		cb = cdata[i].blist;

		if (cdata[i].error)
			continue;
		if (cb == NULL)
			continue;

		for (j = 0; j < cdata[i].sides; j++, cb = cb->right) {

			ret = 0;

			if (v != cb->boundary->v1) {
				v = cb->boundary->v1;
				ret = 1;
			}
			else {
				if (cb->boundary->v2 != NULL)
					if (v != cb->boundary->v2) {
						v = cb->boundary->v2;
						ret = 1;
					}
			}

			if (ret) {
				cx += v->x;
				cy += v->y;
				s++;
			}
		}

		site_out[i].x = cx / s;
		site_out[i].y = cy / s;
	}

	/* store results if requested */

	if (actual_vnum_out != NULL)
		*actual_vnum_out = actual_vnum;

	if (v_out != NULL && b_out != NULL && c_out != NULL &&
		vnum_out != NULL && bnum_out != NULL) {
		*v_out = vdata;
		*vnum_out = vnum;
		*b_out = bdata;
		*bnum_out = bnum;
		*c_out = cdata;
	}
	#ifndef FORTUNE_MASTER_ARRAYS
	else {
		Voronoi_cleanup(vdata, bdata, cdata);
	}
	#endif

	return RET_FORTUNE_OK;
}

