/*************************************************************************
 *
 *  Maze algorithm applied to Delaunay triangles and Voronoi diagrams
 *  Copyright (C) 2012-2014 Jussi M. Hukkanen
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *************************************************************************
 *
 *  maze.c: main module
 *
 ************************************************************************/

#include <string.h>		/* declares memset() */
#include "maze.h"
#include "rnd.h"

#if 0
#define DEBUG_MAZE 1//2
#include <stdio.h>
#endif


static mazedata_t *maze;
static int maze_size;
static uint8_t *maze_bits;	/* bitfield with one bit per wall; always
							     one or more full bytes per cell/triangle */
static int maze_bits_size;

static int *maze_edge;		/* index numbers of triangles with at least one
							     NULL neighbour (MAZE_DELAUNAY), cells with
							     at least one infinite neighbour cell
							     (MAZE_VORONOI), or infinite cells
							     (MAZE_VORONOI_INF) */
static int maze_edge_size;	/* number of elements in maze_edge */

static uint8_t source_type;	/* MAZE_DELAUNAY, MAZE_VORONOI or MAZE_VORONOI_INF */
static int source_size;

#ifndef FORTUNE_NO_DELAUNAY
static triangle_t *dsource = NULL;
static triangle_t *dstart;	/* triangle we start from */
static triangle_t *dptr;	/* triangle we are in */
#endif

#ifndef FORTUNE_NO_VORONOI
static celldata_t *vsource = NULL;
static celldata_t *vstart;	/* cell we start from */
static celldata_t *vptr;	/* cell we are in */
#endif

static int prev_index;		/* previous maze cell index */

static vertex_coord_t threshold_squared = 0.0;	/* minimun length (squared)
												     allowed for maze doorways */


#define MAZE_RND(LIMIT) RND_INT(LIMIT)
#define MAZE_RND_2 RND_BITS(1)


/* maze_access()

	returns either of the MAZE_ACCESS_*
	  values depending on the access bit
	  corresponding to the given side
	  in the given maze cell */

char maze_access(mazedata_t *mdata, int side) {

	if (mdata->access[side >> 3] & (1 << (side & 7)))
		return MAZE_ACCESS_OPEN;
	else
		return MAZE_ACCESS_CLOSED;
}


#define LENGTH_SQUARED(X1, Y1, X2, Y2) {	\
	X1 -= X2; Y1 -= Y2;						\
	return (X1 * X1) + (Y1 * Y1);			\
}


#ifndef FORTUNE_NO_VORONOI

/* cellboundary_length_squared()

	returns the length of the given
	  boundary, squared;
	if the boundary is infinite, the
	  return value is negative */

static vertex_coord_t cellboundary_length_squared(cellboundary_t *b) {

	vertex_coord_t x1, y1, x2, y2;

	if (b->boundary->v2 == NULL)
		return -1.0;

	x1 = b->boundary->v1->x; y1 = b->boundary->v1->y;
	x2 = b->boundary->v2->x; y2 = b->boundary->v2->y;

	LENGTH_SQUARED(x1, y1, x2, y2);
}


/* cellboundary_length_check()

	returns RET_MAZE_SMALL_WALL if the
	  given boundary is shorter than
	  allowed or RET_MAZE_OK otherwise;
	threshold_squared must be preset by
	  calling Voronoi_maze_init() or
	  maze_make() */

char cellboundary_length_check(cellboundary_t *b) {

	vertex_coord_t d = cellboundary_length_squared(b);

	if (d < 0.0)	/* infinite boundary */
		return RET_MAZE_OK;

	return (d < threshold_squared) ? RET_MAZE_SMALL_WALL : RET_MAZE_OK;
}

#endif


#ifndef FORTUNE_NO_DELAUNAY

/* connection_length_squared()

	returns the length of the given
	  connection, squared */

static site_coord_t connection_length_squared(connect_t *b) {

	site_coord_t x1, y1, x2, y2;

	x1 = b->s1->x; y1 = b->s1->y;
	x2 = b->s2->x; y2 = b->s2->y;

	LENGTH_SQUARED(x1, y1, x2, y2);
}


/* connection_length_check()

	returns RET_MAZE_SMALL_WALL if the
	  given triangle side is shorter than
	  allowed or RET_MAZE_OK otherwise;
	threshold_squared must be preset by
	  calling Delaunay_maze_init() or
	  maze_make() */

char connection_length_check(connect_t *b) {

	site_coord_t d = connection_length_squared(b);

	return (d < threshold_squared) ? RET_MAZE_SMALL_WALL : RET_MAZE_OK;
}

#endif


/* allocate_maze()
   allocate_maze_edge()

	allocate memory for the maze, maze_bits
	  and maze_edge arrays and initializes
	  them;
	these functions are called from
	  Voronoi_maze_init() and
	  Delaunay_maze_init();
	the functions return RET_MAZE_NO_MEM
	  if there is not enough memory */

static char allocate_maze(void) {

	int i;

	/* allocate memory for maze cells and walls */

	maze = (mazedata_t *)malloc(maze_size * sizeof(mazedata_t));
	if (maze == NULL)
		return RET_MAZE_NO_MEM;

	maze_bits = (uint8_t *)malloc(maze_bits_size);
	if (maze_bits == NULL) {
		free(maze);
		return RET_MAZE_NO_MEM;
	}

	for (i = 0; i < maze_size; i++) {
		maze[i].index = MAZE_INDEX_UNDEFINED;
		maze[i].previous = NULL;
	}

	/* start with every wall open in maze_bits */

	memset(maze_bits, 0xFF, maze_bits_size);

	return RET_MAZE_OK;
}

static char allocate_maze_edge(void) {

	maze_edge = (int *)malloc(maze_edge_size * sizeof(int));

	if (maze_edge == NULL) {
		free(maze);
		free(maze_bits);
		return RET_MAZE_NO_MEM;
	}

	return RET_MAZE_OK;
}


/* set_maze_threshold()

	calculates threshold_squared;
	num is the number of maze cell walls
	  and sum is the sum of their lengths;
	threshold_squared is the average length
	  divided by four, multiplied by mult,
	  and squared */

static void set_maze_threshold(int num, vertex_coord_t sum, vertex_coord_t mult) {

	if (num > 0) {
		threshold_squared = (sum / num) * mult;
		threshold_squared /= 4.0;
		threshold_squared *= threshold_squared;
		#if DEBUG_MAZE > 1
		fprintf(stderr, "average length = %f\n", sum / num);
		fprintf(stderr, "threshold = %f\n", sqrt(threshold_squared));
		#endif
	}
}


#ifndef FORTUNE_NO_VORONOI

/* Voronoi_maze_init()

	type must be MAZE_VORONOI or
	  MAZE_VORONOI_INF;
	input is a list of Voronoi cells */

char Voronoi_maze_init(uint8_t type, celldata_t *input, int input_size, int32_t random_seed,
					   float thresh_mult) {

	char ret = RET_MAZE_OK;
	int i, j, k;
	int m, n;
	cellboundary_t *b;
	mazedata_t *mdata;
	uint8_t *acc;
	uint8_t mask;
	vertex_coord_t len;
	vertex_coord_t total_len = 0.0;
	int total_walls = 0;

	/* the extra fields in input must be NULL */

	for (i = 0; i < input_size; i++)
		if (input[i].extra != NULL)
			return RET_FORTUNE_PARAM_ERROR;

	vsource = input;
	maze_size = input_size;
	source_size = input_size;
	source_type = type;

	/* create maze and maze_bits arrays */

	maze_bits_size = 0;

	switch (source_type) {

		case MAZE_VORONOI:
			for (i = 0; i < source_size; i++) {
				if (vsource[i].infinity)
					maze_size--;
				else
					maze_bits_size += (vsource[i].sides + 7) >> 3;
			}
			break;

		case MAZE_VORONOI_INF:
			for (i = 0; i < source_size; i++) {
				maze_bits_size += (vsource[i].sides + 7) >> 3;
			}
			break;
	}

	ret = allocate_maze();
	if (ret != RET_MAZE_OK)
		return ret;

	/* attach maze cells to source and mark edge cells */

	maze_edge_size = 0;

	switch (source_type) {

		case MAZE_VORONOI:

			for (m = i = j = 0; i < source_size; i++) {

				#if DEBUG_MAZE > 1
				fprintf(stderr, "cell %02x ", i);
				#endif

				if (vsource[i].infinity) {
					vsource[i].extra = NULL;
					#if DEBUG_MAZE > 1
					fprintf(stderr, "infinite\n");
					#endif
					continue;
				}

				vsource[i].extra = &maze[m];
				maze[m].available = vsource[i].sides;
				maze[m].access = &maze_bits[j];

				maze[m].edge = 0;

				for (b = vsource[i].blist, n = vsource[i].sides; n; b = b->right, n--) {

					if (b->neighbour->infinity) {
						#if DEBUG_MAZE > 1
						#if 1
						maze[m].edge = 1;
						maze_edge_size++;
						fprintf(stderr, "-> edge");
						break;
						#else
						if (!maze[m].edge) {
							fprintf(stderr, "-> infinite neighbours");
							maze[m].edge = 1;
							maze_edge_size++;
						}
						for (k = 0; k < source_size; k++)
							if (b->neighbour->s == vsource[k].s) {
								fprintf(stderr, " %02x", k);
								break;
							}
						#endif
						#else
						maze[m].edge = 1;
						maze_edge_size++;
						break;
						#endif
					}

					len = cellboundary_length_squared(b);
					if (len > 0.0) {
						total_len += sqrt(len);
						total_walls++;
					}
				}

				#if DEBUG_MAZE > 1
				fprintf(stderr, "\n");
				#endif

				m++;
				j += (vsource[i].sides + 7) >> 3;
			}

			break;

		case MAZE_VORONOI_INF:

			for (i = j = 0; i < source_size; i++) {

				#if DEBUG_MAZE > 1
				fprintf(stderr, "cell %02x ", i);
				#endif

				vsource[i].extra = &maze[i];
				maze[i].available = vsource[i].sides;
				maze[i].access = &maze_bits[j];

				if (vsource[i].infinity) {
					#if DEBUG_MAZE > 1
					fprintf(stderr, "infinite -> edge");
					#endif
					maze[i].edge = 1;
					maze_edge_size++;
				}
				else {
					maze[i].edge = 0;

					/* some infinite cells contain very long boundaries,
					     which causes the average length to be too high;
					     this is why we only take non-infinite cells
					     into account when determining threshold_squared */

					for (b = vsource[i].blist, n = vsource[i].sides; n; b = b->right, n--) {
						len = cellboundary_length_squared(b);
						if (len > 0.0) {
							total_len += sqrt(len);
							total_walls++;
						}
					}
				}

				#if DEBUG_MAZE > 1
				fprintf(stderr, "\n");
				#endif

				j += (vsource[i].sides + 7) >> 3;
			}

			break;
	}

	/* determine wall threshold value */

	set_maze_threshold(total_walls, total_len, thresh_mult);

	/* create maze edge cell list */

	ret = allocate_maze_edge();
	if (ret != RET_MAZE_OK)
		return ret;

	/* fill maze_edge with numbers of edge cells;
	   also mark small boundaries and edge boundaries */

	switch (source_type) {

		case MAZE_VORONOI:

			for (i = j = 0; i < source_size; i++) {

				mdata = vsource[i].extra;
				if (mdata == NULL)
					continue;

				if (mdata->edge) {
					maze_edge[j] = i;
					j++;
				}

				for (acc = mdata->access, mask = 0x01,
					b = vsource[i].blist, n = vsource[i].sides; n;
					b = b->right, n--, mask <<= 1) {

					if (!mask) {
						mask = 0x01;
						acc++;
					}

					if (cellboundary_length_check(b) != RET_MAZE_OK ||
						b->neighbour->infinity != 0) {
						*acc &= ~mask;
						mdata->available--;
					}
				}
			}

			break;

		case MAZE_VORONOI_INF:

			for (i = j = 0; i < source_size; i++) {

				if (maze[i].edge) {
					maze_edge[j] = i;
					j++;
				}

				for (acc = maze[i].access, mask = 0x01,
					b = vsource[i].blist, n = vsource[i].sides; n;
					b = b->right, n--, mask <<= 1) {

					if (!mask) {
						mask = 0x01;
						acc++;
					}

					if (cellboundary_length_check(b) != RET_MAZE_OK) {
						*acc &= ~mask;
						maze[i].available--;
					}
				}
			}
	}

	#if DEBUG_MAZE
	fprintf(stderr, "%u edge cells:\n", maze_edge_size);
	for (i = 0; i < maze_edge_size; i++)
		fprintf(stderr, "%02x ", maze_edge[i]);
	fprintf(stderr, "\n");
	#endif

    RND_SEED(random_seed);

	/* select one edge cell as starting point */

	k = maze_edge[MAZE_RND(maze_edge_size)];

	vptr = vstart = &vsource[k];

	prev_index = MAZE_INDEX_UNDEFINED;

	#if DEBUG_MAZE > 1
	fprintf(stderr, "start cell = %02x\n", k);
	#endif

	return RET_MAZE_OK;
}


/* Voronoi_maze_advance()

	if ptr is not NULL, it should point to
	  a variable which recieves a pointer
	  to the cell we have reached */

char Voronoi_maze_advance(celldata_t **ptr) {

	char ret = RET_MAZE_OK;
	mazedata_t *mdata;
	cellboundary_t *b;
	boundary_t *bptr = NULL;
	celldata_t *vseg = NULL;
	int a;
	int m, n, nseg;
	uint8_t *acc;
	uint8_t mask;
	vertex_coord_t sq, max_sq;

	if (prev_index == MAZE_INDEX_UNDEFINED) {
		mdata = vptr->extra;
		if (mdata->index == MAZE_INDEX_UNDEFINED) {	/* we have just started */
			mdata->index = prev_index = 0;
			goto update_neighbours_Voronoi;
		}
	}

	mdata = vptr->extra;
	b = NULL;

	if (mdata->available <= 0) {

		/* we've hit a dead end, so walk backwards and
		     find an available neighbour;
		   once we walk all the way back to square one,
		     we're (hopefully) done */

		while (mdata->available <= 0) {

			mdata = vptr->extra;
			vptr = mdata->previous;

			if (vptr == NULL) {

				/* make sure every cell has been visited;
				     if not, find a visited cell with at
				     least one unvisited neighbour and
				     take it from there */

				for (a = 0, nseg = -1, max_sq = 0.0; a < source_size; a++) {

					mdata = vsource[a].extra;
					if (mdata == NULL)
						continue;

					if (mdata->index == MAZE_INDEX_UNDEFINED) {

						/* this cell has not been visited */

						#if DEBUG_MAZE > 1
						if (nseg < 0) {
							fprintf(stderr, "unvisited cell found!\n");
							nseg = 0;
						}
						#endif

						for (b = vsource[a].blist, n = 0; n < vsource[a].sides;
							b = b->right, n++) {

							mdata = b->neighbour->extra;
							if (mdata == NULL)	/* edge boundary */
								continue;

							if (mdata->index != MAZE_INDEX_UNDEFINED) {

								/* this neighbour cell has been visited;
								     find the longest available boundary */

								sq = cellboundary_length_squared(b);
								#if DEBUG_MAZE > 1
								fprintf(stderr, "sq = %8f", sq);
								#endif
								if (sq > max_sq) {
									vseg = &vsource[a];
									nseg = n;
									vptr = b->neighbour;
									bptr = b->boundary;
									max_sq = sq;
									#if DEBUG_MAZE > 1
									fprintf(stderr, " <- new max_sq");
									#endif
								}
								#if DEBUG_MAZE > 1
								fprintf(stderr, "\n");
								#endif
							}
						}
					}
				}

				/* if we have checked every cell and haven't
				     found a new vptr, all is well */

				if (vptr == NULL)
					return RET_MAZE_FINISHED;

				/* vptr has been visited, so break down
					 the wall between it and vseg */

				mdata = vptr->extra;
				prev_index = mdata->index + 1;

				for (b = vptr->blist, m = 0; m < vptr->sides;
					b = b->right, m++)
					if (b->boundary == bptr) {
						mdata->access[m >> 3] |= 1 << (m & 7);
						mdata->available++;
						break;
					}

				/* let the previous pointer in vseg stay
				     NULL instead of pointing it to vptr;
					 this way we avoid having to walk
					 back through the already-filled
					 area */

				mdata = vseg->extra;
				mdata->access[nseg >> 3] |= 1 << (nseg & 7);
				mdata->index = prev_index;

				vptr = vseg;
				ret = RET_MAZE_SEGMENT;

				goto update_neighbours_Voronoi;
			}

			mdata = vptr->extra;
		}

		prev_index = mdata->index;

		goto done_advancing_Voronoi;
	}

	/* select one available boundary at random */

	a = (mdata->available == 1) ? 0 : MAZE_RND(mdata->available);

	for (acc = mdata->access, mask = 0x01,
		b = vptr->blist, n = vptr->sides; n;
		b = b->right, n--, mask <<= 1) {

		if (!mask) {
			mask = 0x01;
			acc++;
		}

		if ((*acc) & mask) {
			mdata = b->neighbour->extra;
			if (mdata->index == MAZE_INDEX_UNDEFINED) {
				if (!a)
					break;
				a--;
			}
		}
	}

	prev_index++;
	mdata = b->neighbour->extra;
	mdata->index = prev_index;
	mdata->previous = vptr;

	vptr = b->neighbour;

	update_neighbours_Voronoi:

	/* update available counts in neighbour cells */

	mdata = vptr->extra;
	if (mdata != NULL) {
		acc = mdata->access;
		for (b = vptr->blist, n = vptr->sides,
			mask = 0x01; n;
			b = b->right, n--, mask <<= 1) {
			if (!mask) {
				mask = 0x01;
				acc++;
			}
			if ((*acc) & mask) {
				mdata = b->neighbour->extra;
				if (mdata != NULL)
					mdata->available--;
			}
		}
	}

	done_advancing_Voronoi:

	if (ptr != NULL)
		*ptr = vptr;

	return ret;
}


/* Voronoi_maze_finish()

	sets up the access bits of the cells
	  and releases memory that is only in
	  use while the algorithm is running */

char Voronoi_maze_finish(uint8_t type) {

	celldata_t *cn;
	cellboundary_t *b;
	mazedata_t *mdata;
	mazedata_t *ndata;
	int i, j;
	int n;
	uint8_t *acc;
	uint8_t mask;
	vertex_coord_t len_sq;
	vertex_coord_t max_len_sq;
	int doorway_side;

	/* place walls at cell edges */

	#if DEBUG_MAZE
	fprintf(stderr, "converting access bitfields to walls...\n");
	#endif

	for (i = 0; i < source_size; i++) {

		mdata = vsource[i].extra;
		if (mdata == NULL)
			continue;

		acc = mdata->access;
		mask = 0x01;

		for (b = vsource[i].blist, n = 0; n < vsource[i].sides;
			b = b->right, n++, mask <<= 1) {

			if (!mask) {
				mask = 0x01;
				acc++;
			}

			cn = b->neighbour;
			if (cn == NULL)
				continue;

			ndata = cn->extra;
			if (ndata == NULL)
				continue;

			if (mdata->index != ndata->index - 1 &&
				mdata->index != ndata->index + 1)
				*acc &= ~mask;
		}
	}

	/* open entry and exit doorways;
		 the latter is in the edge cell with
		 the highest index;
	   usually there are more than one such cell,
	     so we need to find out which one has
	     the longest cell boundary to use as
	     the doorway */

	#if 1

	#if DEBUG_MAZE
	#define REPORT_EDGE_BOUNDARY {			\
		fprintf(stderr,						\
			"best candidate so far: "		\
			"side %d (len_sq = %.6f)\n",	\
			doorway_side, len_sq);			\
	}
	#else
	#define REPORT_EDGE_BOUNDARY { }
	#endif

	/* max_len_sq must be set to zero before using
		 the FIND_LONGEST_VORONOI_EDGE() macro */

	#define FIND_LONGEST_VORONOI_EDGE(CELL, ACTION) {		\
		for (b = CELL->blist, n = 0; n < CELL->sides;		\
			b = b->right, n++) {							\
			if (b->neighbour->infinity) {					\
				len_sq = cellboundary_length_squared(b);	\
				if (len_sq > max_len_sq) {					\
					max_len_sq = len_sq;					\
					doorway_side = n;						\
					ACTION;									\
					REPORT_EDGE_BOUNDARY;					\
				}											\
			}												\
		}													\
	}

	#define OPEN_VORONOI_DOORWAY(WHERE) {		\
		mdata = WHERE->extra;					\
		mdata->access[doorway_side >> 3] |=		\
			(1 << (doorway_side & 7));			\
	}

	if (type == MAZE_VORONOI) {

		for (i = j = 0; i < maze_edge_size; i++) {
			mdata = vsource[maze_edge[i]].extra;
			if (mdata->index > j)
				j = mdata->index;
		}

		#if DEBUG_MAZE
		fprintf(stderr, "exit index: %d\n", j);
		#endif

		for (i = 0, doorway_side = 0, max_len_sq = 0.0, vptr = NULL;
			i < maze_edge_size; i++) {

			cn = &vsource[maze_edge[i]];
			mdata = cn->extra;

			if (mdata->index < j)
				continue;

			#if DEBUG_MAZE
			fprintf(stderr, "possible exit cell: %d\n", maze_edge[i]);
			#endif

			FIND_LONGEST_VORONOI_EDGE(cn, { vptr = cn; } );
		}

		OPEN_VORONOI_DOORWAY(vptr);

		max_len_sq = 0.0;

		#if DEBUG_MAZE
		for (i = 0; i < maze_edge_size; i++)
			if (vstart == &vsource[maze_edge[i]]) {
				fprintf(stderr, "entry cell: %d\n", maze_edge[i]);
				break;
			}
		#endif

		FIND_LONGEST_VORONOI_EDGE(vstart, { } );

		OPEN_VORONOI_DOORWAY(vstart);
	}

	#else	/* the old way based on random chance */

	#if DEBUG_MAZE
	#define REPORT_AVAILABLE_VORONOI_DOORWAYS(TOTAL, LAST) {		\
		fprintf(stderr, "%d edge", TOTAL);							\
		if (TOTAL != 1) fprintf(stderr, "s");						\
		fprintf(stderr,												\
			" to choose from; most recent = %d\n", LAST);			\
	}
	#else
	#define REPORT_AVAILABLE_VORONOI_DOORWAYS(TOTAL, LAST) { }
	#endif

	#define SELECT_AND_OPEN_VORONOI_DOORWAY(P) {					\
		for (b = P->blist, n = i = j = 0;							\
			n < P->sides; b = b->right, n++)						\
			if (b->neighbour->infinity) {							\
				i++;												\
				j = n;												\
			}														\
		REPORT_AVAILABLE_VORONOI_DOORWAYS(i, j);					\
		if (i > 1) {												\
			for (n = MAZE_RND(i), b = P->blist, j = 0;				\
				j < P->sides - 1; b = b->right, j++)				\
				if (b->neighbour->infinity) {						\
					if (!n) break;									\
					n--;											\
				}													\
		}															\
		mdata = P->extra;											\
		mdata->access[j >> 3] |= (1 << (j & 7));					\
	}

	if (type == MAZE_VORONOI) {

		for (i = j = 0, vptr = NULL; i < maze_edge_size; i++) {
			n = maze_edge[i];
			mdata = vsource[n].extra;
			if (mdata->index > j) {
				j = mdata->index;
				vptr = &vsource[n];
				#if DEBUG_MAZE > 1
				fprintf(stderr, "%3d for exit\n", n);
				#endif
			}
		}

		#if DEBUG_MAZE
		mdata = vstart->extra;
		fprintf(stderr, "opening entry doorway... (index %d)...\n", mdata->index);
		#endif

		SELECT_AND_OPEN_VORONOI_DOORWAY(vstart);

		if (vptr != NULL) {

			#if DEBUG_MAZE
			mdata = vptr->extra;
			fprintf(stderr, "opening exit doorway... (index %d)...\n", mdata->index);
			#endif

			SELECT_AND_OPEN_VORONOI_DOORWAY(vptr);
		}
	}

	#endif

	/* destroy maze_edge */

	#if DEBUG_MAZE
	fprintf(stderr, "releasing memory...\n");
	#endif

	if (maze_edge != NULL)
		free(maze_edge);

	#if DEBUG_MAZE
	fprintf(stderr, "Voronoi maze finished\n");
	#endif

	return RET_MAZE_OK;
}


/* Voronoi_maze_cleanup()

	destroys the maze data structures
	  attached to the triangles or cells
	  in input */

char Voronoi_maze_cleanup(celldata_t *input, int input_size) {

	int i;
	mazedata_t *mdata;

	if (input == NULL)
		return RET_MAZE_OK;

	for (i = 0; i < input_size; i++) {
		mdata = input[i].extra;
		if (mdata != NULL) {
			free(mdata->access);
			free(input[i].extra);
			break;
		}
	}

	return RET_MAZE_OK;
}

#endif


#ifndef FORTUNE_NO_DELAUNAY

/* Delaunay_maze_init()

	input is a list of Delaunay triangles */

char Delaunay_maze_init(triangle_t *input, int input_size, int32_t random_seed,
						float thresh_mult) {

	char ret = RET_MAZE_OK;
	int i, j, k;
	mazedata_t *mdata;
	uint8_t *acc;
	char block;
	site_coord_t total_len = SITE_COORD_ZERO;

	/* the extra fields in input must be NULL */

	for (i = 0; i < input_size; i++)
		if (input[i].extra != NULL)
			return RET_FORTUNE_PARAM_ERROR;

	dsource = input;
	maze_size = input_size;
	source_size = input_size;
	source_type = MAZE_DELAUNAY;
	maze_bits_size = maze_size;	/* 3 sides (per triangle) fit in a byte */

	/* create maze and maze_bits arrays */

	ret = allocate_maze();
	if (ret != RET_MAZE_OK)
		return ret;

	/* attach maze cells to source and mark edge cells */

	maze_edge_size = 0;

	for (i = 0; i < source_size; i++) {

		dsource[i].extra = &maze[i];

		maze[i].available = 3;
		maze[i].access = &maze_bits[i];
		maze[i].edge = 0;

		for (k = 0; k < 3; k++) {

			#if DEBUG_MAZE > 1
			fprintf(stderr, "cell %02x ", i);
			#endif

			if (!maze[i].edge) {
				if (dsource[i].neighbour[k] == NULL) {
					maze[i].edge = 1;
					maze_edge_size++;
					#if DEBUG_MAZE > 1
					fprintf(stderr, "-> edge");
					#endif
				}
			}

			total_len += sqrt(connection_length_squared(dsource[i].conn[k]));

			#if DEBUG_MAZE > 1
			fprintf(stderr, "\n");
			#endif
		}
	}

	/* determine wall threshold value */

	set_maze_threshold(source_size * 3, total_len, thresh_mult);

	/* create maze edge cell list */

	ret = allocate_maze_edge();
	if (ret != RET_MAZE_OK)
		return ret;

	/* fill maze_edge with numbers of edge cells;
	   also mark small boundaries and edge boundaries */

	for (i = j = 0; i < source_size; i++) {

		if (maze[i].edge) {
			maze_edge[j] = i;
			j++;
		}

		for (acc = maze[i].access, k = 0; k < 3; k++) {

			block = (dsource[i].neighbour[k] == NULL) ?
				RET_MAZE_SMALL_WALL : RET_MAZE_OK;

			if (block == RET_MAZE_OK)
				block = connection_length_check(dsource[i].conn[k]);

			if (block != RET_MAZE_OK) {
				*acc &= ~(1 << k);
				maze[i].available--;
			}
		}
	}

	#if DEBUG_MAZE > 1
	fprintf(stderr, "%u edge cells:\n", maze_edge_size);
	for (i = 0; i < maze_edge_size; i++)
		fprintf(stderr, "%02x ", maze_edge[i]);
	fprintf(stderr, "\n");
	#endif

    RND_SEED(random_seed);

	/* select one edge cell as starting point */

	k = maze_edge[MAZE_RND(maze_edge_size)];

	dptr = dstart = &dsource[k];

	prev_index = MAZE_INDEX_UNDEFINED;

	#if DEBUG_MAZE > 1
	fprintf(stderr, "start cell = %02x\n", k);
	#endif

	return RET_MAZE_OK;
}


/* Delaunay_maze_advance()

	if ptr is not NULL, it should point to
	  a variable which recieves a pointer
	  to the triangle we have reached */

char Delaunay_maze_advance(triangle_t **ptr) {

	char ret = RET_MAZE_OK;
	mazedata_t *mdata;
	triangle_t *dseg = NULL;
	int a;
	int m, n, nseg, nnseg;
	uint8_t *acc;
	uint8_t mask;
	site_coord_t sq, max_sq;

	if (prev_index == MAZE_INDEX_UNDEFINED) {
		mdata = dptr->extra;
		if (mdata->index == MAZE_INDEX_UNDEFINED) {	/* just starting */
			mdata->index = prev_index = 0;
			goto update_neighbours_Delaunay;
		}
	}

	mdata = dptr->extra;

	if (mdata->available <= 0) {

		/* hit a dead end; walk backwards */

		while (mdata->available <= 0) {

			mdata = dptr->extra;
			dptr = mdata->previous;

			if (dptr == NULL) {

				/* make sure every cell has been visited */

				nseg = -1;
				nnseg = 0;
				max_sq = 0.0;

				for (a = 0; a < source_size; a++) {

					mdata = dsource[a].extra;

					if (mdata->index == MAZE_INDEX_UNDEFINED) {

						/* this cell has not been visited */

						#if DEBUG_MAZE > 1
						if (nseg < 0) {
							fprintf(stderr, "unvisited cell found!\n");
							nseg = 0;
						}
						#endif

						for (n = 0; n < 3; n++) {

							if (dsource[a].neighbour[n] == NULL)
								continue;

							mdata = dsource[a].neighbour[n]->extra;

							if (mdata->index != MAZE_INDEX_UNDEFINED) {

								/* this neighbour cell has been visited */

								sq = connection_length_squared(dsource[a].conn[n]);
								#if DEBUG_MAZE > 1
								fprintf(stderr, "sq = %8f", sq);
								#endif
								if (sq > max_sq) {
									dseg = &dsource[a];
									nseg = n;
									dptr = dsource[a].neighbour[n];
									nnseg = dsource[a].neighbour_side[n];
									max_sq = sq;
									#if DEBUG_MAZE > 1
									fprintf(stderr, " <- new max_sq");
									#endif
								}
								#if DEBUG_MAZE > 1
								fprintf(stderr, "\n");
								#endif
							}
						}
					}
				}

				/* if we have checked every cell and haven't
				     found a new dptr, all is well */

				if (dptr == NULL)
					return RET_MAZE_FINISHED;

				mdata = dptr->extra;
				prev_index = mdata->index + 1;
				mdata->access[nnseg >> 3] |= 1 << (nnseg & 7);
				mdata->available++;

				mdata = dseg->extra;
				mdata->access[nseg >> 3] |= 1 << (nseg & 7);
				mdata->index = prev_index;

				dptr = dseg;
				ret = RET_MAZE_SEGMENT;

				goto update_neighbours_Delaunay;
			}

			mdata = dptr->extra;
		}

		prev_index = mdata->index;

		goto done_advancing_Delaunay;
	}

	/* select one available boundary at random */

	a = (mdata->available == 1) ? 0 : MAZE_RND(mdata->available);

	for (acc = mdata->access, mask = 0x01, n = 0; n < 3; n++, mask <<= 1) {

		if ((*acc) & mask) {
			mdata = dptr->neighbour[n]->extra;
			if (mdata->index == MAZE_INDEX_UNDEFINED) {
				if (!a)
					break;
				a--;
			}
		}
	}

	prev_index++;
	mdata = dptr->neighbour[n]->extra;
	mdata->index = prev_index;
	mdata->previous = dptr;

	dptr = dptr->neighbour[n];

	update_neighbours_Delaunay:

	/* update available counts in neighbour cells */

	mdata = dptr->extra;
	acc = mdata->access;

	for (n = 0, mask = 0x01; n < 3; n++, mask <<= 1) {
		if ((*acc) & mask) {
			mdata = dptr->neighbour[n]->extra;
			if (mdata != NULL)
				mdata->available--;
		}
	}

	done_advancing_Delaunay:

	if (ptr != NULL)
		*ptr = dptr;

	return ret;
}


/* Delaunay_maze_finish()

	sets up the access bits of the cells
	  and releases memory that is only in
	  use while the algorithm is running */

char Delaunay_maze_finish(void) {

	triangle_t *tn;
	mazedata_t *mdata;
	mazedata_t *ndata;
	int i, j;
	int n;
	uint8_t *acc;
	site_coord_t len_sq;
	site_coord_t max_len_sq;
	int doorway_side;

	/* place walls at triangle edges */

	#if DEBUG_MAZE
	fprintf(stderr, "converting access bitfields to walls...\n");
	#endif

	for (i = 0; i < source_size; i++) {

		mdata = dsource[i].extra;
		acc = mdata->access;

		for (n = 0; n < 3; n++) {

			tn = dsource[i].neighbour[n];
			if (tn == NULL)
				continue;

			ndata = tn->extra;

			if (mdata->index != ndata->index - 1 &&
				mdata->index != ndata->index + 1)
				*acc &= ~(1 << n);
		}
	}

	/* open entry and exit doorways, in the same
	     manner as in Voronoi_maze_finish() */

	#if 1

	#if DEBUG_MAZE
	#define REPORT_EDGE_BOUNDARY {			\
		fprintf(stderr,						\
			"best candidate so far: "		\
			"side %d (len_sq = %.6f)\n",	\
			doorway_side, len_sq);			\
	}
	#else
	#define REPORT_EDGE_BOUNDARY { }
	#endif

	/* max_len_sq must be set to zero before using
		 the FIND_LONGEST_DELAUNAY_EDGE() macro */

	#define FIND_LONGEST_DELAUNAY_EDGE(CELL, ACTION) {				\
		for (n = 0; n < 3; n++) {									\
			if (CELL->neighbour[n] == NULL) {						\
				len_sq = connection_length_squared(CELL->conn[n]);	\
				if (len_sq > max_len_sq) {							\
					max_len_sq = len_sq;							\
					doorway_side = n;								\
					ACTION;											\
					REPORT_EDGE_BOUNDARY;							\
				}													\
			}														\
		}															\
	}

	#define OPEN_DELAUNAY_DOORWAY(WHERE) {		\
		mdata = WHERE->extra;					\
		mdata->access[doorway_side >> 3] |=		\
			(1 << (doorway_side & 7));			\
	}

	for (i = j = 0; i < maze_edge_size; i++) {
		mdata = dsource[maze_edge[i]].extra;
		if (mdata->index > j)
			j = mdata->index;
	}

	#if DEBUG_MAZE
	fprintf(stderr, "exit index: %d\n", j);
	#endif

	for (i = 0, doorway_side = 0, max_len_sq = SITE_COORD_ZERO, dptr = NULL;
		i < maze_edge_size; i++) {

		tn = &dsource[maze_edge[i]];
		mdata = tn->extra;

		if (mdata->index < j)
			continue;

		#if DEBUG_MAZE
		fprintf(stderr, "possible exit cell: %d\n", maze_edge[i]);
		#endif

		FIND_LONGEST_DELAUNAY_EDGE(tn, { dptr = tn; } );
	}

	OPEN_DELAUNAY_DOORWAY(dptr);

	max_len_sq = SITE_COORD_ZERO;

	#if DEBUG_MAZE
	for (i = 0; i < maze_edge_size; i++)
		if (dstart == &dsource[maze_edge[i]]) {
			fprintf(stderr, "entry cell: %d\n", maze_edge[i]);
			break;
		}
	#endif

	FIND_LONGEST_DELAUNAY_EDGE(dstart, { } );

	OPEN_DELAUNAY_DOORWAY(dstart);

	#else

	#if DEBUG_MAZE
	#define REPORT_AVAILABLE_DELAUNAY_DOORWAYS(TOTAL, LAST) {		\
		fprintf(stderr, "%d edge", TOTAL);							\
		if (TOTAL != 1) fprintf(stderr, "s");						\
		fprintf(stderr,												\
			" to choose from; most recent = %d\n", LAST);			\
	}
	#else
	#define REPORT_AVAILABLE_DELAUNAY_DOORWAYS(TOTAL, LAST) { }
	#endif

	#define SELECT_AND_OPEN_DELAUNAY_DOORWAY(P) {					\
		for (n = i = j = 0; n < 3; n++)								\
			if (P->neighbour[n] == NULL) {							\
				i++;												\
				j = n;												\
			}														\
		REPORT_AVAILABLE_DELAUNAY_DOORWAYS(i, j);					\
		if (i > 1) {												\
			for (n = MAZE_RND(i), j = 0; j < 2; j++)				\
				if (P->neighbour[j] == NULL) {						\
					if (!n) break;									\
					n--;											\
				}													\
		}															\
		mdata = P->extra;											\
		mdata->access[j >> 3] |= (1 << (j & 7));					\
	}

	for (i = j = 0, dptr = NULL; i < maze_edge_size; i++) {
		n = maze_edge[i];
		mdata = dsource[n].extra;
		if (mdata->index > j) {
			j = mdata->index;
			dptr = &dsource[n];
		}
	}

	#if DEBUG_MAZE
	mdata = dstart->extra;
	fprintf(stderr, "opening entry doorway (index %d)...\n", mdata->index);
	#endif

	SELECT_AND_OPEN_DELAUNAY_DOORWAY(dstart);

	if (dptr != NULL) {

		#if DEBUG_MAZE
		mdata = dptr->extra;
		fprintf(stderr, "opening exit doorway (index %d)...\n", mdata->index);
		#endif

		SELECT_AND_OPEN_DELAUNAY_DOORWAY(dptr);
	}

	#endif

	/* destroy maze_edge */

	#if DEBUG_MAZE
	fprintf(stderr, "releasing memory...\n");
	#endif

	if (maze_edge != NULL)
		free(maze_edge);

	#if DEBUG_MAZE
	fprintf(stderr, "Delaunay maze finished\n");
	#endif

	return RET_MAZE_OK;
}


/* Delaunay_maze_cleanup()

	destroys the maze data structures
	  attached to the triangles or cells
	  in input */

char Delaunay_maze_cleanup(triangle_t *input, int input_size) {

	int i;
	mazedata_t *mdata;

	if (input == NULL)
		return RET_MAZE_OK;

	mdata = input[0].extra;
	if (mdata != NULL) {
		free(mdata->access);
		free(input[0].extra);
	}

	return RET_MAZE_OK;
}

#endif


/* maze_make()

	runs the maze algorithm
	  from start to finish */

char maze_make(uint8_t type, void *input, int input_size, int32_t random_seed,
			   float thresh_mult) {

	char ret = RET_MAZE_OK;

	switch (type) {

		case MAZE_DELAUNAY:

			#ifndef FORTUNE_NO_DELAUNAY
			ret = Delaunay_maze_init(input, input_size,
				random_seed, thresh_mult);

			while (ret == RET_MAZE_OK) {

				ret = Delaunay_maze_advance(NULL);
				if (ret == RET_MAZE_SEGMENT)
					ret = RET_MAZE_OK;
			}

			if (ret == RET_MAZE_FINISHED)
				ret = Delaunay_maze_finish();
			#else
			ret = RET_MAZE_UNSUPPORTED;
			#endif

			break;

		default:

			#ifndef FORTUNE_NO_VORONOI
			ret = Voronoi_maze_init(type, input, input_size,
				random_seed, thresh_mult);

			while (ret == RET_MAZE_OK) {

				ret = Voronoi_maze_advance(NULL);
				if (ret == RET_MAZE_SEGMENT)
					ret = RET_MAZE_OK;
			}

			if (ret == RET_MAZE_FINISHED)
				ret = Voronoi_maze_finish(type);
			#else
			ret = RET_MAZE_UNSUPPORTED;
			#endif

			break;
	}

	return ret;
}


/* maze_cleanup()

	calls Delaunay_maze_cleanup() or
	  Voronoi_maze_cleanup() depending
	  on type */

char maze_cleanup(uint8_t type, void *input, int input_size) {

	if (type == MAZE_DELAUNAY) {
		return
			#ifndef FORTUNE_NO_DELAUNAY
			Delaunay_maze_cleanup(input, input_size);
			#else
			RET_MAZE_UNSUPPORTED;
			#endif
	}
	else {
		return
			#ifndef FORTUNE_NO_VORONOI
			Voronoi_maze_cleanup(input, input_size);
			#else
			RET_MAZE_UNSUPPORTED;
			#endif
	}
}
