/*************************************************************************
 *
 *  Maze algorithm applied to Delaunay triangles and Voronoi diagrams
 *  Copyright (C) 2012-2014 Jussi M. Hukkanen
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *************************************************************************
 *
 *  maze.c: main header file
 *
 ************************************************************************/

#ifndef _MAZE_H_INCLUDED
#define _MAZE_H_INCLUDED

#include "fortune.h"


#define RET_MAZE_OK				 0
#define RET_MAZE_NO_MEM			-2
#define RET_MAZE_SEGMENT		-7	/* if the input data contains walls that are
									     shorter than the given threshold value,
									     there may be closed segments; this
									     return value tells the application that
									     a new segment has been entered */
#define RET_MAZE_SMALL_WALL		-8	/* used in cellboundary_length_check() */
#define RET_MAZE_FINISHED		-9	/* algorithm finished */
#define RET_MAZE_UNSUPPORTED	-10	/* returned by maze_make() and maze_cleanup()
									     if the requested function is not
									     available in the build */


typedef struct mazedata_s {
	int index;			/* index number used by the algorithm */
	void *previous;		/* pointer to previous triangle or cell in solution */
	int available;		/* number of unresolved neighbours */
	char edge;			/* triangle or cell is at the edge of the maze */
	uint8_t *access;	/* one or more bitfields representing walls
						     (1 = open, 0 = blocked) */
} mazedata_t;


#define MAZE_INDEX_UNDEFINED	(-1)


/* function declarations */

#define MAZE_DELAUNAY		1	/* maze made of Delaunay triangles */
#define MAZE_VORONOI		2	/* maze made of non-infinite Voronoi cells */
#define MAZE_VORONOI_INF	3	/* maze made of a complete Voronoi diagram */

/**** TODO: option to convert Gabriel graphs into mazes ****/

#define MAZE_ACCESS_OPEN	1
#define MAZE_ACCESS_CLOSED	0

char maze_access(mazedata_t *, int);

#define maze_wall_open(X, A) (maze_access(X, A) == MAZE_ACCESS_OPEN)
#define maze_wall_closed(X, A) (maze_access(X, A) == MAZE_ACCESS_CLOSED)

char maze_make(uint8_t, void *, int, int32_t, float);
char maze_cleanup(uint8_t, void *, int);

#ifndef FORTUNE_NO_VORONOI

/* either call
     maze_make() with MAZE_VORONOI or MAZE_VORONOI_INF as the type
   or
     Voronoi_maze_init() and
     Voronoi_maze_advance()
       until the return value is RET_MAZE_FINISHED, and then
     Voronoi_maze_finish();
   when done with the data, either call
     maze_cleanup() with the same type
   or
     Voronoi_maze_cleanup() */

char Voronoi_maze_init(uint8_t, celldata_t *, int, int32_t, float);
char Voronoi_maze_advance(celldata_t **);
char Voronoi_maze_finish(uint8_t);
char Voronoi_maze_cleanup(celldata_t *, int);
char cellboundary_length_check(cellboundary_t *);

#endif

#ifndef FORTUNE_NO_DELAUNAY

/* either call
     maze_make() with MAZE_DELAUNAY as the type
   or
     Delaunay_maze_init() and
     Delaunay_maze_advance()
       until the return value is RET_MAZE_FINISHED, and then
     Delaunay_maze_finish();
   when done with the data, either call
     maze_cleanup() with the same type
   or
     Delaunay_maze_cleanup() */

char Delaunay_maze_init(triangle_t *, int, int32_t, float);
char Delaunay_maze_advance(triangle_t **);
char Delaunay_maze_finish(void);
char Delaunay_maze_cleanup(triangle_t *, int);
char connection_length_check(connect_t *);

#endif

#endif /* _MAZE_H_INCLUDED */


