/*************************************************************************
 *
 *  Orbit -- an animation utilising Fortune's algorithm
 *  Copyright (C) 2012-2014 Jussi M. Hukkanen
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *************************************************************************
 *
 *  orbit.c: main module
 *
 ************************************************************************/

#ifndef SDL
#define _GNU_SOURCE
#endif
#include <math.h>       /* declares sincosf() */
#include <stdio.h>
#include <signal.h>
#include <sys/time.h>

#include "fortune.h"
#include "opts.h"
#include "rnd.h"

#ifdef TARGET_WINDOWS   /* Windows version needs SDL timer functions */
#ifndef SDL
#define NO_VGA 1
#endif
#endif

#ifdef H_GFX
#include H_GFX
#endif

#ifdef H_PLATFORM
#include H_PLATFORM
#endif

char *window_title = "orbit";


typedef struct orbit_s {
    float ang;  /* radians */
    float step; /* radians */
} orbit_t;

typedef struct orbiter_s {
    orbit_t x;
    orbit_t y;
    orbit_t z;
    float radius; /* distance from origin */
} orbiter_t;

#define CAMERA_Z_OFFSET (-3.1)

#define MIN_RADIUS 0.5
#define MAX_RADIUS 2.5


#define TIMER_DELAY_USEC  50000 /* 50 milliseconds between ticks */

volatile int timer_flag = 0;

#ifdef VALGRIND
#define DEFAULT_VALGRIND_DURATION 100
int valgrind_frames = DEFAULT_VALGRIND_DURATION;
#endif


#define DEFAULT_ORBITERS    6
#define EXTRA_SITES         8

int orbiters = DEFAULT_ORBITERS;
int sites = DEFAULT_ORBITERS + EXTRA_SITES;

orbiter_t *orb; /* orbiters items */
site_t *site;   /* sites items */


#define DELAUNAY (function & FORTUNE_DELAUNAY_MASK)
#define VORONOI  (function & FORTUNE_VORONOI_MASK)

uint8_t function = 0;
int32_t random_seed = 0;
char *binary_name;


#define SITE_COLOUR        7
#define VERTEX_COLOUR      1


#ifndef FORTUNE_NO_VORONOI
vertex_t *vlist;    /* array of Voronoi vertices */
boundary_t *blist;  /* array of Voronoi boundaries */
celldata_t *cell_list;  /* array of Voronoi cells */
int vnum, bnum;     /* number of vertices and boundaries */
#endif

#ifndef FORTUNE_NO_DELAUNAY
triangle_t *tlist;  /* array of Delaunay triangles */
connect_t *clist;   /* array of Delaunay connections */
int tnum, cnum;     /* number of triangles and connections */
#endif


#ifndef NO_VGA

#if defined(SDL) && !defined(VALGRIND)        /* use SDL for timing if possible */

SDL_TimerID sdl_timer_id = NULL;

/* sighandler()

    sets the timer semaphore */

Uint32 sighandler(Uint32 interval, void *param) {

	timer_flag = 1;

    return interval;
}


/* wait_for_timer()

    suspends the program execution
      until the next tick of the timer */

void wait_for_timer(void) {

	timer_flag = 0;

	while (!timer_flag) { }
}


/* destroy_sdl_timer()

    kills the timer created by init_timer() */

void destroy_sdl_timer(void) {

    if (sdl_timer_id != NULL)
        SDL_RemoveTimer(sdl_timer_id);
}


/* init_timer()

    creates a handler for SIGALRM */

void init_timer(long int sec, long int usec) {

    sdl_timer_id = SDL_AddTimer((sec * 1000) + ((usec + 500) / 1000),
        sighandler, NULL);

    if (sdl_timer_id != NULL)
        atexit(destroy_sdl_timer);
}

#else               /* on Linux, we can use signals */


/* sighandler()

    sets the timer semaphore */

void sighandler(int sig) {

	timer_flag = 1;
}


/* wait_for_timer()

    suspends the program execution
      until the next tick of the timer */

void wait_for_timer(void) {

	sigset_t mask, oldmask;

	timer_flag = 0;

	sigemptyset(&mask);
	sigaddset(&mask, SIGALRM);

	sigprocmask(SIG_BLOCK, &mask, &oldmask);

	while (!timer_flag)
	    sigsuspend(&oldmask);

	sigprocmask(SIG_UNBLOCK, &mask, NULL);

	timer_flag = 0;
}


/* init_timer()

    creates a handler for SIGALRM */

void init_timer(long int sec, long int usec) {

	struct sigaction sa;
	struct itimerval itv;

	sigemptyset(&sa.sa_mask);
	sa.sa_flags = SA_RESTART;
	sa.sa_handler = sighandler;
	sigaction(SIGALRM, &sa, NULL);

	itv.it_interval.tv_sec = sec;
	itv.it_interval.tv_usec = usec;
	itv.it_value.tv_sec = itv.it_interval.tv_sec;
	itv.it_value.tv_usec = itv.it_interval.tv_usec;

	setitimer(ITIMER_REAL, &itv, NULL);
}

#endif


/* plot_site()

    draws a small circle at the given site's coordinates */

char plot_site(site_t *s, uint8_t colour) {

    return plot_dot_ex(PIXEL_X(s->x), PIXEL_Y(s->y),
      PLOT_DOT_HOLLOW, colour, BRIGHT_COLOUR(colour));
}


/* plot_dot()

    draws a dot at the given coordinates */

char plot_dot(float px, float py, uint8_t colour1, uint8_t colour2) {

    return plot_dot_ex(PIXEL_X(px), PIXEL_Y(py),
      PLOT_DOT_PHYSICAL, colour1, colour2);
}


/* plot_pixel()

    plots a single pixel at the given coordinates */

void plot_pixel(float px, float py, uint8_t colour, uint8_t flags) {

    int sx = PIXEL_X(px);
    int sy = PIXEL_Y(py);

    if (sx <       0) return;
    if (sx >=  WIDTH) return;
    if (sy <       0) return;
    if (sy >= HEIGHT) return;

    #ifdef SVGALIB
	if (flags & PLOT_DOT_PHYSICAL) {
        if (scr_frame == NULL)
            gl_setcontext(scr_physical);
        else
            gl_setcontext(scr_frame);
    }
    else {
        gl_setcontext(scr_virtual);
    }
    gl_setpixel(sx, sy, colour);
    #endif

    #ifdef SDL
    pixelColor((flags & PLOT_DOT_PHYSICAL) ? scr_physical : scr_virtual,
        sx, sy, sdl_gfx_colour(colour));
    #endif
}


/* do_trace()
   do_not_trace()

    functions to draw and not draw the
      orbiters as part of the animation */

void do_trace(float x, float y, uint8_t base_colour) {
    plot_pixel(x, y, 16 + base_colour, 0);
    plot_dot(x, y, 17 + base_colour, 15);
}

void do_not_trace(float x, float y, uint8_t base_colour) { }

void (*trace)(float, float, uint8_t);


/* create_orbiters()

    creates a set of orbiters with
      random rotation parameters */

void create_orbiters(void) {

    int i;
    float f;

    #define OFFSET 0.5

    for (i = 0, f = OFFSET; i < orbiters; i++, f += 1.0) {

        orb[i].x.ang  = RND_FLOAT(M_PI);
        orb[i].y.ang  = RND_FLOAT(M_PI * 2.0);
        orb[i].z.ang  = RND_FLOAT(M_PI);

        orb[i].y.step = (0.5 + RND_FLOAT_1) * M_PI / 64.0;

        orb[i].x.step = RND_FLOAT(M_PI) / 200.0;
        orb[i].z.step = RND_FLOAT(M_PI) / 500.0;

        orb[i].radius = MIN_RADIUS +
          ((MAX_RADIUS - MIN_RADIUS) *
          sqrt(f / (orbiters + OFFSET))
          );
    }

    /* add extra sites along screen edges */

    if (sites <= orbiters)
        return;

    #if EXTRA_SITES >= 4
    #define MARGIN_FACTOR 1.25
    site[orbiters    ].x = WINDOW_LEFT   * MARGIN_FACTOR;
    site[orbiters + 2].x = WINDOW_RIGHT  * MARGIN_FACTOR;
    site[orbiters + 1].x = site[orbiters + 3].x = 0.0;
    site[orbiters + 1].y = WINDOW_TOP    * MARGIN_FACTOR;
    site[orbiters + 3].y = WINDOW_BOTTOM * MARGIN_FACTOR;
    site[orbiters    ].y = site[orbiters + 2].y = 0.0;
    #if EXTRA_SITES == 8
    site[orbiters + 4].x = site[orbiters + 6].x = WINDOW_LEFT;
    site[orbiters + 5].x = site[orbiters + 7].x = WINDOW_RIGHT;
    site[orbiters + 4].y = site[orbiters + 5].y = WINDOW_TOP;
    site[orbiters + 6].y = site[orbiters + 7].y = WINDOW_BOTTOM;
    #endif
    #endif
}

#endif


/* parse_options()

    interprets the given list of command-line
      options */

char parse_options(int argc, char *argv[]) {

    #define CMDLINE_OPTC_HELP       'h'
    #define CMDLINE_OPTS_HELP           "help"
    #define CMDLINE_OPTC_DELAUNAY   'd'
    #define CMDLINE_OPTS_DELAUNAY       "delaunay"
    #define CMDLINE_OPTC_VORONOI    'v'
    #define CMDLINE_OPTS_VORONOI        "voronoi"
    #define CMDLINE_OPTC_ORBS       'n'
    #define CMDLINE_OPTS_ORBS           "num"
    #define CMDLINE_OPTC_RNDSEED    's'
    #define CMDLINE_OPTS_RNDSEED        "seed"
    #define CMDLINE_OPTC_TRACE      't'
    #define CMDLINE_OPTS_TRACE          "no-trace"
    #define CMDLINE_OPTC_VGAMODE    'm'
    #define CMDLINE_OPTS_VGAMODE        "mode"
    #define CMDLINE_OPTC_SDLMODE    'w'
    #define CMDLINE_OPTS_SDLMODE        "window"

    #ifdef FORTUNE_NO_DELAUNAY
    #define DEFAULT_FUNCTION        CMDLINE_OPTC_VORONOI
    #elif defined(FORTUNE_NO_VORONOI)
    #define DEFAULT_FUNCTION        CMDLINE_OPTC_DELAUNAY
    #else
    /* if the line below is commented out, neither
         `-d' nor `-v' is made the default function;
         that way, when neither option is given,
         the program only traces the orbiters without
         drawing anything on top */
    #define DEFAULT_FUNCTION        CMDLINE_OPTC_VORONOI
    #endif

    struct option long_opts[] = {
      { CMDLINE_OPTS_HELP,     no_argument,       NULL, CMDLINE_OPTC_HELP     },
      { CMDLINE_OPTS_DELAUNAY, no_argument,       NULL, CMDLINE_OPTC_DELAUNAY },
      { CMDLINE_OPTS_VORONOI,  no_argument,       NULL, CMDLINE_OPTC_VORONOI  },
      { CMDLINE_OPTS_ORBS,     required_argument, NULL, CMDLINE_OPTC_ORBS     },
      { CMDLINE_OPTS_RNDSEED,  required_argument, NULL, CMDLINE_OPTC_RNDSEED  },
      { CMDLINE_OPTS_TRACE,    no_argument,       NULL, CMDLINE_OPTC_TRACE    },
      { CMDLINE_OPTS_VGAMODE,  required_argument, NULL, CMDLINE_OPTC_VGAMODE  },
      { CMDLINE_OPTS_SDLMODE,  required_argument, NULL, CMDLINE_OPTC_SDLMODE  },
      { NULL, 0, NULL, 0 }
    };
    int c;
    int i = 0;
    char *modestr;
    int index = 0;
    char ret = RET_FORTUNE_OK;
    char *optstring;

    optstring = make_opt_string(long_opts);
    if (optstring == NULL) {
      fprintf(stderr, "Out of memory reading command line options!\n");
      return RET_FORTUNE_NO_MEM;
    }

    while (1) {

      c = getopt_long(argc, argv, optstring, long_opts, &index);

      if (c == -1)
        break;

      if (optarg != NULL)
        i = atoi(optarg);

      switch (c) {

        case CMDLINE_OPTC_HELP:
          #define HELP_STR_D "Delaunay triangles "
          #define HELP_STR_V "Voronoi diagrams "
          #define HELP_STR_B "based on their "
          printf("orbit - Copyright (C) 2012-2014 Jussi M. Hukkanen\n");
          printf("Animate a set of orbiters in 3D space and draw\n"
                 #ifndef FORTUNE_NO_DELAUNAY
                 HELP_STR_D
                 #ifndef FORTUNE_NO_VORONOI
                 "or " HELP_STR_V HELP_STR_B "\n"
                 #else
                 HELP_STR_B
                 #endif
                 #else
                 #ifndef FORTUNE_NO_VORONOI
                 HELP_STR_V HELP_STR_B
                 #endif
                 #endif
                 "projections on the screen.\n\n");
          printf("This program comes with ABSOLUTELY NO WARRANTY.\n"
                 "This is free software, and you are welcome to\n"
                 "redistribute it under the conditions of the\n"
                 "GNU General Public License, version 3 or later.\n"
                 "To view the License, see `gpl.txt'.\n\n");

          printf("Usage: %s [options]\n", binary_name);
          #ifndef FORTUNE_NO_DELAUNAY
          #ifndef FORTUNE_NO_VORONOI
          printf("   -%c, --" CMDLINE_OPTS_DELAUNAY "\n", CMDLINE_OPTC_DELAUNAY);
          printf("        creates a set of Delaunay triangles.\n");
          printf("   -%c, --" CMDLINE_OPTS_VORONOI "\n", CMDLINE_OPTC_VORONOI);
          printf("        creates a Voronoi diagram.\n"
                 "The above functions are mutually exclusive"
                 #ifdef DEFAULT_FUNCTION
                 "; if neither is\n"
                 "given, the default is `-%c'.\n",
                 DEFAULT_FUNCTION);
                 #else
                 ".\n");
                 #endif
          #endif
          #endif
          printf("   -%c N, --" CMDLINE_OPTS_ORBS " N\n", CMDLINE_OPTC_ORBS);
          printf("        sets the desired number of orbiters.\n"
                 "        N must be at least three; default is %u.\n",
                          DEFAULT_ORBITERS);
          printf("   -%c N, --" CMDLINE_OPTS_RNDSEED " N\n", CMDLINE_OPTC_RNDSEED);
          printf("        sets the seed value for the random number\n"
                 "        generator. N can be any non-negative integer.\n");
          printf("   -%c, --" CMDLINE_OPTS_TRACE "\n", CMDLINE_OPTC_TRACE);
          printf("        disables the tracing of the orbiters as\n"
                 "        part of the animation.\n");

          #ifndef VALGRIND
          #ifdef SVGALIB
          printf("   -%c MODE, --" CMDLINE_OPTS_VGAMODE " MODE\n", CMDLINE_OPTC_VGAMODE);
          printf("        selects a screen mode from the following:");
          modestr = list_vga_modes();
          if (modestr != NULL)
            printf("        The default is %s.\n", modestr);
          #endif
          #ifdef SDL
          printf("   -%c WIDTHxHEIGHT, --" CMDLINE_OPTS_SDLMODE " WIDTHxHEIGHT\n",
            CMDLINE_OPTC_SDLMODE);
          printf("        selects a screen size. The default is %u" "x%u.\n",
            SCREEN_DEFAULT_W, SCREEN_DEFAULT_H);
          #endif
          #endif

          ret = RET_FORTUNE_NO_TRIANGLES;
          break;

        case CMDLINE_OPTC_DELAUNAY:
          #ifdef FORTUNE_NO_DELAUNAY
          printf("This build does not create Delaunay triangles.\n");
		  printf("Re-compile without the FORTUNE_NO_DELAUNAY flag.\n");
          ret = RET_FORTUNE_PARAM_ERROR;
          #else
          function = FORTUNE_DELAUNAY_MASK;
          #endif
          break;

        case CMDLINE_OPTC_VORONOI:
          #ifdef FORTUNE_NO_VORONOI
          printf("This build does not create Voronoi diagrams.\n");
          printf("Re-compile without the FORTUNE_NO_VORONOI flag.\n");
          ret = RET_FORTUNE_PARAM_ERROR;
          #else
          function = FORTUNE_VORONOI_MASK;
          #endif
          break;

        case CMDLINE_OPTC_ORBS:
          if (i >= 3)
            orbiters = i;
          break;

        case CMDLINE_OPTC_RNDSEED:
          if (optarg != NULL)
            if (i >= 0)
              random_seed = i;
          break;

        case CMDLINE_OPTC_TRACE:
          trace = do_not_trace;
          break;

        #ifdef SVGALIB
        case CMDLINE_OPTC_VGAMODE:
            if (find_vga_mode(optarg) != 0)
                ret = RET_FORTUNE_PARAM_ERROR;
            break;
        #endif
        #ifdef SDL
        case CMDLINE_OPTC_SDLMODE:
            if (parse_sdl_video_mode(optarg) != 0)
                ret = RET_FORTUNE_PARAM_ERROR;
            break;
        #endif
      }
    }

    free(optstring);

    return ret;
}


/* main()
*/

int main(int argc, char *argv[]) {

    int i;
    int c;
    orbiter_t *o;
    float X, Y, Z, Xtmp;
    float r, sinr, cosr;
    site_coord_t sx, sy;
    char ret = RET_FORTUNE_OK;

    binary_name = argv[0];

    function = 0;
    trace = do_trace;

    #ifdef SVGALIB
	vga_disabledriverreport();
    #endif

    i = parse_options(argc, argv);
    if (i != 0)
        return i;

	#ifdef NO_VGA

    fputs("This program requires SDL or svgalib.\nPlease rebuild.\n", stderr);
    return 0;

    #else

    #ifdef DEFAULT_FUNCTION
    if (!function) {
        #if DEFAULT_FUNCTION == CMDLINE_OPTC_VORONOI
        #ifndef FORTUNE_NO_VORONOI
        function = FORTUNE_VORONOI_MASK;
        #endif
        #else
        #ifndef FORTUNE_NO_DELAUNAY
        function = FORTUNE_DELAUNAY_MASK;
        #endif
        #endif
    }
    #endif

    orb = (orbiter_t *)malloc(orbiters * sizeof(orbiter_t));

    if (orb == NULL) {
        fprintf(stderr,
            "Not enough memory for %u orbiters! (" SIZEOF_FORMAT " bytes needed)\n",
            orbiters, orbiters * sizeof(orbiter_t));
        return RET_FORTUNE_NO_MEM;
    }

    sites = (VORONOI) ? orbiters + EXTRA_SITES : orbiters;

    site = (site_t *)malloc(sites * sizeof(site_t));

    if (site == NULL) {
        fprintf(stderr,
            "Not enough memory for %u sites! (" SIZEOF_FORMAT " bytes needed)\n",
            sites, sites * sizeof(site_t));
        free(orb);
        return RET_FORTUNE_NO_MEM;
    }

    /* enter graphics mode */

    i = init_vga(INIT_VGA_DOUBLE_BUFFER);
    if (i != 0) {
        free(orb);
        free(site);
        return i;
    }

    /* create a random set of sites */

    RND_SEED(random_seed);

    create_orbiters();

    /* setup timer */

    get_vga_accel_caps();

    #ifndef VALGRIND
    init_timer(0, TIMER_DELAY_USEC);
    #endif

    fill_screen(scr_virtual, 0);

    #ifdef SVGALIB
    gl_setcontext(scr_virtual);
    #endif

    while (1) {

        copy_screen(scr_virtual, scr_frame);

        for (i = 0; i < orbiters; i++) {

            /* move orbiter */

            o = &orb[i];
            o->x.ang += o->x.step;
            o->y.ang += o->y.step;
            o->z.ang += o->z.step;

            /* translate coordinates */

            r = o->y.ang;
            X = cos(r) * o->radius;
            Z = sin(r) * o->radius;

            r = o->x.ang;
            Y = sin(r) * Z;
            Z = cos(r) * Z;

            r = o->z.ang;
            #if defined(TARGET_WINDOWS) || defined(VALGRIND)
            sinr = sinf(r);
            cosr = cosf(r);
            #else
            sincosf(r, &sinr, &cosr);
            #endif
            Xtmp = X;
            X = (cosr * Xtmp) - (sinr * Y);
            Y = (sinr * Xtmp) + (cosr * Y);

            /* get colour based on z coordinate */

            c = (14.999 * (MAX_RADIUS - Z)) / (MAX_RADIUS * 2.0);

            /* copy screen coordinates to site */

            Z -= CAMERA_Z_OFFSET;

            if (Z < FLT_EPSILON)
                if (Z > -FLT_EPSILON)
                    continue;

            X /= Z;
            Y /= Z;
            site[i].x = X;
            site[i].y = Y;

            /* draw orbiter and trace route */

            trace(X, Y, c);
        }

        if (DELAUNAY) {

            #ifndef FORTUNE_NO_DELAUNAY
            tlist = NULL;
            clist = NULL;

            ret = Delaunay(site, sites, &clist, &cnum, &tlist, &tnum);

            if (ret == RET_FORTUNE_OK) {

                for (i = 0; i < cnum; i++) {
                    draw_boundary(clist[i].s1->x, clist[i].s1->y,
                        clist[i].s2->x, clist[i].s2->y,
                        DRAW_BOUNDARY_PHYSICAL, 1);
                }

                #ifndef FORTUNE_MASTER_ARRAYS
                Delaunay_cleanup(clist, tlist);
                #endif
            }
            else {
                fprintf(stderr, "ERROR: Delaunay() returned %d\n", ret);
                break;
            }
            #endif
        }
        else if (VORONOI) {

            #ifndef FORTUNE_NO_VORONOI
            vlist = NULL;
            blist = NULL;

            ret = Voronoi(site, sites, &vlist, &vnum, &blist, &bnum, NULL, NULL);

            if (ret == RET_FORTUNE_OK) {

                for (i = 0; i < bnum; i++) {
                    if (blist[i].v2 == NULL)
                        draw_boundary(blist[i].v1->x, blist[i].v1->y,
                            blist[i].vector->x, blist[i].vector->y,
                            DRAW_BOUNDARY_PHYSICAL |
                            DRAW_BOUNDARY_RELATIVE, 2);
                    else
                        draw_boundary(blist[i].v1->x, blist[i].v1->y,
                            blist[i].v2->x, blist[i].v2->y,
                            DRAW_BOUNDARY_PHYSICAL, 2);
                }

			    #ifndef FORTUNE_MASTER_ARRAYS
                Voronoi_cleanup(vlist, blist, NULL);
                #endif
            }
            else {
                fprintf(stderr, "ERROR: Voronoi() returned %d\n", ret);
                break;
            }
            #endif
        }

        blit_screen(scr_frame);

        #ifndef VALGRIND
        switch (dot_getkey()) {
            #ifdef SDL
            case -1:
            #endif
            case 10:
            case 27:
            case 'q':
                ret = RET_FORTUNE_OK;
                goto all_done;
            case 32:
                dot_getch();
            default:
                wait_for_timer();
                break;
        }
        #else
        valgrind_frames--;
        if (valgrind_frames <= 0)
            break;
        #endif
    }

    /* free memory, return to text mode and exit */

    all_done:

    #ifdef FORTUNE_MASTER_ARRAYS
    #ifndef FORTUNE_NO_DELAUNAY
    if (DELAUNAY)
        Delaunay_cleanup(clist, tlist);
    #endif
    #ifndef FORTUNE_NO_VORONOI
    if (VORONOI)
        Voronoi_cleanup(vlist, blist, NULL);
    #endif
    #endif

    free(orb);
    free(site);

    exit_vga();

    return ret;

    #endif
}

