/*************************************************************************
 *
 *  An implementation of Fortune's algorithm
 *  Copyright (C) 2012-2014 Jussi M. Hukkanen
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *************************************************************************
 *
 *  platform-win.c: Windows-specific code
 *
 ************************************************************************/


#include "platform-win.h"


/* getline()

	simple replacement for GNU getline()
	  that does its best to handle input of
	  various kinds (i.e. LF, CR, CR+LF and
	  LF+CR line terminators);
	regardless of the input format, this
	  function always appends an LF and a
	  null character into the string */

ssize_t getline(char **lineptr, size_t *n, FILE *f) {

	int c, next;
	size_t line_len = 0;
	size_t max_line_len = 0;
	long start_this, start_next;
	char *p;
	char allocated = 0;
	char ret = 0;

	#define DEBUG_GETLINE 1

	#define READ_BYTE_ONLY(TARGET, WHERE) { TARGET = fgetc(WHERE); }

	#ifdef DEBUG_GETLINE

	#define REPORT_NO_MEM {										\
		fputs("**** No memory available for string buffer\n",	\
			stderr);											\
	}

	#if DEBUG_GETLINE > 1
	#define READ_BYTE(TARGET, WHERE) {			\
		READ_BYTE_ONLY(TARGET, WHERE);			\
		if (TARGET >= 0) {						\
			if (TARGET >= ' ' && TARGET < 0x7f)	\
				putchar(TARGET);				\
			else								\
				printf("[%02X]", TARGET);		\
		}										\
	}
	#else
	#define READ_BYTE	READ_BYTE_ONLY
	#endif

	#else

	#define REPORT_NO_MEM { }
	#define READ_BYTE	READ_BYTE_ONLY

	#endif

	if (lineptr == NULL || f == NULL) {
		#ifdef DEBUG_GETLINE
		fputs("**** Null pointer given as argument\n", stderr);
		#endif
		return -1;
	}
	if (*lineptr != NULL && n == NULL) {
		#ifdef DEBUG_GETLINE
		fputs("**** String buffer of unknown length given as argument\n", stderr);
		#endif
		return -1;
	}

	#define SEGMENT_LEN  40		/* how many bytes to allocate at a time */
	#define STARTING_LEN (SEGMENT_LEN + 2)	/* includes newline + null */

	#define ALLOC_LINEPTR(NEW_SIZE) {							\
		if (*lineptr == NULL)									\
			*lineptr = (char *)malloc(NEW_SIZE);				\
		else													\
			*lineptr = (char *)realloc(*lineptr, NEW_SIZE);		\
		if (*lineptr == NULL) {									\
			REPORT_NO_MEM;										\
			ret = -1;											\
			goto done_getting;									\
		}														\
		if (n != NULL)											\
			*n = NEW_SIZE;										\
		max_line_len = NEW_SIZE - 2;							\
	}

	/* initialize *lineptr if it is empty */

	if (*lineptr == NULL) {
		ALLOC_LINEPTR(STARTING_LEN);
	}
	else {
		max_line_len = *n - 2;
		if (max_line_len <= 0)
			ALLOC_LINEPTR(STARTING_LEN);
	}

	/* read and copy one character at a time */

	line_len = 0;
	p = *lineptr;

	while (1) {

		READ_BYTE(c, f);

		if (c < 0) break;		/* eof */

		if (c != '\n' && c != '\r') {

			while (line_len >= max_line_len) {
				ALLOC_LINEPTR(max_line_len + SEGMENT_LEN);
				p = *lineptr + line_len;
			}

			*p = c;
			p++;
			line_len++;

			continue;
		}

		if (f != stdin) {

			READ_BYTE(next, f);

			if (next >= 0) {

				if (c == '\n') {
					if (next == '\r')
						c = 0;
				}
				else {
					if (next == '\n')
						c = 0;
				}

				if (c)
					fseek(f, -1, SEEK_CUR);
			}
		}

		break;
	}

	#ifdef DEBUG_GETLINE
	#if DEBUG_GETLINE > 1
	printf("\n");
	fprintf(stderr, "line_len = %u\n", line_len);
	#endif
	#endif

	p = *lineptr + line_len;
	p[0] = '\n';
	p[1] = 0;

	ret = line_len + 1;		/* including newline */

	done_getting:

	return ret;
}
