/*************************************************************************
 *
 *  Rubberband algorithm for creating evenly distributed sets of points
 *  Copyright (C) 2012-2014 Jussi M. Hukkanen
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *************************************************************************
 *
 *  rubber.c: main module
 *
 ************************************************************************/

#include "rubber.h"
#include "rnd.h"


/* distance parameters */

static site_coord_t distance_min;
static site_coord_t distance_range;
static site_coord_t distance_range_half;
static site_coord_t distance_avg;   /* average distance */
static site_coord_t distance_avg_2; /* distance_avg * 2.0 */


/* four corners of the area to fill */

static site_t area_tl;
static site_t area_tr;
static site_t area_bl;
static site_t area_br;


/* total number of sites created */

static int total_sites = 0;


/* where the next site goes in the rubberband */

static rubberband_t *rubberband_ptr = NULL;
static rubberband_t *rubberband_ptr_next = NULL;


/* random number generator */

#define RUBBERBAND_RND(LIMIT) RND_FLOAT(LIMIT)


/* distance_squared()

    returns the distance from site s1
      to site s2, squared */

site_coord_t distance_squared(site_t *s1, site_t *s2) {

    site_coord_t dx = s2->x - s1->x;
    site_coord_t dy = s2->y - s1->y;

    return (dx * dx) + (dy * dy);
}


/* distance()

    returns the square root of distance() */

site_coord_t distance(site_t *s1, site_t *s2) {

    return sqrt(distance_squared(s1, s2));
}


/* distance_from_line()

    returns the distance from site sc to the
      line drawn from site sa to site sb */

static site_coord_t distance_from_line(site_t *sa, site_t *sb,
                                       site_t *sc) {

    site_coord_t A = sa->y - sb->y; /* -dy */
    site_coord_t B = sb->x - sa->x; /*  dx */
    site_coord_t C = -(A * sa->x + B * sa->y);
    double numerator = fabs((A * sc->x) + (B * sc->y) + C);
    double denominator = sqrt((A * A) + (B * B));

    return (denominator < DBL_EPSILON) ?
      SITE_COORD_ZERO : numerator / denominator;
}


/* positive_side_of_line()

    returns nonzero if site sc is on the
      'positive' side of a line drawn from
      site sa to site sb;
    the 'positive' side of a line is the
      right side if RUBBERBAND_POSITIVE_Y_DOWN
      is defined and the left side otherwise */

#define LINE_TEST(X1, Y1, X2, Y2, X3, Y3) \
    ((Y1) - (Y2)) * (X3) +                \
    ((X2) - (X1)) * (Y3) +                \
    ((Y2) * (X1)) - ((X2) * (Y1))

static char positive_side_of_line(site_t *sa, site_t *sb,
                                  site_t *sc) {

    site_coord_t x1, y1;
    site_coord_t x2, y2;

    x1 = sa->x; y1 = sa->y;
    x2 = sb->x; y2 = sb->y;

    return (LINE_TEST(x1, y1, x2, y2, sc->x, sc->y) > SITE_COORD_ZERO) ?
      1 : 0;
}


/* set_rubberband_outside_flags()

    sets the appropriate bits in the outside
      field of the given rubberband segment */

static void set_rubberband_outside_flags(rubberband_t *r) {

    site_coord_t x1 = r->site->x;
    site_coord_t y1 = r->site->y;
    site_coord_t x2 = r->right->site->x;
    site_coord_t y2 = r->right->site->y;

    /* set the outside field to zero if both
         endpoints of this segment are inside
         the limit area */

    r->outside = 0x00;

    if (x1 > area_tl.x - SITE_COORD_EPSILON)
      if (x1 < area_tr.x + SITE_COORD_EPSILON)
        if (y1 > area_bl.y - SITE_COORD_EPSILON)
          if (y1 < area_tl.y + SITE_COORD_EPSILON)
            return;

    if (x2 > area_tl.x - SITE_COORD_EPSILON)
      if (x2 < area_tr.x + SITE_COORD_EPSILON)
        if (y2 > area_bl.y - SITE_COORD_EPSILON)
          if (y2 < area_tl.y + SITE_COORD_EPSILON)
            return;

    /* if both points are outside the same
         horizontal or vertical edge, set all bits
         so that no more sites are added here */

    if (x1 < area_tl.x && x2 < area_tl.x) {
      r->outside = RUBBERBAND_OUTSIDE; return;
    }
    if (x1 > area_tr.x && x2 > area_tr.x) {
      r->outside = RUBBERBAND_OUTSIDE; return;
    }
    if (y1 > area_tl.y && y2 > area_tl.y) {
      r->outside = RUBBERBAND_OUTSIDE; return;
    }
    if (y1 < area_bl.y && y2 < area_bl.y) {
      r->outside = RUBBERBAND_OUTSIDE; return;
    }

    /* compare the segment to the four corners
         of the limit area are set the bits
         accordingly */

    if (LINE_TEST(x1, y1, x2, y2, area_tl.x, area_tl.y) < SITE_COORD_ZERO)
      r->outside |= RUBBERBAND_OUTSIDE_TOP_LEFT;

    if (LINE_TEST(x1, y1, x2, y2, area_tr.x, area_tr.y) < SITE_COORD_ZERO)
      r->outside |= RUBBERBAND_OUTSIDE_TOP_RIGHT;

    if (LINE_TEST(x1, y1, x2, y2, area_bl.x, area_bl.y) < SITE_COORD_ZERO)
      r->outside |= RUBBERBAND_OUTSIDE_BOTTOM_LEFT;

    if (LINE_TEST(x1, y1, x2, y2, area_br.x, area_br.y) < SITE_COORD_ZERO)
      r->outside |= RUBBERBAND_OUTSIDE_BOTTOM_RIGHT;
}


/* add_site()

    adds a new site to the list;
    returns a pointer to the list item or NULL
      if there is not enough memory */

static site_list_t *add_site(site_coord_t x, site_coord_t y) {

    site_t *s;
    site_list_t *sl;

    sl = (site_list_t *)malloc(sizeof(site_list_t));
    if (sl == NULL)
      return NULL;

    s = (site_t *)malloc(sizeof(site_t));
    if (s == NULL) {
      free(sl);
      return NULL;
    }

    total_sites++;

    s->x = x;
    s->y = y;

    sl->site = s;
    sl->next = site_list;
    site_list = sl;

    return sl;
}


/* destroy_site_list()

    releases the memory occupied by the site list */

static char destroy_site_list(void) {

    site_list_t *n;

    while (site_list != NULL) {
      n = site_list->next;
      free(site_list->site);
      free(site_list);
      site_list = n;
    }

    return RET_RUBBERBAND_OK;
}


/* add_site_to_rubberband()

    adds a new site to the rubberband;
    returns a pointer to the list item or NULL
      if there is not enough memory */

static rubberband_t *add_site_to_rubberband(site_t *s,
                                            rubberband_t *left) {

    rubberband_t *r;
    rubberband_t *right;

    r = (rubberband_t *)malloc(sizeof(rubberband_t));

    if (r == NULL)
      return NULL;

    r->site = s;
    r->outside = 0x00;
    r->dist_inside = SITE_COORD_ZERO;
    r->left = left;
    r->right = NULL;

    if (left != NULL) {

      right = left->right;
      left->right = r;
      r->right = right;

      if (right != NULL)
        right->left = r;
    }

    return r;
}


/* remove_site_from_rubberband()

    deletes the given rubberband site and
      connects its two neighbours together */

static char remove_site_from_rubberband(rubberband_t *r) {

    if (r == NULL)
      return RET_RUBBERBAND_OK;

    r->left->right = r->right;
    r->right->left = r->left;

    if (r == rubberband_start)
      rubberband_start = r->right;

    if (r == rubberband_ptr)
      rubberband_ptr = r->left;

    if (r == rubberband_ptr_next)
      rubberband_ptr_next = r->right;

    free(r);

    return RET_RUBBERBAND_OK;
}


/* destroy_rubberband()

    releases the memory occupied by the
      rubberband structure */

static char destroy_rubberband(void) {

    rubberband_t *r;
    rubberband_t *rn;
    long len;

    len = 0;
    r = rubberband_start;

    while (r != NULL) {
      len++;
      rn = r->right;
      if (rn == rubberband_start)
        break;
      r = rn;
    }

    for (r = rubberband_start; len; len--) {
      rn = r->right;
      free(r);
      r = rn;
    }

    return RET_RUBBERBAND_OK;
}


/* create_rubberband()

	creates the first two sites at positions
	  x1,y1 and x2,y2 and connects them with
	  the rubberband */

static char create_rubberband(site_coord_t x1, site_coord_t y1,
                              site_coord_t x2, site_coord_t y2) {

    site_list_t *s1;
    site_list_t *s2;

    /* make sure this is a fresh start */

    if (site_list != NULL)
      destroy_site_list();

    if (rubberband_start != NULL)
      destroy_rubberband();

    /* create first two sites */

    s1 = add_site(x1, y1);
    if (s1 == NULL)
      goto out_of_memory_creating_rubberband;

    s2 = add_site(x2, y2);
    if (s2 == NULL)
      goto out_of_memory_creating_rubberband;

    /* create initial rubberband */

    rubberband_start = add_site_to_rubberband(s1->site, NULL);
    if (rubberband_start == NULL)
      goto out_of_memory_creating_rubberband;

    rubberband_start->left  = rubberband_start;
    rubberband_start->right = rubberband_start;

    if (add_site_to_rubberband(s2->site, rubberband_start) == NULL)
      goto out_of_memory_creating_rubberband;

    set_rubberband_outside_flags(rubberband_start);
    set_rubberband_outside_flags(rubberband_start->right);

    /* get ready to add more sites */

    rubberband_ptr = rubberband_start;

    return RET_RUBBERBAND_OK;

    /* something's gone wrong */

    out_of_memory_creating_rubberband:

    rubberband_exit();

    return RET_RUBBERBAND_NO_MEM;
}


/* rubberband_add()

    creates at least one new site and
      adds it to the rubberband;
    if the rubberband segment pointed to
      by rubberband_ptr is long enough,
      more sites are created accordingly;
    the number of new sites is returned
      in the variable pointed to by
      snum_target if it is not NULL;
    returns RET_RUBBERBAND_FINISHED if
      no more sites can be added inside
      the area */

char rubberband_add(int *snum_target) {

    site_coord_t dx, dy;
    site_coord_t cx, cy;
    site_coord_t dist;
    site_coord_t ndist;
    site_coord_t step;
    site_coord_t advance;
    int divisions;
    site_t *s1;
    site_t *s2;
    site_list_t *sl;
    rubberband_t *r;
    rubberband_t *rl;
    rubberband_t *rr;
    rubberband_t *r1_left;
    rubberband_t *r2_right;
    int snum = 0;
    char convex;

    if (rubberband_ptr == NULL)
      return RET_RUBBERBAND_PARAM_ERROR;
    if (rubberband_ptr->right == NULL)
      return RET_RUBBERBAND_PARAM_ERROR;

    /* check if this segment is entirely outside */

    if (rubberband_ptr->outside >= RUBBERBAND_OUTSIDE) {

      for (r = rubberband_ptr->right; ; r = r->right) {

        if (r == rubberband_ptr) {
          if (snum_target != NULL)
            *snum_target = snum;
          return RET_RUBBERBAND_FINISHED;
        }

        if (r->outside < RUBBERBAND_OUTSIDE) {
          rubberband_ptr = r;
          break;
        }
      }
    }

    rubberband_ptr_next = rubberband_ptr->right;

    /* find the endpoints of this segment and
         the distance between them */

    s1 = rubberband_ptr->site;
    s2 = rubberband_ptr->right->site;
    r1_left = rubberband_ptr->left;
    r2_right = rubberband_ptr->right->right;

    dx = s2->x - s1->x;
    dy = s2->y - s1->y;
    dist = distance(s1, s2);

    if (dist < SITE_COORD_EPSILON)
      goto rubberband_add_done;

    /* create one or more new sites as needed;
         dx and dy contain the direction vector
         (normalized to a length of one) */

    rl = rubberband_ptr;

    dx /= dist;
    dy /= dist;
    #define SCALE_DELTA(VALUE, MULT) (VALUE * MULT)
    #define INT_DIV(N, D) (floor(N / D))

    #define NEW_SITE_STEP(D) {                                 \
                                                               \
      cx = s1->x + SCALE_DELTA(dx, (D));                       \
      cy = s1->y + SCALE_DELTA(dy, (D));                       \
                                                               \
      /* select random distance from the line */               \
                                                               \
      ndist = rl->dist_inside;                                 \
      if (ndist < SITE_COORD_EPSILON)                          \
        ndist = distance_min + RUBBERBAND_RND(distance_range); \
      else                                                     \
        ndist = distance_min - ndist +                         \
          RUBBERBAND_RND(distance_range + ndist);              \
                                                               \
      /* add a site on the 'positive' side of the line */      \
                                                               \
      sl = add_site(cx - SCALE_DELTA(dy, ndist),               \
        cy + SCALE_DELTA(dx, ndist));                          \
      if (sl == NULL)                                          \
        return RET_RUBBERBAND_NO_MEM;                          \
                                                               \
      r = add_site_to_rubberband(sl->site, rl);                \
      if (r == NULL)                                           \
        return RET_RUBBERBAND_NO_MEM;                          \
                                                               \
      set_rubberband_outside_flags(r->left);                   \
      set_rubberband_outside_flags(r);                         \
      rl = r;                                                  \
      snum++;                                                  \
    }

    if (dist > distance_avg_2) {

      divisions = INT_DIV(dist, distance_avg);
      dist /= divisions;
      advance = dist - distance_range_half;

      while (1) {

        divisions--;
        if (divisions <= 0)
          break;

        step = advance + RUBBERBAND_RND(distance_range);

        NEW_SITE_STEP(step);
        advance += dist;
      }
    }
    else {

      if (dist < distance_min + distance_range)
        step = dist / 3.0 + RUBBERBAND_RND(dist / 3.0);
      else
        step = distance_min + RUBBERBAND_RND(distance_range);

      NEW_SITE_STEP(step);
    }

    /* make sure the rubberband remains convex */

    convex = 0;

    while (!convex) {

      for (convex = 1, r = rubberband_start; r != NULL; ) {

        rl = r->left;
        rr = r->right;

        if (!positive_side_of_line(rl->site, rr->site, r->site)) {

          ndist = distance_from_line(rl->site, rr->site, r->site);

          remove_site_from_rubberband(r);

          rl->dist_inside = ndist;

          set_rubberband_outside_flags(rl);
          convex = 0;
        }

        if (rr == rubberband_start)
          break;

        r = rr;
      }
    }

    /* move along rubberband */

    rubberband_add_done:

    rubberband_ptr = rubberband_ptr_next;

    if (snum_target != NULL)
      *snum_target = snum;

    return (snum > 0) ? RET_RUBBERBAND_OK : RET_RUBBERBAND_FINISHED;
}


/* rubberband_init()

    starts the rubberband algorithm;
    dist_avg is the preferred average distance
      between sites and dist_dev is the maximum
      allowed deviation from dist_avg (+/-);
    x1, y1, x2 and y2 determine the left, top,
      right and bottom edges of the area to be
      filled;
    random_seed is the value to seed the pseudo-
      random number generator with */

char rubberband_init(site_coord_t dist_avg, site_coord_t dist_dev,
                     site_coord_t x1, site_coord_t y1,
                     site_coord_t x2, site_coord_t y2,
                     int32_t random_seed) {

    site_coord_t x, y;
    site_coord_t dist;
    double d;
    char ret;

    site_list = NULL;
    rubberband_start = NULL;

    RND_SEED(random_seed);

    /* set up local distance parameters */

    if (dist_avg < RUBBERBAND_MIN_RANGE)
      dist_avg = RUBBERBAND_MIN_RANGE;
    if (dist_avg - dist_dev < RUBBERBAND_MIN_RANGE)
      dist_dev = dist_avg - RUBBERBAND_MIN_RANGE;

    distance_min = dist_avg - dist_dev;
    distance_range = dist_dev * 2.0;
    distance_range_half = dist_dev;
    distance_avg = dist_avg;
    distance_avg_2 = dist_avg * 2.0;

    /* set up area to fill */

    area_tl.x = area_bl.x = x1;
    area_tl.y = area_tr.y = y1;
    area_tr.x = area_br.x = x2;
    area_bl.y = area_br.y = y2;

    /* create first two sites */

    x = RUBBERBAND_RND(distance_avg_2) - dist_avg;
    y = RUBBERBAND_RND(distance_avg_2) - dist_avg;
    d = (x < SITE_COORD_ZERO) ?
       ((y < SITE_COORD_ZERO) ? 0.0 : 3.0) :
       ((y < SITE_COORD_ZERO) ? 1.0 : 2.0);
    d = (d + RND_FLOAT_1) * M_PI_2;
    #if 0
    /* place first pair in the center of the area */
    x += (x1 + x2) / 2.0;
    y += (y1 + y2) / 2.0;
    #else
    /* first pair can be anywhere inside the area */
    #if 1
    x = area_tl.x + RUBBERBAND_RND(area_tr.x - area_tl.x - distance_avg_2) + distance_avg;
    y = area_bl.y + RUBBERBAND_RND(area_tl.y - area_bl.y - distance_avg_2) + distance_avg;
    #else
    x = area_tl.x + RUBBERBAND_RND(area_tr.x - area_tl.x - distance_min - distance_min) + distance_min;
    y = area_bl.y + RUBBERBAND_RND(area_tl.y - area_bl.y - distance_min - distance_min) + distance_min;
    #endif
    #endif

    dist = distance_min + RUBBERBAND_RND(distance_range);

    ret = create_rubberband(x, y, x + (cos(d) * dist), y + (sin(d) * dist));
    if (ret != RET_RUBBERBAND_OK)
      return ret;

    return RET_RUBBERBAND_OK;
}


/* rubberband_exit()

    releases all memory allocated by the
      rubberband algorithm */

char rubberband_exit(void) {

    destroy_site_list();
    destroy_rubberband();

    return RET_RUBBERBAND_OK;
}


/* rubberband_copy_sites()

    copies the contents of site_list to site_t
      array pointed to by sout */

char rubberband_copy_sites(site_t **sout, int *snum) {

    site_list_t *n;
    site_t *sa;
    int s;

    if (sout == NULL)
      return RET_RUBBERBAND_PARAM_ERROR;
    if (*sout != NULL)
      return RET_RUBBERBAND_PARAM_ERROR;
    if (snum == NULL)
      return RET_RUBBERBAND_PARAM_ERROR;

    #if 0
    /* find total number of sites */

    for (n = site_list, s = 0; ; n = n->next) {
      if (n == NULL) break;
      if (n->site != NULL) s++; /* no site pointer should be NULL */
    }
    #endif

    /* create array for sites */

    sa = (site_t *)malloc(total_sites * sizeof(site_t));

    if (sa == NULL)
      return RET_RUBBERBAND_NO_MEM;

    /* copy sites from site list to array;
       this must be done backwards so that
         the first sites created by the
         algorithm end up as first items
         in the array */

    for (n = site_list, s = total_sites; n != NULL; n = n->next) {

      if (n->site == NULL)
        continue;

      s--;
      sa[s].x = n->site->x;
      sa[s].y = n->site->y;
    }

    *sout = sa;
    *snum = total_sites;

    return RET_RUBBERBAND_OK;
}


/* rubberband()

    runs the rubberband algorithm;
    sout and snum are sent to
      rubberband_copy_sites(), all other
      arguments are the same as those
      in rubberband_init() */

char rubberband(site_t **sout, int *snum,
                site_coord_t dist_avg, site_coord_t dist_dev,
                site_coord_t x1, site_coord_t y1,
                site_coord_t x2, site_coord_t y2,
                int32_t random_seed) {

    char ret, exit_ret;

    /* initialize */

    ret = rubberband_init(dist_avg, dist_dev, x1, y1, x2, y2, random_seed);

    if (ret != RET_RUBBERBAND_OK)
      return ret;

    /* handle rubberband loop */

    while (1) {

      ret = rubberband_add(NULL);

      if (ret == RET_RUBBERBAND_FINISHED)
        break;

      if (ret != RET_RUBBERBAND_OK)
        break;
    }

    /* copy site list to an array */

    if (ret == RET_RUBBERBAND_FINISHED)
      ret = rubberband_copy_sites(sout, snum);

    /* clean up */

    exit_ret = rubberband_exit();

    return (ret == RET_RUBBERBAND_OK) ? exit_ret : ret;
}
