/*************************************************************************
 *
 *  Rubberband algorithm for creating evenly distributed sets of points
 *  Copyright (C) 2012-2014 Jussi M. Hukkanen
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *************************************************************************
 *
 *  rubber.h: main header file
 *
 ************************************************************************/

#ifndef _RUBBER_H_INCLUDED
#define _RUBBER_H_INCLUDED

#include <stdlib.h>
#include <stdint.h>
#include <math.h>   /* needed for sqrt() */
#include <float.h>  /* needed for FLT_EPSILON */


#define RET_RUBBERBAND_OK              0
#define RET_RUBBERBAND_NO_MEM         -2
#define RET_RUBBERBAND_FINISHED       -5
#define RET_RUBBERBAND_PARAM_ERROR    -7


/* an application may define either the macro
     RUBBERBAND_POSITIVE_Y_DOWN or
     RUBBERBAND_POSITIVE_Y_UP to specify the
     orientation of its coordinate system;
   RUBBERBAND_POSITIVE_Y_UP is the default if
     neither is defined */

#ifdef RUBBERBAND_POSITIVE_Y_DOWN
#ifdef RUBBERBAND_POSITIVE_Y_UP
#undef RUBBERBAND_POSITIVE_Y_UP
#endif
#else
#ifndef RUBBERBAND_POSITIVE_Y_UP
#define RUBBERBAND_POSITIVE_Y_UP
#endif
#endif


/* the application may define RUBBERBAND_DEF_HEADER as
     a file containing the type definitions
     site_coord_t and site_t */

#ifdef RUBBERBAND_DEF_HEADER

#include RUBBERBAND_DEF_HEADER

#else

#ifndef _FORTUNE_H_INCLUDED

typedef float site_coord_t;

typedef struct site_s {
    site_coord_t x;
    site_coord_t y;
} site_t;

#endif

#endif

typedef struct site_list_s {	/* site in a single-linked list */
    site_t *site;
    struct site_list_s *next;
} site_list_t;

site_list_t *site_list;

#define SITE_COORD_EPSILON  FLT_EPSILON
#define SITE_COORD_ZERO 	0.0


typedef struct rubberband_s {
    site_t *site;
    uint8_t outside;	/* combination of RUBBERBAND_OUTSIDE_* bits */
    site_coord_t dist_inside;	/* distance to closest site inside */
    struct rubberband_s *left;
    struct rubberband_s *right;
} rubberband_t;

rubberband_t *rubberband_start;

#define RUBBERBAND_OUTSIDE_TOP_LEFT      0x01
#define RUBBERBAND_OUTSIDE_TOP_RIGHT     0x02
#define RUBBERBAND_OUTSIDE_BOTTOM_LEFT   0x04
#define RUBBERBAND_OUTSIDE_BOTTOM_RIGHT  0x08
#define RUBBERBAND_OUTSIDE 				(RUBBERBAND_OUTSIDE_TOP_LEFT    | \
										 RUBBERBAND_OUTSIDE_TOP_RIGHT   | \
										 RUBBERBAND_OUTSIDE_BOTTOM_LEFT | \
										 RUBBERBAND_OUTSIDE_BOTTOM_RIGHT)


#define RUBBERBAND_MIN_RANGE 0.02   /* the smallest allowable distance */


/* function declarations */

char rubberband_init(site_coord_t, site_coord_t,
    site_coord_t, site_coord_t, site_coord_t, site_coord_t, int32_t);
char rubberband_exit(void);
char rubberband_add(int *);
char rubberband_copy_sites(site_t **, int *);

/* either call
     rubberband()
   or
     rubberband_init(),
     rubberband_add() until it returns RET_RUBBERBAND_FINISHED, and then
     [optionally] rubberband_copy_sites() and
     rubberband_exit() */

char rubberband(site_t **, int *, site_coord_t, site_coord_t,
    site_coord_t, site_coord_t, site_coord_t, site_coord_t, int32_t);

/* distance() and distance_squared() may be useful outside rubber.c */

site_coord_t distance(site_t *s1, site_t *s2);
site_coord_t distance_squared(site_t *s1, site_t *s2);


#endif /* _RUBBER_H_INCLUDED */
