/*************************************************************************
 *
 *  Sites -- a tool for creating arrays of sites `triangle'
 *  Copyright (C) 2012-2014 Jussi M. Hukkanen
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *************************************************************************
 *
 *  sites.c: main interface
 *
 ************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdint.h>
#include <float.h>
#include "opts.h"
#include "rnd.h"

#ifdef H_PLATFORM
#include H_PLATFORM
#endif


#define RET_OK        0
#define RET_NO_SITES -2
#define RET_NO_MEM   -3


#define WINDOW_WIDTH  (WINDOW_RIGHT  - WINDOW_LEFT)
#define WINDOW_HEIGHT (WINDOW_TOP    - WINDOW_BOTTOM)

float WINDOW_LEFT, WINDOW_RIGHT;
float WINDOW_TOP, WINDOW_BOTTOM;

#define SCREEN_DEFAULT_W    640
#define SCREEN_DEFAULT_H    480

int WIDTH = SCREEN_DEFAULT_W;
int HEIGHT = SCREEN_DEFAULT_H;

#define SITE_X_RANGE  WINDOW_WIDTH
#define SITE_Y_RANGE  WINDOW_HEIGHT
#define SITE_X_MIN    WINDOW_LEFT
#define SITE_Y_MIN    WINDOW_BOTTOM

#define SITE_MARGIN   0.1

#define DEFAULT_SITES   12

int sites = DEFAULT_SITES;

#define SITE_COORD_EPSILON   FLT_EPSILON

typedef float site_coord_t;

typedef struct site_s {
	site_coord_t x;
	site_coord_t y;
} site_t;

site_t *site; /* sites items */



#define MIN_DIVISIONS      2
#define MAX_DIVISIONS    100

uint8_t divisions = 0;
int32_t random_seed = 0;


/* create_random_sites()
*/

void create_random_sites(void) {

	uint16_t x, y;
	int i;

	if (!divisions) {
		for (i = 0; i < sites; i++) {
			site[i].x = RND_FLOAT(SITE_X_RANGE) + SITE_X_MIN;
			site[i].y = RND_FLOAT(SITE_Y_RANGE) + SITE_Y_MIN;
		}
	}
	else {
		for (i = 0; i < sites; i++) {
			x = RND_INT(divisions + 1);
			y = RND_INT(divisions + 1);
			site[i].x = (((x * SITE_X_RANGE) / divisions) + SITE_X_MIN);
			site[i].y = (((y * SITE_Y_RANGE) / divisions) + SITE_Y_MIN);
		}
	}
}


/* parse_video_mode()
*/

char parse_video_mode(char *str) {

	if (sscanf(str, "%d" "x%d", &WIDTH, &HEIGHT) != 2)
		return -1;
	if (WIDTH < 1 || HEIGHT < 1)
		return 1;

	return 0;
}


/* parse_options()

	interprets the given list of command-line
	  options */

char parse_options(int argc, char *argv[]) {

	#define CMDLINE_OPTC_HELP       'h'
	#define CMDLINE_OPTS_HELP           "help"
	#define CMDLINE_OPTC_SITES      'n'
	#define CMDLINE_OPTS_SITES          "num"
	#define CMDLINE_OPTC_RNDSEED    's'
	#define CMDLINE_OPTS_RNDSEED        "seed"
	#define CMDLINE_OPTC_GRID       'g'
	#define CMDLINE_OPTS_GRID           "grid"
	#define CMDLINE_OPTC_SDLMODE    'w'
	#define CMDLINE_OPTS_SDLMODE        "window"

	struct option long_opts[] = {
		{ CMDLINE_OPTS_HELP,     no_argument,       NULL, CMDLINE_OPTC_HELP     },
		{ CMDLINE_OPTS_SITES,    required_argument, NULL, CMDLINE_OPTC_SITES    },
		{ CMDLINE_OPTS_RNDSEED,  required_argument, NULL, CMDLINE_OPTC_RNDSEED  },
		{ CMDLINE_OPTS_GRID,     required_argument, NULL, CMDLINE_OPTC_GRID     },
		{ CMDLINE_OPTS_SDLMODE,  required_argument, NULL, CMDLINE_OPTC_SDLMODE  },
		{ NULL, 0, NULL, 0 }
	};
	int c;
	int i = 0;
	int index = 0;
	char ret = RET_OK;
	char *optstring;

	optstring = make_opt_string(long_opts);
	if (optstring == NULL) {
		fprintf(stderr, "Out of memory reading command line options!\n");
		return RET_NO_MEM;
	}

	while (1) {

		c = getopt_long(argc, argv, optstring, long_opts, &index);

		if (c == -1)
			break;

		if (optarg != NULL)
			i = atoi(optarg);

		switch (c) {

			case CMDLINE_OPTC_HELP:
				printf("sites - Copyright (C) 2012-2014 Jussi M. Hukkanen\n");
				printf("Scatter a random set of sites (points) inside a\n"
					   "rectangular area on a plane. The output of this\n"
					   "program can be used as input to `triangle'.\n\n");
				printf("This program comes with ABSOLUTELY NO WARRANTY.\n"
				       "This is free software, and you are welcome to\n"
				       "redistribute it under the conditions of the\n"
				       "GNU General Public License, version 3 or later.\n"
				       "To view the License, see `gpl.txt'.\n\n");

				printf("Usage: %s [options]\n", argv[0]);
				printf("   -%c N, --" CMDLINE_OPTS_SITES " N\n", CMDLINE_OPTC_SITES);
				printf("        sets the desired number of sites (points)\n"
					   "        which are placed at random locations.\n"
					   "        N must be at least three. The default is %u.\n",
								DEFAULT_SITES);
				printf("   -%c N, --" CMDLINE_OPTS_RNDSEED " N\n", CMDLINE_OPTC_RNDSEED);
				printf("        sets the seed value for the random number\n"
					   "        generator. N can be any non-negative integer.\n");
				printf("   -%c N, --" CMDLINE_OPTS_GRID " N\n", CMDLINE_OPTC_GRID);
				printf("        causes the sites to be aligned to a grid,\n"
					   "        i.e. the screen is divided into N by N\n"
					   "        blocks. N must be between %u and %u.\n",
								MIN_DIVISIONS, MAX_DIVISIONS);
				printf("   -%c WIDTHxHEIGHT, --" CMDLINE_OPTS_SDLMODE " WIDTHxHEIGHT\n",
								CMDLINE_OPTC_SDLMODE);
				printf("        selects the size of the area to fill. This\n"
					   "        only affects the aspect ratio of the area,\n"
					   "        not the coordinates themselves.\n"
					   "        The default is %u" "x%u.\n",
								SCREEN_DEFAULT_W, SCREEN_DEFAULT_H);

				ret = RET_NO_SITES;
				break;

			case CMDLINE_OPTC_SITES:
				if (i >= 3)
					sites = i;
				break;

			case CMDLINE_OPTC_RNDSEED:
				if (optarg != NULL)
					if (i >= 0)
						random_seed = i;
				break;

			case CMDLINE_OPTC_GRID:
				if (i >= MIN_DIVISIONS && i <= MAX_DIVISIONS)
					divisions = i;
				break;

			case CMDLINE_OPTC_SDLMODE:
				if (parse_video_mode(optarg) != 0)
					ret = RET_NO_SITES;
				break;
		}
	}

	free(optstring);

	return ret;
}


/* main()
*/

int main(int argc, char *argv[]) {

	int i, j;
	int s;
	site_coord_t xd, yd;

	i = parse_options(argc, argv);
	if (i != 0)
		return i;

	/* set up the WINDOW_* variables */

	if (WIDTH < HEIGHT) {
		WINDOW_TOP = 1.0 - SITE_MARGIN;
		WINDOW_RIGHT = ((1.0 - SITE_MARGIN) * WIDTH) / HEIGHT;
	}
	else {
		WINDOW_RIGHT = 1.0 - SITE_MARGIN;
		WINDOW_TOP = ((1.0 - SITE_MARGIN) * HEIGHT) / WIDTH;
	}

	WINDOW_LEFT = -WINDOW_RIGHT;
	WINDOW_BOTTOM = -WINDOW_TOP;

	/* scatter a set of sites */

	site = (site_t *)malloc(sites * sizeof(site_t));

	if (site == NULL) {
		fprintf(stderr,
			"Not enough memory for %u sites! (" SIZEOF_FORMAT " bytes needed)\n",
			sites, sites * sizeof(site_t));
		return RET_NO_MEM;
	}

	RND_SEED(random_seed);

	make_new_sites:

	create_random_sites();

	/* see if there are any duplicate sites
		 and remove them from the array;
	   if the grid parameter is given, then
		 the higher the requested number of
		 sites, the higher the probability
		 that create_random_sites() will
		 place more than one site in the
		 same coordinates;
	   if duplicates are found, the sites
		 are rearranged in the array and the
		 total number of sites is decreased
		 accordingly, but the memory area
		 allocated for the array stays the
		 same */

	for (i = 0, s = sites; i < s; i++) {

		xd = site[i].x;
		yd = site[i].y;

		for (j = i + 1; j < s; j++) {
			if (site[j].x > xd - SITE_COORD_EPSILON)
				if (site[j].x < xd + SITE_COORD_EPSILON)
					if (site[j].y > yd - SITE_COORD_EPSILON)
						if (site[j].y < yd + SITE_COORD_EPSILON) {
							s--;
							site[j].x = site[s].x;
							site[j].y = site[s].y;
						}
		}
	}

	if (s < 3)
		goto make_new_sites;

	/* print sites into stdout */

	for (i = 0; i < s; i++)
		printf("%10.6f,%10.6f\n", site[i].x, site[i].y);

	/* clean up */

	free(site);

	return RET_OK;
}

