/*************************************************************************
 *
 *  Triangle -- an implementation of Fortune's algorithm
 *  Copyright (C) 2012-2014 Jussi M. Hukkanen
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *************************************************************************
 *
 *  triangle.c: main interface and demo program
 *
 ************************************************************************/

#include <stdio.h>
#include <string.h>     /* declares strdup() */
#include <signal.h>
#include <sys/time.h>

#define FORTUNE_PUBLIC
#include "fortune.h"
#include "maze.h"
#include "gabriel.h"
#include "opts.h"

#ifdef TARGET_WINDOWS   /* Windows version needs SDL timer functions */
#ifndef SDL
#define NO_VGA 1        /* without SDL, only produce the quiet version */
#endif
#endif

#ifdef H_PLATFORM
#include H_PLATFORM
#endif

char *window_title = "triangle";


#define SWEEP_ADVANCE 0.01

#define SWEEP_DELAY_USEC  50000 /* 50 milliseconds between ticks */


volatile int timer_flag = 0;

#ifndef NO_VGA

#ifdef SVGALIB
#define DIRECT_TO_PHYSICAL 1
#endif

#include H_GFX

#if 1       /* draw filled Voronoi cells */
#include "dotcell.h"
#else
#define CELL_BOUNDARIES_ONLY
#endif

char quiet_mode = 0;
char filled_cells = 0;

#if 1
#define VGA_FONT_INTEGRATED 1
uint8_t vga_font[] = {
    #include "thin8x8.txt"
};
#endif

#else   /* NO_VGA */

#if 1
#define WINDOW_RIGHT  (4.0 / 3.0)
#else
#define WINDOW_RIGHT  (1.0)
#endif
#define WINDOW_LEFT   (-WINDOW_RIGHT)
#define WINDOW_TOP    ( 1.0)
#define WINDOW_BOTTOM (-1.0)
#define WINDOW_WIDTH  (WINDOW_RIGHT  - WINDOW_LEFT)
#define WINDOW_HEIGHT (WINDOW_TOP    - WINDOW_BOTTOM)

#define WIDTH         640
#define HEIGHT        480

#define quiet_mode 1

#endif


#define DEFAULT_STROKE_WIDTH    0.00125
#define MAX_STROKE_WIDTH        0.1
#define DEFAULT_MARKER_FREQ     4

float svg_stroke = DEFAULT_STROKE_WIDTH;
uint8_t svg_marker_frequency = 0;
uint8_t svg_circles = 0;


#define DEFAULT_SITES   12

int sites = DEFAULT_SITES;

site_t *site; /* sites items */

typedef struct site_list_s {
    site_coord_t x;
    site_coord_t y;
    struct site_list_s *next;
} site_list_t;

site_list_t *site_list;

char *site_list_file = NULL;
float site_scale = 1.0;


#define DELAUNAY (function & FORTUNE_DELAUNAY_MASK)
#define VORONOI  (function & FORTUNE_VORONOI_MASK)

#define FORTUNE_GABRIEL_MASK (FORTUNE_DELAUNAY_MASK << FORTUNE_VORONOI_MASK)
#define GABRIEL  (function & FORTUNE_GABRIEL_MASK)

uint8_t function;


#define MIN_DIVISIONS      2
#define MAX_DIVISIONS    100

#define MAX_SHRINK_PIXELS 19

uint8_t shrink_width = 1;
uint8_t maze_type = 0;
float maze_short_threshold_mult = 1.0;
int32_t random_seed = 0;


#define MAZE_INTERACTIVE 1  /* repeated keypresses trace the maze progress */


#ifndef NO_VGA

#define SITE_COLOUR        7
#define DOT_COLOUR_EVENT   6
#define DOT_COLOUR_CIRCLE  4
#define VERTEX_COLOUR      1

int *beach; /* SCREEN_WIDTH y coordinates */

#endif


#ifndef FORTUNE_NO_VORONOI
vertex_t *vlist;        /* array of Voronoi vertices */
boundary_t *blist;      /* array of Voronoi boundaries */
celldata_t *cell_list;  /* array of Voronoi cells */
int vnum, bnum;         /* number of vertices and boundaries */
int actual_vnum;        /* number of non-infinite vertices */
#endif

#ifndef FORTUNE_NO_DELAUNAY
triangle_t *tlist;      /* array of Delaunay triangles */
connect_t *clist;       /* array of Delaunay connections */
int tnum, cnum;         /* number of triangles and connections */
#endif


#define BUFFER_LEN 16
char buffer[BUFFER_LEN + 1];


#ifndef NO_VGA      /* without screen output, there's no need for a timer */

#if defined(SDL) && !defined(VALGRIND)        /* use SDL for timing if possible */

SDL_TimerID sdl_timer_id = NULL;

/* sighandler()

    sets the timer semaphore */

Uint32 sighandler(Uint32 interval, void *param) {

	timer_flag = 1;

    return interval;
}


/* wait_for_timer()

    suspends the program execution
      until the next tick of the timer */

void wait_for_timer(void) {

	timer_flag = 0;

	while (!timer_flag) { }
}


/* destroy_sdl_timer()

    kills the timer created by init_timer() */

void destroy_sdl_timer(void) {

    if (sdl_timer_id != NULL)
        SDL_RemoveTimer(sdl_timer_id);
}


/* init_timer()

    creates a handler for SIGALRM */

void init_timer(long int sec, long int usec) {

    sdl_timer_id = SDL_AddTimer((sec * 1000) + ((usec + 500) / 1000),
        sighandler, NULL);

    if (sdl_timer_id != NULL)
        atexit(destroy_sdl_timer);
}

#else               /* on Linux, we can use signals */


/* sighandler()

    sets the timer semaphore */

void sighandler(int sig) {

	timer_flag = 1;
}


/* wait_for_timer()

    suspends the program execution
      until the next tick of the timer */

void wait_for_timer(void) {

	sigset_t mask, oldmask;

	timer_flag = 0;

	sigemptyset(&mask);
	sigaddset(&mask, SIGALRM);

	sigprocmask(SIG_BLOCK, &mask, &oldmask);

	while (!timer_flag)
	    sigsuspend(&oldmask);

	sigprocmask(SIG_UNBLOCK, &mask, NULL);

	timer_flag = 0;
}


/* init_timer()

    creates a handler for SIGALRM */

void init_timer(long int sec, long int usec) {

	struct sigaction sa;
	struct itimerval itv;

	sigemptyset(&sa.sa_mask);
	sa.sa_flags = SA_RESTART;
	sa.sa_handler = sighandler;
	sigaction(SIGALRM, &sa, NULL);

	itv.it_interval.tv_sec = sec;
	itv.it_interval.tv_usec = usec;
	itv.it_value.tv_sec = itv.it_interval.tv_sec;
	itv.it_value.tv_usec = itv.it_interval.tv_usec;

	setitimer(ITIMER_REAL, &itv, NULL);
}

#endif

#endif /* NO_VGA */


#ifndef NO_VGA

/* plot_site()

    draws a small circle at the given site's coordinates
      in scr_virtual */

char plot_site(site_t *s, uint8_t colour) {

    return plot_dot_ex(PIXEL_X(s->x), PIXEL_Y(s->y),
      PLOT_DOT_HOLLOW, colour, BRIGHT_COLOUR(colour));
}


/* plot_dot()

    draws a dot at the given coordinates */

char plot_dot(float px, float py, uint8_t colour) {

    return plot_dot_ex(PIXEL_X(px), PIXEL_Y(py),
      PLOT_DOT_PHYSICAL, colour, BRIGHT_COLOUR(colour));
}


/* draw_line()

    draws the given line l on the screen */

char draw_line(line_t *l, uint8_t colour) {

    float x, y;
    int xs, ys;

    if (l->A > -DBL_EPSILON && l->A < DBL_EPSILON) {

        if (l->B > -DBL_EPSILON && l->B < DBL_EPSILON)
            return 0;

        y = -l->C / l->B;
        ys = PIXEL_Y(y);

        if (ys < 0 || ys >= SCREEN_HEIGHT) return 0;

        #ifdef SVGALIB
        #ifdef DIRECT_TO_PHYSICAL
        gl_setcontext(scr_physical);
        #else
        gl_setcontext(scr_frame);
        #endif
        gl_hline(0, ys, SCREEN_WIDTH - 1, colour);
        #endif

        #ifdef SDL
        hlineColor(scr_physical,
            0, SCREEN_WIDTH - 1, ys,
            sdl_gfx_colour(colour));
        #endif

        return 0;
    }
    else
        return draw_boundary(
            (l->B * (WINDOW_TOP) + l->C) / -l->A, WINDOW_TOP,
            (l->B * (WINDOW_BOTTOM) + l->C) / -l->A, WINDOW_BOTTOM,
            DRAW_BOUNDARY_PHYSICAL, colour);

    return 0;
}


/* draw_parab()

    adds the beach arc ba to the beach array */

char draw_parab(beach_arc_t *ba, float sweep) {

    site_coord_t px;
    float x;
    uint16_t xs, xe;
    int ys;
    float a, b;

    if (ba == NULL)
      return 0;

    px = ba->s->x;

    a = -1.0 / (2.0 * (sweep - ba->s->y));
    b = (ba->s->y + sweep) / 2.0;

    if (a > -FLT_EPSILON && a < FLT_EPSILON)
      return 0;

    #if 0   /* ignore sites outside the screen area */
    if (px < WINDOW_LEFT ) return 0;
    if (px > WINDOW_RIGHT) return 0;
    #endif

    xs = 0;
    xe = SCREEN_WIDTH;

    if (ba->split_left != NULL)
      if (ba->split_left->x > WINDOW_LEFT) {
        xs = PIXEL_X(ba->split_left->x);
      }

    if (ba->split_right != NULL)
      if (ba->split_right->x < WINDOW_RIGHT) {
        xe = PIXEL_X(ba->split_right->x);
        if (xe > SCREEN_WIDTH)
          xe = SCREEN_WIDTH;
      }

    for ( ; xs < xe; xs++) {

       x = WORLD_X_(xs + 0.5) - px;
       ys = PIXEL_Y((a * x * x) + b);

       if (ys >= SCREEN_HEIGHT)
         continue;

       if (beach[xs] > ys)
         beach[xs] = ys;
    }

    return 0;
}


/* draw_beach()

    draws the beach line on the screen */

char draw_beach(void) {

    uint16_t xs, ys;
    uint8_t colour = 0x54;
    uint8_t shift = 0;

    #ifdef SVGALIB
    #ifdef DIRECT_TO_PHYSICAL
	gl_setcontext(scr_physical);
    #else
	gl_setcontext(scr_frame);
    #endif
    #endif

    for (xs = 0; xs < SCREEN_WIDTH; xs++, shift ^= 4) {

        ys = beach[xs];
        if (ys >= SCREEN_HEIGHT) continue;
        if (ys <= 0) continue;

        #ifdef SVGALIB
        gl_setpixel(xs, ys, (colour >> shift) & 0x0F);
        #endif

        #ifdef SDL
        pixelColor(scr_physical, xs, ys,
            sdl_gfx_colour((colour >> shift) & 0x0F));
        #endif
    }

    return 0;
}

#endif


/* read_sites_from_file()

    reads site_list_file and fills site_list
      with coordinates found */

char read_sites_from_file(void) {

    FILE *fh;
    char *line = NULL;
    size_t size;
    ssize_t len;
    int n;
    site_list_t *sl;
    site_coord_t x, y;

    sites = 0;
    site_list = NULL;

	if (site_list_file == NULL) {
        fh = stdin;
    }
    else {
        fh = fopen(site_list_file, "rb");
        if (fh == NULL) {
            fprintf(stderr, "File `%s' could not be read.\n", site_list_file);
            return 1;
        }
    }

    while (1) {

        len = getline(&line, &size, fh);
        if (len <= 0)
            break;

        if (line[0] == '\n')
            break;

        n = sscanf(line, "%f,%f", &x, &y);
        if (n < 2) {
            if (n < 0)
                break;
            continue;
        }

        sl = malloc(sizeof(site_list_t));
        if (sl == NULL) {
            fprintf(stderr, "Out of memory allocating another site!\n");
            break;
        }
        sl->x = x * site_scale;
        sl->y = y * site_scale;
        sl->next = site_list;
        site_list = sl;

        sites++;
    }

    if (fh != stdin)
        fclose(fh);

    if (line != NULL)
        free(line);

    return 0;
}


/* interpret_site_list()

    fills the site array with the coordinate
      pairs in site_list and frees site_list */

void interpret_site_list(void) {

    site_list_t *sn;
    int s;

    for (s = 0; s < sites; s++) {

        sn = site_list->next;
        site[s].x = site_list->x;
        site[s].y = site_list->y;

        free(site_list);

        if (sn == NULL)
            break;

        site_list = sn;
    }
}


#ifndef NO_VGA

/* reset_screen()

    clears the screen;
    plots sites as well if plot_sites
      is nonzero */

void reset_screen(char plot_sites) {

    int i;

    fill_screen(scr_virtual, 0);

    #ifdef SDL
    rectangleColor(scr_virtual, 0, 0, WIDTH - 1, HEIGHT - 1,
        sdl_gfx_colour(1));
    #elif defined(SVGALIB)
    gl_setcontext(scr_virtual);
    gl_hline(0,          0, WIDTH - 1, 1);
    gl_hline(0, HEIGHT - 1, WIDTH - 1, 1);
    gl_line(        0, 1,         0, HEIGHT - 2, 1);
    gl_line(WIDTH - 1, 1, WIDTH - 1, HEIGHT - 2, 1);
    #endif

    if (plot_sites)
        for (i = 0; i < sites; i++)
            plot_site(&site[i], SITE_COLOUR);
}

#endif


#ifndef FORTUNE_NO_VORONOI

/* clip_boundary()

    clips the given infinite boundary against
      the screen edges and returns the endpoint
      in *tx,*ty;
    returns nonzero if the boundary is entirely
      outside the screen */

char clip_boundary(vertex_t *v1, vertex_t *vector,
                   vertex_coord_t *tx, vertex_coord_t *ty) {

    vertex_coord_t x1 = v1->x;
    vertex_coord_t y1 = v1->y;
    vertex_coord_t x2 = vector->x;
    vertex_coord_t y2 = vector->y;
    #ifdef WINDOW_COORDINATES_FLOAT
    float
    #else
    int
    #endif
        ypos, yneg;

    #ifdef FORTUNE_POSITIVE_Y_DOWN
    ypos = WINDOW_BOTTOM;
    yneg = WINDOW_TOP;
    #else
    ypos = WINDOW_TOP;
    yneg = WINDOW_BOTTOM;
    #endif

    if (x2 < 0.0) {         /* left */

        if (x1 < WINDOW_LEFT)
            return 1;

        if (y2 < 0.0) {     /* negative y */

            if (y1 < yneg)
                return 1;

            if (-x2 < -y2) {
                *tx = x1 + (x2 * (yneg - y1) / y2);
                *ty = yneg;
            }
            else {
                *ty = y1 + (y2 * (WINDOW_LEFT - x1) / x2);
                *tx = WINDOW_LEFT;
            }

            return 0;
        }
        else {              /* positive y */

            if (y1 > ypos)
                return 1;

            if (-x2 < y2) {
                *tx = x1 + (x2 * (ypos - y1) / y2);
                *ty = ypos;
            }
            else {
                *ty = y1 + (y2 * (WINDOW_LEFT - x1) / x2);
                *tx = WINDOW_LEFT;
            }

            return 0;
        }
    }
    else {                  /* right */

        if (x1 > WINDOW_RIGHT)
            return 1;

        if (y2 < 0.0) {     /* negative y */

            if (y1 < yneg)
                return 1;

            if (x2 < -y2) {
                *tx = x1 + (x2 * (yneg - y1) / y2);
                *ty = yneg;
            }
            else {
                *ty = y1 + (y2 * (WINDOW_RIGHT - x1) / x2);
                *tx = WINDOW_RIGHT;
            }

            return 0;
        }
        else {              /* positive y */

            if (y1 > ypos)
                return 1;

            if (x2 < y2) {
                *tx = x1 + (x2 * (ypos - y1) / y2);
                *ty = ypos;
            }
            else {
                *ty = y1 + (y2 * (WINDOW_RIGHT - x1) / x2);
                *tx = WINDOW_RIGHT;
            }

            return 0;
        }
    }

    return 1;
}

#endif


/* output_svg()

    prints out a list of Delaunay connections
      (if func == FORTUNE_DELAUNAY_MASK)
      or Voronoi boundaries (if func ==
      FORTUNE_VORONOI_MASK), but not both,
      in SVG file format into stdout */

char output_svg(uint8_t func) {

    int i, j;
    vertex_coord_t ex, ey;
    #ifndef FORTUNE_NO_VORONOI
    cellboundary_t *cb;
    boundary_t *b;
    #endif
    mazedata_t *m;
    mazedata_t *n;
    int svg_scale_factor;
    uint8_t marker = svg_marker_frequency;
    float svg_circle_radius = svg_stroke * 5.0;
    float length;
    uint8_t g;
    #define LENGTH_TBL_SIZE   5
    int length_tbl[LENGTH_TBL_SIZE];

    #define SVG_FMT  "%.6f"
    #define SVG_FMTQ "\"" SVG_FMT "\""

    printf("<svg xmlns=\"http://www.w3.org/2000/svg\"");
    //printf(" version=\"1.1\"");
    printf(" width=\"%u\" height=\"%u\"", WIDTH, HEIGHT);
    printf(">\n");

    printf("\t<g id=\"");
    if (func == FORTUNE_DELAUNAY_MASK)
        printf("Delaunay");
    if (func & FORTUNE_GABRIEL_MASK)
        printf("Gabriel");
    if (func == FORTUNE_VORONOI_MASK)
        printf("Voronoi");
    printf("\" stroke=\"black\" stroke-width=" SVG_FMTQ " ",
        svg_stroke);
    printf("transform=\"");
    printf("translate(%d,%d)", WIDTH >> 1, HEIGHT >> 1);

    svg_scale_factor = (WIDTH < HEIGHT) ? WIDTH : HEIGHT;
    svg_scale_factor >>= 1;
    printf(" "
        "scale(%d,%d)", svg_scale_factor,
        #ifdef FORTUNE_POSITIVE_Y_UP
        -
        #endif
        svg_scale_factor
        );

    printf("\">\n");

    printf("\t\t<rect x=" SVG_FMTQ " width=" SVG_FMTQ
        " y=" SVG_FMTQ " height=" SVG_FMTQ
        " fill=\"white\" stroke=\"none\" />\n",
        WINDOW_LEFT, WINDOW_RIGHT - WINDOW_LEFT,
        #ifdef FORTUNE_POSITIVE_Y_UP
        WINDOW_BOTTOM, WINDOW_TOP - WINDOW_BOTTOM
        #else
        WINDOW_TOP, WINDOW_BOTTOM - WINDOW_TOP
        #endif
        );

    #define SVG_LINE(X1, Y1, X2, Y2) {                      \
        printf("\t\t<line x1=" SVG_FMTQ " y1=" SVG_FMTQ     \
            " x2=" SVG_FMTQ " y2=" SVG_FMTQ,                \
            X1, Y1, X2, Y2);                                \
        marker--;                                           \
        if (!marker) {                                      \
            if (svg_marker_frequency > 0) {                 \
                printf(" stroke=\"red\"");                  \
                marker = svg_marker_frequency;              \
            }                                               \
        }                                                   \
        printf(" />\n");                                    \
        length  = ((X2) - (X1)) * ((X2) - (X1));            \
        length += ((Y2) - (Y1)) * ((Y2) - (Y1));            \
        length = sqrt(length) * svg_scale_factor;           \
        if (length < 5.0) {                                 \
            if (length < 1.0)                               \
                length_tbl[0]++;                            \
            else if (length < 2.0)                          \
                length_tbl[1]++;                            \
            else if (length < 3.0)                          \
                length_tbl[2]++;                            \
            else if (length < 4.0)                          \
                length_tbl[3]++;                            \
            else                                            \
                length_tbl[4]++;                            \
        }                                                   \
    }

    for (i = 0; i < LENGTH_TBL_SIZE; i++)
        length_tbl[i] = 0;

    switch (func) {

        #ifndef FORTUNE_NO_DELAUNAY

        case FORTUNE_DELAUNAY_MASK:
        case FORTUNE_DELAUNAY_MASK | FORTUNE_GABRIEL_MASK:

            if (maze_type > 0) {    /* draw maze */

                for (j = 0; j < tnum; j++) {
                    m = tlist[j].extra;
                    for (i = 0; i < 3; i++) {
                        if (maze_wall_open(m, i))
                            continue;
                        if (tlist[j].neighbour[i] != NULL) {
                            n = tlist[j].neighbour[i]->extra;
                            if (n != NULL)
                                if (n->index < 0)
                                    continue;
                        }
                        SVG_LINE(tlist[j].conn[i]->s1->x, tlist[j].conn[i]->s1->y,
                            tlist[j].conn[i]->s2->x, tlist[j].conn[i]->s2->y);
                    }
                    m->index = -m->index;       /* draw each boundary only once */
                }

                for (j = 0; j < tnum; j++) {   /* restore indices */
                    m = tlist[j].extra;
                    m->index = -m->index;
                }
            }
            else {

                if (GABRIEL) {      /* draw triangles and polygons */
                    for (i = 0; i < cnum; i++) {
                        g = *((uint8_t *)clist[i].extra);
                        if (!g)
                            continue;
                        SVG_LINE(clist[i].s1->x, clist[i].s1->y,
                            clist[i].s2->x, clist[i].s2->y);
                    }
                }
                else {              /* draw triangles */
                    for (i = 0; i < cnum; i++)
                        SVG_LINE(clist[i].s1->x, clist[i].s1->y,
                            clist[i].s2->x, clist[i].s2->y);
                }
            }

            break;

        #endif

        #ifndef FORTUNE_NO_VORONOI

        case FORTUNE_VORONOI_MASK:

            #define HANDLE_SVG_VORONOI_BOUNDARY(BPTR) {             \
                b = BPTR;                                           \
                if (b->v2 == NULL) {                                \
                    if (!clip_boundary(b->v1, b->vector, &ex, &ey)) \
                        SVG_LINE(b->v1->x, b->v1->y, ex, ey);       \
                }                                                   \
                else {                                              \
                    SVG_LINE(b->v1->x, b->v1->y,                    \
                        b->v2->x, b->v2->y);                        \
                }                                                   \
            }

            if (maze_type > 0) {    /* draw maze */

                for (j = 0; j < sites; j++) {
                    m = cell_list[j].extra;
                    if (m == NULL)
                        continue;
                    for (i = 0, cb = cell_list[j].blist; i < cell_list[j].sides;
                        i++, cb = cb->right) {
                        if (maze_wall_open(m, i))
                            continue;
                        if (cb->neighbour != NULL) {
                            n = cb->neighbour->extra;
                            if (n != NULL)
                                if (n->index < 0)
                                    continue;
                        }
                        HANDLE_SVG_VORONOI_BOUNDARY(cb->boundary);
                    }
                    m->index = -m->index;       /* draw each boundary only once */
                }

                for (j = 0; j < sites; j++) {   /* restore indices */
                    m = cell_list[j].extra;
                    if (m != NULL)
                        m->index = -m->index;
                }
            }
            else {                  /* draw cells */

                for (i = 0; i < bnum; i++)
                    HANDLE_SVG_VORONOI_BOUNDARY(&blist[i]);
            }

            break;

        #endif
    }

    if (svg_circles) {

        /* mark sites with circles */

        printf("\t\t<g id=\"sites\" stroke=\"blue\" fill=\"none\">\n");

        for (j = 0; j < sites; j++) {
            printf("\t\t\t<circle cx=" SVG_FMTQ " cy=" SVG_FMTQ
                " r=" SVG_FMTQ " />\n",
                site[j].x, site[j].y, svg_circle_radius);
        }

        printf("\t\t</g>\n");
    }

    printf("\t</g>\n");

    printf("\t<!--\n");
    printf("\t\tshort lines:\n");
    for (i = LENGTH_TBL_SIZE; i; ) {
        i--;
        printf("\t\t");
        printf("[%d.0, %d.0[ pixels: ", i, i + 1);
        printf("%11d lines\n", length_tbl[i]);
    }
    printf("\t-->\n");

    printf("</svg>\n");

    return RET_FORTUNE_OK;
}


/* parse_options()

    interprets the given list of command-line
      options */

char parse_options(int argc, char *argv[]) {

    #define CMDLINE_OPTC_HELP       'h'
    #define CMDLINE_OPTS_HELP           "help"
    #define CMDLINE_OPTC_DELAUNAY   'd'
    #define CMDLINE_OPTS_DELAUNAY       "delaunay"
    #define CMDLINE_OPTC_VORONOI    'v'
    #define CMDLINE_OPTS_VORONOI        "voronoi"
    #define CMDLINE_OPTC_RNDSEED    's'
    #define CMDLINE_OPTS_RNDSEED        "seed"
    #define CMDLINE_OPTC_PAINT      'p'
    #define CMDLINE_OPTS_PAINT          "paint"
    #define CMDLINE_OPTC_SHRINK     'b'
    #define CMDLINE_OPTS_SHRINK         "border"
    #define CMDLINE_OPTC_VGAMODE    'm'
    #define CMDLINE_OPTS_VGAMODE        "mode"
    #define CMDLINE_OPTC_SDLMODE    'w'
    #define CMDLINE_OPTS_SDLMODE        "window"
    #define CMDLINE_OPTC_QUIET      'q'
    #define CMDLINE_OPTS_QUIET          "quiet"
    #define CMDLINE_OPTC_STROKE     'l'
    #define CMDLINE_OPTS_STROKE         "stroke"
    #define CMDLINE_OPTC_MARKER     'c'
    #define CMDLINE_OPTS_MARKER         "marker"
    #define CMDLINE_OPTC_CIRCLE     'C'
    #define CMDLINE_OPTS_CIRCLE         "circle"
    #define CMDLINE_OPTC_SCALE      'f'
    #define CMDLINE_OPTS_SCALE          "scale"
    #define CMDLINE_OPTC_MAZE       'z'
    #define CMDLINE_OPTS_MAZE           "maze"
    #define CMDLINE_OPTC_THRESH     't'
    #define CMDLINE_OPTS_THRESH         "threshold"
    #define CMDLINE_OPTC_GABRIEL    'g'
    #define CMDLINE_OPTS_GABRIEL        "gabriel"

    struct option long_opts[] = {
      { CMDLINE_OPTS_HELP,     no_argument,       NULL, CMDLINE_OPTC_HELP     },
      { CMDLINE_OPTS_DELAUNAY, no_argument,       NULL, CMDLINE_OPTC_DELAUNAY },
      { CMDLINE_OPTS_VORONOI,  no_argument,       NULL, CMDLINE_OPTC_VORONOI  },
      { CMDLINE_OPTS_RNDSEED,  required_argument, NULL, CMDLINE_OPTC_RNDSEED  },
      { CMDLINE_OPTS_PAINT,    no_argument,       NULL, CMDLINE_OPTC_PAINT    },
      { CMDLINE_OPTS_SHRINK,   required_argument, NULL, CMDLINE_OPTC_SHRINK   },
      #ifndef NO_VGA
      #ifndef VALGRIND
      #ifdef SVGALIB
      { CMDLINE_OPTS_VGAMODE,  required_argument, NULL, CMDLINE_OPTC_VGAMODE  },
      #endif
	  #ifdef SDL
      { CMDLINE_OPTS_SDLMODE,  required_argument, NULL, CMDLINE_OPTC_SDLMODE  },
	  #endif
	  #endif
      { CMDLINE_OPTS_QUIET,    no_argument,       NULL, CMDLINE_OPTC_QUIET    },
      #endif
      { CMDLINE_OPTS_STROKE,   required_argument, NULL, CMDLINE_OPTC_STROKE   },
      { CMDLINE_OPTS_MARKER,   optional_argument, NULL, CMDLINE_OPTC_MARKER   },
      { CMDLINE_OPTS_CIRCLE,   no_argument,       NULL, CMDLINE_OPTC_CIRCLE   },
      { CMDLINE_OPTS_SCALE,    required_argument, NULL, CMDLINE_OPTC_SCALE    },
      { CMDLINE_OPTS_MAZE,     optional_argument, NULL, CMDLINE_OPTC_MAZE     },
      { CMDLINE_OPTS_THRESH,   required_argument, NULL, CMDLINE_OPTC_THRESH   },
      { CMDLINE_OPTS_GABRIEL,  no_argument,       NULL, CMDLINE_OPTC_GABRIEL  },
      { NULL, 0, NULL, 0 }
    };
    int c;
    int i = 0;
    float f = 0.0;
    char *modestr;
    int index = 0;
    char ret = RET_FORTUNE_OK;
    char *optstring;

    optstring = make_opt_string(long_opts);
    if (optstring == NULL) {
      fprintf(stderr, "Out of memory reading command line options!\n");
      return RET_FORTUNE_NO_MEM;
    }

    site_list_file = get_input_file_name(argc, argv);

    while (1) {

      c = getopt_long(argc, argv, optstring, long_opts, &index);

      if (c == -1)
        break;

      if (optarg != NULL)
        i = atoi(optarg);

      switch (c) {

        case CMDLINE_OPTC_HELP:
          printf("triangle - Copyright (C) 2012-2014 Jussi M. Hukkanen\n");
          #define HELP_STR_A "based on a given list "
          #define HELP_STR_B "of points on a plane"
          #ifdef FORTUNE_NO_DELAUNAY
          #ifndef FORTUNE_NO_VORONOI
          printf("Draw a Voronoi diagram " HELP_STR_A "\n" HELP_STR_B);
          #endif
          #else
          printf("Draw a set of Delaunay triangles "
                 #ifndef FORTUNE_NO_VORONOI
                 "and/or a Voronoi\ndiagram " HELP_STR_A HELP_STR_B
                 #else
                 HELP_STR_A "\n" HELP_STR_B
                 #endif
                 );
          #endif
          printf(".\n\n");
          printf("This program comes with ABSOLUTELY NO WARRANTY.\n"
                 "This is free software, and you are welcome to\n"
                 "redistribute it under the conditions of the\n"
                 "GNU General Public License, version 3 or later.\n"
                 "To view the License, see `gpl.txt'.\n\n");

          printf("Usage: %s [FILE] [options]\n", argv[0]);
          printf("FILE should contain a list of sites as one pair\n"
                 "of x,y coordinates per line. If no FILE is given,\n"
                 "the list is read from standard input. The output\n"
                 "from `sites', `bitmap', `even' and `dot' is\n"
                 "suitable for piping into this program.\n");
          #ifndef FORTUNE_NO_DELAUNAY
          #ifndef FORTUNE_NO_VORONOI
          printf("   -%c, --" CMDLINE_OPTS_DELAUNAY "\n", CMDLINE_OPTC_DELAUNAY);
          printf("        creates a set of Delaunay triangles.\n");
          printf("   -%c, --" CMDLINE_OPTS_VORONOI "\n", CMDLINE_OPTC_VORONOI);
          printf("        creates a Voronoi diagram.\n"
                 "The default function is --" CMDLINE_OPTS_DELAUNAY ".\n");
          #endif
          printf("   -%c, --" CMDLINE_OPTS_GABRIEL "\n", CMDLINE_OPTC_GABRIEL);
          printf("        "
                 #ifdef FORTUNE_NO_VORONOI
                 #define HELP_STR_GAB_SPACE_A " "
                 #define HELP_STR_GAB_SPACE_B "\n        "
                 #else
                 "(only with --" CMDLINE_OPTS_DELAUNAY ") "
                 #define HELP_STR_GAB_SPACE_A "\n        "
                 #define HELP_STR_GAB_SPACE_B " "
                 #endif
                 "converts the resulting" HELP_STR_GAB_SPACE_A
                 "Delaunay triangles into" HELP_STR_GAB_SPACE_B
                 "a Gabriel graph.\n");
          #endif
          printf("   -%c N, --" CMDLINE_OPTS_SCALE " N\n", CMDLINE_OPTC_SCALE);
          printf("        scales the coordinates of the sites by factor\n"
                 "        N, which must be greater than zero. Values of\n"
                 "        N less than one shrink the set of sites, while\n"
                 "        values greater than one expand it.\n");
          #define HELP_STR_VORONOI_MAZE \
                 "        is immediately followed by an equal sign and\n"   \
                 "        an argument X, which can be any string,\n"        \
                 "        infinite cells are included in the maze."
          #ifndef FORTUNE_NO_DELAUNAY
          #ifdef FORTUNE_NO_VORONOI
          printf("   -%c, --" CMDLINE_OPTS_MAZE "\n", CMDLINE_OPTC_MAZE);
          #else
          printf("   -%c[=X], --" CMDLINE_OPTS_MAZE "[=X]\n", CMDLINE_OPTC_MAZE);
          #endif
          printf("        turns the output diagram into a maze.");
          #ifndef FORTUNE_NO_VORONOI
          printf(" If the\n"
                 "        --" CMDLINE_OPTS_VORONOI " function is selected and --" CMDLINE_OPTS_MAZE "\n"
                 HELP_STR_VORONOI_MAZE);
          #endif
          printf("\n");
          printf("        Note that currently Gabriel graphs cannot be\n"
                 "        used to create mazes.\n");
          #else
          #ifndef FORTUNE_NO_VORONOI
          printf("   -%c[=X], --" CMDLINE_OPTS_MAZE "[=X]\n", CMDLINE_OPTC_MAZE);
          printf("        turns the results into a maze. If --" CMDLINE_OPTS_MAZE "\n"
                 HELP_STR_VORONOI_MAZE);
          printf("\n");
          #endif
          #endif
          printf("   -%c N, --" CMDLINE_OPTS_THRESH " N\n", CMDLINE_OPTC_THRESH);
          printf("        (only with --" CMDLINE_OPTS_MAZE ") controls the maze creating\n"
                 "        algorithm. Cell boundaries shorter than the\n"
                 "        threshold value will not be used for doorways\n"
                 "        unless absolutely necessary. The default is\n"
                 "        one quarter of the average boundary length.\n"
                 "        If --" CMDLINE_OPTS_THRESH " is used, the default value is\n"
                 "        multiplied by N. N must be greater than zero.\n");
          printf("   -%c N, --" CMDLINE_OPTS_RNDSEED " N\n", CMDLINE_OPTC_RNDSEED);
          printf("        sets the seed value for the random number\n"
                 "        generator used by the maze algorithm. N can be\n"
                 "        any non-negative integer.\n");

          #ifndef NO_VGA
          #ifndef VALGRIND
          #ifndef CELL_BOUNDARIES_ONLY
          printf("   -%c, --" CMDLINE_OPTS_PAINT "\n", CMDLINE_OPTC_PAINT);
          printf("        (only with --" CMDLINE_OPTS_VORONOI ") draws filled cells.\n"
                 "        The colour of each cell depends on the number\n"
                 "        of sides the cell has.\n");
          printf("   -%c N, --" CMDLINE_OPTS_SHRINK " N\n", CMDLINE_OPTC_SHRINK);
          printf("        (only with --" CMDLINE_OPTS_VORONOI " and --" CMDLINE_OPTS_PAINT ") draws a\n"
                 "        black border around the cells. N is the width\n"
                 "        in pixels; the default is to draw a one-pixel-\n"
                 "        wide border. N must be either an odd integer\n"
                 "        between 1 and %u, or zero, which disables this\n"
                 "        feature.\n",
                          MAX_SHRINK_PIXELS);
          #endif
          #ifdef SVGALIB
          printf("   -%c MODE, --" CMDLINE_OPTS_VGAMODE " MODE\n", CMDLINE_OPTC_VGAMODE);
          printf("        selects a screen mode from the following:");
          modestr = list_vga_modes();
          if (modestr != NULL)
            printf("        The default is %s.\n", modestr);
          #endif
          #ifdef SDL
          printf("   -%c WIDTHxHEIGHT, --" CMDLINE_OPTS_SDLMODE " WIDTHxHEIGHT\n",
            CMDLINE_OPTC_SDLMODE);
          printf("        selects a screen size. The default is %u" "x%u.\n",
            SCREEN_DEFAULT_W, SCREEN_DEFAULT_H);
          #endif
          #endif    /* VALGRIND */
          printf("   -%c, --" CMDLINE_OPTS_QUIET "\n", CMDLINE_OPTC_QUIET);
          printf("        disables graphic output. The results are\n"
                 "        instead written to standard output in Scalable\n"
                 "        Vector Graphics (SVG) format.\n"
                 "These options are available only when --" CMDLINE_OPTS_QUIET
                 " is used:\n");
          #endif    /* NO_VGA */

          printf("   -%c N, --" CMDLINE_OPTS_STROKE " N\n", CMDLINE_OPTC_STROKE);
          printf("        sets the stroke width for the output image.\n"
                 "        N must be greater than zero and less than or\n"
                 "        equal to %.3f. N defaults to %f.\n",
                          MAX_STROKE_WIDTH, DEFAULT_STROKE_WIDTH);
          printf("   -%c[=N], --" CMDLINE_OPTS_MARKER "[=N]\n", CMDLINE_OPTC_MARKER);
          printf("        draws every Nth line in the output image with\n"
                 "        a red pen instead of the default black.\n"
                 "        If --" CMDLINE_OPTS_MARKER " is given without a value for N,\n"
                 "        the default is %u.\n",
                          DEFAULT_MARKER_FREQ);
          printf("   -%c, --" CMDLINE_OPTS_CIRCLE "\n", CMDLINE_OPTC_CIRCLE);
          printf("        draws a blue circle around each site in the\n"
                 "        output image.\n");

          ret = RET_FORTUNE_NO_TRIANGLES;
          break;

        case CMDLINE_OPTC_DELAUNAY:
          #ifdef FORTUNE_NO_DELAUNAY
          printf("This build does not create Delaunay triangles.\n");
		  printf("Re-compile without the FORTUNE_NO_DELAUNAY flag.\n");
          ret = RET_FORTUNE_PARAM_ERROR;
          #else
          function |= FORTUNE_DELAUNAY_MASK;
          #endif
          break;

        case CMDLINE_OPTC_VORONOI:
          #ifdef FORTUNE_NO_VORONOI
          printf("This build does not create Voronoi diagrams.\n");
		  printf("Re-compile without the FORTUNE_NO_VORONOI flag.\n");
          ret = RET_FORTUNE_PARAM_ERROR;
          #else
          function |= FORTUNE_VORONOI_MASK;
          #endif
          break;

        case CMDLINE_OPTC_GABRIEL:
          #ifdef FORTUNE_NO_DELAUNAY
          printf("This build does not create Delaunay triangles,\n");
          printf("and therefore does not accept the --" CMDLINE_OPTS_GABRIEL " option.\n");
		  printf("Re-compile without the FORTUNE_NO_DELAUNAY flag.\n");
          ret = RET_FORTUNE_PARAM_ERROR;
          #else
          function |= FORTUNE_GABRIEL_MASK;
          #endif
          break;

        case CMDLINE_OPTC_RNDSEED:
          if (optarg != NULL)
            if (i >= 0)
              random_seed = i;
          break;

        #ifndef CELL_BOUNDARIES_ONLY
        case CMDLINE_OPTC_PAINT:
          filled_cells ^= 1;
          break;

        case CMDLINE_OPTC_SHRINK:
          if (i >= 0 && i <= MAX_SHRINK_PIXELS)
            shrink_width = (i + 1) >> 1;
          break;
        #endif

        case CMDLINE_OPTC_SCALE:
          if (optarg != NULL) {
            f = strtod(optarg, NULL);
            if (f > 0.0)
              site_scale = f;
          }
          break;

        case CMDLINE_OPTC_MAZE:
          maze_type = (optarg == NULL) ? MAZE_VORONOI : MAZE_VORONOI_INF;
          break;

        case CMDLINE_OPTC_THRESH:
          if (optarg != NULL) {
            f = strtod(optarg, NULL);
            if (f > 0.0)
              maze_short_threshold_mult = f;
          }
          break;

        #ifndef NO_VGA

        #ifdef SVGALIB
        case CMDLINE_OPTC_VGAMODE:
          if (find_vga_mode(optarg) != 0)
            ret = RET_FORTUNE_PARAM_ERROR;
          break;
        #endif
        #ifdef SDL
        case CMDLINE_OPTC_SDLMODE:
            if (parse_sdl_video_mode(optarg) != 0)
                ret = RET_FORTUNE_PARAM_ERROR;
        break;
        #endif

        case CMDLINE_OPTC_QUIET:
            quiet_mode = 1;
            break;

        #endif  /* NO_VGA */

        case CMDLINE_OPTC_STROKE:
            if (optarg != NULL) {
                f = strtod(optarg, NULL);
                if (f > 0.0) {
                    if (f > MAX_STROKE_WIDTH)
                        f = MAX_STROKE_WIDTH;
                    svg_stroke = f;
                }
            }
            break;

        case CMDLINE_OPTC_MARKER:
            svg_marker_frequency = DEFAULT_MARKER_FREQ;
            if (optarg != NULL) {
                i = (optarg[0] == '=') ?
                    atoi(optarg + 1) : atoi(optarg);
                if (i >= 1)
                    svg_marker_frequency = i;
            }
            break;

        case CMDLINE_OPTC_CIRCLE:
            svg_circles = 1;
            break;
      }
    }

    free(optstring);

    return ret;
}


/* main()
*/

int main(int argc, char *argv[]) {

    uint16_t x;
    site_coord_t sweep_y;
    int i, j;
    int s;
    site_coord_t xd, yd;
    event_t *e;
    line_t line;
    beach_arc_t *b;
    site_t ss; /* should be vertex_t, but plot_site() expects a site_t */
    triangle_list_t *trl;
    #ifndef FORTUNE_NO_DELAUNAY
    triangle_t *tm;
    #endif
    #ifndef FORTUNE_NO_VORONOI
    boundary_t *bp;
    cellboundary_t *cb;
    celldata_t *cm;
    boundary_loop_t *bound;
    #endif
    mazedata_t *maze_cell;
    mazedata_t *maze_cell_b;
    mazedata_t *maze_cell_n;
    uint8_t b1, b2, b3;
    char animate = 1;

    function = 0;

    #ifndef NO_VGA
    #ifdef SVGALIB
	vga_disabledriverreport();
    #endif
    #endif

    i = parse_options(argc, argv);
    if (i != 0)
        return i;

    if (!(function & (FORTUNE_DELAUNAY_MASK | FORTUNE_VORONOI_MASK))) {
        #ifndef FORTUNE_NO_DELAUNAY
        function |= FORTUNE_DELAUNAY_MASK;
        #else
        function |= FORTUNE_VORONOI_MASK;
        #endif
    }

    if (maze_type > 0) {
        if (DELAUNAY)
            maze_type = MAZE_DELAUNAY;
        if (GABRIEL)
            function &= ~FORTUNE_GABRIEL_MASK;
    }

    if (GABRIEL) {
        if (!DELAUNAY)
            function &= ~FORTUNE_GABRIEL_MASK;
    }

    i = read_sites_from_file();
    if (i != 0)
        return i;

    site = (site_t *)malloc(sites * sizeof(site_t));

    if (site == NULL) {
        fprintf(stderr,
            "Not enough memory for %u sites! (" SIZEOF_FORMAT " bytes needed)\n",
            sites, sites * sizeof(site_t));
        return RET_FORTUNE_NO_MEM;
    }

    #ifndef NO_VGA
    #ifdef SDL

    /* we need the WINDOW_* values, which the SDL
         version has to initialize by calling
         sdl_get_window_size() either separately
         or via init_vga() */

    sdl_get_window_size();

    #endif
    #endif

    /* read a list of sites from a file */

    interpret_site_list();

    /* see if there are any duplicate sites
         and remove them from the array;
       if duplicates are found, the sites
         are rearranged in the array and the
         total number of sites is decreased
         accordingly, but the memory area
         allocated for the array stays the
         same */

    remove_duplicate_sites:

    for (i = 0, s = sites; i < s; i++) {

        xd = site[i].x;
        yd = site[i].y;

        for (j = i + 1; j < s; j++) {
            if (site[j].x > xd - SITE_COORD_EPSILON)
                if (site[j].x < xd + SITE_COORD_EPSILON)
                    if (site[j].y > yd - SITE_COORD_EPSILON)
                        if (site[j].y < yd + SITE_COORD_EPSILON) {
                            s--;
                            site[j].x = site[s].x;
                            site[j].y = site[s].y;
                        }
        }
    }

    if (s < 3) {
        fprintf(stderr, "Less than three valid coordinate pairs entered!\n");
        free(site);
        return RET_FORTUNE_NO_TRIANGLES;
    }

    sites = s;

    if (quiet_mode) {

        /* only run the algorithm and print out an SVG file;
             if both functions have been selected, unselect
             FORTUNE_VORONOI_MASK */

        i = RET_FORTUNE_OK;

        if (function & FORTUNE_DELAUNAY_MASK) {

            #ifndef FORTUNE_NO_DELAUNAY
            tlist = NULL;
            clist = NULL;

            i = Delaunay(site, sites, &clist, &cnum, &tlist, &tnum);

            if (i == RET_FORTUNE_OK) {
                if (maze_type == MAZE_DELAUNAY) {
                    if (maze_make(MAZE_DELAUNAY, tlist, tnum, random_seed,
                        maze_short_threshold_mult) != RET_MAZE_OK)
                        i = RET_FORTUNE_PARAM_ERROR;
                }
                else if (GABRIEL) {
                    if (Delaunay_to_Gabriel(tlist, tnum, clist, cnum) != RET_FORTUNE_OK)
                        i = RET_FORTUNE_PARAM_ERROR;
                }
            }

            if (i == RET_FORTUNE_OK)
                i = output_svg(function & (FORTUNE_DELAUNAY_MASK | FORTUNE_GABRIEL_MASK));

            if (maze_type == MAZE_DELAUNAY)
                maze_cleanup(MAZE_DELAUNAY, tlist, tnum);
            else if (GABRIEL)
                Gabriel_cleanup(clist);

            Delaunay_cleanup(clist, tlist);
            #endif
        }
        else {

            #ifndef FORTUNE_NO_VORONOI
            vlist = NULL;
            blist = NULL;
            cell_list = NULL;

            i = Voronoi(site, sites, &vlist, &vnum, &blist, &bnum, &cell_list, NULL);

            if (maze_type == MAZE_VORONOI || maze_type == MAZE_VORONOI_INF) {
                if (i == RET_FORTUNE_OK) {
                    if (maze_make(maze_type, cell_list, sites, random_seed,
                        maze_short_threshold_mult) != RET_MAZE_OK)
                        i = RET_FORTUNE_PARAM_ERROR;
                }
            }

            if (i == RET_FORTUNE_OK)
                i = output_svg(FORTUNE_VORONOI_MASK);

            if (maze_type == MAZE_VORONOI || maze_type == MAZE_VORONOI_INF)
                maze_cleanup(maze_type, cell_list, sites);

            Voronoi_cleanup(vlist, blist, cell_list);
            #endif
        }

        goto all_done;
    }

    #ifndef NO_VGA

    /* enter graphics mode */

    i = init_vga(INIT_VGA_DRAW_FRAME
        #ifndef DIRECT_TO_PHYSICAL
        | INIT_VGA_DOUBLE_BUFFER
        #endif
        );
    if (i != 0) {
        free(site);
        return 1;
    }

    get_vga_accel_caps();

	#ifdef SDL
    #ifdef VGA_FONT_INTEGRATED
	gfxPrimitivesSetFont(vga_font, 8, 8);
    #else
	gfxPrimitivesSetFont(NULL, 0, 0);
    #endif
    #endif

    /* allocate memory for visualization of beach line */

    beach = (int *)malloc(SCREEN_WIDTH * sizeof(int));

    if (beach == NULL) {
        exit_vga();
        fprintf(stderr,
            "Not enough memory for beach line! (" SIZEOF_FORMAT " bytes needed)\n",
            SCREEN_WIDTH * sizeof(int));
        free(site);
        return RET_FORTUNE_NO_MEM;
    }

    line.A = line.B = line.C = 0.0;

    /* draw sites onto virtual screen */

    for (i = 0; i < sites; i++)
        plot_site(&site[i], SITE_COLOUR);

    blit_screen(scr_virtual);

    #ifndef VALGRIND
    i = dot_getch();
    if (i == 27)
        goto all_done;
    if (i == 10)
        animate = 0;
    #else
    animate = 0;
    #endif

    /* setup timer */

    if (animate)
        init_timer(0, SWEEP_DELAY_USEC);

    #endif

    i = Fortune_init(site, sites, function);

    if (i != RET_FORTUNE_OK)
        goto all_done;

    #ifdef FORTUNE_POSITIVE_Y_DOWN
    sweep_y = WINDOW_TOP;
    #else
    sweep_y = WINDOW_BOTTOM;
    #endif

    #ifdef NO_VGA
    #define VERBOSE 0
    #else
    //#define VERBOSE 1
    #endif

    while (1) {

        #ifndef NO_VGA

        e = event_queue;

        if (e == NULL)
            break;

        yd = e->y;

        /* make sure the sweep line doesn't run away when
             several events need to be handled at the same time */

        i = (yd < WINDOW_TOP && sweep_y < yd) ? animate : 0;

        while (i) {

            blit_screen(scr_virtual);

            #ifndef DIRECT_TO_PHYSICAL
            copy_screen(scr_virtual, scr_frame);
            #endif

            #if 1
            for (x = SCREEN_WIDTH; x; )
                beach[--x] = SCREEN_HEIGHT;

            for (b = beach_line; b; b = b->right) {
                draw_parab(b, sweep_y);
                #if 0   /* green laser beams mark possible cell boundaries */
                if (b->right != NULL)
                    draw_line(&(b->line), 191);
                #endif
                #if defined(FORTUNE_DEBUG_MODE) && defined(SDL)
                #if 1   /* print debug_id's of beach line segments */
                b1 = snprintf(buffer, BUFFER_LEN, "%X", b->debug_id) << 2;
                stringColor(scr_physical,
                    PIXEL_X(b->s->x) - b1,
                    PIXEL_Y(sweep_y) + (b->debug_id << 2),
                    buffer, sdl_gfx_colour(14));
                #endif
                #endif
            }

            draw_beach();
            #endif

            line.A = 0.0; line.B = 1.0;
            line.C = -sweep_y;
            draw_line(&line, 7);

            for (e = event_queue; e; e = e->next)
                if (e->type == FORTUNE_CIRCLE_EVENT) {
                    plot_dot(e->x, e->cdata->center_y, DOT_COLOUR_CIRCLE);
                    plot_dot(e->x, e->y, DOT_COLOUR_EVENT);
                }

            switch (dot_getkey()) {
                #ifdef SDL
                case -1:                    /* user closed the window */
                    goto algorithm_done;
                #endif
                case 32:                    /* space: pause */
                    #ifdef SDL
                    blit_screen(scr_frame);
                    #endif
                    dot_getch();
                case 0:                     /* no key pressed */
                    #ifndef DIRECT_TO_PHYSICAL
                    blit_screen(scr_frame);
                    #endif
                    wait_for_timer();
                    sweep_y += SWEEP_ADVANCE;
                    if (sweep_y >= yd)
                        i = 0;
                    break;
                case 27:                    /* esc: skip rest of animation */
                    animate = 0;
                default:                    /* any other key: jump forward */
                    i = 0;
                    while (sweep_y < yd)
                        sweep_y += SWEEP_ADVANCE;
                    break;
            }
        }

        #endif

        /* check the type of next event */

        b3 = event_queue->type;

        #if VERBOSE
        #if VERBOSE > 1
        printf("%5u -> %c%10.6f,%10.6f",
            Fortune_events_left,
            (b3 == FORTUNE_CIRCLE_EVENT) ? 'C' : 'S',
            event_queue->x, event_queue->y);
        #endif
        #ifdef FORTUNE_DEBUG_MODE
        if (b3 == FORTUNE_CIRCLE_EVENT) {
            printf(" %X/%X/%X",
                event_queue->cdata->left->debug_id,
                event_queue->cdata->middle->debug_id,
                event_queue->cdata->right->debug_id);
        }
        else {
            printf("      ");
        }
        #endif
        #endif

        /* handle the next event */

        i = Fortune_handle_event();

        #if VERBOSE
        #ifdef FORTUNE_DEBUG_MODE /* list beach line segments */
        printf(" ~");
        for (b = beach_line; b; b = b->right)
            printf(" %X", b->debug_id);
        #endif
        #if VERBOSE > 1 || defined(FORTUNE_DEBUG_MODE)
        printf("\n");
        #if 1   /* pause */
        if (animate) {
            if (dot_getch() == 27)
                animate = 0;
        }
        #endif
        #endif
        #endif

        #ifndef NO_VGA

        if (b3 == FORTUNE_CIRCLE_EVENT && animate != 0) {

            trl = triangle_list;
            if (trl != NULL) {

                if (DELAUNAY) {
                    draw_boundary(trl->site[0]->x, trl->site[0]->y,
                                  trl->site[1]->x, trl->site[1]->y, 0, 8);
                    draw_boundary(trl->site[1]->x, trl->site[1]->y,
                                  trl->site[2]->x, trl->site[2]->y, 0, 8);
                    draw_boundary(trl->site[2]->x, trl->site[2]->y,
                                  trl->site[0]->x, trl->site[0]->y, 0, 8);
                }

                #ifndef FORTUNE_NO_VORONOI
                if (VORONOI) {
                    ss.x = trl->v->x;
                    ss.y = trl->v->y;
                    plot_site(&ss, VERTEX_COLOUR);
                    #if VERBOSE
                    #if 0   /* print index of vertex */
                    #ifdef SDL
                    snprintf(buffer, BUFFER_LEN, "%d", trl->v->index);
                    stringColor(scr_virtual, PIXEL_X(ss.x) + 2, PIXEL_Y(ss.y) - 4,
                        buffer, sdl_gfx_colour(11));
                    #endif
                    #endif
                    #if 0
                    printf("t%d.", trl->v->index);
                    for (b1 = 0; b1 < 3; b1++) {
                        bound = trl->cell[b1]->bloop;
                        if (bound == NULL) continue;
                        printf(" ");
                        for (b3 = 0; bound->left != NULL; ) {
                            bound = bound->left;
                            if (bound == trl->cell[b1]->bloop) {
                                b3 = 1;
                                break;
                            }
                        }
                        for (b2 = ' '; bound; b2 = ';') {
                            putchar(b2);
                            if (bound->tri != NULL)
                                printf("c%d-t%d", bound->conn_l->index, bound->tri->v->index);
                            printf("-c%d", bound->conn_r->index);
                            bound = bound->right;
                            if (b3)
                                if (bound == trl->cell[b1]->bloop)
                                    break;
                        }
                    }
                    printf("\n");
                    #elif 0
                    printf("V%4u. %9.6f,%9.6f\n", trl->v->index, trl->v->x, trl->v->y);
                    #endif
                    #endif
                }
                #endif
            }
        }

        #endif

        if (i == RET_FORTUNE_QUEUE_EMPTY)
            break;

        if (i == RET_FORTUNE_NO_MEM) {
            #ifndef NO_VGA
            exit_vga();
            #endif
            fprintf(stderr, "OUT OF MEMORY\n");
            goto algorithm_done;
        }
    }

    #ifndef NO_VGA
    #if 0
    #ifndef FORTUNE_NO_VORONOI
    #if VERBOSE
    if (VORONOI)
        if (animate) {
            blit_screen(scr_virtual);
            dot_getch();
        }
    #endif
    #endif
    #endif
    i = 1;
    #ifndef CELL_BOUNDARIES_ONLY
    if (filled_cells)
        i = 0;
    #endif
    reset_screen(i);
    #endif

    #ifndef FORTUNE_NO_DELAUNAY

    if (DELAUNAY) {

        tlist = NULL;
        clist = NULL;

        i = Delaunay_finish(&clist, &cnum, &tlist, &tnum);

        if (i != RET_FORTUNE_OK) {
            #ifndef NO_VGA
            exit_vga();
            #endif
            fprintf(stderr, "ERROR: Delaunay_finish() returned %d\n", i);
            goto algorithm_done;
        }

        #ifndef NO_VGA

        #if 1
        if (GABRIEL) {
            i = Delaunay_to_Gabriel(tlist, tnum, clist, cnum);
            if (i == RET_FORTUNE_OK) {
                for (i = 0; i < cnum; i++) {
                    b1 = *((uint8_t *)clist[i].extra);
                    if (b1 != 0)
                        draw_boundary(clist[i].s1->x, clist[i].s1->y,
                                      clist[i].s2->x, clist[i].s2->y, 0, 4);
                }
                Gabriel_cleanup(clist);
            }
            else {
                #ifndef NO_VGA
                exit_vga();
                #endif
                fprintf(stderr, "ERROR: Delaunay_to_Gabriel() returned %d\n", i);
                goto algorithm_done;
            }
        }
        else
            for (i = 0; i < cnum; i++)
                draw_boundary(clist[i].s1->x, clist[i].s1->y,
                              clist[i].s2->x, clist[i].s2->y, 0, 4);
        #if 0
        for (i = 0; i < tnum; i++) {
            b3 = tlist[i].shape;
            if (!b3)
                continue;
            for (b1 = 0, b2 = 1; b1 < 3; b1++, b2 = (b2 + 1) % 3, b3 >>= 1)
                if (b3 & 1)
                    draw_boundary_ex(tlist[i].site[b1]->x,
                                     tlist[i].site[b1]->y,
                                     tlist[i].site[b2]->x,
                                     tlist[i].site[b2]->y, 0, 14, 0.5);
        }
        #endif
        #elif 0
        for (i = 0, b3 = 1; i < tnum; i++)
            for (b1 = 0; b1 < 3; b1++) {
                if (tlist[i].site[b1]->x < WINDOW_LEFT)
                    if (tlist[i].site[(b1 + 1) % 3]->x < WINDOW_LEFT) continue;
                if (tlist[i].site[b1]->x > WINDOW_RIGHT)
                    if (tlist[i].site[(b1 + 1) % 3]->x > WINDOW_RIGHT) continue;
                if (tlist[i].site[b1]->y > WINDOW_TOP)
                    if (tlist[i].site[(b1 + 1) % 3]->y > WINDOW_TOP) continue;
                if (tlist[i].site[b1]->y < WINDOW_BOTTOM)
                    if (tlist[i].site[(b1 + 1) % 3]->y < WINDOW_BOTTOM) continue;
                draw_boundary(tlist[i].conn[b1]->s1->x,
                              tlist[i].conn[b1]->s1->y,
                              tlist[i].conn[b1]->s2->x,
                              tlist[i].conn[b1]->s2->y, 0, 14);
                if (b3) {
                    blit_screen(scr_virtual);
                    if (dot_getch() == 27) b3 = 0;
                }
                if (tlist[i].neighbour[b1] == NULL) continue;
                b2 = tlist[i].neighbour_side[b1];
                draw_boundary(tlist[i].neighbour[b1]->conn[b2]->s1->x,
                              tlist[i].neighbour[b1]->conn[b2]->s1->y,
                              tlist[i].neighbour[b1]->conn[b2]->s2->x,
                              tlist[i].neighbour[b1]->conn[b2]->s2->y, 0, 12);
                if (b3) {
                    blit_screen(scr_virtual);
                    if (dot_getch() == 27) b3 = 0;
                }
            }
        if (!b3) {
            blit_screen(scr_virtual);
            dot_getch();
        }
        #endif

        if (maze_type == MAZE_DELAUNAY) {

            #ifdef MAZE_INTERACTIVE   /* interactive mode */

            if (Delaunay_maze_init(tlist, tnum, random_seed,
                maze_short_threshold_mult) == RET_MAZE_OK) {

                #if 1
                #ifndef NO_VGA
                for (i = 0; i < tnum; i++) {
                    maze_cell = tlist[i].extra;
                    for (j = 0; j < 3; j++)
                        if (maze_wall_closed(maze_cell, j))   /* small wall */
                            draw_boundary(
                                tlist[i].conn[j]->s1->x, tlist[i].conn[j]->s1->y,
                                tlist[i].conn[j]->s2->x, tlist[i].conn[j]->s2->y,
                                0, 15);
                }
                #endif
                #endif

                for (i = RET_MAZE_OK, b2 = 1; ; ) {
                    i = Delaunay_maze_advance(&tm);
                    if (i == RET_MAZE_FINISHED)
                        break;
                    #ifndef NO_VGA
                    if (!b2)
                        continue;
                    for (j = 0; j < tnum; j++) {
                        maze_cell = tlist[j].extra;
                        #if 0
                        b1 = j & 15; b3 = b1;
                        #else
                        b1 = b3 = (maze_cell->index >= 0) ? 4 : 0;
                        if (tm == &tlist[j])
                            b1 = b3 = 14;
                        else {
                            maze_cell_b = tm->extra;
                            if (maze_cell_b != NULL)
                                if (maze_cell_b->previous == &tlist[j])
                                    b3 = 12;
                        }
                        #endif
                        #ifndef CELL_BOUNDARIES_ONLY
                        paint_triangle(&tlist[j], 0, 3, b1, b3, NULL, NULL);
                        #endif
                        #ifdef SDL
                        #if 1   /* display indices */
                        snprintf(buffer, BUFFER_LEN, "%d", maze_cell->index);
                        #else   /* display numbers of available neighbours */
                        snprintf(buffer, BUFFER_LEN, "%d", maze_cell->available);
                        #endif
                        xd = (tlist[j].site[0]->x +
                              tlist[j].site[1]->x +
                              tlist[j].site[2]->x) / 3.0;
                        yd = (tlist[j].site[0]->y +
                              tlist[j].site[1]->y +
                              tlist[j].site[2]->y) / 3.0;
                        stringColor(scr_virtual,
                            PIXEL_X(xd) - (strlen(buffer) << 2), PIXEL_Y(yd),
                            buffer, 0xaa5500ff);
                        #endif
                    }
                    j = 0;
                    if (j == 0 && b2 > 0) {
                        blit_screen(scr_virtual);
                        j = dot_getch();
                        if (j == 10) b2 = 0;
                        if (j != 27) j = 0;
                    }
                    if (j > 0)
                        break;
                    #endif
                }

                Delaunay_maze_finish();

                #ifndef NO_VGA
                //dot_getch();
                reset_screen(0);
                for (j = 0; j < tnum; j++) {
                    maze_cell = tlist[j].extra;
                    for (i = 0; i < 3; i++) {
                        if (maze_wall_open(maze_cell, i))
                            continue;
                        if (tlist[j].neighbour[i] != NULL) {
                            /* don't draw the same boundary twice: part 1 */
                            maze_cell_n = tlist[j].neighbour[i]->extra;
                            if (maze_cell_n != NULL)
                                if (maze_cell_n->index == MAZE_INDEX_UNDEFINED)
                                    continue;
                        }
                        draw_boundary(
                            tlist[j].conn[i]->s1->x, tlist[j].conn[i]->s1->y,
                            tlist[j].conn[i]->s2->x, tlist[j].conn[i]->s2->y,
                            0, 4);
                    }
                    #if 0       /* display final indices */
                    #ifdef SDL
                    snprintf(buffer, BUFFER_LEN, "%d", maze_cell->index);
                    xd = (tlist[j].site[0]->x +
                          tlist[j].site[1]->x +
                          tlist[j].site[2]->x) / 3.0;
                    yd = (tlist[j].site[0]->y +
                          tlist[j].site[1]->y +
                          tlist[j].site[2]->y) / 3.0;
                    stringColor(scr_virtual,
                        PIXEL_X(xd) - (strlen(buffer) << 2), PIXEL_Y(yd),
                        buffer, 0xffff55ff);
                    #endif
                    #endif
                    /* don't draw the same boundary twice: part 2 */
                    maze_cell->index = MAZE_INDEX_UNDEFINED;
                }
                #endif

                Delaunay_maze_cleanup(tlist, tnum);
            }

            #else   /* non-interactive mode */

            if (maze_make(MAZE_DELAUNAY, tlist, tnum, random_seed,
                maze_short_threshold_mult) == RET_MAZE_OK) {
                #ifndef NO_VGA
                reset_screen(0);
                for (j = 0; j < tnum; j++) {
                    maze_cell = tlist[j].extra;
                    for (i = 0; i < 3; i++) {
                        if (maze_wall_open(maze_cell, i))
                            continue;
                        if (tlist[j].neighbour[i] != NULL) {
                            maze_cell_n = tlist[j].neighbour[i]->extra;
                            if (maze_cell_n != NULL)
                                if (maze_cell_n->index == MAZE_INDEX_UNDEFINED)
                                    continue;
                        }
                        draw_boundary(
                            tlist[j].conn[i]->s1->x, tlist[j].conn[i]->s1->y,
                            tlist[j].conn[i]->s2->x, tlist[j].conn[i]->s2->y,
                            0, 4);
                    }
                    #if 1
                    #ifdef SDL
                    snprintf(buffer, BUFFER_LEN, "%d", maze_cell->index);
                    xd = (tlist[j].site[0]->x +
                          tlist[j].site[1]->x +
                          tlist[j].site[2]->x) / 3.0;
                    yd = (tlist[j].site[0]->y +
                          tlist[j].site[1]->y +
                          tlist[j].site[2]->y) / 3.0;
                    stringColor(scr_virtual,
                        PIXEL_X(xd) - (strlen(buffer) << 2), PIXEL_Y(yd),
                        buffer, 0xffff55ff);
                    #endif
                    #endif
                    maze_cell->index = MAZE_INDEX_UNDEFINED;
                }
                #endif
                maze_cleanup(MAZE_DELAUNAY, tlist, tnum);
            }

            #endif
        }

        #else
        for (i = 0; i < tnum; i++)
            printf("T%4u. %u - %u - %u\n", i,
                tlist[i].site[0] - site,
                tlist[i].site[1] - site,
                tlist[i].site[2] - site);
        #endif

        Delaunay_cleanup(clist, tlist);
    }

    #endif

    #ifndef FORTUNE_NO_VORONOI

    if (VORONOI) {

        vlist = NULL;
        blist = NULL;
        cell_list = NULL;

        i = Voronoi_finish(&vlist, &vnum, &blist, &bnum, &cell_list, &actual_vnum);

        if (i != RET_FORTUNE_OK) {
            #ifndef NO_VGA
            exit_vga();
            #endif
            fprintf(stderr, "ERROR: Voronoi_finish() returned %d\n", i);
            goto algorithm_done;
        }

        #ifndef NO_VGA

        #define DO_BOUNDARY(BPTR, COLOUR) {                     \
            bp = BPTR;                                          \
            if (bp->v2 == NULL)                                 \
                draw_boundary(bp->v1->x, bp->v1->y,             \
                              bp->vector->x, bp->vector->y,     \
                              DRAW_BOUNDARY_RELATIVE, COLOUR);  \
            else                                                \
                draw_boundary(bp->v1->x, bp->v1->y,             \
                              bp->v2->x, bp->v2->y, 0, COLOUR); \
        }

        #ifndef CELL_BOUNDARIES_ONLY
        if (filled_cells) {                     /* draw filled cells */

            b2 =
                0;    /* paint all cells, then pause */
                //1;    /* pause after painting each cell */
            for (i = 0; i < sites; i++) {
                b1 = cell_list[i].sides;
                j = paint_cell(&cell_list[i], 0,
                    shrink_width,
                    ((b1 - 1) % 15) + 1,
                    (b1 == 8) ? 7 : ((b1 > 8) ? 15 : (b1 | 8)),
                    NULL, NULL);
                if (j == 0 && b2 > 0) {
                    #if 0
                    for (j = 0, cb = cell_list[i].blist; j < cell_list[i].sides;
                        j++, cb = cb->right)
                        DO_BOUNDARY(cb->boundary, 10);
                    #endif
                    blit_screen(scr_virtual);
                    j = dot_getch();
                    if (j == 10) b2 = 0;
                    if (j != 27) j = 0;
                }
                if (j > 0)
                    break;
            }
        }
        else                                    /* draw cell boundaries */
        #endif
        {

            #if 1

            for (i = 0; i < actual_vnum; i++) {
                ss.x = vlist[i].x;
                ss.y = vlist[i].y;
                plot_site(&ss, 1);
            }
            if (maze_type == MAZE_VORONOI) {
                for (i = 0; i < bnum; i++)
                    if (blist[i].v2 != NULL) {
                        DO_BOUNDARY(&blist[i], 2);
                    }
            }
            else {
                for (i = 0; i < bnum; i++) {
                    DO_BOUNDARY(&blist[i], 2);
                    #if 0   /* print boundary numbers */
                    #ifdef SDL
                    b3 = snprintf(buffer, BUFFER_LEN, "%d", i) << 2;
                    if (blist[i].v2 == NULL)
                        stringColor(scr_virtual,
                            PIXEL_X(blist[i].v1->x + (blist[i].vector->x / 3.0)) - b3,
                            PIXEL_Y(blist[i].v1->y + (blist[i].vector->y / 3.0)) - 4,
                            buffer, sdl_gfx_colour(12));
                    else
                        stringColor(scr_virtual,
                            PIXEL_X((blist[i].v1->x + blist[i].v2->x) / 2.0) - b3,
                            PIXEL_Y((blist[i].v1->y + blist[i].v2->y) / 2.0) - 4,
                            buffer, sdl_gfx_colour(10));
                    #endif
                    #endif
                }
            }

            #if 0   /* print site numbers */
            for (i = 0; i < sites; i++) {
                #ifdef SDL
                snprintf(buffer, BUFFER_LEN, "%d", i);
                stringColor(scr_virtual,
                    PIXEL_X(site[i].x) + 2, PIXEL_Y(site[i].y) - 4,
                    buffer, sdl_gfx_colour(15));
                #endif
                #if VERBOSE
                #if 1   /* print triangle and connection numbers */
                {
                    bound = cell_list[i].bloop;
                    for (j = cell_list[i].sides; j > 0; j--, bound = bound->right)
                        if (bound->first) break;
                    printf("%5d.", i);
                    if (cell_list[i].infinity)
                        printf("inf");
                    else
                        printf("   ");
                    for (j = cell_list[i].sides; j > 0; j--, bound = bound->right) {
                        printf("  ");
                        if (bound->tri != NULL) {
                            printf("c%d-t%d", bound->conn_l->index, bound->tri->v->index);
                            #ifdef SDL
                            b3 = snprintf(buffer, BUFFER_LEN, "%d", bound->tri->v->index) << 2;
                            stringColor(scr_virtual,
                                PIXEL_X((bound->tri->site[0]->x + bound->tri->site[1]->x + bound->tri->site[2]->x) / 3.0) - b3,
                                PIXEL_Y((bound->tri->site[0]->y + bound->tri->site[1]->y + bound->tri->site[2]->y) / 3.0) - 4,
                                buffer, sdl_gfx_colour(3));
                            #endif
                        }
                        printf("-c%d", bound->conn_r->index);
                        #ifdef SDL
                        b3 = snprintf(buffer, BUFFER_LEN, "%d", bound->conn_r->index) << 2;
                        stringColor(scr_virtual,
                            PIXEL_X((site[i].x + bound->conn_r->s1->x + bound->conn_r->s2->x) / 3.0) - b3,
                            PIXEL_Y((site[i].y + bound->conn_r->s1->y + bound->conn_r->s2->y) / 3.0) - 4,
                            buffer, sdl_gfx_colour(14));
                        #endif
                    }
                    printf("\n");
                }
                #endif
                #endif
            }
            #endif

            #else

            for (i = b3 = 0; i < sites; i++) {
                if (cell_list[i].blist == NULL)
                    continue;
                if (cell_list[i].infinity) {
                    for (j = 0, cb = cell_list[i].blist; j < cell_list[i].sides;
                        j++, cb = cb->right) {
                        DO_BOUNDARY(cb->boundary, 31 - ((j << 1) & 30));
                    }
                    blit_screen(scr_virtual);
                    if (dot_getch() == 27) break;
                }
                for (j = 0, cb = cell_list[i].blist; j < cell_list[i].sides;
                    j++, cb = cb->right)
                    DO_BOUNDARY(cb->boundary, 2);
            }

            #endif
        }

        if (maze_type == MAZE_VORONOI || maze_type == MAZE_VORONOI_INF) {

            #ifdef MAZE_INTERACTIVE   /* interactive mode */

            if (Voronoi_maze_init(maze_type, cell_list, sites, random_seed,
                maze_short_threshold_mult) == RET_MAZE_OK) {

                #if 1
                for (j = 0; j < sites; j++) {
                    maze_cell = cell_list[j].extra;
                    if (maze_cell == NULL)
                        continue;
                    for (i = 0, cb = cell_list[j].blist; i < cell_list[j].sides;
                        i++, cb = cb->right) {
                        if (maze_wall_closed(maze_cell, i))   /* small wall */
                            DO_BOUNDARY(cb->boundary, 15);
                    }
                }
                #endif

                for (i = RET_MAZE_OK, b2 = 1; ; ) {
                    i = Voronoi_maze_advance(&cm);
                    if (i == RET_MAZE_FINISHED)
                        break;
                    #ifndef NO_VGA
                    #if 0
                    if (i == RET_MAZE_SEGMENT) {    /* redraw */
                        for (j = 0; j < sites; j++) {
                            maze_cell = cell_list[j].extra;
                            if (maze_cell == NULL)
                                continue;
                            for (i = 0, cb = cell_list[j].blist; i < cell_list[j].sides;
                                i++, cb = cb->right) {
                                if (maze_wall_open(maze_cell, i))
                                    DO_BOUNDARY(cb->boundary, 2);
                            }
                        }
                    }
                    #endif
                    if (!b2)
                        continue;
                    for (j = 0; j < sites; j++) {
                        maze_cell = cell_list[j].extra;
                        if (maze_cell == NULL)
                            continue;
                        #if 0
                        b1 = j & 15; b3 = b1;
                        #else
                        b1 = b3 = (maze_cell->index >= 0) ? 2 : 0;
                        if (cm == &cell_list[j])
                            b1 = b3 = 14;
                        else {
                            maze_cell_b = cm->extra;
                            if (maze_cell_b != NULL)
                                if (maze_cell_b->previous == &cell_list[j])
                                    b3 = 10;
                        }
                        #endif
                        #ifndef CELL_BOUNDARIES_ONLY
                        paint_cell(&cell_list[j], 0, 3, b1, b3, NULL, NULL);
                        #endif
                        #ifdef SDL
                        #if 1
                        #if 1   /* display indices */
                        snprintf(buffer, BUFFER_LEN, "%d", maze_cell->index);
                        #else   /* display numbers of available neighbours */
                        snprintf(buffer, BUFFER_LEN, "%d", maze_cell->available);
                        #endif
                        stringColor(scr_virtual,
                            PIXEL_X(cell_list[j].s->x) - (strlen(buffer) << 2), PIXEL_Y(cell_list[j].s->y) - 4,
                            buffer, 0x5555ffff);
                        #endif
                        #endif
                    }
                    j = 0;
                    if (j == 0 && b2 > 0) {
                        blit_screen(scr_virtual);
                        j = dot_getch();
                        if (j == 10) b2 = 0;
                        if (j != 27) j = 0;
                    }
                    if (j > 0)
                        break;
                    #endif
                }

                Voronoi_maze_finish(maze_type);

                #ifndef NO_VGA
                reset_screen(0);
                for (j = 0; j < sites; j++) {
                    maze_cell = cell_list[j].extra;
                    if (maze_cell == NULL)
                        continue;
                    for (i = 0, cb = cell_list[j].blist; i < cell_list[j].sides;
                        i++, cb = cb->right) {
                        if (maze_wall_open(maze_cell, i))
                            continue;
                        if (cb->neighbour != NULL) {
                            /* don't draw the same boundary twice: part 1 */
                            maze_cell_n = cb->neighbour->extra;
                            if (maze_cell_n != NULL)
                                if (maze_cell_n->index == MAZE_INDEX_UNDEFINED)
                                    continue;
                        }
                        DO_BOUNDARY(cb->boundary, 2);
                    }
                    #if 0       /* display final indices */
                    #ifdef SDL
                    snprintf(buffer, BUFFER_LEN, "%d", maze_cell->index);
                    stringColor(scr_virtual,
                        PIXEL_X(cell_list[j].s->x) - (strlen(buffer) << 2), PIXEL_Y(cell_list[j].s->y) - 4,
                        buffer, 0xffff55ff);
                    #endif
                    #endif
                    /* don't draw the same boundary twice: part 2 */
                    maze_cell->index = MAZE_INDEX_UNDEFINED;
                }
                #endif

                Voronoi_maze_cleanup(cell_list, sites);
            }

            #else   /* non-interactive mode */

            if (maze_make(maze_type, cell_list, sites, random_seed,
                maze_short_threshold_mult) == RET_MAZE_OK) {
                #ifndef NO_VGA
                reset_screen(0);
                for (j = 0; j < sites; j++) {
                    maze_cell = cell_list[j].extra;
                    if (maze_cell == NULL)
                        continue;
                    for (i = 0, cb = cell_list[j].blist; i < cell_list[j].sides;
                        i++, cb = cb->right) {
                        if (maze_wall_open(maze_cell, i))
                            continue;
                        if (cb->neighbour != NULL) {
                            maze_cell_n = cb->neighbour->extra;
                            if (maze_cell_n != NULL)
                                if (maze_cell_n->index == MAZE_INDEX_UNDEFINED)
                                    continue;
                        }
                        b2 = 2;
                        DO_BOUNDARY(cb->boundary, b2);
                    }
                    #if 0
                    #ifdef SDL
                    snprintf(buffer, BUFFER_LEN, "%d", maze_cell->index);
                    stringColor(scr_virtual,
                        PIXEL_X(cell_list[j].s->x) - (strlen(buffer) << 2), PIXEL_Y(cell_list[j].s->y) - 4,
                        buffer, 0xffff55ff);
                    #endif
                    #endif
                    maze_cell->index = MAZE_INDEX_UNDEFINED;
                }
                #endif
                maze_cleanup(maze_type, cell_list, sites);
            }

            #endif
        }

        #endif

        #ifdef NO_VGA
        for (i = 0; i < actual_vnum; i++)
            printf("V%4u. %9.6f,%9.6f\n", i, vlist[i].x, vlist[i].y);
        for ( ; i < vnum; i++)
            printf("V%4u. %9.6f,%9.6f (Inf)\n", i, vlist[i].x, vlist[i].y);
        for (i = 0; i < bnum; i++) {
            printf("B%4u. %9.6f,%9.6f -> ", i, blist[i].v1->x, blist[i].v1->y);
            if (blist[i].v2 != NULL)
                printf("%9.6f,%9.6f", blist[i].v2->x, blist[i].v2->y);
            else {
                if (blist[i].vector->x > -VERTEX_COORD_EPSILON &&
                    blist[i].vector->x <  VERTEX_COORD_EPSILON)
                    printf(" 0.0");
                else
                    printf("%cInf", (blist[i].vector->x < 0.0) ? '-' : '+');
                printf(", ");
                if (blist[i].vector->y > -VERTEX_COORD_EPSILON &&
                    blist[i].vector->y <  VERTEX_COORD_EPSILON)
                    printf(" 0.0");
                else
                    printf("%cInf", (blist[i].vector->y < 0.0) ? '-' : '+');
            }
            printf("\n");
        }
        #endif

        #if 0
        fprintf(stderr, "cell boundaries:\n");
        for (i = 0; i < sites; i++) {
            int z;
            fprintf(stderr, "C%4u.", i);
            if (cell_list[i].infinity)
                fprintf(stderr, "inf ");
            else
                fprintf(stderr, "    ");
            for (j = cell_list[i].sides, cb = cell_list[i].blist;
                j; j--) {
                for (z = 0; z < bnum; z++) {
                    if (cb->boundary == &blist[z]) {
                        fprintf(stderr, " %u", z);
                        break;
                    }
                }
                cb = cb->right;
            }
            fprintf(stderr, "\n");
        }
        #endif
        #if 0
        fprintf(stderr, "number of links from cells to each boundary:\n");
        for (j = 0; j < bnum; j++) {
            int z, zz;
            for (i = z = 0; i < sites; i++) {
                for (zz = cell_list[i].sides, cb = cell_list[i].blist;
                    zz; zz--, cb = cb->right)
                    if (cb->boundary == &blist[j]) z++;
            }
            fprintf(stderr, "B%u:%d ", j, z);
        }
        fprintf(stderr, "\n");
        #endif
        #if 0
        fprintf(stderr, "number of links from cells to each vertex:\n");
        for (j = 0; j < actual_vnum; j++) {
            int z, zz;
            for (i = z = 0; i < sites; i++) {
                for (zz = cell_list[i].sides, cb = cell_list[i].blist;
                    zz; zz--, cb = cb->right) {
                    if (cb->boundary->v1 == &vlist[j]) z++;
                    if (cb->boundary->v2 == &vlist[j]) z++;
                }
            }
            fprintf(stderr, "V%u:%d ", j, z);
        }
        fprintf(stderr, "\n");
        #endif

        Voronoi_cleanup(vlist, blist, NULL);
    }

    #endif

    #ifndef NO_VGA
    blit_screen(scr_virtual);

    if (dot_getch() == 's') {

        #ifdef SDL
        #ifndef VALGRIND
        SDL_SaveBMP(scr_virtual, "out.bmp");
        #endif
        #endif

        #ifdef SVGALIB
        save_pbm_screenshot("out.pbm", scr_virtual, 4);
        #endif
    }
    #endif

    algorithm_done:

    i = Fortune_exit();

    #ifndef FORTUNE_NO_VORONOI
    if (VORONOI) {
        #if 0
        fprintf(stderr, "cells:\n");
        for (i = 0; i < sites; i++) {
            fprintf(stderr, "C%4u.%5d b ", i, cell_list[i].sides);
            if (cell_list[i].infinity)
                fprintf(stderr, "[Inf]");
            else
                fprintf(stderr, "     ");
            fprintf(stderr, "%10.6f,%10.6f", cell_list[i].s->x, cell_list[i].s->y);
            fprintf(stderr, "\n");
        }
        #endif
        Voronoi_cleanup(NULL, NULL, cell_list);
    }
    #endif

    /* free memory, return to text mode and exit */

    all_done:

    free(site);

    #ifndef NO_VGA
    free(beach);

    exit_vga();
    #endif

    return i;
}

