/*************************************************************************
 *
 *  Triangle -- an implementation of Fortune's algorithm
 *  Copyright (C) 2012-2014 Jussi M. Hukkanen
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *************************************************************************
 *
 *  vsplit.c: fractal demo based on recursive splitting of Voronoi cells
 *
 ************************************************************************/

#include <stdio.h>
#include <string.h>     /* declares strdup() */

#define FORTUNE_PUBLIC
#include "fortune.h"
#include "lloyd.h"
#include "opts.h"
#include "rnd.h"

#ifdef H_PLATFORM
#include H_PLATFORM
#endif

char *window_title = "vsplit";


#define QUIET_NOT           0
#define QUIET_OUTPUT_SVG    1
#define QUIET_OUTPUT_SITES  2


#ifndef NO_VGA

#ifdef SVGALIB
#define DIRECT_TO_PHYSICAL 1
#endif

#include H_GFX

#if 0       /* draw filled Voronoi cells */
#include "dotcell.h"
#else
#define CELL_BOUNDARIES_ONLY
#endif

#if 1
#define VGA_FONT_INTEGRATED 1
uint8_t vga_font[] = {
    #include "thin8x8.txt"
};
#endif

#else   /* NO_VGA */

#if 1
#define WINDOW_RIGHT  (4.0 / 3.0)
#else
#define WINDOW_RIGHT  (1.0)
#endif
#define WINDOW_LEFT   (-WINDOW_RIGHT)
#define WINDOW_TOP    ( 1.0)
#define WINDOW_BOTTOM (-1.0)
#define WINDOW_WIDTH  (WINDOW_RIGHT  - WINDOW_LEFT)
#define WINDOW_HEIGHT (WINDOW_TOP    - WINDOW_BOTTOM)

#define WIDTH         640
#define HEIGHT        480

#endif

char quiet_mode = QUIET_NOT;


#if 1
#define SVG_SUPPORT

#define DEFAULT_STROKE_WIDTH    0.00125
#define MAX_STROKE_WIDTH        0.1

float svg_stroke = DEFAULT_STROKE_WIDTH;
#endif


#define DEFAULT_SITES   12

int sites = DEFAULT_SITES;

site_t *site; /* sites items */
site_t *new_site;

typedef struct site_list_s {
    site_coord_t x;
    site_coord_t y;
    struct site_list_s *next;
} site_list_t;

site_list_t *site_list;

char *site_list_file = NULL;
float site_scale = 1.0;

#define DEFAULT_ITERATIONS    3

int iterations = -1; /* run 2^32 times by default! */


#define LLOYD_ITERATIONS      2


#define MIN_DIVISIONS      2
#define MAX_DIVISIONS    100

#define MAX_SHRINK_PIXELS 19

uint8_t divisions = 0;
uint8_t shrink_width = 0;
int32_t random_seed = 0;


#ifndef NO_VGA

#define SITE_COLOUR        7
#define DOT_COLOUR_EVENT   6
#define DOT_COLOUR_CIRCLE  4
#define VERTEX_COLOUR      1

#endif


vertex_t *vlist;        /* array of Voronoi vertices */
boundary_t *blist;      /* array of Voronoi boundaries */
celldata_t *cell_list;  /* array of Voronoi cells */
int vnum, bnum;         /* number of vertices and boundaries */
int actual_vnum;        /* number of non-infinite vertices */


#define BUFFER_LEN 16
char buffer[BUFFER_LEN + 1];


#ifndef NO_VGA

/* plot_site()

    draws a small circle at the given site's coordinates
      in scr_virtual */

char plot_site(site_t *s, uint8_t colour) {

    return plot_dot_ex(PIXEL_X(s->x), PIXEL_Y(s->y),
      PLOT_DOT_HOLLOW, colour, BRIGHT_COLOUR(colour));
}


/* plot_dot()

    draws a dot at the given coordinates */

char plot_dot(float px, float py, uint8_t colour) {

    return plot_dot_ex(PIXEL_X(px), PIXEL_Y(py),
      PLOT_DOT_PHYSICAL, colour, BRIGHT_COLOUR(colour));
}


/* draw_line()

    draws the given line l on the screen */

char draw_line(line_t *l, uint8_t colour) {

    float x, y;
    int xs, ys;

    if (l->A > -DBL_EPSILON && l->A < DBL_EPSILON) {

        if (l->B > -DBL_EPSILON && l->B < DBL_EPSILON)
            return 0;

        y = -l->C / l->B;
        ys = PIXEL_Y(y);

        if (ys < 0 || ys >= SCREEN_HEIGHT) return 0;

        #ifdef SVGALIB
        #ifdef DIRECT_TO_PHYSICAL
        gl_setcontext(scr_physical);
        #else
        gl_setcontext(scr_frame);
        #endif
        gl_hline(0, ys, SCREEN_WIDTH - 1, colour);
        #endif

        #ifdef SDL
        hlineColor(scr_physical,
            0, SCREEN_WIDTH - 1, ys,
            sdl_gfx_colour(colour));
        #endif

        return 0;
    }
    else
        return draw_boundary(
            (l->B * (WINDOW_TOP) + l->C) / -l->A, WINDOW_TOP,
            (l->B * (WINDOW_BOTTOM) + l->C) / -l->A, WINDOW_BOTTOM,
            DRAW_BOUNDARY_PHYSICAL, colour);

    return 0;
}


#endif


/* read_sites_from_file()

    reads site_list_file and fills site_list
      with coordinates found */

char read_sites_from_file(void) {

    FILE *fh;
    char *line = NULL;
    size_t size;
    ssize_t len;
    int n;
    site_list_t *sl;
    site_coord_t x, y;

    sites = 0;
    site_list = NULL;

    if (site_list_file == NULL) {
        fh = stdin;
    }
    else {
        fh = fopen(site_list_file, "rb");
        if (fh == NULL) {
            fprintf(stderr, "File `%s' could not be read.\n", site_list_file);
            return 1;
        }
    }

    while (1) {

        len = getline(&line, &size, fh);
        if (len <= 0)
            break;

        if (line[0] == '\n')
            break;

        n = sscanf(line, "%f,%f", &x, &y);
        if (n < 2) {
            if (n < 0)
                break;
            continue;
        }

        sl = malloc(sizeof(site_list_t));
        if (sl == NULL) {
            fprintf(stderr, "Out of memory allocating another site!\n");
            break;
        }
        sl->x = x * site_scale;
        sl->y = y * site_scale;
        sl->next = site_list;
        site_list = sl;

        sites++;
    }

    if (fh != stdin)
        fclose(fh);

    if (line != NULL)
        free(line);

    return 0;
}


/* interpret_site_list()

    fills the site array with the coordinate
      pairs in site_list and frees site_list */

void interpret_site_list(void) {

    site_list_t *sn;
    int s;

    for (s = 0; s < sites; s++) {

      sn = site_list->next;
      site[s].x = site_list->x;
      site[s].y = site_list->y;

      free(site_list);

      if (sn == NULL)
        break;

      site_list = sn;
    }
}


#ifndef NO_VGA

/* reset_screen()

    clears the screen;
    plots sites as well if plot_sites
      is nonzero */

void reset_screen(char plot_sites) {

    int i;

    fill_screen(scr_virtual, 0);

    #ifdef SDL
    rectangleColor(scr_virtual, 0, 0, WIDTH - 1, HEIGHT - 1,
        sdl_gfx_colour(1));
    #elif defined(SVGALIB)
    gl_setcontext(scr_virtual);
    gl_hline(0,          0, WIDTH - 1, 1);
    gl_hline(0, HEIGHT - 1, WIDTH - 1, 1);
    gl_line(        0, 1,         0, HEIGHT - 2, 1);
    gl_line(WIDTH - 1, 1, WIDTH - 1, HEIGHT - 2, 1);
    #endif

    if (plot_sites)
        for (i = 0; i < sites; i++)
            plot_site(&site[i], SITE_COLOUR);
}

#endif


#ifdef SVG_SUPPORT
#ifndef FORTUNE_NO_VORONOI

/* clip_boundary()

    clips the given infinite boundary against
      the screen edges and returns the endpoint
      in *tx,*ty;
    returns nonzero if the boundary is entirely
      outside the screen */

char clip_boundary(vertex_t *v1, vertex_t *vector,
                   vertex_coord_t *tx, vertex_coord_t *ty) {

    vertex_coord_t x1 = v1->x;
    vertex_coord_t y1 = v1->y;
    vertex_coord_t x2 = vector->x;
    vertex_coord_t y2 = vector->y;
    #ifdef WINDOW_COORDINATES_FLOAT
    float
    #else
    int
    #endif
        ypos, yneg;

    #ifdef FORTUNE_POSITIVE_Y_DOWN
    ypos = WINDOW_BOTTOM;
    yneg = WINDOW_TOP;
    #else
    ypos = WINDOW_TOP;
    yneg = WINDOW_BOTTOM;
    #endif

    if (x2 < 0.0) {         /* left */

        if (x1 < WINDOW_LEFT)
            return 1;

        if (y2 < 0.0) {     /* negative y */

            if (y1 < yneg)
                return 1;

            if (-x2 < -y2) {
                *tx = x1 + (x2 * (yneg - y1) / y2);
                *ty = yneg;
            }
            else {
                *ty = y1 + (y2 * (WINDOW_LEFT - x1) / x2);
                *tx = WINDOW_LEFT;
            }

            return 0;
        }
        else {              /* positive y */

            if (y1 > ypos)
                return 1;

            if (-x2 < y2) {
                *tx = x1 + (x2 * (ypos - y1) / y2);
                *ty = ypos;
            }
            else {
                *ty = y1 + (y2 * (WINDOW_LEFT - x1) / x2);
                *tx = WINDOW_LEFT;
            }

            return 0;
        }
    }
    else {                  /* right */

        if (x1 > WINDOW_RIGHT)
            return 1;

        if (y2 < 0.0) {     /* negative y */

            if (y1 < yneg)
                return 1;

            if (x2 < -y2) {
                *tx = x1 + (x2 * (yneg - y1) / y2);
                *ty = yneg;
            }
            else {
                *ty = y1 + (y2 * (WINDOW_RIGHT - x1) / x2);
                *tx = WINDOW_RIGHT;
            }

            return 0;
        }
        else {              /* positive y */

            if (y1 > ypos)
                return 1;

            if (x2 < y2) {
                *tx = x1 + (x2 * (ypos - y1) / y2);
                *ty = ypos;
            }
            else {
                *ty = y1 + (y2 * (WINDOW_RIGHT - x1) / x2);
                *tx = WINDOW_RIGHT;
            }

            return 0;
        }
    }

    return 1;
}

#endif


/* output_svg()

    prints out a list of Voronoi boundaries
      in SVG file format into stdout */

char output_svg(uint8_t func) {

    int i, j;
    vertex_coord_t ex, ey;
    cellboundary_t *cb;
    boundary_t *b;

    #define SVG_FMT  "%.6f"
    #define SVG_FMTQ "\"" SVG_FMT "\""

    printf("<svg xmlns=\"http://www.w3.org/2000/svg\"");
    //printf(" version=\"1.1\"");
    printf(" width=\"%u\" height=\"%u\"", WIDTH, HEIGHT);
    printf(">\n");

    printf("\t<g id=\"");
    printf("Voronoi");
    printf("\" stroke=\"black\" stroke-width=" SVG_FMTQ " ",
        svg_stroke);
    printf("transform=\"");
    printf("translate(%d,%d)", WIDTH >> 1, HEIGHT >> 1);

    i = (WIDTH < HEIGHT) ? WIDTH : HEIGHT;
    i >>= 1;
    printf(" "
        "scale(%d,%d)", i,
        #ifdef FORTUNE_POSITIVE_Y_UP
        -
        #endif
        i
        );

    printf("\">\n");

    printf("\t\t<rect x=" SVG_FMTQ " width=" SVG_FMTQ
        " y=" SVG_FMTQ " height=" SVG_FMTQ
        " fill=\"white\" stroke=\"none\" />\n",
        WINDOW_LEFT, WINDOW_RIGHT - WINDOW_LEFT,
        #ifdef FORTUNE_POSITIVE_Y_UP
        WINDOW_BOTTOM, WINDOW_TOP - WINDOW_BOTTOM
        #else
        WINDOW_TOP, WINDOW_BOTTOM - WINDOW_TOP
        #endif
        );

    #define SVG_LINE(X1, Y1, X2, Y2) {                      \
        printf("\t\t<line x1=" SVG_FMTQ " y1=" SVG_FMTQ     \
            " x2=" SVG_FMTQ " y2=" SVG_FMTQ " />\n",        \
            X1, Y1, X2, Y2);                                \
    }

    #define HANDLE_SVG_VORONOI_BOUNDARY(BPTR) {             \
        b = BPTR;                                           \
        if (b->v2 == NULL) {                                \
            if (!clip_boundary(b->v1, b->vector, &ex, &ey)) \
                SVG_LINE(b->v1->x, b->v1->y, ex, ey);       \
        }                                                   \
        else {                                              \
            SVG_LINE(b->v1->x, b->v1->y,                    \
                b->v2->x, b->v2->y);                        \
        }                                                   \
    }

    for (i = 0; i < bnum; i++)
        HANDLE_SVG_VORONOI_BOUNDARY(&blist[i]);

    printf("\t</g>\n");
    printf("</svg>\n");

    return RET_FORTUNE_OK;
}
#endif


/* parse_options()

    interprets the given list of command-line
      options */

char parse_options(int argc, char *argv[]) {

    #define CMDLINE_OPTC_HELP       'h'
    #define CMDLINE_OPTS_HELP           "help"
    #define CMDLINE_OPTC_VGAMODE    'm'
    #define CMDLINE_OPTS_VGAMODE        "mode"
    #define CMDLINE_OPTC_SDLMODE    'w'
    #define CMDLINE_OPTS_SDLMODE        "window"
    #define CMDLINE_OPTC_SHRINK     'b'
    #define CMDLINE_OPTS_SHRINK         "border"
    #define CMDLINE_OPTC_LIST       's'
    #define CMDLINE_OPTS_LIST           "sites"
    #define CMDLINE_OPTC_QUIET      'q'
    #define CMDLINE_OPTS_QUIET          "quiet"
    #define CMDLINE_OPTC_STROKE     'l'
    #define CMDLINE_OPTS_STROKE         "stroke"
    #define CMDLINE_OPTC_SCALE      'f'
    #define CMDLINE_OPTS_SCALE          "scale"
    #define CMDLINE_OPTC_RUNS       'r'
    #define CMDLINE_OPTS_RUNS           "runs"

    struct option long_opts[] = {
      { CMDLINE_OPTS_HELP,     no_argument,       NULL, CMDLINE_OPTC_HELP     },
      #ifndef NO_VGA
      #ifndef VALGRIND
      #ifdef SVGALIB
      { CMDLINE_OPTS_VGAMODE,  required_argument, NULL, CMDLINE_OPTC_VGAMODE  },
      #endif
	  #ifdef SDL
      { CMDLINE_OPTS_SDLMODE,  required_argument, NULL, CMDLINE_OPTC_SDLMODE  },
	  #endif
	  #endif
      #ifndef CELL_BOUNDARIES_ONLY
      { CMDLINE_OPTS_SHRINK,   required_argument, NULL, CMDLINE_OPTC_SHRINK   },
      #endif
      #endif
      { CMDLINE_OPTS_LIST,     no_argument,       NULL, CMDLINE_OPTC_LIST     },
      #ifdef SVG_SUPPORT
      { CMDLINE_OPTS_QUIET,    no_argument,       NULL, CMDLINE_OPTC_QUIET    },
      { CMDLINE_OPTS_STROKE,   required_argument, NULL, CMDLINE_OPTC_STROKE   },
      #endif
      { CMDLINE_OPTS_SCALE,    required_argument, NULL, CMDLINE_OPTC_SCALE    },
      { CMDLINE_OPTS_RUNS,     required_argument, NULL, CMDLINE_OPTC_RUNS     },
      { NULL, 0, NULL, 0 }
    };
    int c;
    int i = 0;
    float f = 0.0;
    char *modestr;
    int index = 0;
    char ret = RET_FORTUNE_OK;
    char *optstring;

    optstring = make_opt_string(long_opts);
    if (optstring == NULL) {
      fprintf(stderr, "Out of memory reading command line options!\n");
      return RET_FORTUNE_NO_MEM;
    }

	site_list_file = get_input_file_name(argc, argv);

    while (1) {

      c = getopt_long(argc, argv, optstring, long_opts, &index);

      if (c == -1)
        break;

      if (optarg != NULL)
        i = atoi(optarg);

      switch (c) {

        case CMDLINE_OPTC_HELP:
          printf("vsplit - Copyright (C) 2012-2014 Jussi M. Hukkanen\n");
          printf("Draw a Voronoi diagram based on a given list of\n"
                 "points on a plane, and then divide the Voronoi cells\n"
                 "into smaller ones.");
          #ifdef SVG_SUPPORT
          #ifdef NO_VGA
          printf(" The resulting diagram is written\n"
                 "to standard output as a Scalable Vector Graphics\n"
                 "(SVG) image.");
          #endif
          #endif
          printf("\n\n");
          printf("This program comes with ABSOLUTELY NO WARRANTY.\n"
                 "This is free software, and you are welcome to\n"
                 "redistribute it under the conditions of the\n"
                 "GNU General Public License, version 3 or later.\n"
                 "To view the License, see `gpl.txt'.\n\n");

          printf("Usage: %s [FILE] [options]\n", argv[0]);
          printf("FILE should contain a list of sites as one pair\n"
                 "of x,y coordinates per line. If no FILE is given,\n"
                 "the list is read from standard input.\n");
          printf("   -%c N, --" CMDLINE_OPTS_SCALE " N\n", CMDLINE_OPTC_SCALE);
          printf("        scales the coordinates of the sites by factor\n"
                 "        N, which must be greater than zero. Values of\n"
                 "        N less than one shrink the set of sites, while\n"
                 "        values greater than one expand it.\n");

          #ifndef NO_VGA
          #ifndef VALGRIND
          #ifndef CELL_BOUNDARIES_ONLY
          printf("   -%c N, --" CMDLINE_OPTS_SHRINK " N\n", CMDLINE_OPTC_SHRINK);
          printf("        draws a border around the cells;\n"
                 "        N is the width in pixels and must be an odd\n"
                 "        integer between 1 and %u.\n",
                          MAX_SHRINK_PIXELS);
          #endif
          #ifdef SVGALIB
          printf("   -%c MODE, --" CMDLINE_OPTS_VGAMODE " MODE\n", CMDLINE_OPTC_VGAMODE);
          printf("        selects a screen mode from the following:");
          modestr = list_vga_modes();
          if (modestr != NULL)
            printf("        The default is %s.\n", modestr);
          #endif
          #ifdef SDL
          printf("   -%c WIDTHxHEIGHT, --" CMDLINE_OPTS_SDLMODE " WIDTHxHEIGHT\n",
            CMDLINE_OPTC_SDLMODE);
          printf("        selects a screen size. The default is %u" "x%u.\n",
            SCREEN_DEFAULT_W, SCREEN_DEFAULT_H);
          #endif
          #endif
          #endif

          printf("   -%c, --" CMDLINE_OPTS_LIST "\n", CMDLINE_OPTC_LIST);
          printf("        disables graphic output. The resulting list of\n"
                 "        sites is written to standard output.\n");
          #ifdef SVG_SUPPORT
          #ifndef NO_VGA
          printf("   -%c, --" CMDLINE_OPTS_QUIET "\n", CMDLINE_OPTC_QUIET);
          printf("        disables graphic output. The resulting Voronoi\n"
                 "        diagram is instead written to standard output\n"
                 "        in Scalable Vector Graphics (SVG) format.\n");
          #endif
          printf("   -%c N, --" CMDLINE_OPTS_STROKE " N\n", CMDLINE_OPTC_STROKE);
          printf("        sets the stroke width for the SVG image.\n"
                 "        N must be greater than zero and less than or\n"
                 "        equal to %.3f. N defaults to %f.\n",
                          MAX_STROKE_WIDTH, DEFAULT_STROKE_WIDTH);
          #endif

          printf("   -%c N, --" CMDLINE_OPTS_RUNS " N\n", CMDLINE_OPTC_RUNS);
          printf("        sets the number of iterations. ");
          #ifdef NO_VGA
          printf("The default is %u",
                          DEFAULT_ITERATIONS);
          #else
          printf("Without this\n"
                 "        option, the algorithm is run until the user\n"
                 "        interrupts it");
          #ifdef SVG_SUPPORT
          printf(" or, with --" CMDLINE_OPTS_QUIET ", %u times",
                          DEFAULT_ITERATIONS);
          #endif
          #endif
          printf(".\n");
          printf("        Note that running only one iteration yields\n"
                 "        the same result as running `triangle -v'\n"
                 "        with the same list of input sites.\n");

          ret = RET_FORTUNE_NO_TRIANGLES;
          break;

        #ifndef CELL_BOUNDARIES_ONLY
        case CMDLINE_OPTC_SHRINK:
          if (i >= 1 && i <= MAX_SHRINK_PIXELS)
            shrink_width = (i + 1) >> 1;
          break;
        #endif

        case CMDLINE_OPTC_SCALE:
          if (optarg != NULL) {
            f = strtod(optarg, NULL);
            if (f > 0.0)
              site_scale = f;
          }
          break;

        #ifndef NO_VGA
        #ifdef SVGALIB
        case CMDLINE_OPTC_VGAMODE:
          if (find_vga_mode(optarg) != 0)
            ret = RET_FORTUNE_PARAM_ERROR;
          break;
        #endif
        #ifdef SDL
        case CMDLINE_OPTC_SDLMODE:
            if (parse_sdl_video_mode(optarg) != 0)
                ret = RET_FORTUNE_PARAM_ERROR;
        break;
        #endif
        #endif

        case CMDLINE_OPTC_LIST:
            quiet_mode = QUIET_OUTPUT_SITES;
            break;

        #ifdef SVG_SUPPORT
        case CMDLINE_OPTC_QUIET:
            quiet_mode = QUIET_OUTPUT_SVG;
            break;

        case CMDLINE_OPTC_STROKE:
            if (optarg != NULL) {
                f = strtod(optarg, NULL);
                if (f > 0.0) {
                    if (f > MAX_STROKE_WIDTH)
                        f = MAX_STROKE_WIDTH;
                    svg_stroke = f;
                }
            }
            break;
        #endif

        case CMDLINE_OPTC_RUNS:
          if (i > 0)
            iterations = i;
          break;
      }
    }

    free(optstring);

    return ret;
}


/* main()
*/

int main(int argc, char *argv[]) {

    uint16_t x;
    int i, j;
    int s;
    int dup;
    site_coord_t xd, yd;
    site_t ss; /* should be vertex_t, but plot_site() expects a site_t */
    triangle_list_t *trl;
    boundary_t *bp;
    cellboundary_t *cb;
    uint8_t b1, b2, b3;

    divisions = 0;

    #ifndef NO_VGA
    #ifdef SVGALIB
	vga_disabledriverreport();
    #endif
    #endif

    i = parse_options(argc, argv);
    if (i != 0)
        return i;

    if (iterations < 0) {
        #ifndef VALGRIND
        if (quiet_mode)
        #endif
            iterations = DEFAULT_ITERATIONS;
    }

    //fprintf(stderr, "Running %d iterations.\n", iterations);

    i = read_sites_from_file();
    if (i != 0)
        return i;

    site = (site_t *)malloc(sites * sizeof(site_t));

    if (site == NULL) {
        fprintf(stderr,
            "Not enough memory for %u sites! (" SIZEOF_FORMAT " bytes needed)\n",
            sites, sites * sizeof(site_t));
        return 1;
    }

    #ifndef NO_VGA
    #ifdef SDL

    /* we need the WINDOW_* values, which the SDL
         version has to initialize by calling
         sdl_get_window_size() either separately
         or via init_vga() */

    sdl_get_window_size();

    #endif
    #endif

    /* read a list of sites from a file */

    interpret_site_list();

    /* see if there are any duplicate sites
         and remove them from the array;
       if duplicates are found, the sites
         are rearranged in the array and the
         total number of sites is decreased
         accordingly, but the memory area
         allocated for the array stays the
         same */

    remove_duplicate_sites:

    for (i = 0, s = sites; i < s; i++) {

        xd = site[i].x;
        yd = site[i].y;

        for (j = i + 1; j < s; j++) {
            if (site[j].x > xd - SITE_COORD_EPSILON)
                if (site[j].x < xd + SITE_COORD_EPSILON)
                    if (site[j].y > yd - SITE_COORD_EPSILON)
                        if (site[j].y < yd + SITE_COORD_EPSILON) {
                            s--;
                            site[j].x = site[s].x;
                            site[j].y = site[s].y;
                        }
        }
    }

    if (s < 3) {
        fprintf(stderr, "Less than three valid coordinate pairs entered!\n");
        free(site);
        return 0;
    }

    sites = s;

    #ifndef NO_VGA

    /* enter graphics mode */

    if (!quiet_mode) {

        i = init_vga(0
            #ifndef DIRECT_TO_PHYSICAL
            | INIT_VGA_DOUBLE_BUFFER
            #endif
            );
        if (i != 0) {
            free(site);
            return 1;
        }

        get_vga_accel_caps();

        #ifdef SDL
        #ifdef VGA_FONT_INTEGRATED
        gfxPrimitivesSetFont(vga_font, 8, 8);
        #else
        gfxPrimitivesSetFont(NULL, 0, 0);
        #endif
        #endif
    }

    #endif

    vlist = NULL;
    blist = NULL;
    cell_list = NULL;

    while (1) {

        i = Voronoi(site, sites, &vlist, &vnum, &blist, &bnum, &cell_list, &actual_vnum);

        if (i != RET_FORTUNE_OK) {
            #ifndef NO_VGA
            exit_vga();
            #endif
            fprintf(stderr, "ERROR: Voronoi() returned %d\n", i);
            break;
        }

        #define SKIP_VERTEX_OUTSIDE(N) {                        \
            if (vlist[N].x < WINDOW_LEFT) continue;             \
            if (vlist[N].x > WINDOW_RIGHT) continue;            \
            if (Y_ABOVE(vlist[N].y, WINDOW_TOP)) continue;      \
            if (Y_BELOW(vlist[N].y, WINDOW_BOTTOM)) continue;   \
        }

        #define DUP_THRESHOLD   (VERTEX_COORD_EPSILON * 2.0)
        #define SKIP_DUPLICATE_VERTEX(N) {                                  \
            for (dup = 0; dup < N; dup++) {                                 \
                if (vlist[N].x < vlist[dup].x - DUP_THRESHOLD) continue;    \
                if (vlist[N].x > vlist[dup].x + DUP_THRESHOLD) continue;    \
                if (vlist[N].y < vlist[dup].y - DUP_THRESHOLD) continue;    \
                if (vlist[N].y > vlist[dup].y + DUP_THRESHOLD) continue;    \
                break;                                                      \
            }                                                               \
            if (dup < N)                                                    \
                continue;                                                   \
        }

        #if 0
        #define PRINT_NUMBER_OF_SITES 1
        printf("%d", sites);
        fflush(stdout);
        #endif

        #ifndef VALGRIND
        #ifndef NO_VGA

        if (!quiet_mode) {

            reset_screen(1);

            #define DO_BOUNDARY(BPTR, COLOUR) {                     \
                bp = BPTR;                                          \
                if (bp->v2 == NULL)                                 \
                    draw_boundary(bp->v1->x, bp->v1->y,             \
                                  bp->vector->x, bp->vector->y,     \
                                  DRAW_BOUNDARY_RELATIVE, COLOUR);  \
                else                                                \
                    draw_boundary(bp->v1->x, bp->v1->y,             \
                                  bp->v2->x, bp->v2->y, 0, COLOUR); \
            }

            #ifdef CELL_BOUNDARIES_ONLY
            for (i = 0; i < bnum; i++)
                DO_BOUNDARY(&blist[i], 2);
            #else
            for (i = 0, b2 = 1; i < sites; i++) {
                b1 = cell_list[i].sides;
                j = paint_cell(&cell_list[i], 0,
                    shrink_width,
                    ((b1 - 1) % 15) + 1,
                    (b1 == 8) ? 7 : ((b1 > 8) ? 15 : (b1 | 8)),
                    NULL, NULL);
                if (j == 0 && b2 > 0) {
                    blit_screen(scr_virtual);
                    j = dot_getch();
                    if (j == 10) b2 = 0;
                    if (j != 27) j = 0;
                }
                if (j > 0)
                    break;
            }
            #endif

            #if 0       /* print numbers of old and new sites */
            #ifdef SDL
            for (i = 0; i < sites; i++) {
                snprintf(buffer, BUFFER_LEN, "%d", i);
                stringColor(scr_virtual,
                    PIXEL_X(site[i].x) + 2, PIXEL_Y(site[i].y) - 4,
                    buffer, sdl_gfx_colour(15));
            }
            for (i = 0, s = sites; i < actual_vnum; i++, s++) {
                SKIP_VERTEX_OUTSIDE(i);
                snprintf(buffer, BUFFER_LEN, "%d", s);
                stringColor(scr_virtual,
                    PIXEL_X(vlist[i].x) + 2, PIXEL_Y(vlist[i].y) - 4,
                    buffer, sdl_gfx_colour(11));
            }
            #endif
            #endif

            blit_screen(scr_virtual);

            j = dot_getch();

            if (j == 's') {
                #ifdef SDL
                #ifndef VALGRIND
                SDL_SaveBMP(scr_virtual, "out.bmp");
                #endif
                #endif
                #ifdef SVGALIB
                save_pbm_screenshot("out.pbm", scr_virtual, 4);
                #endif
                j = 0;
            }
            if (j == 27 || j == 'q' || j < 0)
                break;
        }

        #endif
        #endif

        iterations--;

        if (!iterations)
            break;

        /* add new sites at vertex coordinates */

        for (i = s = 0; i < actual_vnum; i++) {
            SKIP_VERTEX_OUTSIDE(i);
            SKIP_DUPLICATE_VERTEX(i);
            s++;
        }

        new_site = realloc(site, sizeof(site_t) * (sites + s));

        if (new_site == NULL) {
            i = RET_FORTUNE_NO_MEM;
            j = -1;
            break;
        }

        for (i = 0; i < actual_vnum; i++) {
            SKIP_VERTEX_OUTSIDE(i);
            SKIP_DUPLICATE_VERTEX(i);
            new_site[sites].x = vlist[i].x;
            new_site[sites].y = vlist[i].y;
            sites++;
        }

        site = new_site;

        Voronoi_cleanup(vlist, blist, cell_list);

        #ifdef PRINT_NUMBER_OF_SITES
        printf("\t-> %d sites\n", sites);
        #endif

        /* run Lloyd's algorithm */

        /**** TODO: there should be a variant of Lloyd()
                we could use here on the first iteration
                that uses the existing set of Voronoi cells
                instead of calling Voronoi() again ****/

        for (j = LLOYD_ITERATIONS, i = RET_FORTUNE_OK; j > 0 && i == RET_FORTUNE_OK; j--) {
            i = (j > 1) ?
                Lloyd(site, sites, NULL, NULL, NULL, NULL, NULL, NULL, NULL) :
                Lloyd(site, sites, NULL, &vlist, &vnum, &blist, &bnum, &cell_list, &actual_vnum);
        }

        if (i != RET_FORTUNE_OK) {
            #ifndef NO_VGA
            exit_vga();
            #endif
            fprintf(stderr, "ERROR: Lloyd() returned %d\n", i);
            break;
        }
    }

    #ifdef PRINT_NUMBER_OF_SITES
    printf("\n");
    #endif

    if (i == RET_FORTUNE_OK) {
        #ifdef SVG_SUPPORT
        if (quiet_mode == QUIET_OUTPUT_SVG)
            i = output_svg(FORTUNE_VORONOI_MASK);
        #endif
        if (quiet_mode == QUIET_OUTPUT_SITES) {
            for (j = 0; j < sites; j++)
                printf("%10.6f,%10.6f\n", site[j].x, site[j].y);
        }
    }

    Voronoi_cleanup(vlist, blist, cell_list);

    /* free memory, return to text mode and exit */

    all_done:

    free(site);

    #ifndef NO_VGA
    exit_vga();
    #endif

    return i;
}
